theory FiniteMap
  imports Main
  abbrevs "(=" = "\<subseteq>\<^sub>f"
begin

typedef ('a, 'b) fmap (infixr "\<rightharpoonup>\<^sub>f" 0) = "{m :: ('a \<rightharpoonup> 'b). finite (dom m)}" using finite_dom_map_of by auto 
setup_lifting type_definition_fmap
lift_definition fmap_empty :: "('a, 'b) fmap" is "Map.empty" by simp
lift_definition fmap_lookup :: "('a, 'b) fmap \<Rightarrow> 'a \<Rightarrow> 'b option" is "\<lambda>m k. m k" done
lift_definition fmap_update :: "('a, 'b) fmap \<Rightarrow> 'a \<Rightarrow> 'b option \<Rightarrow> ('a, 'b) fmap" is "\<lambda>m k v. m(k := v)" by simp
lift_definition fmap_delete :: "'a \<Rightarrow> ('a, 'b) fmap \<Rightarrow> ('a, 'b) fmap" is "\<lambda>k m. m |` (-{k})" by simp
lift_definition fmap_dom :: "('a, 'b) fmap \<Rightarrow> 'a set" is dom done
lift_definition fmap_ran :: "('a, 'b) fmap \<Rightarrow> 'b set" is ran done
lift_definition fmap_add :: "('a, 'b) fmap \<Rightarrow> ('a, 'b) fmap \<Rightarrow> ('a, 'b) fmap" is "\<lambda>m\<^sub>1 m\<^sub>2. m\<^sub>1++m\<^sub>2" by simp
lift_definition fmap_le :: "('a, 'b) fmap \<Rightarrow> ('a, 'b) fmap \<Rightarrow> bool" is "\<lambda>m\<^sub>1 m\<^sub>2. m\<^sub>1\<subseteq>\<^sub>mm\<^sub>2" done
lift_definition restrict_fmap :: "('a, 'b) fmap \<Rightarrow> 'a set \<Rightarrow> ('a, 'b) fmap" is "\<lambda>m d. m|`d" by force
notation fmap_add (infixl "++\<^sub>f" 100)
notation fmap_le (infixl "\<subseteq>\<^sub>f" 50)
notation restrict_fmap (infixl "|`\<^sub>f" 110)
abbreviation fmap_ge (infix "\<supseteq>\<^sub>f" 50) where "h \<supseteq>\<^sub>f h\<^sub>0 \<equiv> h\<^sub>0 \<subseteq>\<^sub>f h"

lemmas fmap_dom_simp = fmap_dom.rep_eq[simp]
lemmas fmap_ran_simp = fmap_ran.rep_eq[simp]
lemmas fmap_empty_simp = fmap_empty.rep_eq[simp]
lemmas fmap_add_simp = fmap_add.rep_eq[simp]
lemmas fmap_le_simp = fmap_le.rep_eq[simp]
lemmas fmap_lookup_simp = fmap_lookup.rep_eq[simp]
lemmas fmap_update_simp = fmap_update.rep_eq[simp]
lemmas restrict_fmap_simp = restrict_fmap.rep_eq[simp]

lemma fmap_empty_upd_none[simp]: "fmap_update fmap_empty x  None = fmap_empty"
  using Rep_fmap_inject by fastforce

lemma fmap_upd_triv: "fmap_lookup t k = x \<Longrightarrow> fmap_update t k x = t"
  using Rep_fmap_inject by force

lemma fmap_upd_nonempty[simp]: "fmap_update t k (Some x) \<noteq> fmap_empty"
  by (metis fmap_empty_simp fmap_update_simp fun_upd_same option.discI)

lemma fmap_upd_eqD1:
  assumes "fmap_update m a x = fmap_update n a y"
    shows "x = y"
  by (metis assms fmap_update_simp fun_upd_eqD)

lemma fmap_upd_Some_unfold:
  "(fmap_lookup (fmap_update m a (Some b)) x = Some y) = (x = a \<and> b = y \<or> x \<noteq>a \<and> fmap_lookup m x = Some y)" by force

lemma fmap_add_empty[simp]: "m ++\<^sub>f fmap_empty = m"
  by (metis Rep_fmap_inverse all_not_in_conv dom_eq_empty_conv fmap_add_simp fmap_empty.rep_eq map_add_subsumed2 map_le_def)

lemma empty_fmap_add[simp]: "fmap_empty++\<^sub>fm=m"
  by (metis Rep_fmap_inverse all_not_in_conv dom_eq_empty_conv fmap_add_simp fmap_empty.rep_eq map_add_subsumed1 map_le_def)

lemma fmap_add_assoc[simp]: "(h\<^sub>1 ++\<^sub>f h\<^sub>2) ++\<^sub>f h\<^sub>3 = h\<^sub>1 ++\<^sub>f (h\<^sub>2 ++\<^sub>f h\<^sub>3)"
  using Rep_fmap_inject by force

lemma fmap_add_Some_iff: "(fmap_lookup (m ++\<^sub>f n) k = Some x) = (fmap_lookup n k = Some x \<or> fmap_lookup n k = None \<and> fmap_lookup m k = Some x)"
  by (simp add: map_add_Some_iff)

lemma fmap_add_SomeD[dest!]: "fmap_lookup (m ++\<^sub>f n) k = Some x \<Longrightarrow> fmap_lookup n k = Some x \<or> fmap_lookup n k = None \<and> fmap_lookup m k = Some x"
  by auto

lemma fmap_add_find_right[simp]: "fmap_lookup n k = Some v \<Longrightarrow> fmap_lookup (m ++\<^sub>f n) k = Some v"
  by fastforce

lemma fmap_add_None[iff]: "(fmap_lookup (m ++\<^sub>f n) k = None) = (fmap_lookup n k = None \<and> fmap_lookup m k = None)"
  by auto

lemma fmap_add_upd[simp]: "f ++\<^sub>f (fmap_update g x (Some y)) = fmap_update (f ++\<^sub>f g) x (Some y)"
  using Rep_fmap_inject by force

lemma fmap_add_upd_left: "m\<notin>fmap_dom e2 \<Longrightarrow> fmap_update e1 m (Some u1) ++\<^sub>f e2 = fmap_update (e1 ++\<^sub>f e2) m (Some u1)"
  by (metis Rep_fmap_inject fmap_add_simp fmap_dom_simp fmap_update_simp map_add_upd_left)



lemma restrict_fmap_to_empty[simp]: "m|`\<^sub>f{} = fmap_empty"
  using Rep_fmap_inject by force

lemma restrict_fmap_empty[simp]: "fmap_empty|`\<^sub>fD = fmap_empty"
  using Rep_fmap_inject by force

lemma fmap_restrict_in[simp]: "x\<in>A \<Longrightarrow> fmap_lookup (m|`\<^sub>fA) x = fmap_lookup m x"
  by simp

lemma fmap_restrict_out[simp]: "x\<notin>A \<Longrightarrow> fmap_lookup (m|`\<^sub>fA) x = None"
  by fastforce

lemma fmap_ran_restrictD: "y \<in> fmap_ran (m|`\<^sub>fA) \<Longrightarrow> \<exists>x\<in>A. fmap_lookup m x = Some y"
  by (simp add: ran_restrictD)

lemma fmap_dom_restrict[simp]: "fmap_dom (m|`\<^sub>fA) = fmap_dom m \<inter> A"
  by auto

lemma fmap_restrict_upd_same[simp]: "fmap_update m x (Some y)|`\<^sub>f(-{x}) = m|`\<^sub>f(-{x})"
  using Rep_fmap_inject by fastforce

lemma fmap_restrict_restrict[simp]: "m|`\<^sub>fA|`\<^sub>fB = m|`\<^sub>f(A\<inter>B)"
  using Rep_fmap_inject by fastforce

lemma fmap_restrict_fun_upd[simp]: "fmap_update m x y|`\<^sub>fD = (if x\<in>D then fmap_update (m|`\<^sub>f(D-{x})) x y else m|`\<^sub>fD)"
  using Rep_fmap_inject by force

lemma fmap_fun_upd_None_restrict[simp]: "fmap_update (m|`\<^sub>fD) x None = (if x\<in>D then m|`\<^sub>f(D - {x}) else m|`\<^sub>fD)"
  using Rep_fmap_inject by fastforce

lemma fmap_fun_upd_restrict: "fmap_update (m|`\<^sub>fD) x y = fmap_update (m|`\<^sub>f(D-{x})) x y"
  using Rep_fmap_inject by (metis fmap_update_simp fun_upd_restrict restrict_fmap_simp)

lemma fmap_fun_upd_restrict_conv[simp]: "x\<in>D \<Longrightarrow> fmap_update (m|`\<^sub>fD) x y = fmap_update (m|`\<^sub>f(D-{x})) x y"
  using Rep_fmap_inject by fastforce

lemma fmap_restrict_complement_singleton_eq:
  "f|`\<^sub>f(-{x}) = fmap_update f x None"
  using Rep_fmap_inject by fastforce





lemma fmap_le_fmap_add[simp]:
  "f \<subseteq>\<^sub>f g ++\<^sub>f f" by simp

lemma fmap_add_subsumed1:
  "f \<subseteq>\<^sub>f g \<Longrightarrow> f ++\<^sub>f g = g" by (metis Rep_fmap_inject fmap_add_simp fmap_le_simp map_add_subsumed1)

lemma fmap_le_refl[simp]:
  "f \<subseteq>\<^sub>f f" by simp

lemma fmap_le_iff_fmap_add_commute:
  "(f \<subseteq>\<^sub>f f ++\<^sub>f g) = (f ++\<^sub>f g = g ++\<^sub>f f)" by (metis Rep_fmap_inject fmap_add_simp fmap_le_simp map_le_iff_map_add_commute)

lemma fmap_le_imp_upd_le[simp]:
  "m\<^sub>1 \<subseteq>\<^sub>f m\<^sub>2 \<Longrightarrow> fmap_update m\<^sub>1 x None \<subseteq>\<^sub>f fmap_update m\<^sub>2 x (Some y)" by fastforce

lemma fmap_add_dom_app_simps:
  "m \<in> fmap_dom l\<^sub>2 \<Longrightarrow> fmap_lookup (l\<^sub>1 ++\<^sub>f l\<^sub>2) m = fmap_lookup l\<^sub>2 m"
  "m \<notin> fmap_dom l\<^sub>1 \<Longrightarrow> fmap_lookup (l\<^sub>1 ++\<^sub>f l\<^sub>2) m = fmap_lookup l\<^sub>2 m"
  "m \<notin> fmap_dom l\<^sub>2 \<Longrightarrow> fmap_lookup (l\<^sub>1 ++\<^sub>f l\<^sub>2) m = fmap_lookup l\<^sub>1 m"
    apply auto[1] apply (simp add: map_add_dom_app_simps(2)) by (simp add: map_add_dom_app_simps(3))

lemma upd_None_fmap_le[simp]:
  "fmap_update f x  None \<subseteq>\<^sub>f f" by fastforce

lemma fmap_add_comm: "fmap_dom m1 \<inter> fmap_dom m2 = {} \<Longrightarrow> m1 ++\<^sub>f m2 = m2 ++\<^sub>f m1"
  using Rep_fmap_inject map_add_comm by fastforce

lemma fmap_le_implies_dom_le: "(f \<subseteq>\<^sub>f g) \<Longrightarrow> (fmap_dom f \<subseteq> fmap_dom g)"
  by (fastforce simp add: map_le_def dom_def)

lemma fmap_add_le_fmapE: "f ++\<^sub>f g \<subseteq>\<^sub>f h \<Longrightarrow> g \<subseteq>\<^sub>f h"
  by (fastforce simp: map_le_def map_add_def dom_def)

lemma fmap_le_antisym: "f\<subseteq>\<^sub>fg \<Longrightarrow> g\<subseteq>\<^sub>ff \<Longrightarrow> f=g"
  using Rep_fmap_inject map_le_antisym by auto

(*declare fmap_dom_simp[simp del]
declare fmap_add_simp[simp del]
declare fmap_le_simp[simp del]
declare fmap_lookup_simp[simp del]
declare fmap_update_simp[simp del]*)

definition compat :: "('a \<rightharpoonup>\<^sub>f 'b) \<Rightarrow> ('a \<rightharpoonup>\<^sub>f 'b) \<Rightarrow> bool" (infix "\<bottom>" 51) where
 "h\<^sub>1 \<bottom> h\<^sub>2 \<longleftrightarrow> fmap_dom h\<^sub>1 \<inter> fmap_dom h\<^sub>2 = {}"

lemma compat_comm:
  "h \<bottom> h' \<Longrightarrow> h' \<bottom> h" unfolding compat_def by blast

lemma compat_distr:
  "h \<bottom> h' \<Longrightarrow> h = h\<^sub>1++\<^sub>fh\<^sub>2 \<Longrightarrow> h\<^sub>1 \<bottom> h' \<and> h\<^sub>2 \<bottom> h'" unfolding compat_def by auto

lemma compat_heap_union_comm:
  "h\<^sub>1 \<bottom> h\<^sub>2 \<Longrightarrow> h\<^sub>1++\<^sub>fh\<^sub>2 = h\<^sub>2++\<^sub>fh\<^sub>1" unfolding compat_def by (metis Rep_fmap_inject fmap_add_simp fmap_dom_simp map_add_comm)

lemma heap_union_assoc:"(h\<^sub>1++h\<^sub>2)++h\<^sub>3=h\<^sub>1++(h\<^sub>2++h\<^sub>3)" by simp

lemma compat_subset:
  "h\<^sub>1\<bottom>h\<^sub>2 \<Longrightarrow> h=h\<^sub>1++\<^sub>fh\<^sub>2 \<Longrightarrow> h\<^sub>2\<subseteq>\<^sub>fh" by simp

lemma compat_union: "a\<bottom>b \<Longrightarrow> a\<bottom>c \<Longrightarrow> a\<bottom>(b++\<^sub>fc)" unfolding compat_def by auto

lemma compat_heap_added_equal: "\<And>h\<^sub>1 h\<^sub>2 h\<^sub>2'. h\<^sub>1\<bottom>h\<^sub>2 \<Longrightarrow> h\<^sub>1\<bottom>h\<^sub>2' \<Longrightarrow> h\<^sub>2++\<^sub>fh\<^sub>1=h\<^sub>2'++\<^sub>fh\<^sub>1 \<Longrightarrow> h\<^sub>2=h\<^sub>2'" unfolding compat_def apply simp
proof(-)
  fix h\<^sub>1 h\<^sub>2 h\<^sub>2' :: "('a, 'b) fmap"
  assume "dom (Rep_fmap h\<^sub>1) \<inter> dom (Rep_fmap h\<^sub>2) = {}" "dom (Rep_fmap h\<^sub>1) \<inter> dom (Rep_fmap h\<^sub>2') = {}" "h\<^sub>2 ++\<^sub>f h\<^sub>1 = h\<^sub>2' ++\<^sub>f h\<^sub>1"
  then have "\<forall>n. fmap_lookup h\<^sub>2' n = fmap_lookup h\<^sub>2 n" by (metis (no_types, lifting) IntI empty_iff fmap_add_simp fmap_lookup.rep_eq map_add_dom_app_simps(2) map_add_dom_app_simps(3))
  then show "h\<^sub>2 = h\<^sub>2'" using Rep_fmap_inject by force
qed

lemma split_fmap:
  "h=h|`\<^sub>f(-A)++\<^sub>fh|`\<^sub>fA" 
proof(-)
  have "h|`\<^sub>f(-A) \<subseteq>\<^sub>f h" by (simp add: map_le_def)
  moreover have "h|`\<^sub>fA \<subseteq>\<^sub>f h" by (simp add: map_le_def)
  ultimately have "h|`\<^sub>f(-A)++\<^sub>fh|`\<^sub>fA \<subseteq>\<^sub>f h" by (simp add: map_add_le_mapI)
  moreover have "h\<subseteq>\<^sub>fh|`\<^sub>f(-A)++\<^sub>fh|`\<^sub>fA" apply simp by (smt (verit, ccfv_SIG) Diff_Compl Diff_iff domIff dom_restrict map_add_dom_app_simps(1) map_add_dom_app_simps(3) map_le_def restrict_map_def)
  ultimately show ?thesis by (simp add: fmap_le_antisym)
qed

end