theory HeaplessProgramCommands
  imports Main
begin

(*type_synonym val = nat*)
type_synonym ('var, 'val) state = "'var \<Rightarrow> 'val"
type_synonym ('var, 'val) bexp = "('var, 'val) state set"
type_synonym ('var, 'val) aexp = "('var, 'val) state \<Rightarrow> 'val"

datatype ('var, 'val) cmd = 
 Assign 'var "('var, 'val) aexp"
|Seq "('var, 'val) cmd" "('var, 'val) cmd"
|If "('var, 'val) cmd" "('var, 'val) cmd"
|Skip
|Havoc 'var
|Assume "('var, 'val) bexp"
|While "('var, 'val) cmd"

type_synonym ('var, 'val) conf = "('var, 'val) cmd \<times> ('var, 'val) state"

subsection Semantics

inductive red :: "('var, 'val) conf \<Rightarrow> ('var, 'val) conf \<Rightarrow> bool" ("_ \<rightarrow> _" [0,0] 81) where
 RedAssign: "(Assign x e, s) \<rightarrow> (Skip, s(x:=e s))"
|RedSeq1: "(Seq Skip C\<^sub>2, s) \<rightarrow> (C\<^sub>2, s)"
|RedSeq2: "(C\<^sub>1, s) \<rightarrow> (C\<^sub>1', s') \<Longrightarrow> (Seq C\<^sub>1 C\<^sub>2, s) \<rightarrow> (Seq C\<^sub>1' C\<^sub>2, s')"
|RedIf1: "(If C\<^sub>1 C\<^sub>2, s) \<rightarrow> (C\<^sub>1, s)"
|RedIf2: "(If C\<^sub>1 C\<^sub>2, s) \<rightarrow> (C\<^sub>2, s)"
|RedHavoc: "(Havoc x, s) \<rightarrow> (Skip, s(x:=v))"
|RedAssume: "s\<in>b \<Longrightarrow> (Assume b, s) \<rightarrow> (Skip, s)"
|RedWhile: "(While C, s) \<rightarrow> (If (Seq C (While C)) Skip, s)"

inductive_cases red_Seq_elim[elim!]: "(Seq C\<^sub>1 C\<^sub>2, s) \<rightarrow> (C', s')"
inductive_cases red_Seq_elim2[elim!]: "(Seq Skip C\<^sub>2, s) \<rightarrow> (C\<^sub>2, s')"
inductive_cases red_While_elim: "(While C, s) \<rightarrow> (C', s')"
inductive_cases red_If_elim[elim!]: "(If C\<^sub>1 C\<^sub>2, s) \<rightarrow> (C', s')"
inductive_cases red_Assume_elim[elim!]: "(Assume b, s) \<rightarrow> (Skip, s')"
inductive_cases red_Assign_elim[elim!]: "(Assign x e, s) \<rightarrow> (Skip, s')"
inductive_cases red_Havoc_elim[elim!]: "(Havoc x, s) \<rightarrow> (Skip, s')"

inductive rtred :: "('var, 'val) conf \<Rightarrow> ('var, 'val) conf \<Rightarrow> bool" ("_ \<rightarrow>* _" [0,0] 81) where
 RedRefl: "c \<rightarrow>* c"
|RedTran: "c \<rightarrow> c' \<Longrightarrow> c' \<rightarrow>* c'' \<Longrightarrow> c \<rightarrow>* c''"

lemma rtred_trans: "c \<rightarrow>* c' \<Longrightarrow> c' \<rightarrow>* c'' \<Longrightarrow> c \<rightarrow>* c''"
  apply(induction rule: rtred.induct)
   apply (simp add: RedRefl RedTran)
  by (simp add: RedTran)

fun red_iter :: "('var, 'val) conf \<Rightarrow> nat \<Rightarrow> ('var, 'val) conf \<Rightarrow> bool" where
 "red_iter c 0 c' \<longleftrightarrow> c=c'"
|"red_iter c (Suc n) c'' \<longleftrightarrow> (\<exists>c'. (c \<rightarrow> c') \<and> red_iter c' n c'')"

lemma rtred_eq_red_iter_bkw: "red_iter c n c' \<Longrightarrow> c \<rightarrow>* c'"
  apply(induction n arbitrary: c c')
   apply(simp add: RedRefl)
  by (meson red_iter.simps(2) rtred.simps)

definition red_iter_closed :: "('var, 'val) conf \<Rightarrow> ('var, 'val) conf \<Rightarrow> bool" where
 "red_iter_closed c c' \<longleftrightarrow> (\<exists>n. red_iter c n c')"

lemma rtred_eq_red_iter_closed_fwd: "c \<rightarrow>* c' \<Longrightarrow> red_iter_closed c c'"
  apply(induction rule: rtred.induct)
  using red_iter.simps(1) red_iter_closed_def apply blast
  by (meson red_iter.simps(2) red_iter_closed_def)

lemma rtred_eq_red_iter_closed_bkw: "red_iter_closed c c' \<Longrightarrow> c \<rightarrow>* c'"
  using red_iter_closed_def rtred_eq_red_iter_bkw by blast

lemma rtred_eq_red_iter_closed: "(c \<rightarrow>* c') \<longleftrightarrow> red_iter_closed c c'"
  using rtred_eq_red_iter_closed_bkw rtred_eq_red_iter_closed_fwd by blast

definition dsem :: "('var, 'val) cmd \<Rightarrow> (('var, 'val) state \<times> ('var, 'val) state) set" ("\<lbrakk>_\<rbrakk>" [0] 81) where
 "\<lbrakk>C\<rbrakk> = {(s, s')| s s'. ((C, s) \<rightarrow>* (Skip, s'))}"

lemma dsem_char:
  "(s, s')\<in>\<lbrakk>C\<rbrakk> \<longleftrightarrow> ((C, s) \<rightarrow>* (Skip, s'))"
  using dsem_def by blast

(*tacitly: a valid path with length n, note that there are paths with length 0*)
definition a_path :: "('var, 'val) conf list \<Rightarrow> nat \<Rightarrow> bool" where
 "a_path l n \<longleftrightarrow> length l = n+1 \<and> (\<forall>k<n. (l!k \<rightarrow> l!(k+1)))"

lemma a_pathI:
  assumes "length l = n+1"
      and "\<And>k. k<n \<Longrightarrow> l!k \<rightarrow> l!(k+1)"
    shows "a_path l n"
  using assms a_path_def by blast

(*tacitly: a successful path with length n, note that there are paths with length 0*)
definition is_path :: "('var, 'val) conf list \<Rightarrow> ('var, 'val) conf \<Rightarrow> nat \<Rightarrow> ('var, 'val) conf \<Rightarrow> bool" where
 "is_path l c n c' \<longleftrightarrow> a_path l n \<and> l!0 = c \<and> l!n = c'"

lemma red_iter's_path:
  assumes "red_iter (C, s) n (C'', s'')"
  shows "\<exists>l. is_path l (C, s) n (C'', s'')" 
  using assms
proof(induction "(C, s)" n "(C'', s'')" arbitrary: C s s'' rule: red_iter.induct)
  case 1
  then have "(C, s) = (C'', s'')" by simp
  moreover obtain l where "l=[(C, s)]" by auto
  ultimately show ?case using a_path_def is_path_def by (metis Suc_eq_plus1 length_Cons list.size(3) not_less_zero nth_Cons_0)
next
  case (2 n)
  then obtain C' s' where 0: "(C, s)\<rightarrow>(C', s')" "red_iter (C', s') n (C'', s'')" by auto
  then obtain l where l_def: "is_path l (C', s') n (C'', s'')" using "2.hyps" by blast
  then obtain l' where l'_def: "l'=(C, s)#l" by auto
  have "a_path l' (Suc n)"
  proof(rule a_pathI)
    have "length l = n+1" using l_def a_path_def is_path_def by blast
    then show "length l' = Suc n + 1" using l'_def by simp
  next
    fix k
    assume "k<Suc n"
    show "l' ! k \<rightarrow> l' ! (k + 1)"
    proof(cases "k=0")
      case True
      then show ?thesis using "0"(1) by (metis Suc_eq_plus1 is_path_def l'_def l_def nth_Cons_0 nth_Cons_Suc)
    next
      case False
      then show ?thesis using l_def is_path_def a_path_def
        by (metis \<open>k < Suc n\<close> add_Suc_shift add_cancel_right_right add_diff_cancel_left' l'_def less_Suc_eq_0_disj nth_Cons_pos plus_1_eq_Suc)
    qed
  qed
  then show ?case using l_def l'_def is_path_def by (metis nth_Cons_0 nth_Cons_Suc)
qed

lemma dsem_path_left_fixed_long:
  assumes "length l = n+1"
      and "\<forall>k'<n. (l!k' \<rightarrow> l!(k'+1))"
      and "k\<le>n"
    shows "l!0 \<rightarrow>* l!k"
  using assms
proof(induction n arbitrary: l k)
  case 0
  then show ?case
    by (simp add: RedRefl)
next
  case (Suc n)
  then obtain c l' where l'_def: "l = c # l'"
    by (metis Suc_eq_plus1 length_Suc_conv)
  then have same: "\<forall>k\<le>n. l'!k = l!(k+1)"
    by auto
  then have "length l' = n + 1" "\<forall>k'<n. (l'!k' \<rightarrow> l'!(k'+1))" "n \<le> Suc n" 
      apply (metis Suc.prems(1) Suc_eq_plus1 diff_Suc_1 l'_def length_Cons)
     using Suc.prems(2) Suc_eq_plus1 Suc_leI Suc_less_eq less_or_eq_imp_le same apply presburger
     by simp
   then have "l'!0 \<rightarrow>* l'!(k-1)"
     using Suc.IH Suc.prems(3) by force
  then show ?case
    by (metis (no_types) Suc.prems(2) l'_def le_add2 le_add_same_cancel2 nth_Cons' rtred.simps same zero_less_Suc)
qed

lemma dsem_path_long:
  assumes "length l = n+1"
      and "\<forall>k'<n. (l!k' \<rightarrow> l!(k'+1))"
      and "k\<^sub>l\<le>k\<^sub>r"
      and "k\<^sub>r\<le>n"
    shows "l!k\<^sub>l \<rightarrow>* l!k\<^sub>r"
  using assms
proof(-)
  obtain l' where l'_def: "l' = drop k\<^sub>l l"
    by simp
  then have same: "\<forall>k\<le>n-k\<^sub>l. l'!k = l!(k+k\<^sub>l)"
    by (metis (no_types, lifting) add.commute assms(1) assms(3) assms(4) le_add1 le_trans nth_drop)
  have props: "length l' = n-k\<^sub>l+1" "\<forall>k'<n-k\<^sub>l. (l'!k' \<rightarrow> l'!(k'+1))" "k\<^sub>r-k\<^sub>l\<le>n-k\<^sub>l" 
      apply (metis Nat.add_diff_assoc2 assms(1) assms(3) assms(4) l'_def le_trans length_drop)
    using same assms(1) assms(2) assms(4) diff_le_mono by auto
  then have "l'!0 \<rightarrow>* l'!(k\<^sub>r-k\<^sub>l)"
    by (meson eq_imp_le dsem_path_left_fixed_long)
  thus "l!k\<^sub>l \<rightarrow>* l!k\<^sub>r" using same
    by (simp add: assms(3) props(3))
qed

lemma dsem_path:
  assumes "a_path l n"
      and "k\<^sub>l\<le>k\<^sub>r"
      and "k\<^sub>r\<le>n"
    shows "l!k\<^sub>l \<rightarrow>* l!k\<^sub>r"
  using assms by (simp add: a_path_def dsem_path_long)

lemma dsem_path_left_fixed_same_len_long:
  assumes "length l = n+1"
      and "\<forall>k'<n. (l!k' \<rightarrow> l!(k'+1))"
      and "k\<le>n"
    shows "red_iter (l!0) k (l!k)"
  using assms
proof(induction n arbitrary: l k)
  case 0
  then show ?case
    by (simp add: RedRefl)
next
  case (Suc n)
  then obtain c l' where l'_def: "l = c # l'"
    by (metis Suc_eq_plus1 length_Suc_conv)
  then have same: "\<forall>k\<le>n. l'!k = l!(k+1)"
    by auto
  then have "length l' = n + 1" "\<forall>k'<n. (l'!k' \<rightarrow> l'!(k'+1))" "n \<le> Suc n" 
      apply (metis Suc.prems(1) Suc_eq_plus1 diff_Suc_1 l'_def length_Cons)
     using Suc.prems(2) Suc_eq_plus1 Suc_leI Suc_less_eq less_or_eq_imp_le same apply presburger
     by simp
   then have "red_iter (l'!0) (k-1) (l'!(k-1))"
     using Suc.IH Suc.prems(3) by simp
  then show ?case by (metis (no_types, lifting) Suc.prems(2) diff_Suc_1 l'_def le_add2 le_add_same_cancel2 nth_Cons' red_iter.elims(3) red_iter.simps(1) same zero_less_Suc)
qed

lemma dsem_path_same_len_long:
  assumes "length l = n+1"
      and "\<forall>k'<n. (l!k' \<rightarrow> l!(k'+1))"
      and "k\<^sub>l\<le>k\<^sub>r"
      and "k\<^sub>r\<le>n"
    shows "red_iter (l!k\<^sub>l) (k\<^sub>r-k\<^sub>l) (l!k\<^sub>r)"
  using assms
proof(-)
  obtain l' where l'_def: "l' = drop k\<^sub>l l"
    by simp
  then have same: "\<forall>k\<le>n-k\<^sub>l. l'!k = l!(k+k\<^sub>l)"
    by (metis (no_types, lifting) add.commute assms(1) assms(3) assms(4) le_add1 le_trans nth_drop)
  have props: "length l' = n-k\<^sub>l+1" "\<forall>k'<n-k\<^sub>l. (l'!k' \<rightarrow> l'!(k'+1))" "k\<^sub>r-k\<^sub>l\<le>n-k\<^sub>l" 
      apply (metis Nat.add_diff_assoc2 assms(1) assms(3) assms(4) l'_def le_trans length_drop)
    using same assms(1) assms(2) assms(4) diff_le_mono by auto
  then have "red_iter (l'!0) (k\<^sub>r-k\<^sub>l) (l'!(k\<^sub>r-k\<^sub>l))" using dsem_path_left_fixed_same_len_long by blast
  thus "red_iter (l!k\<^sub>l) (k\<^sub>r-k\<^sub>l) (l!k\<^sub>r)" using same
    by (simp add: assms(3) props(3))
qed

lemma dsem_path_same_len:
  assumes "a_path l n"
      and "k\<^sub>l\<le>k\<^sub>r"
      and "k\<^sub>r\<le>n"
    shows "red_iter (l!k\<^sub>l) (k\<^sub>r-k\<^sub>l) (l!k\<^sub>r)"
  using assms by (simp add: a_path_def dsem_path_same_len_long)

lemma skip_deadend: 
  assumes "c \<rightarrow> c'"
    shows "fst c \<noteq> Skip"
  using assms
  apply(induction rule: red.induct)
  by auto

(*declare [[show_types]]*)

lemma dsem_skip: "\<lbrakk>Skip :: ('var, 'val) cmd\<rbrakk> = {(s, s)| s :: ('var, 'val) state. True}"
proof
  show "\<lbrakk>Skip\<rbrakk> \<supseteq> {(s, s)| s. True}" using RedRefl dsem_def by auto
  show "\<lbrakk>Skip :: ('var, 'val) cmd\<rbrakk> \<subseteq> {(s, s)| s :: ('var, 'val) state. True}"
  proof
    fix io
    assume "io\<in>\<lbrakk>Skip :: ('var, 'val) cmd\<rbrakk>"
    moreover have "\<lbrakk>Skip\<rbrakk> = {(s, s')| s s'. ((Skip, s) \<rightarrow>* (Skip, s'))}" using dsem_def by blast
    then have "\<lbrakk>Skip\<rbrakk> = {(s, s')| s s'. ((Skip, s) \<rightarrow>* (Skip, s')) \<and> s=s'}" by (smt (verit, ccfv_threshold) Collect_cong fstI rtred.cases skip_deadend sndI)
    ultimately show "io\<in>{(s, s)| s :: ('var, 'val) state. True}" by blast
  qed
qed
    
lemma seq1_unique:
  assumes "(Seq Skip C\<^sub>2, s) \<rightarrow> c'"
    shows "fst c' = C\<^sub>2"
  using assms
  apply(induction "(Seq Skip C\<^sub>2, s)" c' rule: red.induct)
             apply(auto)
  by (meson fst_conv skip_deadend)

lemma seq1_preserves_dsem:
  "(((Seq Skip C\<^sub>2), s) \<rightarrow>* (Skip, s')) \<longleftrightarrow> ((C\<^sub>2, s) \<rightarrow>* (Skip, s'))" 
  by (smt (verit) RedSeq1 RedTran cmd.distinct(15) prod.exhaust_sel red_Seq_elim2 rtred.cases seq1_unique)
(*  by (smt (verit) RedSeq1 RedTran cmd.distinct(23) prod.exhaust_sel red_Seq_elim2 rtred.cases seq1_unique)*)

lemma dsem_seq1: "\<lbrakk>Seq Skip C\<^sub>2\<rbrakk> = \<lbrakk>C\<^sub>2\<rbrakk>" using dsem_def seq1_preserves_dsem by blast

lemma seq_trcl:
  assumes "(C\<^sub>1, s) \<rightarrow>* (C\<^sub>1', s')"
    shows "(Seq C\<^sub>1 C\<^sub>2, s) \<rightarrow>* (Seq C\<^sub>1' C\<^sub>2, s')"
  using assms
  apply(induction "(C\<^sub>1, s)" "(C\<^sub>1', s')" arbitrary: C\<^sub>1 s s' rule: rtred.induct)
   apply (simp add: RedRefl)
  using RedSeq2 rtred.simps by fastforce

(* Opposite direction not true: (x++;(x++)*, (x=0)) \<rightarrow>\<^sup>5 (skip;(x++), (x=2)) *)
(* (x++;(x++)*, (x=0)) \<rightarrow> (skip;(x++)*, (x=1)) \<rightarrow> ((x++)*, (x=1)) \<rightarrow> (x++;(x++)* + skip, (x=1)) \<rightarrow> (x++;(x++)*, (x=1)) \<rightarrow> (skip;(x++)*, (x=2)) *)
(* However, check seq_rev_path for particular case, where the opposite direction holds *)
lemma seq_trcl_same_len:
  "red_iter (C\<^sub>1, s) n (C\<^sub>1'', s'') \<Longrightarrow> red_iter (Seq C\<^sub>1 C\<^sub>2, s) n (Seq C\<^sub>1'' C\<^sub>2, s'')" 
proof(induction "(C\<^sub>1, s)" n "(C\<^sub>1'', s'')" arbitrary: C\<^sub>1 s s'' rule: red_iter.induct)
  case 1
  then show ?case by fastforce
next
  case (2 n)
  then obtain C\<^sub>1' s' where "(C\<^sub>1, s) \<rightarrow> (C\<^sub>1', s')" "red_iter (C\<^sub>1', s') n (C\<^sub>1'', s'')" using red_iter.simps(2) by auto
  then have "red_iter (Seq C\<^sub>1' C\<^sub>2, s') n (Seq C\<^sub>1'' C\<^sub>2, s'')" by (simp add: "2.hyps")
  moreover have "(Seq C\<^sub>1 C\<^sub>2, s) \<rightarrow> (Seq C\<^sub>1' C\<^sub>2, s')" using \<open>(C\<^sub>1, s) \<rightarrow> (C\<^sub>1', s')\<close> by (simp add: RedSeq2)
  ultimately show ?case by (meson red_iter.simps(2))
qed

lemma seq2_non_skip:
  assumes "(Seq C\<^sub>1 C\<^sub>2, s) \<rightarrow> c'"
      and "C\<^sub>1 \<noteq> Skip"
    shows "\<exists>C\<^sub>1'. fst c' = Seq C\<^sub>1' C\<^sub>2"
  using assms
  apply(induction "(Seq C\<^sub>1 C\<^sub>2, s)" c' rule: red.induct)
  by auto

lemma seq_has_skip_in_path_long:
  assumes "length l = n+1"
      and "l!0 = (Seq C\<^sub>1 C\<^sub>2, s)"
      and "l!n = c''" 
      and "\<forall>k<n. (l!k \<rightarrow> l!(k+1))"
      and "\<forall>C\<^sub>1''. fst c'' \<noteq> Seq C\<^sub>1'' C\<^sub>2"
      and "C\<^sub>1 \<noteq> Skip"
    shows "\<exists>k<n. fst (l!k) = Seq Skip C\<^sub>2 \<and> (\<forall>k'<k. \<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip)"
  using assms
proof(induction n arbitrary: l C\<^sub>1 s c'')
  case 0
  then show ?case
    by auto
next
  case (Suc n)
  then obtain C\<^sub>1' s' where C1'_def: "fst (l!1) = Seq C\<^sub>1' C\<^sub>2" "snd (l!1) = s'"
    by (metis One_nat_def add.commute plus_1_eq_Suc seq2_non_skip zero_less_Suc)
  then show ?case
  proof(cases "C\<^sub>1' = Skip")
    case True
    then show ?thesis
      using C1'_def Suc.prems(2) Suc.prems(6) by (metis One_nat_def Suc.prems(3) Suc.prems(5) fst_eqD less_one not_less_eq) 
  next
    case False
    obtain l' where l'_def: "l = (Seq C\<^sub>1 C\<^sub>2, s) # l'"
      by (metis Cons_nth_drop_Suc Suc.prems(1) Suc.prems(2) add_gr_0 drop0 gr0_conv_Suc)
    then have same: "\<forall>k\<le>n. l'!k = l!(k+1)"
      by auto
    then have l'_prop: "length l' = n+1" "l'!0 = (Seq C\<^sub>1' C\<^sub>2, s')"  "l'!n = c''"  "\<forall>k<n. (l'!k \<rightarrow> l'!(k+1))"
         apply (metis l'_def Suc.prems(1) Suc_eq_plus1 diff_Suc_1 length_Cons)
        apply (metis C1'_def(1) C1'_def(2) One_nat_def \<open>l = (Seq C\<^sub>1 C\<^sub>2, s) # l'\<close> nth_Cons_Suc prod.collapse)
       using Suc.prems(3) l'_def apply force
      by (metis Suc.prems(4) Suc_eq_plus1 leI not_less_eq order.strict_implies_order same)
    then obtain k where k_def: "fst (l'!k) = Seq Skip C\<^sub>2" "\<forall>k'<k. \<exists>C\<^sub>1'. fst (l'!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip" "k<Suc n"
      using False Suc.IH Suc.prems(5) by (metis plus_1_eq_Suc trans_less_add2)
    have "\<forall>k'<(Suc k). \<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip"
    proof
      fix k'
      show "k' < Suc k \<longrightarrow> (\<exists>C\<^sub>1'. fst (l ! k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip)"
      proof
        assume "k' < Suc k"
        show "\<exists>C\<^sub>1'. fst (l ! k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip"
        proof(cases "k' = k")
          case True
          then show ?thesis
            by (metis C1'_def(1) C1'_def(2) False \<open>k' < Suc k\<close> l'_prop(2) cmd.inject(2) k_def(1) k_def(2) l'_def less_Suc_eq_0_disj nth_Cons_Suc prod.collapse)
        next
          case False
          then show ?thesis
            using Suc.prems(6) \<open>k' < Suc k\<close> k_def(2) l'_def less_Suc_eq_0_disj by fastforce
        qed
      qed
    qed
    then show ?thesis using \<open>k<Suc n\<close> by (smt (verit, del_insts) Suc.prems(5) Suc_mono add_Suc_right add_cancel_right_right ex_least_nat_less k_def(1) l'_prop(3) le_SucE le_antisym le_numeral_extra(4) less_antisym less_imp_le_nat less_numeral_extra(4) less_one same)
  qed
qed

lemma seq_has_skip_in_path:
  assumes "is_path l (Seq C\<^sub>1 C\<^sub>2, s) n c''" 
      and "\<forall>C\<^sub>1''. fst c'' \<noteq> Seq C\<^sub>1'' C\<^sub>2"
    shows "\<exists>k<n. fst (l!k) = Seq Skip C\<^sub>2 \<and> (\<forall>k'<k. \<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip)"
proof(cases "C\<^sub>1=Skip")
  case True
  then show ?thesis by (metis assms(1) assms(2) fst_conv is_path_def less_nat_zero_code neq0_conv)
next
  case False
  then show ?thesis using assms seq_has_skip_in_path_long a_path_def is_path_def by blast
qed

lemma seq_rev:
  assumes "(Seq C\<^sub>1 C\<^sub>2, s\<^sub>0) \<rightarrow> (Seq C\<^sub>1' C\<^sub>2, s\<^sub>2)"
    shows "(C\<^sub>1, s\<^sub>0) \<rightarrow> (C\<^sub>1', s\<^sub>2)"
  using assms
  apply(induction "(Seq C\<^sub>1 C\<^sub>2, s\<^sub>0)" "(Seq C\<^sub>1' C\<^sub>2, s\<^sub>2)" rule: red.induct)
  by auto

lemma seq_rev_path_long:
  assumes "length l = n+1"
      and "l!0 = (Seq C\<^sub>1 C\<^sub>2, s)"
      and "l!n = (Seq Skip C\<^sub>2, s'')" 
      and "\<forall>k<n. (l!k \<rightarrow> l!(k+1))"
      and "\<forall>k<n. \<exists>C\<^sub>1'. fst (l!k) = Seq C\<^sub>1' C\<^sub>2"
    shows "red_iter (C\<^sub>1, s) n (Skip, s'')"
  using assms
proof(induction n arbitrary: l C\<^sub>1 s s'')
  case 0
  then show ?case
    by simp
next
  case (Suc n)
  then obtain C\<^sub>1' s' where l1: "l!1 = (Seq C\<^sub>1' C\<^sub>2, s')"
    by (metis Suc_eq_plus1 fst_conv linorder_neqE_nat not_add_less2 surj_pair)
  have "length l > 0"
    by (simp add: Suc.prems(1))
  then obtain l' where l'_def: "l = (Seq C\<^sub>1 C\<^sub>2, s) # l'"
    by (metis Suc.prems(2) hd_conv_nth length_greater_0_conv list.collapse)
  then have same: "\<forall>k\<le>n. l'!k = l!(k+1)"
    by auto
  then have l'_prop: "\<forall>k<n. (l'!k \<rightarrow> l'!(k+1))" "\<forall>k<n. \<exists>C\<^sub>1'. fst (l'!k) = Seq C\<^sub>1' C\<^sub>2"
     apply (metis Suc.prems(4) Suc_eq_plus1 Suc_less_eq less_eq_Suc_le less_or_eq_imp_le)
    using Suc.prems(5) l'_def by auto
  then have "red_iter (C\<^sub>1', s') n (Skip, s'')"
    by (metis One_nat_def Suc.IH Suc.prems(1) Suc.prems(3) Suc_eq_plus1 add_left_cancel l'_def l1 length_Cons nth_Cons_Suc plus_1_eq_Suc)
  then show ?case
    by (metis One_nat_def Suc.prems(2) Suc.prems(4) add.commute l1 plus_1_eq_Suc red_iter.simps(2) seq_rev zero_less_Suc)
qed

lemma seq_rev_path:
  assumes "is_path l (Seq C\<^sub>1 C\<^sub>2, s) n (Seq Skip C\<^sub>2, s'')"
      and "\<forall>k<n. \<exists>C\<^sub>1'. fst (l!k) = Seq C\<^sub>1' C\<^sub>2"
    shows "red_iter (C\<^sub>1, s) n (Skip, s'')"
  using assms seq_rev_path_long a_path_def is_path_def by blast

lemma seq_has_skip:
  assumes "(Seq C\<^sub>1 C\<^sub>2, s) \<rightarrow>* (Skip, s'')"
    shows "\<exists>s'. ((C\<^sub>1, s) \<rightarrow>* (Skip, s')) \<and> ((Seq Skip C\<^sub>2, s') \<rightarrow>* (Skip, s''))"
proof(-)
  have "red_iter_closed (Seq C\<^sub>1 C\<^sub>2, s) (Skip, s'')" using assms by (simp add: rtred_eq_red_iter_closed)
  then obtain n where n_def: "red_iter (Seq C\<^sub>1 C\<^sub>2, s) n (Skip, s'')" using red_iter_closed_def by blast
  then obtain l where l_def: "is_path l (Seq C\<^sub>1 C\<^sub>2, s) n (Skip, s'')" using red_iter's_path by blast
  then obtain k where k_def: "k<n" "fst (l!k) = Seq Skip C\<^sub>2" "\<forall>k'<k. \<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip" using seq_has_skip_in_path by force
  obtain l' where l'_def: "l'=take (k+1) l" by blast
  have "\<forall>k'<k. (l!k' \<rightarrow> l!(k'+1))" using l_def is_path_def a_path_def k_def(1) by (metis add_lessD1 canonically_ordered_monoid_add_class.lessE)
  moreover have "length l = n+1" using l_def is_path_def a_path_def by blast
  then have "length l' = k+1" using k_def(1) l'_def \<open>length l = n + 1\<close> by force
  ultimately have "a_path l' k" using a_path_def[of l' k] by (simp add: l'_def)
  moreover have "l'!0 = (Seq C\<^sub>1 C\<^sub>2, s)" using is_path_def l'_def l_def by fastforce
  ultimately have "is_path l' (Seq C\<^sub>1 C\<^sub>2, s) k (l!k)" using is_path_def[of l' "(Seq C\<^sub>1 C\<^sub>2, s)" k] using l'_def by simp
  moreover have "\<forall>k'<k. \<exists>C\<^sub>1'. fst (l'!k') = Seq C\<^sub>1' C\<^sub>2" using k_def(3) l'_def by auto
  then have "red_iter (C\<^sub>1, s) k (Skip, snd (l!k))" using seq_rev_path by (metis calculation k_def(2) prod.collapse)
  then show ?thesis by (metis eq_imp_le is_path_def k_def(1) k_def(2) l_def less_imp_le_nat prod.collapse rtred_eq_red_iter_bkw dsem_path)
qed

lemma seq1'_preserves_dsem:
 "((Seq C Skip, s) \<rightarrow>* (Skip, s')) \<longleftrightarrow> ((C, s) \<rightarrow>* (Skip, s'))"
  by (meson RedRefl rtred_trans seq1_preserves_dsem seq_has_skip seq_trcl)

lemma dsem_seq1': "\<lbrakk>Seq C\<^sub>2 Skip\<rbrakk> = \<lbrakk>C\<^sub>2\<rbrakk>" using dsem_def seq1'_preserves_dsem by blast

lemma skip_comm: "\<lbrakk>Seq C Skip\<rbrakk> = \<lbrakk>Seq Skip C\<rbrakk>" using dsem_seq1 dsem_seq1' by auto

lemma dsem_seq:
  "\<lbrakk>Seq C\<^sub>1 C\<^sub>2\<rbrakk> = {(s, s'')| s s''. \<exists>s'. (s, s')\<in>\<lbrakk>C\<^sub>1\<rbrakk> \<and> (s', s'')\<in>\<lbrakk>C\<^sub>2\<rbrakk>}" 
proof(-)
  have "\<lbrakk>Seq C\<^sub>1 C\<^sub>2\<rbrakk> = {(s, s'')| s s''. ((Seq C\<^sub>1 C\<^sub>2), s) \<rightarrow>* (Skip, s'')}" using dsem_def by blast
  then have "\<lbrakk>Seq C\<^sub>1 C\<^sub>2\<rbrakk> = {(s, s'')| s s''. \<exists>s'. ((C\<^sub>1, s) \<rightarrow>* (Skip, s')) \<and> ((Seq Skip C\<^sub>2, s') \<rightarrow>* (Skip, s''))}" by (smt (verit, ccfv_SIG) Collect_cong case_prodI2 eq_fst_iff fst_eqD rtred_trans seq_has_skip seq_trcl)
  then show ?thesis using dsem_char by (smt (verit, ccfv_SIG) Collect_cong dsem_seq1)
qed

lemma seq_assoc:
  "\<lbrakk>Seq (Seq C\<^sub>1 C\<^sub>2) C\<^sub>3\<rbrakk> = \<lbrakk>Seq C\<^sub>1 (Seq C\<^sub>2 C\<^sub>3)\<rbrakk>" (is "?X=?Y")
proof(-)
  have "?X = {(s, s''')| s s'''. \<exists>s''. (s, s'')\<in>\<lbrakk>Seq C\<^sub>1 C\<^sub>2\<rbrakk> \<and> (s'', s''')\<in>\<lbrakk>C\<^sub>3\<rbrakk>}" using dsem_seq by blast
  then have "?X = {(s, s''')| s s'''. \<exists>s''. (s, s'')\<in>{(s, s'')| s s''. \<exists>s'. (s, s')\<in>\<lbrakk>C\<^sub>1\<rbrakk> \<and> (s', s'')\<in>\<lbrakk>C\<^sub>2\<rbrakk>} \<and> (s'', s''')\<in>\<lbrakk>C\<^sub>3\<rbrakk>}" using dsem_seq by fastforce
  then have "?X = {(s, s''')| s s' s'' s'''. (s, s')\<in>\<lbrakk>C\<^sub>1\<rbrakk> \<and> (s', s'')\<in>\<lbrakk>C\<^sub>2\<rbrakk> \<and> (s'', s''')\<in>\<lbrakk>C\<^sub>3\<rbrakk>}" by auto
  moreover have "?Y = {(s, s''')| s s'''. \<exists>s'. (s, s')\<in>\<lbrakk>C\<^sub>1\<rbrakk> \<and> (s', s''')\<in>\<lbrakk>Seq C\<^sub>2 C\<^sub>3\<rbrakk>}" using dsem_seq by blast
  then have "?Y = {(s, s''')| s s'''. \<exists>s'. (s, s')\<in>\<lbrakk>C\<^sub>1\<rbrakk> \<and> (s', s''')\<in>{(s', s''')| s' s'''. \<exists>s''. (s', s'')\<in>\<lbrakk>C\<^sub>2\<rbrakk> \<and> (s'', s''')\<in>\<lbrakk>C\<^sub>3\<rbrakk>}}" using dsem_seq by fast
  then have "?Y = {(s, s''')| s s' s'' s'''. (s, s')\<in>\<lbrakk>C\<^sub>1\<rbrakk> \<and> (s', s'')\<in>\<lbrakk>C\<^sub>2\<rbrakk> \<and> (s'', s''')\<in>\<lbrakk>C\<^sub>3\<rbrakk>}" by auto
  ultimately show ?thesis by argo
qed

lemma assume_preserves_dsem:
  "((Assume e, s)\<rightarrow>*(Skip, s')) \<longleftrightarrow> s=s' \<and> s\<in>e"
proof
  show "s=s' \<and> s\<in>e \<Longrightarrow> ((Assume e, s)\<rightarrow>*(Skip, s'))" by (meson RedAssume RedRefl RedTran)
  show "((Assume e, s)\<rightarrow>*(Skip, s')) \<Longrightarrow> s=s' \<and> s\<in>e"
  proof(-)
    have "\<And>s s'. ((Assume e, s)\<rightarrow>(Skip, s')) \<longleftrightarrow> s=s' \<and> s\<in>e" using RedAssume by auto
    assume "(Assume e, s)\<rightarrow>*(Skip, s')"
    then have "red_iter_closed (Assume e, s) (Skip, s')" by (simp add: rtred_eq_red_iter_closed)
    then obtain n where n_def: "red_iter (Assume e, s) n (Skip, s')" using red_iter_closed_def by blast
    then obtain l where l_def: "is_path l (Assume e, s) n (Skip, s')" using red_iter's_path by blast
    then have 0: "\<forall>k<n. (l!k\<rightarrow>l!(k+1))" using is_path_def a_path_def by blast
    moreover have "l!0 = (Assume e, s)" using is_path_def l_def by fast
    moreover have "\<And>c. (Assume e, s)\<rightarrow>c \<Longrightarrow> fst c = Skip" using red.cases by fastforce
    ultimately have "fst (l!1) = Skip" by (metis RedSeq1 add_0 cmd.distinct(33) n_def not_gr_zero red_iter.simps(1) seq1_unique)
    then have "n=1" by (metis "0" cmd.distinct(33) fst_conv is_path_def l_def less_one linorder_neqE_nat skip_deadend)
    then show "s=s' \<and> s\<in>e" by (metis "0" add_0 fact_0 fact_gt_zero is_path_def l_def red_Assume_elim)
  qed
qed

lemma dsem_assume:
  "\<lbrakk>Assume e\<rbrakk> = {(s, s)| s. s\<in>e}"
  by (smt (verit, ccfv_SIG) Collect_cong assume_preserves_dsem dsem_def)

lemma if_preserves_dsem:
  "((If C\<^sub>1 C\<^sub>2, s) \<rightarrow>* (Skip, s')) \<longleftrightarrow> ((C\<^sub>1, s) \<rightarrow>* (Skip, s')) \<or> ((C\<^sub>2, s) \<rightarrow>* (Skip, s'))" (is "?X \<longleftrightarrow> ?Y")
proof
  show "?X \<Longrightarrow> ?Y" using rtred.cases by fastforce
  show "?Y \<Longrightarrow> ?X" by (meson RedIf1 RedIf2 RedTran)
qed

lemma dsem_if:
  "\<lbrakk>If C\<^sub>1 C\<^sub>2\<rbrakk> = \<lbrakk>C\<^sub>1\<rbrakk> \<union> \<lbrakk>C\<^sub>2\<rbrakk>" 
proof(-)
  have "\<lbrakk>If C\<^sub>1 C\<^sub>2\<rbrakk> = {(s, s')| s s'. ((If C\<^sub>1 C\<^sub>2, s) \<rightarrow>* (Skip, s'))}" using dsem_def by blast
  then have "\<lbrakk>If C\<^sub>1 C\<^sub>2\<rbrakk> = {(s, s')| s s'. ((C\<^sub>1, s) \<rightarrow>* (Skip, s')) \<or> ((C\<^sub>2, s) \<rightarrow>* (Skip, s'))}" using if_preserves_dsem by fastforce
  then have "\<lbrakk>If C\<^sub>1 C\<^sub>2\<rbrakk> = {(s, s')| s s'. ((C\<^sub>1, s) \<rightarrow>* (Skip, s'))} \<union> {(s, s')| s s'. ((C\<^sub>2, s) \<rightarrow>* (Skip, s'))}" by auto
  then show ?thesis using dsem_def by auto
qed

lemma Iter_0_step: "red_iter (While C, s) 0 c' \<Longrightarrow> c'=(While C, s)" by simp

lemma Iter_1_step: "red_iter (While C, s) 1 c' \<Longrightarrow> c'=(If (Seq C (While C)) Skip, s)"
  by (metis (no_types, opaque_lifting) One_nat_def red_While_elim red_iter.simps(1) red_iter.simps(2) surj_pair)

lemma Iter_2_step: "red_iter (While C, s) 2 c' \<Longrightarrow> c'=(Skip, s) \<or> c'=(Seq C (While C), s)"
  using Iter_1_step by (smt (verit) One_nat_def Suc_1 old.prod.exhaust red_If_elim red_iter.simps(1) red_iter.simps(2))

lemma Successfull_iter_is_ge2: "red_iter (While C, s) n (Skip, s') \<Longrightarrow> n>1"
  using Iter_0_step Iter_1_step using nat_neq_iff by blast

lemma Iter_ge2_steps: "is_path l (While C, s) n c'' \<Longrightarrow> n>2 \<Longrightarrow> l!2=(Seq C (While C), s)"
proof(-)
  assume asm: "is_path l (While C, s) n c''" "n>2"
  then have "red_iter (While C, s) 2 (l!2)" using is_path_def dsem_path_same_len by (metis (no_types, lifting) diff_zero less_or_eq_imp_le zero_le_numeral)
  moreover have "is_path l (While C, s) n c'' \<Longrightarrow> n>2 \<Longrightarrow> l!2\<noteq>(Skip, s)"
  proof(-)
    assume asm: "is_path l (While C, s) n c''" "n>2"
    then have "length l > 3" using is_path_def a_path_def by (metis add_less_cancel_right nat_1_add_1 numeral_Bit1 numerals(1))
    then have "l!2 \<rightarrow> l!3" using asm is_path_def a_path_def by (metis (no_types, lifting) inc.simps(2) numeral_inc)
    then show ?thesis by (metis fst_conv skip_deadend)
  qed
  ultimately show ?thesis using Iter_2_step asm by blast
qed

(* Seems useless for now *)
lemma iter_has_next:
  assumes "is_path l (While C, s) n (Skip, s'')"
      and "n>2"
    shows "\<exists>k<n. k>2 \<and> fst (l!k) = (While C)" (*k<n.. even k<n-1 or n-2 or n-3.. w/e*)
proof(-)
  obtain l' where l'_def: "l' = drop 2 l" by fastforce
  moreover have "length l > 3" using a_path_def is_path_def assms by (metis add.commute nat_add_left_cancel_less numeral_Bit0 numeral_Bit1)
  ultimately have shifted_by_2: "\<forall>k. l'!k = l!(k+2)" by simp
  moreover have l_path: "\<forall>k<n. (l!k \<rightarrow> l!(k+1))" using is_path_def a_path_def assms(1) by blast
  ultimately have l'_path: "\<forall>k<n-2. (l'!k \<rightarrow> l'!(k+1))" by simp
  moreover have l_second: "l!2=(Seq C (While C), s)" using Iter_ge2_steps assms by blast
  then have "l'!0=(Seq C (While C), s)" using shifted_by_2 by (metis add_0)
  moreover have "l!n=(Skip, s'')" using assms(1) is_path_def by blast
  then have l'_last: "l'!(n-2)=(Skip, s'')" using shifted_by_2 by (metis assms(2) le_add_diff_inverse2 less_imp_le_nat)
  ultimately have "is_path l' (Seq C (While C), s) (n-2) (Skip, s'')"
    using is_path_def a_path_def by (smt (verit, del_insts) Nat.diff_diff_right \<open>l' = drop 2 l\<close> assms diff_diff_left le_add_diff_inverse2 length_drop less_diff_conv less_or_eq_imp_le nat_1_add_1)
  moreover have "\<forall>C\<^sub>1''. fst (Skip, s'') \<noteq> Seq C\<^sub>1'' (While C)" by auto
  show ?thesis
  proof(cases "C=Skip")
    case True
    then show ?thesis by (metis RedSeq1 Suc_eq_plus1 Suc_lessI \<open>l ! n = (Skip, s'')\<close> assms(2) cmd.distinct(35) l_path l_second lessI seq1_unique)
  next
    case False
    obtain k where k_def: "fst (l'!k) = Seq Skip (While C)" "\<forall>k'<k. \<exists>C\<^sub>1'. fst (l'!k') = Seq C\<^sub>1' (While C) \<and> C\<^sub>1' \<noteq> Skip" "k<n-2"
      using seq_has_skip_in_path[of l' C "While C" s "n-2" "(Skip, s'')"] using False calculation by auto
    then have "fst (l'!(k+1)) = (While C)" by (metis eq_fst_iff l'_path seq1_unique)
    moreover have "k+3\<le>n" using k_def(3) by linarith
    ultimately have "\<exists>k\<le>n. k>2 \<and> fst (l!k) = (While C)" using \<open>3 < length l\<close> l'_def verit_comp_simplify1(3) k_def(1) by fastforce
    then show ?thesis using \<open>l ! n = (Skip, s'')\<close> fst_conv by fastforce
  qed
qed

lemma iter_one:
  assumes "is_path l (While C, s) n (Skip, s'')"
      and "n>2"
    shows "\<exists>k<n. k>2 \<and> fst (l!(k-1)) = Seq Skip (While C) \<and> fst (l!k) = While C \<and> (\<forall>k'<k. k'=0 \<or> fst (l!k')\<noteq>While C) \<and> fst (l!(k-1)) = Seq Skip (While C) \<and> (\<forall>k'<k. k'\<ge>2 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' (While C)))"
proof(-)
  obtain l' where l'_def: "l' = drop 2 l" by fastforce
  moreover have "length l > 3" using a_path_def is_path_def assms by (metis add.commute nat_add_left_cancel_less numeral_Bit0 numeral_Bit1)
  ultimately have shifted_by_2: "\<forall>k. l'!k = l!(k+2)" by simp
  moreover have l_path: "\<forall>k<n. (l!k \<rightarrow> l!(k+1))" using is_path_def a_path_def assms(1) by blast
  ultimately have l'_path: "\<forall>k<n-2. (l'!k \<rightarrow> l'!(k+1))" by simp
  moreover have l_second: "l!2=(Seq C (While C), s)" using Iter_ge2_steps assms by blast
  then have "l'!0=(Seq C (While C), s)" using shifted_by_2 by (metis add_0)
  moreover have "l!n=(Skip, s'')" using assms(1) is_path_def by blast
  then have l'_last: "l'!(n-2)=(Skip, s'')" using shifted_by_2 by (metis assms(2) le_add_diff_inverse2 less_imp_le_nat)
  ultimately have "is_path l' (Seq C (While C), s) (n-2) (Skip, s'')"
    using is_path_def a_path_def by (smt (verit, del_insts) Nat.diff_diff_right \<open>l' = drop 2 l\<close> assms diff_diff_left le_add_diff_inverse2 length_drop less_diff_conv less_or_eq_imp_le nat_1_add_1)
  moreover have "\<forall>C\<^sub>1''. fst (Skip, s'') \<noteq> Seq C\<^sub>1'' (While C)" by auto
  show ?thesis
  proof(cases "C=Skip")
    case True
    then have "fst (l ! 3) = While C" using Suc_eq_plus1 Suc_lessI \<open>l ! n = (Skip, s'')\<close> assms(2) l_path l_second lessI seq1_unique by (metis One_nat_def Suc_1 numeral_3_eq_3)
    moreover have "fst (l!1) \<noteq> While C" by (metis add_lessD1 assms(2) cmd.distinct(13) eq_fst_iff l_path l_second one_add_one red_While_elim)
    moreover have "fst (l!2) \<noteq> While C" by (simp add: l_second)
    moreover have "3<n" by (metis One_nat_def Suc_1 Suc_lessI assms(1) assms(2) iter_has_next not_less_eq numeral_3_eq_3)
    moreover have "\<forall>k'<(3::nat). k'=0 \<or> k'=1 \<or> k'=2" by auto
    ultimately have "3<n \<and> 3>(2::nat) \<and> fst (l!3) = While C \<and> (\<forall>k'<3. k'=0 \<or> fst (l!k')\<noteq>While C) \<and> (\<forall>k'<3. k'\<ge>2 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' (While C)))" using l_second numeral_Bit1 by auto
    then show ?thesis using True l_second by auto
  next
    case False
    obtain k where k_def: "fst (l'!k) = Seq Skip (While C)" "\<forall>k'<k. \<exists>C\<^sub>1'. fst (l'!k') = Seq C\<^sub>1' (While C) \<and> C\<^sub>1' \<noteq> Skip" "k<n-2"
      using seq_has_skip_in_path[of l' C "While C" s "n-2" "(Skip, s'')"] using False calculation by auto
    then have "fst (l'!(k+1)) = (While C)" by (metis eq_fst_iff l'_path seq1_unique)
    moreover have "k+3\<le>n" using k_def(3) by linarith
    moreover have "\<forall>k'\<le>k+2. k'=0 \<or> fst (l!k')\<noteq>While C"
    proof(-)
      have "\<forall>k'\<le>k. fst (l'!k')\<noteq>While C" using k_def(2) k_def(1) nat_less_le by fastforce
      then have "\<forall>k'\<le>k+2. k'<2 \<or> fst (l!k')\<noteq>While C" by (metis add.commute le_diff_conv linordered_semidom_class.add_diff_inverse shifted_by_2)
      moreover have "fst (l!1) \<noteq> While C" by (metis add_lessD1 assms(2) cmd.distinct(13) eq_fst_iff l_path l_second one_add_one red_While_elim)
      moreover have "fst (l!2) \<noteq> While C" by (simp add: l_second)
      ultimately show ?thesis by (metis diff_is_0_eq le_antisym less_diff_conv2 less_one one_add_one verit_comp_simplify1(3))
    qed
    moreover have "fst (l!(k+3)) = (While C)" by (metis add.assoc calculation(1) nat_1_add_1 numeral_Bit1 numerals(1) shifted_by_2)
    moreover have "fst (l!(k+2)) = Seq Skip (While C)" using k_def(1) shifted_by_2 by auto
    moreover have "\<forall>k'<k+2. k'\<ge>2 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' (While C))" using k_def(2) by (metis le_add_diff_inverse2 less_diff_conv2 shifted_by_2)
    then have "\<forall>k'<k+3. k'\<ge>2 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' (While C))" using \<open>fst (l!(k+2)) = Seq Skip (While C)\<close> by (smt (verit, ccfv_threshold) add.commute add.left_commute add_diff_inverse_nat less_diff_conv2 less_one numeral_Bit0 numeral_Bit1 plus_nat.add_0 verit_comp_simplify1(3))
    ultimately have "k+3\<le>n \<and> k+3>0 \<and> fst (l!(k+3)) = (While C) \<and> (\<forall>k'<k+3. k'=0 \<or> fst (l!k')\<noteq>While C) \<and> fst (l!(k+2)) = Seq Skip (While C) \<and> (\<forall>k'<k+3. k'\<ge>2 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' (While C)))" by force
    then have "\<exists>k\<le>n. k>2 \<and> fst (l!k) = (While C) \<and> (\<forall>k'<k. k'=0 \<or> fst (l!k')\<noteq>While C) \<and> fst (l!(k-1)) = Seq Skip (While C) \<and> (\<forall>k'<k. k'\<ge>2 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' (While C)))" by auto
    then show ?thesis using \<open>l ! n = (Skip, s'')\<close> fst_conv by (metis cmd.distinct(35) le_neq_implies_less)
  qed
qed

(* the world's ugliest proof *)
lemma iter_preserves_prop_intermediate:
  assumes "is_path l (While C, s) n (While C, s'')"
      and "P s"
      and "\<And>s s'. P s \<Longrightarrow> (C, s) \<rightarrow>* (Skip, s') \<Longrightarrow> P s'"
    shows "P s''"
  using assms
proof(induction n arbitrary: l s rule: nat_less_induct)
  case (1 n)
  then show ?case
  proof(cases "n=0")
    case True
    then show ?thesis by (metis "1.prems"(1) "1.prems"(2) Pair_inject is_path_def)
  next
    case False
    then have "red_iter (l!0) 1 (l!1)" using dsem_path_same_len[of l n 0 1] "1.prems"(1) is_path_def by (metis bot_nat_0.extremum diff_zero leI less_one)
    then have "l!1 = (If (Seq C (While C)) Skip, s)" by (metis "1.prems"(1) Iter_1_step is_path_def)
    then have "n>1" by (metis "1.prems"(1) False cmd.distinct(29) is_path_def leI less_one nat_less_le prod.inject)
    then have "red_iter (l!0) 2 (l!2)" using dsem_path_same_len[of l n 0 2] "1.prems"(1) is_path_def by fastforce
    then have l_seq_beg: "l!2 = (Seq C (While C), s)" by (metis (mono_tags, lifting) "1.prems"(1) Iter_2_step Iter_ge2_steps RedIf2 Suc_1 Suc_lessI \<open>1 < n\<close> cmd.distinct(36) is_path_def red_If_elim)
    then have "n>2" by (metis "1.prems"(1) \<open>1 < n\<close> add_less_imp_less_left canonically_ordered_monoid_add_class.lessE cmd.distinct(22) is_path_def less_one linorder_neqE_nat nat_1_add_1 prod.inject)
    
    obtain l' where l'_def: "l' = drop 2 l" by fast
    moreover have "length l = n+1" using is_path_def a_path_def "1.prems"(1) by blast
    then have "length l > 3" using is_path_def a_path_def "1.prems"(1) \<open>n>2\<close> by linarith
    ultimately have shifted_by_2: "\<forall>k. l'!k = l!(k+2)" by simp
    then have l_path: "\<forall>k<n. (l!k \<rightarrow> l!(k+1))" using is_path_def a_path_def "1.prems"(1) by blast

    have l'_beg: "l'!0 = (Seq C (While C), s)" using l_seq_beg shifted_by_2 \<open>n>1\<close> by (metis add_0)
    moreover have "l'!(n-2) = (While C, s'')" using l_seq_beg shifted_by_2 by (metis "1.prems"(1) \<open>2 < n\<close> is_path_def le_add_diff_inverse2 less_imp_le_nat)
    moreover have l'_path: "\<forall>k<n-2. (l'!k \<rightarrow> l'!(k+1))" using l_path shifted_by_2 by fastforce
    ultimately have l'_is_path: "is_path l' (Seq C (While C), s) (n-2) (While C, s'')"
      using a_path_def is_path_def by (smt (verit) \<open>2 < n\<close> \<open>length l = n + 1\<close> add.commute add.left_commute add_diff_cancel_left' canonically_ordered_monoid_add_class.lessE l'_def length_drop)
    moreover have "\<forall>C\<^sub>1''. fst (While C, s'') \<noteq> Seq C\<^sub>1'' (While C)" by fastforce
    ultimately obtain k where k_def: "fst (l'!k) = Seq Skip (While C)" "\<forall>k'<k. \<exists>C\<^sub>1'. fst (l'!k') = Seq C\<^sub>1' (While C) \<and> C\<^sub>1' \<noteq> Skip" "k<n-2"
      using seq_has_skip_in_path[of l' C "(While C)" s "n-2" "(While C, s'')"] by blast
    then obtain l'' where l''_def: "l'' = take (k+1) l'" by fast
    then have l''_prefix: "\<forall>n<k+1. l'!n=l''!n" by simp

    have "l''!0 = (Seq C (While C), s)" using l''_prefix l'_beg by auto
    moreover have "l''!k = l'!k" using l''_prefix by fastforce
    moreover have "\<forall>k'<k. (l''!k' \<rightarrow> l''!(k'+1))"
      using l'_path l''_prefix \<open>k<n-2\<close> by (smt (verit, ccfv_SIG) add.commute le_add_diff_inverse nat_add_left_cancel_less nat_less_le one_add_one trans_less_add2)
    moreover have "length l'' = k+1" using \<open>k < n - 2\<close> \<open>length l = n + 1\<close> l'_def l''_def by force
    ultimately have l''_path: "is_path l'' (Seq C (While C), s) k (l''!k)" using is_path_def a_path_def by blast
    then have "red_iter (C, s) k (Skip, snd (l'!k))"
      using seq_rev_path[of l'' C "While C" s k] by (metis \<open>l'' ! k = l' ! k\<close> add.commute k_def(1) k_def(2) l''_prefix prod.collapse trans_less_add2)
    then have "P (snd (l'!k))" using "1.prems"(2) "1.prems"(3) rtred_eq_red_iter_bkw by blast
    moreover have "fst (l'!(k+1)) = While C" by (metis k_def(1) k_def(3) l'_path prod.collapse seq1_unique)
    moreover have "snd (l'!k) = snd (l'!(k+1))" using k_def(1) k_def(3) l'_path by (metis calculation(2) prod.collapse red_Seq_elim2)
    then have "P (snd (l'!(k+1)))" using \<open>P (snd (l'!k))\<close> by simp
    ultimately have 0: "fst (l'!(k+1)) = While C \<and> P (snd (l'!(k+1)))" by fast

    obtain l\<^sub>0 where l\<^sub>0_def: "l\<^sub>0 = drop (k+1) l'" by blast
    moreover have "length l' > k+1" using is_path_def a_path_def \<open>length l = n + 1\<close> l'_def using k_def(3) by auto
    ultimately have shifted: "\<forall>k'. l\<^sub>0!k' = l'!(k'+k+1)" by (simp add: add.commute)
    then have "\<forall>k'<n-2-(k+1). (l\<^sub>0!k' \<rightarrow> l\<^sub>0!(k'+1))" using l'_path add.commute by auto
    moreover have "length l\<^sub>0 = n-1 - (k+1)" using is_path_def a_path_def \<open>length l = n + 1\<close> l'_def l\<^sub>0_def by simp
    ultimately have "a_path l\<^sub>0 (n-2 - (k+1))" using a_path_def[of l\<^sub>0 "n-2 - (k+1)"] k_def(3) by auto
    moreover have "n-2\<ge>k+1 \<and> (\<forall>k'\<ge>k+1. l\<^sub>0!(k'-(k+1)) = l'!(k'))" using \<open>k<n-2\<close> \<open>\<forall>k'. l\<^sub>0!k' = l'!(k'+k+1)\<close> by force
    then have "l\<^sub>0!(n-2-(k+1)) = l'!(n-2)" by blast
    moreover have "l\<^sub>0!0 = l'!(k+1)" using \<open>\<forall>k'. l\<^sub>0!k' = l'!(k'+k+1)\<close> by fastforce
    ultimately have "is_path l\<^sub>0 (l'!(k+1)) (n-2-(k+1)) (l'!(n-2))" using is_path_def[of l\<^sub>0 "l'!(k+1)" "n-2-(k+1)" "l'!(n-2)"] by fast
    moreover have \<open>\<forall>m<n. \<forall>s'. is_path l\<^sub>0 (While C, s') m (While C, s'') \<longrightarrow> P s' \<longrightarrow> (\<forall>x. P x \<longrightarrow> (\<forall>xb. ((C, x) \<rightarrow>* (Skip, xb)) \<longrightarrow> P xb)) \<longrightarrow> P s''\<close> using "1.IH" by blast
    
    ultimately show ?thesis 
      by (metis (no_types, lifting) "1.prems"(3) False "0" \<open>l' ! (n - 2) = (While C, s'')\<close> add_gr_0 bot_nat_0.not_eq_extremum diff_diff_left diff_less prod.collapse zero_less_one)
  qed
qed

lemma iter_preserves_prop:
  assumes "is_path l (While C, s) n (Skip, s'')"
      and "P s"
      and "\<And>s s'. P s \<Longrightarrow> (C, s) \<rightarrow>* (Skip, s') \<Longrightarrow> P s'"
    shows "P s''"
  using assms
proof(induction n arbitrary: l s rule: nat_less_induct)
  case (1 n)
  have "red_iter (l!0) 1 (l!1)" using dsem_path_same_len[of l n 0 1] "1.prems"(1) is_path_def by (metis bot_nat_0.extremum cmd.distinct(35) diff_zero leI less_one prod.inject)
  then have "l!1 = (If (Seq C (While C)) Skip, s)" by (metis "1.prems"(1) Iter_1_step is_path_def)
  then have "n>1" by (metis "1.prems"(1) cmd.distinct(24) cmd.distinct(35) is_path_def less_one linorder_le_less_linear nat_less_le prod.inject)
  then have "red_iter (l!0) 2 (l!2)" using dsem_path_same_len[of l n 0 2] "1.prems"(1) is_path_def by fastforce
  then have l_two: "l!2 = (Seq C (While C), s) \<or> l!2 = (Skip, s)" by (metis (no_types, lifting) "1.prems"(1) Iter_2_step is_path_def)
  then show ?case 
  proof(cases "l!2 = (Skip, s)")
    case True
    then show ?thesis
      by (metis (no_types, lifting) "1.prems"(1) "1.prems"(2) Iter_ge2_steps Pair_inject Suc_1 Suc_lessI \<open>1 < n\<close> cmd.distinct(16) is_path_def)
  next
    case False
    then have "l!2 = (Seq C (While C), s)" using l_two by presburger
    then have "n>2" by (metis "1.prems"(1) Suc_1 Suc_lessI \<open>1 < n\<close> cmd.distinct(16) fst_conv is_path_def)
    then obtain k where k_def:  "k<n" "k>2" "fst (l!k) = (While C)" using iter_has_next using "1.prems"(1) by blast
    then obtain l' where l'_def: "l' = drop k l" by blast
    moreover have "length l = n+1" using is_path_def a_path_def "1.prems"(1) by blast
    ultimately have l'_shifted: "\<forall>k'. l'!k' = l!(k'+k)" by (metis add.commute k_def(1) less_imp_le_nat nth_drop trans_le_add2)
    
    then have "\<forall>k'<n-k. (l'!k' \<rightarrow> l'!(k'+1))" using is_path_def a_path_def "1.prems"(1) by (smt (verit) add.assoc add.commute less_diff_conv)
    moreover have "l'!(n-k) = (Skip, s'')" using l'_shifted "1.prems"(1) by (simp add: is_path_def k_def(1) less_imp_le_nat)
    moreover have "length l' = n+1-k" by (simp add: \<open>length l = n + 1\<close> l'_def)
    ultimately have "is_path l' (l'!0) (n-k) (Skip, s'')" using a_path_def is_path_def by (metis Nat.add_diff_assoc2 k_def(1) less_imp_le_nat)

    moreover have "fst (l'!0) = While C" using l'_shifted k_def(3) by simp
    moreover have "P (snd (l!k))"
    proof(-)
      obtain l\<^sub>0 where l\<^sub>0_def: "l\<^sub>0 = take (k+1) l" by blast
      then have l\<^sub>0_shifted: "\<forall>k'\<le>k. l\<^sub>0!k'=l!k'" by auto
      then have "\<forall>k'<k. (l\<^sub>0!k' \<rightarrow> l\<^sub>0!(k'+1))" using is_path_def a_path_def "1.prems"(1) by (smt (verit) add.commute k_def(1) l\<^sub>0_def le_antisym le_trans nat_add_left_cancel_less nat_less_le nth_take)
      moreover have "k\<noteq>n" using \<open>fst (l' ! 0) = While C\<close> \<open>l' ! (n - k) = (Skip, s'')\<close> by auto
      then have "length l \<ge> k+1" using \<open>k<n\<close> \<open>length l = n + 1\<close> by simp
      then have "length l\<^sub>0 = k+1" using is_path_def a_path_def "1.prems"(1) l\<^sub>0_def by simp
      ultimately have "a_path l\<^sub>0 k" using a_path_def by blast
      then have "is_path l\<^sub>0 (l\<^sub>0!0) k (l\<^sub>0!k)" using is_path_def[of l\<^sub>0 "l\<^sub>0!0" k "l\<^sub>0!k"] by blast
      moreover have "snd (l\<^sub>0!0) = s" by (metis "1.prems"(1) is_path_def l\<^sub>0_shifted less_eq_nat.simps(1) snd_conv)
      then have "P (snd (l\<^sub>0!0))" using "1.prems"(2) by blast
      moreover have "fst (l\<^sub>0!0) = While C" by (metis "1.prems"(1) bot_nat_0.extremum fst_eqD is_path_def l\<^sub>0_shifted)
      moreover have "fst (l\<^sub>0!k) = While C" using k_def(3) l\<^sub>0_shifted by fastforce
      ultimately show ?thesis by (smt (verit) "1.prems"(3) iter_preserves_prop_intermediate l\<^sub>0_shifted order_refl prod.collapse)
    qed
    then have "P (snd (l'!0))" using l'_shifted by simp
    ultimately show ?thesis
      by (metis (no_types, lifting) "1.IH" "1.prems"(3) \<open>1 < n\<close> diff_less k_def(2) nat_1_add_1 order_less_trans prod.collapse zero_less_one zero_less_two)
  qed
qed

lemma dsem_add_seqL:
  assumes "\<lbrakk>C'\<rbrakk> = \<lbrakk>C''\<rbrakk>"
    shows "\<lbrakk>Seq C C'\<rbrakk> = \<lbrakk>Seq C C''\<rbrakk>"
  using assms dsem_def by (simp add: dsem_seq)

lemma dsem_add_seqR:
  assumes "\<lbrakk>C'\<rbrakk> = \<lbrakk>C''\<rbrakk>"
    shows "\<lbrakk>Seq C' C\<rbrakk> = \<lbrakk>Seq C'' C\<rbrakk>"
  using assms dsem_def by (simp add: dsem_seq)

fun iter :: "('var, 'val) cmd \<Rightarrow> nat \<Rightarrow> ('var, 'val) cmd" where
 "iter C 0 = Skip"
|"iter C (Suc n) = (Seq C (iter C n))"

fun iter' :: "('var, 'val) cmd \<Rightarrow> nat \<Rightarrow> ('var, 'val) cmd" where
 "iter' C 0 = Skip"
|"iter' C (Suc n) = (Seq (iter' C n) C)"

lemma iter_inject: "(iter C m) = (iter C m') \<Longrightarrow> m=m'"
proof(induction m arbitrary: m')
  case 0
  then show ?case using cmd.distinct(16) by (metis iter.elims iter.simps(1))
next
  case (Suc m)
  then show ?case by (smt (verit) iter.elims iter.simps(2) cmd.distinct(16) cmd.inject(2))
qed

lemma add_Seq_to_iterL:
  "\<lbrakk>iter C (Suc n)\<rbrakk> = \<lbrakk>Seq C (iter C n)\<rbrakk>" by simp

lemma add_Seq_to_iterR:
  "\<lbrakk>iter C (Suc n)\<rbrakk> = \<lbrakk>Seq (iter C n) C\<rbrakk>"
proof(induction n)
  case 0
  then show ?case using dsem_seq1 dsem_seq1' iter.simps by metis
next
  case (Suc n)
  then show ?case by (smt (z3) iter.simps(2) dsem_add_seqL seq_assoc)
qed

lemma iter_iter':
  "\<lbrakk>iter C n\<rbrakk> = \<lbrakk>iter' C n\<rbrakk>"
proof(induction n)
  case 0
  then show ?case by simp
next
  case (Suc n)
  then show ?case by (metis add_Seq_to_iterR iter'.simps(2) dsem_add_seqR)
qed

lemma "\<lbrakk>iter C 0\<rbrakk> = \<lbrakk>Skip\<rbrakk>" by simp
lemma "\<lbrakk>iter C 1\<rbrakk> = \<lbrakk>C\<rbrakk>" using dsem_seq1' by auto
lemma "\<lbrakk>iter C 2\<rbrakk> = \<lbrakk>Seq C C\<rbrakk>" using iter.simps dsem_seq1' seq_assoc by (metis numeral_2_eq_2)

lemma iter_breakdown:
  assumes "(iter C (Suc n), s) \<rightarrow>* (Skip, s'')"
  shows "\<exists>s'. ((iter C n, s) \<rightarrow>* (Skip, s')) \<and> ((C, s') \<rightarrow>* (Skip, s''))"
  using assms by (metis add_Seq_to_iterR dsem_char seq1_preserves_dsem seq_has_skip)

lemma iter_breakdown2:
  assumes "(iter C (Suc n), s) \<rightarrow>* (Skip, s'')"
  shows "\<exists>s'. ((C, s) \<rightarrow>* (Skip, s')) \<and> ((iter C n, s') \<rightarrow>* (Skip, s''))"
  by (metis add_Seq_to_iterL assms dsem_char seq1_preserves_dsem seq_has_skip)


lemma every_iter_is_while:
  assumes "(iter C n, s) \<rightarrow>* (Skip, s'')"
    shows "(While C, s) \<rightarrow>* (Skip, s'')"
  using assms
proof(induction n arbitrary: s)
  case 0
  then show ?case using RedTran RedWhile if_preserves_dsem by fastforce
next
  case (Suc n)
  then obtain s' where s'_def: "((C, s) \<rightarrow>* (Skip, s'))" "((iter C n, s') \<rightarrow>* (Skip, s''))" using iter_breakdown2 by blast
  then have "(While C, s') \<rightarrow>* (Skip, s'')" using Suc.IH by fast
  moreover have "\<And>s s' s'' C. (C, s) \<rightarrow>* (Skip, s') \<Longrightarrow> (While C, s') \<rightarrow>* (Skip, s'') \<Longrightarrow> (While C, s) \<rightarrow>* (Skip, s'')" by (meson RedIf1 RedSeq1 RedTran RedWhile rtred_trans seq_trcl)
  ultimately show ?case using s'_def(1) by auto
qed

fun iter_count :: "('var, 'val) cmd \<Rightarrow> ('var, 'val) conf list \<Rightarrow> nat" where
 "iter_count C [] = 0"
|"iter_count C (c#l) = (if fst c=While C then Suc (iter_count C l) else (iter_count C l))"

lemma iter_count_mono_aux:
  "iter_count C l \<le> iter_count C (c#l)" 
  using iter_count.cases by simp

lemma iter_count_mono:
  "iter_count C (drop n l) \<le> iter_count C l"
proof(induction n)
  case 0
  then show ?case by simp
next
  case (Suc n)
  then show ?case using iter_count_mono_aux by (metis Cons_nth_drop_Suc drop_eq_Nil2 dual_order.trans le_SucI linorder_le_less_linear)
qed

lemma iter_count_ge2:
  assumes "is_path l (While C, s) len (Skip, s')"
    shows "len>2 \<longleftrightarrow> iter_count C l \<ge> 2"
proof
  show "len>2 \<Longrightarrow> iter_count C l \<ge> 2"
  proof(-)
    assume "len>2"
    then obtain k where k_def: "k<len" "k>2" "fst (l!k) = (While C)" using assms iter_has_next by meson
    moreover obtain l' where l'_def: "l' = drop k l" by blast
    then have "fst (l'!0) = While C" using k_def(3) assms(1) is_path_def a_path_def by (metis (mono_tags, lifting) add.right_neutral add_mono_thms_linordered_semiring(1) k_def(1) le0 nth_drop less_imp_le_nat) 
    moreover have "length l' > 0" using l'_def k_def(1) is_path_def a_path_def by (metis add_leE assms(1) length_drop verit_comp_simplify1(3) zero_less_diff)
    moreover obtain l'' where "l'' = drop 1 l'" by force
    then have "l'=(l'!0)#l''" using Cons_nth_drop_Suc calculation(5) by force
    ultimately have "iter_count C l' > 0" by (metis iter_count.simps(2) less_Suc_eq_0_disj)
    moreover have "l' = drop (k-1) (drop 1 l)" using k_def(2) l'_def by force
    then have "iter_count C (drop 1 l) \<ge> iter_count C l'" using iter_count_mono by fast
    ultimately have "iter_count C (drop 1 l) \<ge> 1" by linarith
    moreover have "l=(l!0)#(drop 1 l)" by (metis Cons_nth_drop_Suc One_nat_def \<open>0 < length l'\<close> diff_le_self drop0 l'_def length_drop less_eq_Suc_le not_gr_zero not_less_eq_eq)
    moreover have "fst (l!0) = While C" using assms(1) is_path_def by force
    ultimately show ?thesis using iter_count.simps(2) by (metis Suc_1 Suc_leI le_imp_less_Suc)
  qed
  show "iter_count C l \<ge> 2 \<Longrightarrow> len>2"
  proof(rule ccontr)
    assume cntr: "iter_count C l \<ge> 2" "\<not> len>2"
    then show "False" 
    proof(cases "len<2")
      case True
      moreover have "red_iter (While C, s) len (Skip, s')" using assms is_path_def dsem_path_same_len by fastforce
      ultimately show ?thesis using Successfull_iter_is_ge2 by fastforce
    next
      case False
      then have "len=2" using cntr(2) by auto
      then have "red_iter (While C, s) 2 (Skip, s')" using assms is_path_def dsem_path_same_len by fastforce
      then have 2: "fst (l!2) = Skip" by (metis \<open>len = 2\<close> assms fst_conv is_path_def)
      have "fst (l!0) = While C" using assms is_path_def by force
      moreover have "l!0 \<rightarrow> l!1" using assms is_path_def a_path_def by (metis (no_types, lifting) \<open>len = 2\<close> add_0 nat_1_add_1 zero_less_two)
      ultimately have 1: "fst (l!1) = (If (Seq C (While C)) Skip)" by (metis eq_fst_iff red_While_elim)
      have 0: "fst (l!0) = While C" using assms is_path_def by force
      have "length (drop 2 l) = 1" using \<open>len=2\<close> assms is_path_def a_path_def by (metis add.commute add_diff_cancel_right' length_drop)
      moreover have "l!2 = (drop 2 l)!0" using calculation by auto
      ultimately have "drop 2 l = [l!2]" by (metis (no_types, lifting) Cons_nth_drop_Suc One_nat_def diff_Suc_1 drop0 length_0_conv length_drop less_numeral_extra(1))
      moreover have "drop 1 l = (l!1)#(drop 2 l)" by (metis Cons_nth_drop_Suc Suc_1 \<open>length (drop 2 l) = 1\<close> diff_Suc_1 length_drop less_imp_diff_less less_numeral_extra(1) zero_less_diff)
      ultimately have "drop 1 l = [l!1, l!2]" by presburger
      moreover have "l=(l!0)#(drop 1 l)" by (metis Cons_nth_drop_Suc \<open>length (drop 2 l) = 1\<close> drop_0 length_drop less_nat_zero_code not_gr_zero numeral_1_eq_Suc_0 numeral_One zero_less_diff zero_less_one)
      ultimately have "l=[l!0, l!1, l!2]" by presburger
      then have "iter_count C l = 1" using "0" "1" "2"  by (metis One_nat_def cmd.distinct(29) cmd.distinct(35) iter_count.simps(1) iter_count.simps(2))
      then show ?thesis using cntr(1) by auto
    qed
  qed
qed

lemma iter_count_1:
  assumes "is_path l (While C, s) len (Skip, s')"
      and "iter_count C l = 1"
    shows "len=2 \<and> s=s'"
proof(-)
  have "len\<le>2" using assms iter_count_ge2 by fastforce
  moreover have "len\<noteq>0" using assms by (metis cmd.distinct(35) fst_eqD is_path_def)
  moreover have "red_iter (While C, s) len (Skip, s')" using assms(1) is_path_def dsem_path_same_len by fastforce
  moreover have "len\<noteq>1" using calculation(3) Iter_1_step by fast
  ultimately show ?thesis using Iter_2_step[of C s "(Skip, s')"] assms(1) by (metis Nat.add_0_right add_diff_inverse_nat le_diff_conv less_one nat_1_add_1 old.prod.inject order_le_neq_trans)
qed

lemma concat_successfull_paths:
  assumes "(C, s) \<rightarrow>* (Skip, s')"
      and "(C', s') \<rightarrow>* (Skip, s'')"
    shows "(Seq C C', s) \<rightarrow>* (Skip, s'')"
  using assms by (meson RedSeq1 RedTran rtred_trans seq_trcl)

lemma iter_combine:
  assumes "(iter C n, s) \<rightarrow>* (Skip, s')"
      and "(C, s') \<rightarrow>* (Skip, s'')"
    shows "(iter C (Suc n), s) \<rightarrow>* (Skip, s'')"
  using add_Seq_to_iterR assms(1) assms(2) concat_successfull_paths dsem_char by blast

lemma iter_combine2:
  assumes "(C, s) \<rightarrow>* (Skip, s')"
      and "(iter C n, s') \<rightarrow>* (Skip, s'')"
    shows "(iter C (Suc n), s) \<rightarrow>* (Skip, s'')"
  using assms concat_successfull_paths add_Seq_to_iterL dsem_char by blast

value "take 2 [x, y, z]"

(*non-precise*)
lemma while_to_seq_and_while:
  assumes "is_path l (While C, s) len (Skip, s'')"
      and "iter_count C l = n"
      and "n\<ge>2"
    shows "\<exists>k<len. ((C, s) \<rightarrow>* (Skip, snd (l!k))) \<and> (is_path (drop k l) (While C, snd (l!k)) (len-k) (Skip, s'')) \<and> iter_count C (drop k l) \<le> n-1"
proof(-)
  have "len>2" using assms iter_count_ge2 by fast
  then obtain k where k_def: "k<len" "k>2" "fst (l!k) = While C"
                             "\<forall>k'<k. k'=0 \<or> fst (l!k')\<noteq>While C"
                             "fst (l!(k-1)) = Seq Skip (While C)"
                             "\<forall>k'<k. k'\<ge>2 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' (While C))" using assms(1) iter_one by blast
  obtain l' where l'_def: "l' = take (Suc k) l" by fast
  moreover have l_len: "length l = len+1" using assms(1) is_path_def a_path_def by blast
  ultimately have l'_len: "length l' = k+1" using k_def(1) by fastforce
  have ll'_same: "\<forall>k'\<le>k. l!k'=l'!k'" using l'_def by auto
  moreover have "\<forall>k'<len. (l!k' \<rightarrow> l!(k'+1))" using assms(1) is_path_def a_path_def by blast
  ultimately have "\<forall>k'<k. (l'!k' \<rightarrow> l'!(k'+1))" using k_def(1) by (metis Suc_eq_plus1 Suc_lessI leI less_le_not_le order.strict_trans2)
  then have "a_path l' k" using l'_len a_path_def by blast
  then have "is_path l' (While C, s) k (l!k)" using ll'_same by (metis assms(1) bot_nat_0.extremum eq_imp_le is_path_def)

  then have "fst (l'!2) = (Seq C (While C))" using Iter_ge2_steps by (metis fst_conv k_def(2))
  obtain l'\<^sub>2 where l'\<^sub>2_def: "l'\<^sub>2 = drop 2 l'" by blast
  then have l'\<^sub>2_len: "length l'\<^sub>2 = k-1" using l'_len by fastforce
  moreover have l'l'\<^sub>2_same: "\<forall>k'\<le>k-2. l'\<^sub>2!k'=l'!(2+k')" by (metis add_increasing2 k_def(2) l'\<^sub>2_def l'_len less_SucI less_Suc_eq_le linordered_nonzero_semiring_class.zero_le_one nth_drop)
  then have "\<forall>k'<k-2. (l'\<^sub>2!k' \<rightarrow> l'\<^sub>2!(k'+1))" using \<open>\<forall>k'<k. (l' ! k' \<rightarrow> l' ! (k' + 1))\<close> by fastforce
  ultimately have "a_path l'\<^sub>2 (k-2)" using a_path_def by (metis Nat.add_diff_assoc2 k_def(2) l'\<^sub>2_def l'_len length_drop less_imp_le_nat)
  then have "is_path l'\<^sub>2 (Seq C (While C), snd (l!2)) (k-2) (l!k)" by (metis (no_types, opaque_lifting) Nat.add_0_right \<open>fst (l' ! 2) = Seq C (While C)\<close> add_diff_inverse_nat is_path_def k_def(2) l'l'\<^sub>2_same le0 less_or_eq_imp_le ll'_same prod.collapse verit_comp_simplify1(3))

  obtain l'\<^sub>3 where l'\<^sub>3_def: "l'\<^sub>3 = butlast l'\<^sub>2" by simp
  then have l'\<^sub>3_len: "length l'\<^sub>3 = k-2" by (simp add: l'\<^sub>2_len)
  moreover have l'\<^sub>2l'\<^sub>3_same: "\<forall>k'\<le>k-3. l'\<^sub>2!k'=l'\<^sub>3!k'" by (smt (verit) add_diff_cancel_right cancel_comm_monoid_add_class.diff_cancel diff_add diff_is_0_eq' k_def(2) l'\<^sub>3_def l'\<^sub>3_len less_add_one less_imp_diff_less nat_1_add_1 not_less_iff_gr_or_eq nth_butlast numeral_Bit1 numerals(1) order.strict_implies_order zero_less_diff)
  then have "\<forall>k'<k-3. (l'\<^sub>3!k' \<rightarrow> l'\<^sub>3!(k'+1))" using \<open>\<forall>k'<k-2. (l'\<^sub>2!k' \<rightarrow> l'\<^sub>2!(k'+1))\<close> by (smt (verit) ab_semigroup_add_class.add_ac(1) add.commute add_lessD1 l'\<^sub>3_def l'\<^sub>3_len less_diff_conv nth_butlast numeral_Bit0 numeral_Bit1)
  ultimately have "a_path l'\<^sub>3 (k-3)" using a_path_def by (smt (verit, best) diff_add diff_diff_left diff_is_0_eq k_def(2) less_one nat_1_add_1 numeral_Bit1 numerals(1) verit_comp_simplify1(3))
  moreover have "l'\<^sub>3!0 = (Seq C (While C), snd (l!2))" by (metis \<open>is_path l'\<^sub>2 (Seq C (While C), snd (l ! 2)) (k - 2) (l ! k)\<close> is_path_def l'\<^sub>2l'\<^sub>3_same zero_le)
  moreover have "l'\<^sub>3!(k-3) = l!(k-1)" by (smt (verit) \<open>a_path l'\<^sub>2 (k - 2)\<close> a_path_def add.commute add_numeral_left calculation(1) diff_le_self l'\<^sub>2_len l'\<^sub>3_def l'\<^sub>3_len l'l'\<^sub>2_same less_add_one less_imp_le_nat ll'_same nth_butlast numeral_One semiring_norm(2))
  ultimately have l'\<^sub>3_path: "is_path l'\<^sub>3 (Seq C (While C), snd (l!2)) (k-3) (Seq Skip (While C), snd (l!(k-1)))" using k_def(5) by (metis is_path_def prod.collapse)
  have "\<forall>k'\<le>k-3. l'\<^sub>3!k'=l'!(2+k')" by (metis add.commute add_lessD1 k_def(2) l'\<^sub>2_def l'\<^sub>2l'\<^sub>3_same l'_len less_or_eq_imp_le nat_1_add_1 nat_add_left_cancel_le nth_drop)
  have "\<forall>k'<k. k'\<ge>2 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' (While C))" using k_def(6) by blast
  have "\<forall>k'<k-3. \<exists>C\<^sub>1'. fst (l'\<^sub>3!k') = Seq C\<^sub>1' (While C)" 
  proof
    fix k'
    show "k'<k-3 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l'\<^sub>3!k') = Seq C\<^sub>1' (While C))"
    proof
      assume "k'<k-3"
      have "k'+2\<ge>2" by simp
      moreover have "\<exists>C\<^sub>1'. fst (l!(k'+2)) = Seq C\<^sub>1' (While C)" using \<open>k' < k - 3\<close> k_def(6) by force
      then show "\<exists>C\<^sub>1'. fst (l'\<^sub>3!k') = Seq C\<^sub>1' (While C)" using \<open>\<forall>k'\<le>k - 3. l'\<^sub>3 ! k' = l' ! (2 + k')\<close> \<open>k' < k - 3\<close> ll'_same by force
    qed
  qed
  then have "red_iter (C, snd (l!2)) (k-3) (Skip, snd (l!(k-1)))" using seq_rev_path l'\<^sub>3_path by blast
  note \<open>fst (l!(k-1)) = Seq Skip (While C)\<close>
  moreover note \<open>fst (l!k) = While C\<close>
  moreover have "l!(k-1) \<rightarrow> l!k" by (metis (full_types) One_nat_def Suc_leI \<open>\<forall>k'<k. (l' ! k' \<rightarrow> l' ! (k' + 1))\<close> diff_add diff_le_self diff_less k_def(2) le_refl ll'_same nat_1_add_1 order_less_trans zero_less_one zero_less_two)
  ultimately have "snd (l!(k-1)) = snd (l!k)" by (metis prod.collapse red_Seq_elim2)
  have "l!0 = (While C, s)" using assms(1) is_path_def by blast
  moreover have "l!2=(Seq C (While C), s)" using Iter_ge2_steps \<open>2 < len\<close> assms(1) by blast
  ultimately have "snd (l!0) = snd (l!2)" by auto
  then have "red_iter (C, snd (l!0)) (k-3) (Skip, snd (l!k))" using \<open>red_iter (C, snd (l!2)) (k-3) (Skip, snd (l!(k-1)))\<close> \<open>snd (l ! (k - 1)) = snd (l ! k)\<close> by auto
  then have 0: "(C, s) \<rightarrow>* (Skip, snd (l!k))" by (simp add: \<open>l ! 0 = (While C, s)\<close> rtred_eq_red_iter_bkw)

  obtain l'' where l''_def: "l'' = drop k l" by fast
  have l''_len: "length l'' = len-k+1" by (metis Nat.add_diff_assoc2 k_def(1) l''_def l_len length_drop less_or_eq_imp_le)
  have ll''_same: "\<forall>k'\<le>len-k. l!(k'+k)=l''!k'" by (metis add.commute k_def(1) l''_def l_len less_or_eq_imp_le nth_drop trans_le_add2)
  moreover note \<open>\<forall>k'<len. (l!k' \<rightarrow> l!(k'+1))\<close>
  ultimately have "\<forall>k'<len-k. (l''!k' \<rightarrow> l''!(k'+1))" using k_def(1) by (simp add: l''_def l_len)
  then have "a_path l'' (len-k)" using a_path_def l''_len by auto
  then have 1: "is_path l'' (l!k) (len-k) (Skip, s'')" by (metis (no_types, lifting) assms(1) is_path_def k_def(1) le_add_diff_inverse2 le_add_same_cancel2 le_numeral_extra(3) less_imp_le_nat ll''_same plus_nat.add_0)

  have "fst (l!0) = While C" using assms(1) is_path_def by force
  then have "iter_count C (drop 1 l) = n-1" using assms(2) by (metis Suc_diff_1 diff_0_eq_0 diff_Suc_1 diff_self_eq_0 drop0 drop_Nil drop_Suc_Cons iter_count.elims less_one nth_Cons_0)
  moreover have "drop k l = drop (k-1) (drop 1 l)" using k_def(2) by force
  ultimately have "iter_count C (drop k l) \<le> n-1" using iter_count_mono by metis

  then have "((C, s) \<rightarrow>* (Skip, snd (l!k))) \<and> (is_path (drop k l) (While C, snd (l!k)) (len-k) (Skip, s'')) \<and> iter_count C (drop k l) \<le> n-1" by (metis "0" "1" k_def(3) l''_def prod.collapse)
  then show ?thesis using k_def(1) by blast
qed

lemma every_while_is_some_iter_path:
  assumes "is_path l (While C, s) len (Skip, s'')"
      and "iter_count C l = Suc n"
    shows "\<exists>N. ((iter C N, s) \<rightarrow>* (Skip, s''))"
  using assms
proof(induction n arbitrary: l s len rule: less_induct)
  case (less n)
  then show ?case
  proof(cases "n=0")
    case True
    then show ?thesis using assms iter_count_1 RedRefl less.prems by (metis iter.simps(1) One_nat_def)
  next
    case False
    
    then obtain k where k_def: "k<len" "(C, s) \<rightarrow>* (Skip, snd (l!k))"
                               "is_path (drop k l) (While C, snd (l!k)) (len-k) (Skip, s'')"
                               "iter_count C (drop k l) \<le> n"
      using while_to_seq_and_while by (metis Suc_1 diff_Suc_1 diff_is_0_eq' less.prems(1) less.prems(2) not_less_eq_eq)
    moreover have "iter_count C (drop k l) \<le> n \<longleftrightarrow> iter_count C (drop k l) - 1 < n" using False by linarith
    have 0: "length l>0 \<Longrightarrow> fst (l!0) = While C \<Longrightarrow> iter_count C l > 0" using iter_count.elims by force
    have 1: "length (drop k l) > 0" using k_def(3) is_path_def a_path_def by (metis Suc_eq_plus1 Zero_not_Suc bot_nat_0.not_eq_extremum)
    have 2: "fst ((drop k l)!0) = While C" using k_def(3) is_path_def by force
    moreover have "iter_count C (drop k l)>0" using "0" "1" "2" by (metis (no_types, lifting) iter_count.elims length_greater_0_conv less_Suc_eq_0_disj nth_Cons_0)
    then have "iter_count C (drop k l) = Suc (iter_count C (drop k l) - 1)" by force
    ultimately obtain N where "(iter C N, snd (l ! k)) \<rightarrow>* (Skip, s'')" using less.IH[of "iter_count C (drop k l) - 1" "drop k l" "snd (l!k)" "len-k"] by fastforce
    then show ?thesis using k_def(2) iter_combine2 by blast
  qed
qed

theorem while_iff_union_iter:
  "((While C, s) \<rightarrow>* (Skip, s')) \<longleftrightarrow> (\<exists>n. ((iter C n, s) \<rightarrow>* (Skip, s')))"
proof
  show "(While C, s) \<rightarrow>* (Skip, s') \<Longrightarrow> \<exists>n. ((iter C n, s) \<rightarrow>* (Skip, s'))"
  proof(-)
    assume "(While C, s) \<rightarrow>* (Skip, s')"
    then obtain len where "red_iter (While C, s) len (Skip, s')" using red_iter_closed_def rtred_eq_red_iter_closed_fwd by blast
    then obtain l where "is_path l (While C, s) len (Skip, s')" using red_iter's_path by blast
    then have "fst (l!0) = While C" by (simp add: is_path_def)
    moreover have "length l>0"
    proof(rule ccontr)
      assume "\<not> (length l>0)"
      then have "length l=0" by fast
      then have "(While C, s) = (Skip, s')" by (metis Suc_eq_plus1 Zero_not_Suc \<open>is_path l (While C, s) len (Skip, s')\<close> a_path_def is_path_def)
      then show "False" by simp
    qed
    ultimately have "iter_count C l \<ge> 1" by (metis One_nat_def Suc_leI iter_count.elims length_greater_0_conv less_Suc_eq_0_disj nth_Cons_0)
    then show "\<exists>n. ((iter C n, s) \<rightarrow>* (Skip, s'))" by (metis One_nat_def \<open>is_path l (While C, s) len (Skip, s')\<close> every_while_is_some_iter_path gr0_implies_Suc linorder_not_less not_less_eq_eq)
  qed
  show "\<exists>n. ((iter C n, s) \<rightarrow>* (Skip, s')) \<Longrightarrow> (While C, s) \<rightarrow>* (Skip, s')" using every_iter_is_while by auto
qed

corollary dsem_while:
  "(\<Union>n. \<lbrakk>iter C n\<rbrakk>) = \<lbrakk>While C\<rbrakk>" 
proof(-)
  have "(\<Union>n. \<lbrakk>iter C n\<rbrakk>) = (\<Union>n. {(s, s')| s s'. (iter C n, s) \<rightarrow>* (Skip, s')})" using dsem_def by fast
  then have "(\<Union>n. \<lbrakk>iter C n\<rbrakk>) = {(s, s')| s s'. \<exists>n. ((iter C n, s) \<rightarrow>* (Skip, s'))}" by blast
  then have "(\<Union>n. \<lbrakk>iter C n\<rbrakk>) = {(s, s')| s s'. (While C, s) \<rightarrow>* (Skip, s')}" using while_iff_union_iter by force
  then show ?thesis using dsem_def by auto
qed

lemma dsem_iter:
  "\<lbrakk>iter C (Suc n)\<rbrakk> = {(\<sigma>, \<sigma>'')| \<sigma> \<sigma>' \<sigma>''. (\<sigma>, \<sigma>')\<in>\<lbrakk>iter C n\<rbrakk> \<and> (\<sigma>', \<sigma>'')\<in>\<lbrakk>C\<rbrakk>}"
  by (smt (verit, best) Collect_cong add_Seq_to_iterR dsem_seq)

lemma dsem_iter':
  "\<lbrakk>iter C (Suc n)\<rbrakk> = {(\<sigma>, \<sigma>'')| \<sigma> \<sigma>' \<sigma>''. (\<sigma>, \<sigma>')\<in>\<lbrakk>C\<rbrakk> \<and> (\<sigma>', \<sigma>'')\<in>\<lbrakk>iter C n\<rbrakk>}"
  by (smt (verit, best) Collect_cong add_Seq_to_iterL dsem_seq)

lemma "{s'. \<exists>s\<in>p. (s, s')\<in>\<lbrakk>Skip\<rbrakk>} = p" using dsem_skip by auto

lemma assign_uniqe:
  assumes "(Assign x e, s) \<rightarrow>* (Skip, s')"
  shows "s'=s(x:=e s)"
  using assms by (smt (verit) RedAssign cmd.distinct(1) cmd.distinct(3) fst_conv red.simps red_Assign_elim red_While_elim rtred.simps skip_deadend)

lemma dsem_assign:
  "\<lbrakk>Assign x e\<rbrakk> = {(s, s(x:=e s))| s. True}" 
proof(-)
  have "\<lbrakk>Assign x e\<rbrakk> = {(s, s')| s s'. (Assign x e, s) \<rightarrow>* (Skip, s')}" using dsem_def by blast
  moreover have "\<And>x e s s'. ((Assign x e, s) \<rightarrow>* (Skip, s')) \<longleftrightarrow> s'=s(x:=e s)" by (meson RedAssign RedRefl RedTran assign_uniqe)
  ultimately show ?thesis by fast
qed

lemma havoc_unique:
  assumes "(Havoc x, s) \<rightarrow>* (Skip, s')"
  shows "\<exists>n. s'=s(x:=n)"
  using assms by (smt (verit) Pair_inject cmd.distinct(17) cmd.distinct(25) cmd.distinct(31) cmd.distinct(37) cmd.distinct(39) cmd.distinct(7) cmd.inject(4) fst_conv red.simps rtred.simps skip_deadend)

lemma dsem_havoc:
  "\<lbrakk>Havoc x\<rbrakk> = {(s, s(x:=n))| s n. True}"
proof(-)
  have "\<lbrakk>Havoc x\<rbrakk> = {(s, s')| s s'. (Havoc x, s) \<rightarrow>* (Skip, s')}" using dsem_def by blast
  moreover have "\<And>x s s'. ((Havoc x, s) \<rightarrow>* (Skip, s')) \<longleftrightarrow> (\<exists>n. s'=s(x:=n))" by (meson RedHavoc havoc_unique rtred.simps)
  ultimately show ?thesis by fast
qed

theorem dsem_properties:
  "\<lbrakk>Skip\<rbrakk> = {(s, s)| s. True}"
  "\<lbrakk>Assign x e\<rbrakk> = {(s, s(x:=e s))| s. True}"
  "\<lbrakk>Havoc x\<rbrakk> = {(s, s(x:=n))| s n. True}"
  "\<lbrakk>Assume b\<rbrakk> = {(s, s)| s. s\<in>b}"
  "\<lbrakk>Seq C\<^sub>1 C\<^sub>2\<rbrakk> = \<lbrakk>C\<^sub>1\<rbrakk> O \<lbrakk>C\<^sub>2\<rbrakk>"
  "\<lbrakk>If C\<^sub>1 C\<^sub>2\<rbrakk> = \<lbrakk>C\<^sub>1\<rbrakk> \<union> \<lbrakk>C\<^sub>2\<rbrakk>"
  "\<lbrakk>While C\<rbrakk> = (\<Union>n. \<lbrakk>iter C n\<rbrakk>)"
  unfolding dsem_skip dsem_assign dsem_havoc dsem_assume dsem_seq dsem_if dsem_while by auto

fun md :: "('var, 'val) cmd \<Rightarrow> 'var set" where
 "md Skip = {}"
|"md (Assign x e) = {x}"
|"md (Havoc x) = {x}"
|"md (Assume b) = {}"
|"md (Seq C\<^sub>1 C\<^sub>2) = md C\<^sub>1 \<union> md C\<^sub>2"
|"md (If C\<^sub>1 C\<^sub>2) = md C\<^sub>1 \<union> md C\<^sub>2"
|"md (While C) = md C"

end