theory HeaplessSyntax
  imports HeaplessProgramCommands
begin

datatype 'var AExp =
 AZero
|AOne
|AVar 'var
|APlus "'var AExp" "'var AExp"
|ATimes "'var AExp" "'var AExp"

datatype 'var SynSAsrt =
 STop
|SBot
|SLess "'var AExp" "'var AExp"
|SEqual "'var AExp" "'var AExp"
|SAnd "'var SynSAsrt" "'var SynSAsrt"
|SOr "'var SynSAsrt" "'var SynSAsrt"
|SImpl "'var SynSAsrt" "'var SynSAsrt"
|SEquiv "'var SynSAsrt" "'var SynSAsrt"
|SNot "'var SynSAsrt"
|SEx 'var "'var SynSAsrt"
|SFa 'var "'var SynSAsrt"

datatype 'var SynSRAsrt =
 DAsrt "'var SynSAsrt" "'var SynSAsrt"
|RAnd "'var SynSRAsrt" "'var SynSRAsrt"
|ROr "'var SynSRAsrt" "'var SynSRAsrt"
|RImpl "'var SynSRAsrt" "'var SynSRAsrt"
|RBimp "'var SynSRAsrt" "'var SynSRAsrt"
|RNeg "'var SynSRAsrt"
|REx 'var "'var SynSRAsrt"
|RFa 'var "'var SynSRAsrt"

datatype ('svar, 'pvar) SynSHAsrt = 
 Top
|Bot
|Asrt 'svar "'pvar SynSAsrt"
|RAsrt 'svar 'svar "'pvar SynSRAsrt"
|ExN 'pvar "('svar, 'pvar) SynSHAsrt"
|FaN 'pvar "('svar, 'pvar) SynSHAsrt"
|ExS 'svar "('svar, 'pvar) SynSHAsrt"
|FaS 'svar "('svar, 'pvar) SynSHAsrt"
|And "('svar, 'pvar) SynSHAsrt" "('svar, 'pvar) SynSHAsrt"
|Or "('svar, 'pvar) SynSHAsrt" "('svar, 'pvar) SynSHAsrt"

fun AExp_fv :: "'var AExp \<Rightarrow> 'var set" where
 "AExp_fv AZero = {}"
|"AExp_fv AOne = {}"
|"AExp_fv (AVar x) = {x}"
|"AExp_fv (APlus e\<^sub>1 e\<^sub>2) = AExp_fv e\<^sub>1 \<union> AExp_fv e\<^sub>2"
|"AExp_fv (ATimes e\<^sub>1 e\<^sub>2) = AExp_fv e\<^sub>1 \<union> AExp_fv e\<^sub>2"

fun AExp_eval :: "'var AExp \<Rightarrow> ('var \<Rightarrow> nat) \<Rightarrow> nat" ("\<lbrakk>_\<rbrakk> [_]" [0,0] 81) where
 "\<lbrakk>AZero\<rbrakk>[s] = 0"
|"\<lbrakk>AOne\<rbrakk>[s] = 1"
|"\<lbrakk>AVar x\<rbrakk>[s] = s x"
|"\<lbrakk>APlus e\<^sub>1 e\<^sub>2\<rbrakk>[s] = \<lbrakk>e\<^sub>1\<rbrakk>[s]+\<lbrakk>e\<^sub>2\<rbrakk>[s]"
|"\<lbrakk>ATimes e\<^sub>1 e\<^sub>2\<rbrakk>[s] = \<lbrakk>e\<^sub>1\<rbrakk>[s]*\<lbrakk>e\<^sub>2\<rbrakk>[s]"

lemma AExp_fv_invariant:
  assumes "x\<notin>AExp_fv e"
    shows "\<lbrakk>e\<rbrakk>[s] = \<lbrakk>e\<rbrakk>[s(x:=n)]"
  using assms apply(induction e) by auto

fun AExp_sub :: "'var AExp \<Rightarrow> 'var AExp \<Rightarrow> 'var \<Rightarrow> 'var AExp" where
 "AExp_sub AZero e x = AZero"
|"AExp_sub AOne e x = AOne"
|"AExp_sub (AVar y) e x = (if x=y then e else (AVar y))"
|"AExp_sub (APlus e\<^sub>1 e\<^sub>2) e x = APlus (AExp_sub e\<^sub>1 e x) (AExp_sub e\<^sub>2 e x)"
|"AExp_sub (ATimes e\<^sub>1 e\<^sub>2) e x = ATimes (AExp_sub e\<^sub>1 e x) (AExp_sub e\<^sub>2 e x)"

lemma AExp_sub_prop:
  "\<lbrakk>AExp_sub E e x\<rbrakk>[s] = \<lbrakk>E\<rbrakk>[s(x:=\<lbrakk>e\<rbrakk>[s])]"
  apply(induction E)
  by auto

lemma AExp_sub_var:
  "\<lbrakk>AExp_sub E (AVar y) x\<rbrakk>[s] = \<lbrakk>E\<rbrakk>[s(x:=s y)]"
  apply(induction E) by auto

fun SynSAsrt_fv :: "'var SynSAsrt \<Rightarrow> 'var set" where
 "SynSAsrt_fv STop = {}"
|"SynSAsrt_fv SBot = {}"
|"SynSAsrt_fv (SLess e\<^sub>1 e\<^sub>2) = AExp_fv e\<^sub>1 \<union> AExp_fv e\<^sub>2"
|"SynSAsrt_fv (SEqual e\<^sub>1 e\<^sub>2) = AExp_fv e\<^sub>1 \<union> AExp_fv e\<^sub>2"
|"SynSAsrt_fv (SAnd p\<^sub>1 p\<^sub>2) = SynSAsrt_fv p\<^sub>1 \<union> SynSAsrt_fv p\<^sub>2"
|"SynSAsrt_fv (SOr p\<^sub>1 p\<^sub>2) = SynSAsrt_fv p\<^sub>1 \<union> SynSAsrt_fv p\<^sub>2"
|"SynSAsrt_fv (SImpl p\<^sub>1 p\<^sub>2) = SynSAsrt_fv p\<^sub>1 \<union> SynSAsrt_fv p\<^sub>2"
|"SynSAsrt_fv (SEquiv p\<^sub>1 p\<^sub>2) = SynSAsrt_fv p\<^sub>1 \<union> SynSAsrt_fv p\<^sub>2"
|"SynSAsrt_fv (SNot p) = SynSAsrt_fv p"
|"SynSAsrt_fv (SEx y p) = (SynSAsrt_fv p)-{y}"
|"SynSAsrt_fv (SFa y p) = (SynSAsrt_fv p)-{y}"

fun SynSAsrt_bd :: "'var SynSAsrt \<Rightarrow> 'var set" where
 "SynSAsrt_bd STop = {}"
|"SynSAsrt_bd SBot = {}"
|"SynSAsrt_bd (SLess e\<^sub>1 e\<^sub>2) = {}"
|"SynSAsrt_bd (SEqual e\<^sub>1 e\<^sub>2) = {}"
|"SynSAsrt_bd (SAnd p\<^sub>1 p\<^sub>2) = SynSAsrt_bd p\<^sub>1 \<union> SynSAsrt_bd p\<^sub>2"
|"SynSAsrt_bd (SOr p\<^sub>1 p\<^sub>2) = SynSAsrt_bd p\<^sub>1 \<union> SynSAsrt_bd p\<^sub>2"
|"SynSAsrt_bd (SImpl p\<^sub>1 p\<^sub>2) = SynSAsrt_bd p\<^sub>1 \<union> SynSAsrt_bd p\<^sub>2"
|"SynSAsrt_bd (SEquiv p\<^sub>1 p\<^sub>2) = SynSAsrt_bd p\<^sub>1 \<union> SynSAsrt_bd p\<^sub>2"
|"SynSAsrt_bd (SNot p) = SynSAsrt_bd p"
|"SynSAsrt_bd (SEx y p) = {y} \<union> SynSAsrt_bd p"
|"SynSAsrt_bd (SFa y p) = {y} \<union> SynSAsrt_bd p"

fun SynSAsrt_sc :: "'var SynSAsrt \<Rightarrow> 'var \<Rightarrow> 'var set" where
 "SynSAsrt_sc STop x = {}"
|"SynSAsrt_sc SBot x = {}"
|"SynSAsrt_sc (SLess e\<^sub>1 e\<^sub>2) x = {}"
|"SynSAsrt_sc (SEqual e\<^sub>1 e\<^sub>2) x = {}"
|"SynSAsrt_sc (SAnd p\<^sub>1 p\<^sub>2) x = SynSAsrt_sc p\<^sub>1 x \<union> SynSAsrt_sc p\<^sub>2 x"
|"SynSAsrt_sc (SOr p\<^sub>1 p\<^sub>2) x = SynSAsrt_sc p\<^sub>1 x \<union> SynSAsrt_sc p\<^sub>2 x"
|"SynSAsrt_sc (SImpl p\<^sub>1 p\<^sub>2) x = SynSAsrt_sc p\<^sub>1 x \<union> SynSAsrt_sc p\<^sub>2 x"
|"SynSAsrt_sc (SEquiv p\<^sub>1 p\<^sub>2) x = SynSAsrt_sc p\<^sub>1 x \<union> SynSAsrt_sc p\<^sub>2 x"
|"SynSAsrt_sc (SNot p) x = SynSAsrt_sc p x"
|"SynSAsrt_sc (SEx y p) x = (if x\<in>SynSAsrt_fv p then {y}\<union>SynSAsrt_sc p x else SynSAsrt_sc p x)"
|"SynSAsrt_sc (SFa y p) x = (if x\<in>SynSAsrt_fv p then {y}\<union>SynSAsrt_sc p x else SynSAsrt_sc p x)"

lemma Scope_subset_of_bounded:
  "SynSAsrt_sc p x \<subseteq> SynSAsrt_bd p"
  apply(induction p) by auto

(*The subscript stands for assertion*)
fun SynSAsrt_eval :: "'var SynSAsrt \<Rightarrow> ('var \<Rightarrow> nat) set" ("\<lbrakk>_\<rbrakk>\<^sub>a" [0] 81) where
 "\<lbrakk>STop\<rbrakk>\<^sub>a = UNIV"
|"\<lbrakk>SBot\<rbrakk>\<^sub>a = {}"
|"\<lbrakk>SLess e\<^sub>1 e\<^sub>2\<rbrakk>\<^sub>a = {s. \<lbrakk>e\<^sub>1\<rbrakk>[s]<\<lbrakk>e\<^sub>2\<rbrakk>[s]}"
|"\<lbrakk>SEqual e\<^sub>1 e\<^sub>2\<rbrakk>\<^sub>a = {s. \<lbrakk>e\<^sub>1\<rbrakk>[s]=\<lbrakk>e\<^sub>2\<rbrakk>[s]}"
|"\<lbrakk>SAnd p\<^sub>1 p\<^sub>2\<rbrakk>\<^sub>a = \<lbrakk>p\<^sub>1\<rbrakk>\<^sub>a \<inter> \<lbrakk>p\<^sub>2\<rbrakk>\<^sub>a"
|"\<lbrakk>SOr p\<^sub>1 p\<^sub>2\<rbrakk>\<^sub>a = \<lbrakk>p\<^sub>1\<rbrakk>\<^sub>a \<union> \<lbrakk>p\<^sub>2\<rbrakk>\<^sub>a"
|"\<lbrakk>SImpl p\<^sub>1 p\<^sub>2\<rbrakk>\<^sub>a = (UNIV-\<lbrakk>p\<^sub>1\<rbrakk>\<^sub>a) \<union> \<lbrakk>p\<^sub>2\<rbrakk>\<^sub>a"
|"\<lbrakk>SEquiv p\<^sub>1 p\<^sub>2\<rbrakk>\<^sub>a = (\<lbrakk>p\<^sub>1\<rbrakk>\<^sub>a \<inter> \<lbrakk>p\<^sub>2\<rbrakk>\<^sub>a) \<union> ((UNIV-\<lbrakk>p\<^sub>1\<rbrakk>\<^sub>a) \<inter> (UNIV-\<lbrakk>p\<^sub>2\<rbrakk>\<^sub>a))"
|"\<lbrakk>SNot p\<rbrakk>\<^sub>a = UNIV-\<lbrakk>p\<rbrakk>\<^sub>a"
|"\<lbrakk>SEx x p\<rbrakk>\<^sub>a = {s. \<exists>n. s(x:=n)\<in>\<lbrakk>p\<rbrakk>\<^sub>a}"
|"\<lbrakk>SFa x p\<rbrakk>\<^sub>a = {s. \<forall>n. s(x:=n)\<in>\<lbrakk>p\<rbrakk>\<^sub>a}"

fun SynSAsrt_sub :: "'var SynSAsrt \<Rightarrow> 'var AExp \<Rightarrow> 'var \<Rightarrow> 'var SynSAsrt" where
 "SynSAsrt_sub STop e x = STop"
|"SynSAsrt_sub SBot e x = SBot"
|"SynSAsrt_sub (SLess e\<^sub>1 e\<^sub>2) e x = SLess (AExp_sub e\<^sub>1 e x) (AExp_sub e\<^sub>2 e x)"
|"SynSAsrt_sub (SEqual e\<^sub>1 e\<^sub>2) e x = SEqual (AExp_sub e\<^sub>1 e x) (AExp_sub e\<^sub>2 e x)"
|"SynSAsrt_sub (SAnd p\<^sub>1 p\<^sub>2) e x = SAnd (SynSAsrt_sub p\<^sub>1 e x) (SynSAsrt_sub p\<^sub>2 e x)"
|"SynSAsrt_sub (SOr p\<^sub>1 p\<^sub>2) e x = SOr (SynSAsrt_sub p\<^sub>1 e x) (SynSAsrt_sub p\<^sub>2 e x)"
|"SynSAsrt_sub (SImpl p\<^sub>1 p\<^sub>2) e x = SImpl (SynSAsrt_sub p\<^sub>1 e x) (SynSAsrt_sub p\<^sub>2 e x)"
|"SynSAsrt_sub (SEquiv p\<^sub>1 p\<^sub>2) e x = SEquiv (SynSAsrt_sub p\<^sub>1 e x) (SynSAsrt_sub p\<^sub>2 e x)"
|"SynSAsrt_sub (SNot p) e x = SNot (SynSAsrt_sub p e x)"
|"SynSAsrt_sub (SEx y p) e x = SEx y (if y=x then p else SynSAsrt_sub p e x)"
|"SynSAsrt_sub (SFa y p) e x = SFa y (if y=x then p else SynSAsrt_sub p e x)"

lemma SEx_unfold: "\<lbrakk>SEx y p\<rbrakk>\<^sub>a = {s(y:=n)| s n. s\<in>\<lbrakk>p\<rbrakk>\<^sub>a}"
proof(-)
  obtain bb :: "'a \<Rightarrow> 'a SynSAsrt \<Rightarrow> ('a \<Rightarrow> nat) \<Rightarrow> bool" where
    f1: "\<forall>X0 X1 E_x. bb X0 X1 E_x = (\<exists>Y0. E_x(X0 := Y0) \<in> \<lbrakk>X1\<rbrakk>\<^sub>a)"
    by moura
  then have f2: "\<forall>a b. \<lbrakk>SEx a b\<rbrakk>\<^sub>a = Collect (bb a b)"
    by fastforce
  obtain bba :: "('a \<Rightarrow> nat) \<Rightarrow> bool" where
    f3: "\<forall>X1. bba X1 = (\<exists>Y0 Y1. X1 = Y0(y := Y1) \<and> Y0 \<in> \<lbrakk>p\<rbrakk>\<^sub>a)"
    by moura
  then have "\<lbrakk>SEx y p\<rbrakk>\<^sub>a = Collect bba"
    using f2 f1 by (metis (no_types) Collect_cong fun_upd_triv fun_upd_upd)
  then show ?thesis using f3 by presburger
qed

lemma SynSAsrt_fv_invariant:
  assumes "x\<notin>SynSAsrt_fv p"
    shows "s\<in>\<lbrakk>p\<rbrakk>\<^sub>a \<longleftrightarrow> s(x:=n)\<in>\<lbrakk>p\<rbrakk>\<^sub>a"
  using assms 
proof(induction p arbitrary: s)
  case STop
  then show ?case by simp
next
  case SBot
  then show ?case by auto
next
  case (SLess e\<^sub>1 e\<^sub>2)
  then have "x\<notin>AExp_fv e\<^sub>1" "x\<notin>AExp_fv e\<^sub>2" by auto 
  then show ?case using AExp_fv_invariant by (metis (no_types, lifting) SynSAsrt_eval.simps(3) mem_Collect_eq)
next
  case (SEqual e\<^sub>1 e\<^sub>2)
  then have "x\<notin>AExp_fv e\<^sub>1" "x\<notin>AExp_fv e\<^sub>2" by auto 
  then show ?case using AExp_fv_invariant by (metis (no_types, lifting) SynSAsrt_eval.simps(4) mem_Collect_eq)
next
  case (SAnd p\<^sub>1 p\<^sub>2)
  then show ?case by force
next
  case (SOr p\<^sub>1 p\<^sub>2)
  then show ?case by force
next
  case (SImpl p\<^sub>1 p\<^sub>2)
  then show ?case by force
next
  case (SEquiv p\<^sub>1 p\<^sub>2)
  then show ?case by simp
next
  case (SNot p) 
  then show ?case by force
next
  case (SEx y p)
  then show ?case
  proof(cases "y=x")
    case True
    then show ?thesis by (smt (verit, best) SynSAsrt_eval.simps(10) fun_upd_upd mem_Collect_eq)
  next
    case False
    show ?thesis
    proof
      show "s\<in>\<lbrakk>SEx y p\<rbrakk>\<^sub>a \<Longrightarrow> s(x:=n)\<in>\<lbrakk>SEx y p\<rbrakk>\<^sub>a"
      proof(-)
        fix s
        assume "s\<in>\<lbrakk>SEx y p\<rbrakk>\<^sub>a"
        then obtain n\<^sub>y where "s(y:=n\<^sub>y)\<in>\<lbrakk>p\<rbrakk>\<^sub>a" by auto
        then have "(s(y:=n\<^sub>y))(x:=n)\<in>\<lbrakk>p\<rbrakk>\<^sub>a" by (metis SEx.IH SEx.prems SynSAsrt_fv.simps(10) DiffI False emptyE insertE)
        then have "(s(x:=n))(y:=n\<^sub>y)\<in>\<lbrakk>p\<rbrakk>\<^sub>a" using False by (simp add: fun_upd_twist)
        then show "s(x:=n)\<in>\<lbrakk>SEx y p\<rbrakk>\<^sub>a" by fastforce
      qed
      show "s(x:=n)\<in>\<lbrakk>SEx y p\<rbrakk>\<^sub>a \<Longrightarrow> s\<in>\<lbrakk>SEx y p\<rbrakk>\<^sub>a"
      proof(-)
        fix s
        assume "s(x:=n)\<in>\<lbrakk>SEx y p\<rbrakk>\<^sub>a"
        then obtain n\<^sub>y where "(s(x:=n))(y:=n\<^sub>y)\<in>\<lbrakk>p\<rbrakk>\<^sub>a" by fastforce
        then have "(s(y:=n\<^sub>y))(x:=n)\<in>\<lbrakk>p\<rbrakk>\<^sub>a" using False by (simp add: fun_upd_twist)
        then have "s(y:=n\<^sub>y)\<in>\<lbrakk>p\<rbrakk>\<^sub>a" by (metis SEx.IH SEx.prems SynSAsrt_fv.simps(10) DiffI False emptyE insertE)
        then show "s\<in>\<lbrakk>SEx y p\<rbrakk>\<^sub>a" by auto
      qed
    qed
  qed
next
  case (SFa y p)
  then show ?case
  proof(cases "y=x")
    case True
    then show ?thesis by (smt (verit) SynSAsrt_eval.simps(11) fun_upd_upd mem_Collect_eq)
  next
    case False
    show ?thesis
    proof
      show "s\<in>\<lbrakk>SFa y p\<rbrakk>\<^sub>a \<Longrightarrow> s(x:=n)\<in>\<lbrakk>SFa y p\<rbrakk>\<^sub>a"
      proof(-)
        fix s
        assume "s\<in>\<lbrakk>SFa y p\<rbrakk>\<^sub>a"
        then have "\<forall>n\<^sub>y. s(y:=n\<^sub>y)\<in>\<lbrakk>p\<rbrakk>\<^sub>a" by auto
        then have "\<forall>n\<^sub>y. (s(y:=n\<^sub>y))(x:=n)\<in>\<lbrakk>p\<rbrakk>\<^sub>a" by (metis SFa.IH SFa.prems SynSAsrt_fv.simps(11) DiffI False emptyE insertE)
        then have "\<forall>n\<^sub>y. (s(x:=n))(y:=n\<^sub>y)\<in>\<lbrakk>p\<rbrakk>\<^sub>a" using False by (simp add: fun_upd_twist)
        then show "s(x:=n)\<in>\<lbrakk>SFa y p\<rbrakk>\<^sub>a" by fastforce
      qed
      show "s(x:=n)\<in>\<lbrakk>SFa y p\<rbrakk>\<^sub>a \<Longrightarrow> s\<in>\<lbrakk>SFa y p\<rbrakk>\<^sub>a"
      proof(-)
        fix s
        assume "s(x:=n)\<in>\<lbrakk>SFa y p\<rbrakk>\<^sub>a"
        then have "\<forall>n\<^sub>y. (s(x:=n))(y:=n\<^sub>y)\<in>\<lbrakk>p\<rbrakk>\<^sub>a" by fastforce
        then have "\<forall>n\<^sub>y. (s(y:=n\<^sub>y))(x:=n)\<in>\<lbrakk>p\<rbrakk>\<^sub>a" using False by (simp add: fun_upd_twist)
        then have "\<forall>n\<^sub>y. s(y:=n\<^sub>y)\<in>\<lbrakk>p\<rbrakk>\<^sub>a" by (metis SFa.IH SFa.prems SynSAsrt_fv.simps(11) DiffI False emptyE insertE)
        then show "s\<in>\<lbrakk>SFa y p\<rbrakk>\<^sub>a" by auto
      qed
    qed
  qed
qed

lemma Hoare_assign_axiom_semantically:
  assumes "AExp_fv e \<inter> SynSAsrt_sc q x = {}"
  shows "\<lbrakk>SynSAsrt_sub q e x\<rbrakk>\<^sub>a = {s. s(x:=\<lbrakk>e\<rbrakk>[s])\<in>\<lbrakk>q\<rbrakk>\<^sub>a}"
  using assms
proof(induction q)
  case STop
  then show ?case by fastforce
next
  case SBot
  then show ?case by simp
next
  case (SLess e\<^sub>1 e\<^sub>2)
  then show ?case by (smt (verit) AExp_sub_prop SynSAsrt_eval.simps(3) SynSAsrt_sub.simps(3) Collect_cong Orderings.order_eq_iff mem_Collect_eq)
next
  case (SEqual e\<^sub>1 e\<^sub>2)
  then show ?case by (smt (verit) AExp_sub_prop SynSAsrt_eval.simps(4) SynSAsrt_sub.simps(4) Collect_cong mem_Collect_eq)
next
  case (SAnd q\<^sub>1 q\<^sub>2)
  then show ?case by auto
next
  case (SOr q\<^sub>1 q\<^sub>2)
  then show ?case by auto
next
  case (SImpl q\<^sub>1 q\<^sub>2)
  then show ?case by auto
next
  case (SEquiv q\<^sub>1 q\<^sub>2)
  then show ?case by auto
next
  case (SNot q)
  then show ?case by auto
next
  case (SEx y q)
  then show ?case
  proof(cases "y=x")
    case True
    then have "\<lbrakk>SynSAsrt_sub (SEx y q) e x\<rbrakk>\<^sub>a = \<lbrakk>SEx y q\<rbrakk>\<^sub>a" by force
    moreover from True have "\<lbrakk>SEx y q\<rbrakk>\<^sub>a = {s. s(x:=\<lbrakk>e\<rbrakk>[s])\<in>\<lbrakk>SEx y q\<rbrakk>\<^sub>a}" by auto
    ultimately show ?thesis by blast
  next
    case False
    show ?thesis (is "?X=?Y") 
    proof
      show "?X\<subseteq>?Y"
      proof
        fix s
        assume "s\<in>?X"
        then obtain n\<^sub>y where "s(y:=n\<^sub>y)\<in>\<lbrakk>SynSAsrt_sub q e x\<rbrakk>\<^sub>a" using False by auto
        then have "(s(y:=n\<^sub>y))(x:=\<lbrakk>e\<rbrakk>[s(y:=n\<^sub>y)])\<in>\<lbrakk>q\<rbrakk>\<^sub>a" using SEx.IH SEx.prems by fastforce
        then have "(s(x:=\<lbrakk>e\<rbrakk>[s(y:=n\<^sub>y)]))(y:=n\<^sub>y)\<in>\<lbrakk>q\<rbrakk>\<^sub>a" using False by (simp add: fun_upd_twist)
        then have "s(x:=\<lbrakk>e\<rbrakk>[s(y:=n\<^sub>y)])\<in>\<lbrakk>SEx y q\<rbrakk>\<^sub>a" by auto
        then have "y\<notin>AExp_fv e \<Longrightarrow> s(x:=\<lbrakk>e\<rbrakk>[s])\<in>\<lbrakk>SEx y q\<rbrakk>\<^sub>a" using AExp_fv_invariant by metis
        moreover have "x\<notin>SynSAsrt_fv q \<Longrightarrow> s(x:=\<lbrakk>e\<rbrakk>[s])\<in>\<lbrakk>SEx y q\<rbrakk>\<^sub>a" by (metis DiffE SynSAsrt_fv.simps(10) SynSAsrt_fv_invariant \<open>s(x := \<lbrakk>e\<rbrakk> [s(y := n\<^sub>y)]) \<in> \<lbrakk>SEx y q\<rbrakk>\<^sub>a\<close>)
        moreover have "y\<notin>AExp_fv e \<or> x\<notin>SynSAsrt_fv q" using SEx.prems by fastforce
        ultimately show "s\<in>?Y" by blast
      qed
      show "?Y\<subseteq>?X" 
      proof
        fix s
        assume "s\<in>?Y"
        then have "s(x:=\<lbrakk>e\<rbrakk>[s])\<in>\<lbrakk>SEx y q\<rbrakk>\<^sub>a" by blast
        then obtain n\<^sub>y where "(s(x:=\<lbrakk>e\<rbrakk>[s]))(y:=n\<^sub>y)\<in>\<lbrakk>q\<rbrakk>\<^sub>a" by fastforce
        then have "(s(y:=n\<^sub>y))(x:=\<lbrakk>e\<rbrakk>[s])\<in>\<lbrakk>q\<rbrakk>\<^sub>a" using False by (simp add: fun_upd_twist)
        then have "y\<notin>AExp_fv e \<Longrightarrow> (s(y:=n\<^sub>y))(x:=\<lbrakk>e\<rbrakk>[s(y:=n\<^sub>y)])\<in>\<lbrakk>q\<rbrakk>\<^sub>a" using AExp_fv_invariant by fastforce
        moreover have "x\<notin>SynSAsrt_fv q \<Longrightarrow> (s(y:=n\<^sub>y))(x:=\<lbrakk>e\<rbrakk>[s(y:=n\<^sub>y)])\<in>\<lbrakk>q\<rbrakk>\<^sub>a" using SynSAsrt_fv_invariant \<open>s(y := n\<^sub>y, x := \<lbrakk>e\<rbrakk> [s]) \<in> \<lbrakk>q\<rbrakk>\<^sub>a\<close> by fastforce
        moreover have "y\<notin>AExp_fv e \<or> x\<notin>SynSAsrt_fv q" using SEx.prems by auto
        ultimately have "(s(y:=n\<^sub>y))(x:=\<lbrakk>e\<rbrakk>[s(y:=n\<^sub>y)])\<in>\<lbrakk>q\<rbrakk>\<^sub>a" using AExp_fv_invariant by fastforce
        then have "s(y:=n\<^sub>y)\<in>\<lbrakk>SynSAsrt_sub q e x\<rbrakk>\<^sub>a" using SEx.IH SEx.prems by fastforce
        then show "s\<in>?X" using False by auto
      qed
    qed
  qed
next
  case (SFa y q)
  then show ?case
  proof(cases "y=x")
    case True
    then have "\<lbrakk>SynSAsrt_sub (SFa y q) e x\<rbrakk>\<^sub>a = \<lbrakk>SFa y q\<rbrakk>\<^sub>a" by force
    moreover from True have "\<lbrakk>SFa y q\<rbrakk>\<^sub>a = {s. s(x:=\<lbrakk>e\<rbrakk>[s])\<in>\<lbrakk>SFa y q\<rbrakk>\<^sub>a}" by auto
    ultimately show ?thesis by blast
  next
    case False
    show ?thesis (is "?X=?Y") 
    proof
      show "?X\<subseteq>?Y"
      proof
        fix s
        assume "s\<in>?X"
        then have "\<forall>n\<^sub>y. s(y:=n\<^sub>y)\<in>\<lbrakk>SynSAsrt_sub q e x\<rbrakk>\<^sub>a" using False by auto
        then have "\<forall>n\<^sub>y. (s(y:=n\<^sub>y))(x:=\<lbrakk>e\<rbrakk>[s(y:=n\<^sub>y)])\<in>\<lbrakk>q\<rbrakk>\<^sub>a" using SFa.IH SFa.prems by fastforce
        then have "\<forall>n\<^sub>y. (s(x:=\<lbrakk>e\<rbrakk>[s(y:=n\<^sub>y)]))(y:=n\<^sub>y)\<in>\<lbrakk>q\<rbrakk>\<^sub>a" using False by (simp add: fun_upd_twist)
        then have "\<forall>n\<^sub>y. y\<notin>AExp_fv e \<longrightarrow> s(x:=\<lbrakk>e\<rbrakk>[s(y:=n\<^sub>y)])\<in>\<lbrakk>SFa y q\<rbrakk>\<^sub>a" using AExp_fv_invariant SFa.prems by fastforce
        moreover have "\<forall>n\<^sub>y. x\<notin>SynSAsrt_fv q \<longrightarrow> s(x:=\<lbrakk>e\<rbrakk>[s(y:=n\<^sub>y)])\<in>\<lbrakk>SFa y q\<rbrakk>\<^sub>a" using SynSAsrt_fv_invariant \<open>\<forall>n\<^sub>y. s(y := n\<^sub>y, x := \<lbrakk>e\<rbrakk> [s(y := n\<^sub>y)]) \<in> \<lbrakk>q\<rbrakk>\<^sub>a\<close> by fastforce
        moreover have "y\<notin>AExp_fv e \<or> x\<notin>SynSAsrt_fv q" using SFa.prems by auto
        ultimately have "s(x:=\<lbrakk>e\<rbrakk>[s])\<in>\<lbrakk>SFa y q\<rbrakk>\<^sub>a" by (metis fun_upd_triv)
        then show "s\<in>?Y" by blast
      qed
      show "?Y\<subseteq>?X" 
      proof
        fix s
        assume "s\<in>?Y"
        then have "s(x:=\<lbrakk>e\<rbrakk>[s])\<in>\<lbrakk>SFa y q\<rbrakk>\<^sub>a" by blast
        then have "\<forall>n\<^sub>y. (s(x:=\<lbrakk>e\<rbrakk>[s]))(y:=n\<^sub>y)\<in>\<lbrakk>q\<rbrakk>\<^sub>a" by fastforce
        then have "\<forall>n\<^sub>y. y\<notin>AExp_fv e \<longrightarrow> (s(y:=n\<^sub>y))(x:=\<lbrakk>e\<rbrakk>[s])\<in>\<lbrakk>q\<rbrakk>\<^sub>a" using False by (simp add: fun_upd_twist)
        moreover have "\<forall>n\<^sub>y. x\<notin>SynSAsrt_fv q \<longrightarrow> (s(y:=n\<^sub>y))(x:=\<lbrakk>e\<rbrakk>[s])\<in>\<lbrakk>q\<rbrakk>\<^sub>a" by (metis False \<open>\<forall>n\<^sub>y. s(x := \<lbrakk>e\<rbrakk> [s], y := n\<^sub>y) \<in> \<lbrakk>q\<rbrakk>\<^sub>a\<close> fun_upd_twist)
        moreover have "y\<notin>AExp_fv e \<or> x\<notin>SynSAsrt_fv q" using SFa.prems by auto
        ultimately have "\<forall>n\<^sub>y. (s(y:=n\<^sub>y))(x:=\<lbrakk>e\<rbrakk>[s(y:=n\<^sub>y)])\<in>\<lbrakk>q\<rbrakk>\<^sub>a" by (metis AExp_fv_invariant SynSAsrt_fv_invariant)
        then have "\<forall>n\<^sub>y. s(y:=n\<^sub>y)\<in>\<lbrakk>SynSAsrt_sub q e x\<rbrakk>\<^sub>a" using SFa.IH SFa.prems by fastforce
        then show "s\<in>?X" using False by auto
      qed
    qed
  qed
qed

lemma Floyd_assign_axiom_semantically:
  assumes "x\<^sub>i\<^sub>n\<notin>SynSAsrt_sc p x \<union> SynSAsrt_fv p \<union> AExp_fv e \<union> {x}"
  shows "{s(x:=\<lbrakk>e\<rbrakk>[s])| s. s\<in>\<lbrakk>p\<rbrakk>\<^sub>a} = \<lbrakk>SEx x\<^sub>i\<^sub>n (SAnd (SynSAsrt_sub p (AVar x\<^sub>i\<^sub>n) x) (SEqual (AVar x) (AExp_sub e (AVar x\<^sub>i\<^sub>n) x)))\<rbrakk>\<^sub>a" (is "?X=?Y")
proof(-)
  have "?Y = {s(x\<^sub>i\<^sub>n:=n)| s n. s\<in>\<lbrakk>SAnd (SynSAsrt_sub p (AVar x\<^sub>i\<^sub>n) x) (SEqual (AVar x) (AExp_sub e (AVar x\<^sub>i\<^sub>n) x))\<rbrakk>\<^sub>a}" using SEx_unfold by fast
  then have "?Y = {s(x\<^sub>i\<^sub>n:=n)| s n. s\<in>\<lbrakk>SynSAsrt_sub p (AVar x\<^sub>i\<^sub>n) x\<rbrakk>\<^sub>a\<inter>\<lbrakk>SEqual (AVar x) (AExp_sub e (AVar x\<^sub>i\<^sub>n) x)\<rbrakk>\<^sub>a}" by auto
  moreover have "x\<^sub>i\<^sub>n \<notin> SynSAsrt_sc p x" using assms by auto
  then have "\<lbrakk>SynSAsrt_sub p (AVar x\<^sub>i\<^sub>n) x\<rbrakk>\<^sub>a = {s. s(x:=\<lbrakk>AVar x\<^sub>i\<^sub>n\<rbrakk>[s])\<in>\<lbrakk>p\<rbrakk>\<^sub>a}" using Hoare_assign_axiom_semantically by force
  ultimately have "?Y = {s(x\<^sub>i\<^sub>n:=n)| s n. s\<in>{s. s(x:=\<lbrakk>AVar x\<^sub>i\<^sub>n\<rbrakk>[s])\<in>\<lbrakk>p\<rbrakk>\<^sub>a}\<inter>\<lbrakk>SEqual (AVar x) (AExp_sub e (AVar x\<^sub>i\<^sub>n) x)\<rbrakk>\<^sub>a}" by presburger
  moreover have "\<lbrakk>SEqual (AVar x) (AExp_sub e (AVar x\<^sub>i\<^sub>n) x)\<rbrakk>\<^sub>a = {s. s x = \<lbrakk>AExp_sub e (AVar x\<^sub>i\<^sub>n) x\<rbrakk>[s]}" by simp
  then have "\<lbrakk>SEqual (AVar x) (AExp_sub e (AVar x\<^sub>i\<^sub>n) x)\<rbrakk>\<^sub>a = {s. s x = \<lbrakk>e\<rbrakk>[s(x:=s x\<^sub>i\<^sub>n)]}" using AExp_sub_var by fastforce
  ultimately have "?Y = {s(x\<^sub>i\<^sub>n:=n)| s n. s(x:=\<lbrakk>AVar x\<^sub>i\<^sub>n\<rbrakk>[s])\<in>\<lbrakk>p\<rbrakk>\<^sub>a \<and> s x = \<lbrakk>e\<rbrakk>[s(x:=s x\<^sub>i\<^sub>n)]}" by blast
  then have Y_equiv: "?Y = {s(x\<^sub>i\<^sub>n:=n)| s n. s(x:=s x\<^sub>i\<^sub>n)\<in>\<lbrakk>p\<rbrakk>\<^sub>a \<and> s x = \<lbrakk>e\<rbrakk>[s(x:=s x\<^sub>i\<^sub>n)]}" by auto
  moreover have X_equiv: "?X = {(s(x:=\<lbrakk>e\<rbrakk>[s]))(x\<^sub>i\<^sub>n:=n)| s n. s\<in>\<lbrakk>p\<rbrakk>\<^sub>a}"
  proof
    show "?X \<subseteq> {(s(x:=\<lbrakk>e\<rbrakk>[s]))(x\<^sub>i\<^sub>n:=n)| s n. s\<in>\<lbrakk>p\<rbrakk>\<^sub>a}" by (smt (verit, ccfv_SIG) fun_upd_triv mem_Collect_eq subset_eq)
    show "?X \<supseteq> {(s(x:=\<lbrakk>e\<rbrakk>[s]))(x\<^sub>i\<^sub>n:=n)| s n. s\<in>\<lbrakk>p\<rbrakk>\<^sub>a}"
    proof
      fix s
      assume asm: "s\<in>{(s(x:=\<lbrakk>e\<rbrakk>[s]))(x\<^sub>i\<^sub>n:=n)| s n. s\<in>\<lbrakk>p\<rbrakk>\<^sub>a}"
      then obtain n s' where ns'_def: "s=(s'(x:=\<lbrakk>e\<rbrakk>[s']))(x\<^sub>i\<^sub>n:=n)" "s'\<in>\<lbrakk>p\<rbrakk>\<^sub>a" by blast
      then have "s=(s'(x\<^sub>i\<^sub>n:=n))(x:=\<lbrakk>e\<rbrakk>[s'])" using assms by fastforce
      moreover have "s'(x\<^sub>i\<^sub>n:=n)\<in>\<lbrakk>p\<rbrakk>\<^sub>a" by (meson SynSAsrt_fv_invariant UnI1 UnI2 assms ns'_def(2))
      moreover have "\<lbrakk>e\<rbrakk>[s'] = \<lbrakk>e\<rbrakk>[s'(x\<^sub>i\<^sub>n:=n)]" by (meson AExp_fv_invariant UnCI assms)
      ultimately show "s\<in>?X" by auto
    qed
  qed
  moreover have "{(s(x:=\<lbrakk>e\<rbrakk>[s]))(x\<^sub>i\<^sub>n:=n)| s n. s\<in>\<lbrakk>p\<rbrakk>\<^sub>a} = {s(x\<^sub>i\<^sub>n:=n)| s n. s(x:=s x\<^sub>i\<^sub>n)\<in>\<lbrakk>p\<rbrakk>\<^sub>a \<and> s x = \<lbrakk>e\<rbrakk>[s(x:=s x\<^sub>i\<^sub>n)]}"
  proof
    show "{(s(x:=\<lbrakk>e\<rbrakk>[s]))(x\<^sub>i\<^sub>n:=n)| s n. s\<in>\<lbrakk>p\<rbrakk>\<^sub>a} \<subseteq> {s(x\<^sub>i\<^sub>n:=n)| s n. s(x:=s x\<^sub>i\<^sub>n)\<in>\<lbrakk>p\<rbrakk>\<^sub>a \<and> s x = \<lbrakk>e\<rbrakk>[s(x:=s x\<^sub>i\<^sub>n)]}"
    proof
      fix s
      assume asm: "s\<in>{(s(x:=\<lbrakk>e\<rbrakk>[s]))(x\<^sub>i\<^sub>n:=n)| s n. s\<in>\<lbrakk>p\<rbrakk>\<^sub>a}"
      then obtain s' n\<^sub>i\<^sub>n where s'n\<^sub>i\<^sub>n_def: "s=(s'(x:=\<lbrakk>e\<rbrakk>[s']))(x\<^sub>i\<^sub>n:=n\<^sub>i\<^sub>n)" "s'\<in>\<lbrakk>p\<rbrakk>\<^sub>a" by blast
      moreover obtain s'' where s''_def: "s''=s(x\<^sub>i\<^sub>n:=s' x)" by blast
      ultimately have "s''(x:=s'' x\<^sub>i\<^sub>n)\<in>\<lbrakk>p\<rbrakk>\<^sub>a"  by (smt (verit, ccfv_threshold) SynSAsrt_fv_invariant UnCI assms fun_upd_same fun_upd_triv fun_upd_twist fun_upd_upd)
      moreover have "s'' x = \<lbrakk>e\<rbrakk>[s''(x:=s'' x\<^sub>i\<^sub>n)]" by (smt (verit, best) AExp_fv_invariant UnI1 UnI2 s'n\<^sub>i\<^sub>n_def(1) s''_def assms fun_upd_same fun_upd_triv fun_upd_twist fun_upd_upd insert_iff)
      moreover have "s=s''(x\<^sub>i\<^sub>n:=n\<^sub>i\<^sub>n)" using s''_def s'n\<^sub>i\<^sub>n_def(1) by auto
      ultimately show "s\<in>{s(x\<^sub>i\<^sub>n:=n)| s n. s(x:=s x\<^sub>i\<^sub>n)\<in>\<lbrakk>p\<rbrakk>\<^sub>a \<and> s x = \<lbrakk>e\<rbrakk>[s(x:=s x\<^sub>i\<^sub>n)]}" by blast
    qed
    show "{s(x\<^sub>i\<^sub>n:=n)| s n. s(x:=s x\<^sub>i\<^sub>n)\<in>\<lbrakk>p\<rbrakk>\<^sub>a \<and> s x = \<lbrakk>e\<rbrakk>[s(x:=s x\<^sub>i\<^sub>n)]} \<subseteq> {(s(x:=\<lbrakk>e\<rbrakk>[s]))(x\<^sub>i\<^sub>n:=n)| s n. s\<in>\<lbrakk>p\<rbrakk>\<^sub>a}" by (smt (verit) Collect_mono_iff fun_upd_triv fun_upd_upd)
  qed
  ultimately show ?thesis by argo
qed

fun SynSRAsrt_fv :: "'var SynSRAsrt \<Rightarrow> 'var set" where
 "SynSRAsrt_fv (DAsrt p q) = SynSAsrt_fv p \<union> SynSAsrt_fv q"
|"SynSRAsrt_fv (RAnd r s) = SynSRAsrt_fv r \<inter> SynSRAsrt_fv s"
|"SynSRAsrt_fv (ROr r s) = SynSRAsrt_fv r \<union> SynSRAsrt_fv s"
|"SynSRAsrt_fv (RImpl r s) = SynSRAsrt_fv r \<union> SynSRAsrt_fv s"
|"SynSRAsrt_fv (RBimp r s) = SynSRAsrt_fv r \<union> SynSRAsrt_fv s"
|"SynSRAsrt_fv (RNeg r) = SynSRAsrt_fv r"
|"SynSRAsrt_fv (REx x r) = SynSRAsrt_fv r-{x}"
|"SynSRAsrt_fv (RFa x r) = SynSRAsrt_fv r-{x}"

(*The subscript stands for rassertion*)
fun rinter :: "'var SynSRAsrt \<Rightarrow> (('var, nat) state \<times> ('var, nat) state) set" ("\<lbrakk>_\<rbrakk>\<^sub>r" [0] 81) where
 "\<lbrakk>DAsrt p q\<rbrakk>\<^sub>r = {(s\<^sub>1, s\<^sub>2)| s\<^sub>1 s\<^sub>2. s\<^sub>1\<in>\<lbrakk>p\<rbrakk>\<^sub>a \<and> s\<^sub>2\<in>\<lbrakk>q\<rbrakk>\<^sub>a}"
|"\<lbrakk>RAnd r s\<rbrakk>\<^sub>r = \<lbrakk>r\<rbrakk>\<^sub>r \<inter> \<lbrakk>s\<rbrakk>\<^sub>r"
|"\<lbrakk>ROr r s\<rbrakk>\<^sub>r = \<lbrakk>r\<rbrakk>\<^sub>r \<union> \<lbrakk>s\<rbrakk>\<^sub>r"
|"\<lbrakk>RImpl r s\<rbrakk>\<^sub>r = (UNIV-\<lbrakk>r\<rbrakk>\<^sub>r) \<union> \<lbrakk>s\<rbrakk>\<^sub>r"
|"\<lbrakk>RBimp r s\<rbrakk>\<^sub>r = (\<lbrakk>r\<rbrakk>\<^sub>r \<inter> \<lbrakk>s\<rbrakk>\<^sub>r) \<union> ((UNIV-\<lbrakk>r\<rbrakk>\<^sub>r) \<inter> (UNIV-\<lbrakk>s\<rbrakk>\<^sub>r))"
|"\<lbrakk>RNeg r\<rbrakk>\<^sub>r = UNIV-\<lbrakk>r\<rbrakk>\<^sub>r"
|"\<lbrakk>REx x r\<rbrakk>\<^sub>r = {(s\<^sub>1, s\<^sub>2)| s\<^sub>1 s\<^sub>2. \<exists>n. (s\<^sub>1(x:=n), s\<^sub>2(x:=n))\<in>\<lbrakk>r\<rbrakk>\<^sub>r}"
|"\<lbrakk>RFa x r\<rbrakk>\<^sub>r = {(s\<^sub>1, s\<^sub>2)| s\<^sub>1 s\<^sub>2. \<forall>n. (s\<^sub>1(x:=n), s\<^sub>2(x:=n))\<in>\<lbrakk>r\<rbrakk>\<^sub>r}"

lemma "\<lbrakk>RImpl r s\<rbrakk>\<^sub>r = \<lbrakk>ROr (RNeg r) s\<rbrakk>\<^sub>r" by simp
lemma "\<lbrakk>RBimp r s\<rbrakk>\<^sub>r = \<lbrakk>RAnd (RImpl r s) (RImpl s r)\<rbrakk>\<^sub>r" by auto

fun SynSHAsrt_fv\<^sub>1 :: "('svar, 'pvar) SynSHAsrt \<Rightarrow> 'pvar set" where
 "SynSHAsrt_fv\<^sub>1 Top = {}"
|"SynSHAsrt_fv\<^sub>1 Bot = {}"
|"SynSHAsrt_fv\<^sub>1 (Asrt s p) = SynSAsrt_fv p"
|"SynSHAsrt_fv\<^sub>1 (RAsrt s\<^sub>1 s\<^sub>2 r) = SynSRAsrt_fv r"
|"SynSHAsrt_fv\<^sub>1 (ExN n P) = SynSHAsrt_fv\<^sub>1 P-{n}"
|"SynSHAsrt_fv\<^sub>1 (FaN n P) = SynSHAsrt_fv\<^sub>1 P-{n}"
|"SynSHAsrt_fv\<^sub>1 (ExS s P) = SynSHAsrt_fv\<^sub>1 P"
|"SynSHAsrt_fv\<^sub>1 (FaS s P) = SynSHAsrt_fv\<^sub>1 P"
|"SynSHAsrt_fv\<^sub>1 (And P\<^sub>1 P\<^sub>2) = SynSHAsrt_fv\<^sub>1 P\<^sub>1 \<union> SynSHAsrt_fv\<^sub>1 P\<^sub>2"
|"SynSHAsrt_fv\<^sub>1 (Or P\<^sub>1 P\<^sub>2) = SynSHAsrt_fv\<^sub>1 P\<^sub>1 \<union> SynSHAsrt_fv\<^sub>1 P\<^sub>2"

fun SynSHAsrt_fv\<^sub>2 :: "('svar, 'pvar) SynSHAsrt \<Rightarrow> 'svar set" where
 "SynSHAsrt_fv\<^sub>2 Top = {}"
|"SynSHAsrt_fv\<^sub>2 Bot = {}"
|"SynSHAsrt_fv\<^sub>2 (Asrt s p) = {s}"
|"SynSHAsrt_fv\<^sub>2 (RAsrt s\<^sub>1 s\<^sub>2 r) = {s\<^sub>1, s\<^sub>2}"
|"SynSHAsrt_fv\<^sub>2 (ExN n P) = SynSHAsrt_fv\<^sub>2 P"
|"SynSHAsrt_fv\<^sub>2 (FaN n P) = SynSHAsrt_fv\<^sub>2 P"
|"SynSHAsrt_fv\<^sub>2 (ExS s P) = SynSHAsrt_fv\<^sub>2 P-{s}"
|"SynSHAsrt_fv\<^sub>2 (FaS s P) = SynSHAsrt_fv\<^sub>2 P-{s}"
|"SynSHAsrt_fv\<^sub>2 (And P\<^sub>1 P\<^sub>2) = SynSHAsrt_fv\<^sub>2 P\<^sub>1 \<union> SynSHAsrt_fv\<^sub>2 P\<^sub>2"
|"SynSHAsrt_fv\<^sub>2 (Or P\<^sub>1 P\<^sub>2) = SynSHAsrt_fv\<^sub>2 P\<^sub>1 \<union> SynSHAsrt_fv\<^sub>2 P\<^sub>2"

fun inter :: "('svar, 'pvar) SynSHAsrt \<Rightarrow> ('svar \<Rightarrow> ('pvar, nat) state) \<Rightarrow> ('pvar, nat) state set set" ("\<lbrakk>_\<rbrakk>\<^sub>_" [51,81] 91) where
 "\<lbrakk>Top\<rbrakk>\<^sub>\<Sigma> = UNIV"
|"\<lbrakk>Bot\<rbrakk>\<^sub>\<Sigma> = {}"
|"\<lbrakk>Asrt s p\<rbrakk>\<^sub>\<Sigma> = {S. \<Sigma> s\<in>\<lbrakk>p\<rbrakk>\<^sub>a}"
|"\<lbrakk>RAsrt s\<^sub>1 s\<^sub>2 r\<rbrakk>\<^sub>\<Sigma> = {S. (\<Sigma> s\<^sub>1, \<Sigma> s\<^sub>2)\<in>\<lbrakk>r\<rbrakk>\<^sub>r}"
|"\<lbrakk>ExN x P\<rbrakk>\<^sub>\<Sigma> = {S. \<exists>v. {s(x:=v)| s. s\<in>S}\<in>\<lbrakk>P\<rbrakk>\<^sub>(\<lambda>s. (\<Sigma> s)(x:=v))}"
|"\<lbrakk>FaN x P\<rbrakk>\<^sub>\<Sigma> = {S. \<forall>v. {s(x:=v)| s. s\<in>S}\<in>\<lbrakk>P\<rbrakk>\<^sub>(\<lambda>s. (\<Sigma> s)(x:=v))}"
|"\<lbrakk>ExS s P\<rbrakk>\<^sub>\<Sigma> = {S. \<exists>\<nu>\<in>S. S\<in>\<lbrakk>P\<rbrakk>\<^sub>\<Sigma>(s:=\<nu>)}"
|"\<lbrakk>FaS s P\<rbrakk>\<^sub>\<Sigma> = {S. \<forall>\<nu>\<in>S. S\<in>\<lbrakk>P\<rbrakk>\<^sub>\<Sigma>(s:=\<nu>)}"
|"\<lbrakk>And P Q\<rbrakk>\<^sub>\<Sigma> = \<lbrakk>P\<rbrakk>\<^sub>\<Sigma> \<inter> \<lbrakk>Q\<rbrakk>\<^sub>\<Sigma>"
|"\<lbrakk>Or P Q\<rbrakk>\<^sub>\<Sigma> = \<lbrakk>P\<rbrakk>\<^sub>\<Sigma> \<union> \<lbrakk>Q\<rbrakk>\<^sub>\<Sigma>"

lemma inter_depends_on_free2:
  assumes "\<And>s. s\<in>SynSHAsrt_fv\<^sub>2 P \<Longrightarrow> \<Sigma> s=\<Sigma>' s"
    shows "\<lbrakk>P\<rbrakk>\<^sub>\<Sigma> = \<lbrakk>P\<rbrakk>\<^sub>\<Sigma>'"
  using assms
proof(induction P arbitrary: \<Sigma> \<Sigma>')
  case Top
  then show ?case by simp
next
  case Bot
  then show ?case by simp
next
  case (Asrt s p)
  then show ?case by auto
next
  case (RAsrt s\<^sub>1 s\<^sub>2 r)
  then show ?case by simp
next
  case (ExN x P)
  moreover have "\<lbrakk>ExN x P\<rbrakk>\<^sub>\<Sigma> = {S. \<exists>v. {s(x:=v)| s. s\<in>S}\<in>\<lbrakk>P\<rbrakk>\<^sub>(\<lambda>s. (\<Sigma> s)(x:=v))}" by fastforce
  moreover have "\<lbrakk>ExN x P\<rbrakk>\<^sub>\<Sigma>' = {S. \<exists>v. {s(x:=v)| s. s\<in>S}\<in>\<lbrakk>P\<rbrakk>\<^sub>(\<lambda>s. (\<Sigma>' s)(x:=v))}" by fastforce
  ultimately show ?case by (smt (verit, ccfv_SIG) Collect_cong SynSHAsrt_fv\<^sub>2.simps(5))
next
  case (FaN x P)
  moreover have "\<lbrakk>FaN x P\<rbrakk>\<^sub>\<Sigma> = {S. \<forall>v. {s(x:=v)| s. s\<in>S}\<in>\<lbrakk>P\<rbrakk>\<^sub>(\<lambda>s. (\<Sigma> s)(x:=v))}" by fastforce
  moreover have "\<lbrakk>FaN x P\<rbrakk>\<^sub>\<Sigma>' = {S. \<forall>v. {s(x:=v)| s. s\<in>S}\<in>\<lbrakk>P\<rbrakk>\<^sub>(\<lambda>s. (\<Sigma>' s)(x:=v))}" by fastforce
  ultimately show ?case by (smt (verit, best) Collect_cong SynSHAsrt_fv\<^sub>2.simps(6))
next
  case (ExS s\<^sub>0 P)
  moreover have "\<lbrakk>ExS s\<^sub>0 P\<rbrakk>\<^sub>\<Sigma> = {S. \<exists>\<nu>\<in>S. S\<in>\<lbrakk>P\<rbrakk>\<^sub>\<Sigma>(s\<^sub>0:=\<nu>)}" by simp
  moreover have "\<lbrakk>ExS s\<^sub>0 P\<rbrakk>\<^sub>\<Sigma>' = {S. \<exists>\<nu>\<in>S. S\<in>\<lbrakk>P\<rbrakk>\<^sub>\<Sigma>'(s\<^sub>0:=\<nu>)}" by simp
  moreover have "\<And>s. s\<in>SynSHAsrt_fv\<^sub>2 P \<Longrightarrow> \<Sigma> s=\<Sigma>' s \<or> s=s\<^sub>0" using ExS(2) by fastforce
  then have "\<And>s \<nu>. s\<in>SynSHAsrt_fv\<^sub>2 P \<Longrightarrow> (\<Sigma>(s\<^sub>0:=\<nu>)) s=(\<Sigma>'(s\<^sub>0:=\<nu>)) s" by auto
  ultimately show ?case by (smt (verit, best) Collect_cong)
next
  case (FaS s\<^sub>0 P)
  moreover have "\<lbrakk>FaS s\<^sub>0 P\<rbrakk>\<^sub>\<Sigma> = {S. \<forall>\<nu>\<in>S. S\<in>\<lbrakk>P\<rbrakk>\<^sub>\<Sigma>(s\<^sub>0:=\<nu>)}" by simp
  moreover have "\<lbrakk>FaS s\<^sub>0 P\<rbrakk>\<^sub>\<Sigma>' = {S. \<forall>\<nu>\<in>S. S\<in>\<lbrakk>P\<rbrakk>\<^sub>\<Sigma>'(s\<^sub>0:=\<nu>)}" by simp
  moreover have "\<And>s. s\<in>SynSHAsrt_fv\<^sub>2 P \<Longrightarrow> \<Sigma> s=\<Sigma>' s \<or> s=s\<^sub>0" using FaS(2) by fastforce
  then have "\<And>s \<nu>. s\<in>SynSHAsrt_fv\<^sub>2 P \<Longrightarrow> (\<Sigma>(s\<^sub>0:=\<nu>)) s=(\<Sigma>'(s\<^sub>0:=\<nu>)) s" by auto
  ultimately show ?case by (smt (verit, best) Collect_cong)
next
  case (And P\<^sub>1 P\<^sub>2)
  then show ?case by (metis (no_types, opaque_lifting) UnI1 UnI2 SynSHAsrt_fv\<^sub>2.simps(9) inter.simps(9))
next
  case (Or P1 P2)
  then show ?case by (metis (no_types, opaque_lifting) UnI1 UnI2 SynSHAsrt_fv\<^sub>2.simps(10) inter.simps(10))
qed

lemma closed2_independent_sigma:
  assumes "SynSHAsrt_fv\<^sub>2 P={}"
    shows "\<lbrakk>P\<rbrakk>\<^sub>\<Sigma>=\<lbrakk>P\<rbrakk>\<^sub>\<Sigma>'"
  by (metis assms empty_iff inter_depends_on_free2)


lemma "\<lbrakk>ExN x (ExS s (Asrt s p))\<rbrakk>\<^sub>\<Sigma> = {S. \<exists>s v. s\<in>S \<and> s(x:=v)\<in>\<lbrakk>p\<rbrakk>\<^sub>a}" by auto
lemma "\<lbrakk>ExS s (ExN x (Asrt s p))\<rbrakk>\<^sub>\<Sigma> = {S. \<exists>s v. s\<in>S \<and> s(x:=v)\<in>\<lbrakk>p\<rbrakk>\<^sub>a}" by auto
lemma commute_update_fun: "(\<lambda>s. (\<Sigma> s)(x:=v))(s\<^sub>0:=s\<^sub>s\<^sub>e\<^sub>m(x:=v)) = (\<lambda>s. ((\<Sigma>(s\<^sub>0:=s\<^sub>s\<^sub>e\<^sub>m)) s)(x:=v))" by force
lemma ex_comm: "\<lbrakk>ExN x (ExS s P)\<rbrakk>\<^sub>\<Sigma> = \<lbrakk>ExS s (ExN x P)\<rbrakk>\<^sub>\<Sigma>" using commute_update_fun[of \<Sigma> x, where ?s\<^sub>0=s] by force
lemma fa_comm: "\<lbrakk>FaN x (FaS s P)\<rbrakk>\<^sub>\<Sigma> = \<lbrakk>FaS s (FaN x P)\<rbrakk>\<^sub>\<Sigma>" using commute_update_fun[of \<Sigma> x, where ?s\<^sub>0=s] by force
lemma ex_distr_asrt: "\<lbrakk>ExN x (Asrt s p)\<rbrakk>\<^sub>\<Sigma> = \<lbrakk>Asrt s (SEx x p)\<rbrakk>\<^sub>\<Sigma>" by simp
lemma fa_distr_asrt: "\<lbrakk>FaN x (Asrt s p)\<rbrakk>\<^sub>\<Sigma> = \<lbrakk>Asrt s (SFa x p)\<rbrakk>\<^sub>\<Sigma>" by simp
lemma ex_distr_rasrt: "\<lbrakk>ExN x (RAsrt s\<^sub>1 s\<^sub>2 r)\<rbrakk>\<^sub>\<Sigma> = \<lbrakk>RAsrt s\<^sub>1 s\<^sub>2 (REx x r)\<rbrakk>\<^sub>\<Sigma>" by simp
lemma fa_distr_rasrt: "\<lbrakk>FaN x (RAsrt s\<^sub>1 s\<^sub>2 r)\<rbrakk>\<^sub>\<Sigma> = \<lbrakk>RAsrt s\<^sub>1 s\<^sub>2 (RFa x r)\<rbrakk>\<^sub>\<Sigma>" by simp
lemma "\<lbrakk>Or (FaS s (Asrt s p)) (FaS s (Asrt s q))\<rbrakk>\<^sub>\<Sigma> \<subseteq> \<lbrakk>FaS s (Asrt s (SOr p q))\<rbrakk>\<^sub>\<Sigma>" by auto
lemma "\<lbrakk>FaS s (Asrt s p)\<rbrakk>\<^sub>\<Sigma> = {S. S\<subseteq>\<lbrakk>p\<rbrakk>\<^sub>a}" by auto
lemma "\<lbrakk>ExS s (Asrt s p)\<rbrakk>\<^sub>\<Sigma> = {S. S\<inter>\<lbrakk>p\<rbrakk>\<^sub>a\<noteq>{}}" by auto
lemma HHL_expresses_HL_1: "(\<forall>S\<subseteq>p. C``S\<subseteq>q) \<longleftrightarrow> (\<forall>s\<in>p. C``{s}\<subseteq>q)" by auto
lemma HHL_expresses_HL_2: "(\<forall>S\<subseteq>p. C``S\<subseteq>q) \<longleftrightarrow> (\<forall>s\<in>p. \<forall>s'. (s,s')\<in>C \<longrightarrow> s'\<in>q)" by auto
lemma HHL_expresses_SIL_1: "(\<forall>S. S\<inter>p\<noteq>{} \<longrightarrow> (C``S)\<inter>q\<noteq>{}) \<longleftrightarrow> (\<forall>s\<in>p. (C``{s})\<inter>q\<noteq>{})" by auto
lemma HHL_expresses_SIL_2:"(\<forall>S. S\<inter>p\<noteq>{} \<longrightarrow> (C``S)\<inter>q\<noteq>{}) \<longleftrightarrow> (\<forall>s\<in>p. \<exists>s'. (s, s')\<in>C \<and> s'\<in>q)" by auto
lemma "\<lbrakk>SNot (SEqual (AVar t) AOne)\<rbrakk>\<^sub>a={s. s t\<noteq>1}" by auto
lemma RHL_encoding_1:"s\<^sub>1\<noteq>s\<^sub>2 \<Longrightarrow> \<lbrakk>FaS s\<^sub>1 (FaS s\<^sub>2 (Or (Or (Asrt s\<^sub>1 (SNot (SEqual (AVar t) AOne))) (Asrt s\<^sub>2 (SNot (SEqual (AVar t) (APlus AOne AOne))))) P))\<rbrakk>\<^sub>\<Sigma> = {S. \<forall>\<nu>\<^sub>1\<in>S. \<forall>\<nu>\<^sub>2\<in>S. S\<in>(if \<nu>\<^sub>1 t\<noteq>1 \<or> \<nu>\<^sub>2 t\<noteq>2 then UNIV else \<lbrakk>P\<rbrakk>\<^sub>\<Sigma>(s\<^sub>1:=\<nu>\<^sub>1,s\<^sub>2:=\<nu>\<^sub>2))}" by auto
lemma RHL_encoding_2: "s\<^sub>1\<noteq>s\<^sub>2 \<Longrightarrow> \<lbrakk>FaS s\<^sub>1 (FaS s\<^sub>2 (Or (Or (Asrt s\<^sub>1 (SNot (SEqual (AVar t) AOne))) (Asrt s\<^sub>2 (SNot (SEqual (AVar t) (APlus AOne AOne))))) P))\<rbrakk>\<^sub>\<Sigma> = {S. \<forall>\<nu>\<^sub>1\<in>S. \<forall>\<nu>\<^sub>2\<in>S. \<nu>\<^sub>1 t=1 \<longrightarrow> \<nu>\<^sub>2 t=2 \<longrightarrow> S\<in>\<lbrakk>P\<rbrakk>\<^sub>\<Sigma>(s\<^sub>1:=\<nu>\<^sub>1,s\<^sub>2:=\<nu>\<^sub>2)}" by auto
lemma RHL_eq_enc: "{S. \<forall>s\<^sub>1\<in>S. \<forall>s\<^sub>2\<in>S. s\<^sub>1 t=1 \<longrightarrow> s\<^sub>2 t=2 \<longrightarrow> (s\<^sub>1, s\<^sub>2)\<in>r} = {S. {s\<in>S. s t=1}\<times>{s\<in>S. s t=2}\<subseteq>r}" by blast
lemma HHL_expresses_RHL:
  assumes "\<forall>(s, s')\<in>C. s t=s' t"
      and "\<forall>(s\<^sub>1, s\<^sub>2)\<in>r. s\<^sub>1 t=1 \<and> s\<^sub>2 t=2"
    shows "(\<forall>S. {s\<in>S. s t=(1::nat)}\<times>{s\<in>S. s t=2}\<subseteq>r \<longrightarrow> {s\<in>C``S. s t=1}\<times>{s\<in>C``S. s t=2}\<subseteq>s) \<longleftrightarrow> (\<forall>(s\<^sub>1, s\<^sub>2)\<in>r. \<forall>s\<^sub>1'\<in>C``{s\<^sub>1}. \<forall>s\<^sub>2'\<in>C``{s\<^sub>2}. (s\<^sub>1', s\<^sub>2')\<in>s)"
proof(-)
  have a1: "\<And>S n. {s\<in>C``S. s t=n} = {s\<in>C``{s\<in>S. s t=n}. s t=n}" using assms(1) by fastforce
  have "\<And>S. {s\<in>S. s t=1}\<times>{s\<in>S. s t=2}\<subseteq>r \<longrightarrow> {s\<in>C``{s\<in>S. s t=1}. s t=1}\<times>{s\<in>C``{s\<in>S. s t=2}. s t=2}\<subseteq>s \<Longrightarrow> {s\<in>S. s t=1}\<times>{s\<in>S. s t=2}\<subseteq>r \<longrightarrow> {s\<in>C``S. s t=1}\<times>{s\<in>C``S. s t=2}\<subseteq>s"
  proof(-)
    fix S
    assume asm1: "{s\<in>S. s t=1}\<times>{s\<in>S. s t=2}\<subseteq>r \<longrightarrow> {s\<in>C``{s\<in>S. s t=1}. s t=1}\<times>{s\<in>C``{s\<in>S. s t=2}. s t=2}\<subseteq>s"
    show "{s\<in>S. s t=1}\<times>{s\<in>S. s t=2}\<subseteq>r \<longrightarrow> {s\<in>C``S. s t=1}\<times>{s\<in>C``S. s t=2}\<subseteq>s"
    proof
      assume asm2: "{s\<in>S. s t=1}\<times>{s\<in>S. s t=2}\<subseteq>r"
      show "{s\<in>C``S. s t=1}\<times>{s\<in>C``S. s t=2}\<subseteq>s" using a1 asm1 asm2 by presburger
    qed
  qed
  then have "\<forall>S. {s\<in>S. s t=1}\<times>{s\<in>S. s t=2}\<subseteq>r \<longrightarrow> {s\<in>C``{s\<in>S. s t=1}. s t=1}\<times>{s\<in>C``{s\<in>S. s t=2}. s t=2}\<subseteq>s \<Longrightarrow> \<forall>S. {s\<in>S. s t=1}\<times>{s\<in>S. s t=2}\<subseteq>r \<longrightarrow> {s\<in>C``S. s t=1}\<times>{s\<in>C``S. s t=2}\<subseteq>s" by presburger
  moreover have "\<forall>S. {s\<in>S. s t=1}\<times>{s\<in>S. s t=2}\<subseteq>r \<longrightarrow> {s\<in>C``S. s t=1}\<times>{s\<in>C``S. s t=2}\<subseteq>s \<Longrightarrow> \<forall>S. {s\<in>S. s t=1}\<times>{s\<in>S. s t=2}\<subseteq>r \<longrightarrow> {s\<in>C``{s\<in>S. s t=1}. s t=1}\<times>{s\<in>C``{s\<in>S. s t=2}. s t=2}\<subseteq>s" by auto
  ultimately have 0: "(\<forall>S. {s\<in>S. s t=1}\<times>{s\<in>S. s t=2}\<subseteq>r \<longrightarrow> {s\<in>C``S. s t=1}\<times>{s\<in>C``S. s t=2}\<subseteq>s) \<longleftrightarrow> (\<forall>S. {s\<in>S. s t=1}\<times>{s\<in>S. s t=2}\<subseteq>r \<longrightarrow> {s\<in>C``{s\<in>S. s t=1}. s t=1}\<times>{s\<in>C``{s\<in>S. s t=2}. s t=2}\<subseteq>s)" by argo

  have "\<forall>S. {s\<in>S. s t=1}\<times>{s\<in>S. s t=2}\<subseteq>r \<longrightarrow> {s\<in>C``{s\<in>S. s t=1}. s t=1}\<times>{s\<in>C``{s\<in>S. s t=2}. s t=2}\<subseteq>s \<Longrightarrow> \<forall>(s\<^sub>1, s\<^sub>2)\<in>r. \<forall>s\<^sub>1'\<in>C``{s\<^sub>1}. \<forall>s\<^sub>2'\<in>C``{s\<^sub>2}. (s\<^sub>1', s\<^sub>2')\<in>s" 
  proof(-)
    assume asm1: "\<forall>S. {s\<in>S. s t=1}\<times>{s\<in>S. s t=2}\<subseteq>r \<longrightarrow> {s\<in>C``{s\<in>S. s t=1}. s t=1}\<times>{s\<in>C``{s\<in>S. s t=2}. s t=2}\<subseteq>s"
    have "\<And>s\<^sub>1 s\<^sub>2 s\<^sub>1' s\<^sub>2'. (s\<^sub>1, s\<^sub>2)\<in>r \<Longrightarrow> s\<^sub>1'\<in>C``{s\<^sub>1} \<Longrightarrow> s\<^sub>2'\<in>C``{s\<^sub>2} \<Longrightarrow> (s\<^sub>1', s\<^sub>2')\<in>s"
    proof(-)
      fix s\<^sub>1 s\<^sub>2 s\<^sub>1' s\<^sub>2'
      assume asm2: "(s\<^sub>1, s\<^sub>2)\<in>r" "s\<^sub>1'\<in>C``{s\<^sub>1}" "s\<^sub>2'\<in>C``{s\<^sub>2}"
      then have "s\<^sub>1 t=1" "s\<^sub>2 t=2" "(s\<^sub>1, s\<^sub>1')\<in>C" "(s\<^sub>2, s\<^sub>2')\<in>C" using assms(2) by auto
      then have "s\<^sub>1' t=1" "s\<^sub>2' t=2" using assms(1) by auto
      obtain S where "S={s\<^sub>1, s\<^sub>2}" by auto
      then have "{s\<in>S. s t=1}\<times>{s\<in>S. s t=2}\<subseteq>r" using asm2(1) \<open>s\<^sub>1 t=1\<close> \<open>s\<^sub>2 t=2\<close> \<open>S = {s\<^sub>1, s\<^sub>2}\<close> by auto
      then have "{s\<in>C``{s\<in>S. s t=1}. s t=1}\<times>{s\<in>C``{s\<in>S. s t=2}. s t=2}\<subseteq>s" using asm1 by presburger
      moreover have "s\<^sub>1'\<in>{s\<in>C``{s\<in>S. s t=1}. s t=1}" using \<open>(s\<^sub>1, s\<^sub>1')\<in>C\<close> \<open>S={s\<^sub>1, s\<^sub>2}\<close> \<open>s\<^sub>1 t=1\<close> \<open>s\<^sub>1' t=1\<close> by auto
      moreover have "s\<^sub>2'\<in>{s\<in>C``{s\<in>S. s t=2}. s t=2}" using \<open>(s\<^sub>2, s\<^sub>2')\<in>C\<close> \<open>S={s\<^sub>1, s\<^sub>2}\<close> \<open>s\<^sub>2 t=2\<close> \<open>s\<^sub>2' t=2\<close> by auto
      ultimately show "(s\<^sub>1', s\<^sub>2')\<in>s" by blast
    qed
    then show "\<forall>(s\<^sub>1, s\<^sub>2)\<in>r. \<forall>s\<^sub>1'\<in>C``{s\<^sub>1}. \<forall>s\<^sub>2'\<in>C``{s\<^sub>2}. (s\<^sub>1', s\<^sub>2')\<in>s" by blast
  qed
  moreover have "\<forall>(s\<^sub>1, s\<^sub>2)\<in>r. \<forall>s\<^sub>1'\<in>C``{s\<^sub>1}. \<forall>s\<^sub>2'\<in>C``{s\<^sub>2}. (s\<^sub>1', s\<^sub>2')\<in>s \<Longrightarrow> \<forall>S. {s\<in>S. s t=1}\<times>{s\<in>S. s t=2}\<subseteq>r \<longrightarrow> {s\<in>C``{s\<in>S. s t=1}. s t=1}\<times>{s\<in>C``{s\<in>S. s t=2}. s t=2}\<subseteq>s" by blast
  ultimately have 1: "(\<forall>S. {s\<in>S. s t=1}\<times>{s\<in>S. s t=2}\<subseteq>r \<longrightarrow> {s\<in>C``{s\<in>S. s t=1}. s t=1}\<times>{s\<in>C``{s\<in>S. s t=2}. s t=2}\<subseteq>s) \<longleftrightarrow> (\<forall>(s\<^sub>1, s\<^sub>2)\<in>r. \<forall>s\<^sub>1'\<in>C``{s\<^sub>1}. \<forall>s\<^sub>2'\<in>C``{s\<^sub>2}. (s\<^sub>1', s\<^sub>2')\<in>s)" by argo

  then show ?thesis using "0" by argo
qed

lemma "x\<noteq>y \<Longrightarrow> \<lbrakk>REx y (DAsrt (SEqual (AVar x) (AVar y)) (SEqual (AVar x) (AVar y)))\<rbrakk>\<^sub>r = {(s\<^sub>1, s\<^sub>2). s\<^sub>1 x=s\<^sub>2 x}" by auto 
lemma "s\<^sub>1\<noteq>s\<^sub>2 \<Longrightarrow> x\<noteq>y \<Longrightarrow> \<lbrakk>RAsrt s\<^sub>1 s\<^sub>2 (REx y (DAsrt (SEqual (AVar x) (AVar y)) (SEqual (AVar x) (AVar y))))\<rbrakk>\<^sub>\<Sigma>(s\<^sub>1:=\<nu>\<^sub>1,s\<^sub>2:=\<nu>\<^sub>2) = {S. \<nu>\<^sub>1 x=\<nu>\<^sub>2 x}" by simp
lemma not_union_closed:
  assumes "s\<^sub>1\<noteq>s\<^sub>2"
      and "x\<noteq>y"
    shows "\<lbrakk>FaS s\<^sub>1 (FaS s\<^sub>2 (Or (Or (Asrt s\<^sub>1 (SNot (SEqual (AVar t) AOne))) (Asrt s\<^sub>2 (SNot (SEqual (AVar t) (APlus AOne AOne))))) (RAsrt s\<^sub>1 s\<^sub>2 (REx y (DAsrt (SEqual (AVar x) (AVar y)) (SEqual (AVar x) (AVar y)))))))\<rbrakk>\<^sub>\<Sigma>
         = {S. \<forall>\<nu>\<^sub>1\<in>S. \<forall>\<nu>\<^sub>2\<in>S. \<nu>\<^sub>1 t=1 \<longrightarrow> \<nu>\<^sub>2 t=2 \<longrightarrow> \<nu>\<^sub>1 x=\<nu>\<^sub>2 x}"(*\<lbrakk>\<forall>\<langle>s\<^sub>1\<rangle>. \<forall>\<langle>s\<^sub>2\<rangle>. s\<^sub>1(\<not>t=1) \<or> s\<^sub>2(\<not>t=2) \<or> [s\<^sub>1, s2](\<exists>y. [x=y, x=y])\<rbrakk>*)
  using assms by auto
lemma implicit_t:
  assumes "s\<^sub>1\<noteq>s\<^sub>2"
      and "x\<noteq>y"
      and "t\<noteq>x \<and> t\<noteq>y"
    shows "\<lbrakk>FaS s\<^sub>1 (FaS s\<^sub>2 (Or (Or (Asrt s\<^sub>1 (SNot (SEqual (AVar t) AOne))) (Asrt s\<^sub>2 (SNot (SEqual (AVar t) (APlus AOne AOne))))) (RAsrt s\<^sub>1 s\<^sub>2 (REx y (DAsrt (SAnd (SEqual (AVar x) (AVar y)) (SEqual (AVar t) AOne)) (SEqual (AVar x) (AVar y)))))))\<rbrakk>\<^sub>\<Sigma>
         = {S. \<forall>\<nu>\<^sub>1\<in>S. \<forall>\<nu>\<^sub>2\<in>S. \<nu>\<^sub>1 t=1 \<longrightarrow> \<nu>\<^sub>2 t=2 \<longrightarrow> \<nu>\<^sub>1 x=\<nu>\<^sub>2 x}" (*it's the same as above, i.e. as in not_union_closed, but note that we want \<nu>\<^sub>1 t=1, but it can be skipped, cuz we have it as assumption*)
  using assms by auto
(*Instead of \<forall>\<langle>s\<^sub>1\<rangle>. \<forall>\<langle>s\<^sub>2\<rangle>. s\<^sub>1(\<not>t=1) \<or> s\<^sub>2(\<not>t=2) \<or> [s\<^sub>1, s\<^sub>2](r), we can write \<forall>\<langle>s\<^sub>1\<rangle>. \<forall>\<langle>s\<^sub>2\<rangle>. [s\<^sub>1, s\<^sub>2](r \<or> [\<not>t=1,\<not>t=2])*)
lemma RSL_encoding_into_the_R_part: (*assuming s\<^sub>1\<noteq>s\<^sub>2, then {S. \<forall>\<nu>\<^sub>1\<in>S. \<forall>\<nu>\<^sub>2\<in>S. \<nu>\<^sub>1 t=1 \<longrightarrow> \<nu>\<^sub>2 t=2 \<longrightarrow> (\<nu>\<^sub>1, \<nu>\<^sub>2)\<in>\<lbrakk>r\<rbrakk>\<^sub>r}*)
  "\<lbrakk>FaS s\<^sub>1 (FaS s\<^sub>2 (Or (Or (Asrt s\<^sub>1 (SNot (SEqual (AVar t) AOne))) (Asrt s\<^sub>2 (SNot (SEqual (AVar t) (APlus AOne AOne))))) (RAsrt s\<^sub>1 s\<^sub>2 r)))\<rbrakk>\<^sub>\<Sigma>
 = \<lbrakk>FaS s\<^sub>1 (FaS s\<^sub>2 (RAsrt s\<^sub>1 s\<^sub>2 (RImpl (DAsrt (SEqual (AVar t) AOne) (SEqual (AVar t) (APlus AOne AOne))) r)))\<rbrakk>\<^sub>\<Sigma>" by force


lemma "\<lbrakk>FaS s (Asrt s (SEqual (AVar x) AOne))\<rbrakk>\<^sub>\<Sigma> = {S. \<forall>s\<in>S. s x=1}" by auto
lemma "{S. \<forall>s\<in>S. s x=1}\<inter>{S. \<forall>\<nu>\<^sub>1\<in>S. \<forall>\<nu>\<^sub>2\<in>S. \<nu>\<^sub>1 t=1 \<longrightarrow> \<nu>\<^sub>2 t=2 \<longrightarrow> \<nu>\<^sub>1 x=\<nu>\<^sub>2 x}={S. \<forall>s\<in>S. s x=1}" by (smt (verit, best) Collect_mono_iff inf.order_iff)


lemma "{S. \<forall>s\<in>S. s x=(1::nat) \<or> s x=2}\<inter>{S. \<forall>\<nu>\<^sub>1\<in>S. \<forall>\<nu>\<^sub>2\<in>S. \<nu>\<^sub>1 t=1 \<longrightarrow> \<nu>\<^sub>2 t=2 \<longrightarrow> \<nu>\<^sub>1 x=\<nu>\<^sub>2 x}={{}}" 

(*proof(-)
  have 0: "\<And>v s. (\<lambda>s\<^sub>s\<^sub>y\<^sub>n. (f s\<^sub>s\<^sub>y\<^sub>n)(x:=v))(s\<^sub>s\<^sub>y\<^sub>n\<^sub>0:=s(x:=v)) = (\<lambda>s\<^sub>s\<^sub>y\<^sub>n. ((f(s\<^sub>s\<^sub>y\<^sub>n\<^sub>0:=s)) s\<^sub>s\<^sub>y\<^sub>n)(x:=v))" by auto
  then show ?thesis by forc
qed*)
(*  have "\<lbrakk>ExN x (ExS s\<^sub>s\<^sub>y\<^sub>n\<^sub>0 P)\<rbrakk>\<^sub>\<Sigma> = {S. \<exists>v. {s(x:=v)| s. s\<in>S}\<in>\<lbrakk>ExS s\<^sub>s\<^sub>y\<^sub>n\<^sub>0 P\<rbrakk>\<^sub>(\<lambda>s\<^sub>s\<^sub>y\<^sub>n. (f s\<^sub>s\<^sub>y\<^sub>n)(x:=v))}" by auto
  also have "... = {S. \<exists>v. {s(x:=v)| s. s\<in>S}\<in>{S'. \<exists>\<nu>\<in>S'. S'\<in>\<lbrakk>P\<rbrakk>\<^sub>(\<lambda>s\<^sub>s\<^sub>y\<^sub>n. (f s\<^sub>s\<^sub>y\<^sub>n)(x:=v))(s\<^sub>s\<^sub>y\<^sub>n\<^sub>0:=\<nu>)}}" by force
  also have "... = {S. \<exists>v. \<exists>\<nu>\<in>{s(x:=v)| s. s\<in>S}. {s(x:=v)| s. s\<in>S}\<in>\<lbrakk>P\<rbrakk>\<^sub>(\<lambda>s\<^sub>s\<^sub>y\<^sub>n. (f s\<^sub>s\<^sub>y\<^sub>n)(x:=v))(s\<^sub>s\<^sub>y\<^sub>n\<^sub>0:=\<nu>)}" by simp
  also have "... = {S. \<exists>s\<in>S. \<exists>v. {s(x:=v)| s. s\<in>S}\<in>\<lbrakk>P\<rbrakk>\<^sub>(\<lambda>s\<^sub>s\<^sub>y\<^sub>n. (f s\<^sub>s\<^sub>y\<^sub>n)(x:=v))(s\<^sub>s\<^sub>y\<^sub>n\<^sub>0:=s(x:=v))}" by fast
  also have "... = {S. \<exists>s\<in>S. \<exists>v. {s(x:=v)| s. s\<in>S}\<in>\<lbrakk>P\<rbrakk>\<^sub>(\<lambda>s\<^sub>s\<^sub>y\<^sub>n. ((f(s\<^sub>s\<^sub>y\<^sub>n\<^sub>0:=s)) s\<^sub>s\<^sub>y\<^sub>n)(x:=v))}" unfolding "0" by blast
  ultimately have "\<lbrakk>ExN x (ExS s\<^sub>s\<^sub>y\<^sub>n\<^sub>0 P)\<rbrakk>\<^sub>\<Sigma> = {S. \<exists>s\<in>S. \<exists>v. {s(x:=v)| s. s\<in>S}\<in>\<lbrakk>P\<rbrakk>\<^sub>(\<lambda>s\<^sub>s\<^sub>y\<^sub>n. ((f(s\<^sub>s\<^sub>y\<^sub>n\<^sub>0:=s)) s\<^sub>s\<^sub>y\<^sub>n)(x:=v))}" by argo
  then show ?thesis by auto
qed*)

lemma "\<forall>\<nu>. \<lbrakk>P\<rbrakk>\<^sub>\<Sigma>(s:=\<nu>) \<subseteq> \<lbrakk>Q\<rbrakk>\<^sub>\<Sigma>(s:=\<nu>) \<Longrightarrow> \<lbrakk>FaS s P\<rbrakk>\<^sub>\<Sigma> \<subseteq> \<lbrakk>FaS s Q\<rbrakk>\<^sub>\<Sigma>" by auto
lemma "\<forall>\<nu>. \<lbrakk>P\<rbrakk>\<^sub>\<Sigma>(s:=\<nu>) \<subseteq> \<lbrakk>Q\<rbrakk>\<^sub>\<Sigma>(s:=\<nu>) \<Longrightarrow> \<lbrakk>ExS s P\<rbrakk>\<^sub>\<Sigma> \<subseteq> \<lbrakk>ExS s Q\<rbrakk>\<^sub>\<Sigma>" by auto
lemma "{S. S\<inter>p\<noteq>{}}\<inter>{S. S\<inter>q\<noteq>{}} \<supseteq> {S. S\<inter>(p\<inter>q)\<noteq>{}}" by blast
lemma "{S. S\<subseteq>p \<and> S\<noteq>{}}\<inter>{S. S\<subseteq>q \<and> S\<noteq>{}} = {S. S\<subseteq>p\<inter>q \<and> S\<noteq>{}}" by auto
lemma "{S. S\<subseteq>p \<and> S\<noteq>{}}\<inter>{S. S\<subseteq>q} = {S. S\<subseteq>p\<inter>q \<and> S\<noteq>{}}" by auto


(*lemma
  assumes "x\<notin>dom f"
    shows "\<lbrakk>ExN x (RAsrt s\<^sub>1 s\<^sub>2 r)\<rbrakk>\<^sub>\<Sigma> \<subseteq> \<lbrakk>RAsrt s\<^sub>1 s\<^sub>2 (REx x r)\<rbrakk>\<^sub>\<Sigma>"
proof(-)
  have "\<lbrakk>RAsrt s\<^sub>1 s\<^sub>2 (REx x r)\<rbrakk>\<^sub>\<Sigma> = {S. (\<forall>x\<in>dom f. the (f x)=the (S s\<^sub>1) x \<and> the (f x)=the (S s\<^sub>2) x) \<and> (the (S s\<^sub>1), the (S s\<^sub>2))\<in>\<lbrakk>REx x r)}" by auto
  moreover have "... = {S. (\<forall>x\<in>dom f. the (f x)=the (S s\<^sub>1) x \<and> the (f x)=the (S s\<^sub>2) x) \<and> (the (S s\<^sub>1), the (S s\<^sub>2))\<in>{(s\<^sub>1, s\<^sub>2)| s\<^sub>1 s\<^sub>2. \<exists>n. (s\<^sub>1(x:=n), s\<^sub>2(x:=n))\<in>rinter r}}" by auto
  moreover have "... = {S. (\<forall>x\<in>dom f. the (f x)=the (S s\<^sub>1) x \<and> the (f x)=the (S s\<^sub>2) x) \<and> (\<exists>n. ((the (S s\<^sub>1))(x:=n), (the (S s\<^sub>2))(x:=n))\<in>rinter r)}" by simp
  moreover have "... = {S. \<exists>n. (\<forall>x\<in>dom f. the (f x)=the (S s\<^sub>1) x \<and> the (f x)=the (S s\<^sub>2) x) \<and> ((the (S s\<^sub>1))(x:=n), (the (S s\<^sub>2))(x:=n))\<in>rinter r}" by auto
  ultimately have 0: "\<lbrakk>RAsrt s\<^sub>1 s\<^sub>2 (REx x r)\<rbrakk>\<^sub>\<Sigma> = {S. \<exists>n. (\<forall>y\<in>dom f. the (f y)=the (S s\<^sub>1) y \<and> the (f y)=the (S s\<^sub>2) y) \<and> ((the (S s\<^sub>1))(x:=n), (the (S s\<^sub>2))(x:=n))\<in>rinter r}" by argo

  have "\<lbrakk>ExN x (RAsrt s\<^sub>1 s\<^sub>2 r)\<rbrakk>\<^sub>\<Sigma> = {S. \<exists>n. S\<in>\<lbrakk>RAsrt s\<^sub>1 s\<^sub>2 r\<rbrakk>\<^sub>\<Sigma>(x\<mapsto>n)}" by simp
  moreover have "... = {S. \<exists>n. S\<in>{S. (\<forall>y\<in>dom (f(x\<mapsto>n)). the ((f(x\<mapsto>n)) y)=the (S s\<^sub>1) y \<and> the ((f(x\<mapsto>n)) y)=the (S s\<^sub>2) y) \<and> (the (S s\<^sub>1), the (S s\<^sub>2))\<in>rinter r}}" by force
  moreover have "... = {S. \<exists>n. (\<forall>y\<in>dom (f(x\<mapsto>n)). the ((f(x\<mapsto>n)) y)=the (S s\<^sub>1) y \<and> the ((f(x\<mapsto>n)) y)=the (S s\<^sub>2) y) \<and> (the (S s\<^sub>1), the (S s\<^sub>2))\<in>rinter r}" by blast
  moreover have "... = {S. \<exists>n. (\<forall>y\<in>dom f. the (f y)=the (S s\<^sub>1) y \<and> the (f y)=the (S s\<^sub>2) y) \<and> the (S s\<^sub>1) x=the (S s\<^sub>2) x \<and> (the (S s\<^sub>1), the (S s\<^sub>2))\<in>rinter r}" using assms by auto
  ultimately have 1: "\<lbrakk>ExN x (RAsrt s\<^sub>1 s\<^sub>2 r)\<rbrakk>\<^sub>\<Sigma> = {S. \<exists>n. (\<forall>y\<in>dom f. the (f y)=the (S s\<^sub>1) y \<and> the (f y)=the (S s\<^sub>2) y) \<and> the (S s\<^sub>1) x=the (S s\<^sub>2) x \<and> (the (S s\<^sub>1), the (S s\<^sub>2))\<in>rinter r}" by presburger

  show ?thesis unfolding "0" "1" by (smt (verit, del_insts) Collect_mono_iff \<open>\<lbrakk>ExN x (RAsrt s\<^sub>1 s\<^sub>2 r)\<rbrakk>\<^sub>\<Sigma> = {S. \<exists>n. S \<in> \<lbrakk>RAsrt s\<^sub>1 s\<^sub>2 r\<rbrakk>\<^sub>\<Sigma> (x \<mapsto> n)}\<close> assms domD domI fun_upd_same fun_upd_triv fun_upd_twist inter.simps(4) mem_Collect_eq)
qed*)

(*lemma ex_comm:
  "\<lbrakk>ExN x (ExS s P)\<rbrakk>\<^sub>\<Sigma> \<subseteq> \<lbrakk>ExS s (ExN x P)\<rbrakk>\<^sub>\<Sigma>" by auto
proof(induction P arbitrary: f)
  case Top
  then show ?case by simp
next
  case Bot
  then show ?case by simp
next
  case (Asrt s\<^sub>0 p)
  then show ?case
  proof(cases "s=s\<^sub>0")
    case True
    then show ?thesis by auto
  next
    case False
    then show ?thesis by auto
  qed
next
  case (RAsrt s\<^sub>1 s\<^sub>2 r)
  have "s=s\<^sub>1 \<Longrightarrow> ?case" by auto
  moreover have "s=s\<^sub>2 \<Longrightarrow> ?case" 
  moreover have "s\<noteq>s\<^sub>1 \<Longrightarrow> s\<noteq>s\<^sub>2 \<Longrightarrow> ?case" by auto
  ultimately show ?case by argo
next
  case (ExN y P)
  then show ?case by auto
next
  case (FaN y P)
  then show ?case by auto
next
  case (ExS s' P)
  then show ?case by auto
next
  case (FaS s' P)
  then show ?case by auto
next
  case (And P Q)
  then show ?case by auto
next
  case (Or P Q)
  then show ?case by auto
qed*)

(*lemma fa_comm:
  "\<lbrakk>FaN x (FaS s P)\<rbrakk>\<^sub>\<Sigma> = \<lbrakk>FaS s (FaN x P)\<rbrakk>\<^sub>\<Sigma>" by auto
proof(induction P arbitrary: f)
  case Top
  then show ?case by simp
next
  case Bot
  then show ?case by simp
next
  case (Asrt s\<^sub>0 p)
  then show ?case
  proof(cases "s=s\<^sub>0")
    case True
    then show ?thesis by auto
  next
    case False
    then show ?thesis by auto
  qed
next
  case (RAsrt s\<^sub>1 s\<^sub>2 r)
  have "s=s\<^sub>1 \<Longrightarrow> ?case" by auto
  moreover have "s=s\<^sub>2 \<Longrightarrow> ?case" by auto 
  moreover have "s\<noteq>s\<^sub>1 \<Longrightarrow> s\<noteq>s\<^sub>2 \<Longrightarrow> ?case" by auto
  ultimately show ?case by argo
next
  case (ExN y P)
  then show ?case by auto
next
  case (FaN y P)
  then show ?case by auto
next
  case (ExS s' P)
  then show ?case by auto
next
  case (FaS s' P)
  then show ?case by auto
next
  case (And P Q)
  then show ?case by auto
next
  case (Or P Q)                            
  then show ?case by auto
qed*)

(*lemma "(\<forall>s s'. (s, s')\<notin>P \<longrightarrow> (\<forall>s''. (s, s'')\<notin>P)) \<longleftrightarrow> (\<forall>s s' s''. (s, s')\<in>P \<or> (s, s'')\<notin>P)" by blast
lemma "(\<forall>s s'. s\<in>P \<or> s'\<notin>P) \<longleftrightarrow> (P={} \<or> P=UNIV)" by blast
lemma "(\<forall>s s' s''. (s, s')\<in>P \<or> (s, s'')\<notin>P) \<longleftrightarrow> (\<forall>s. snd`{(s, s')| s'. (s, s')\<in>P}={} \<or> snd`{(s, s')| s'. (s, s')\<in>P}=UNIV)" by force
lemma "\<nexists>P. \<forall>p q. (C\<subseteq>P p q \<longleftrightarrow> (\<forall>s\<in>p. \<exists>s'\<in>q. (s, s')\<in>C))" nitpick oops
lemma "\<nexists>P. \<forall>p C q. (C\<subseteq>P p q \<longleftrightarrow> (\<forall>s\<in>p. \<exists>s'\<in>q. (s, s')\<in>C))" by blast
lemma "\<forall>P. \<exists>p q. {}\<subseteq>P p q \<and> (\<exists>s\<in>p. \<forall>s'\<in>q. (s, s')\<notin>{})" by blast
lemma "(C\<subseteq>{(s, s')| s s'. s\<in>p \<longrightarrow> s'\<in>q}) \<longleftrightarrow> (\<forall>s\<in>p. \<forall>s'. (s, s')\<in>C \<longrightarrow> s'\<in>q)" by auto
lemma "\<exists>P. \<forall>p C q. (C\<subseteq>P p q) \<longleftrightarrow> (\<forall>s\<in>p. \<forall>s'. (s, s')\<in>C \<longrightarrow> s'\<in>q)"
proof
  show "\<forall>p C q. (C\<subseteq>(\<lambda>p q. {(s, s')| s s'. s\<in>p \<longrightarrow> s'\<in>q}) p q) = (\<forall>s\<in>p. \<forall>s'. (s, s') \<in> C \<longrightarrow> s' \<in> q)" by fast
qed
lemma x: "\<not>(\<exists>P. \<forall>C. (\<forall>s::bool. \<exists>s'::bool. (s, s')\<in>C) \<longrightarrow> ((\<forall>s. \<exists>s'\<noteq>x. (s, s')\<in>C) \<longleftrightarrow> C\<subseteq>P))" by auto

lemma "\<exists>C. (\<forall>s::bool. \<exists>s'::bool. (s, s')\<in>C) \<and> \<not>((\<forall>s. \<exists>s'\<noteq>x. (s, s')\<in>C) \<longleftrightarrow> C\<subseteq>P) \<and> C\<noteq>{}" by auto
lemma xx\<^sub>0: "\<exists>C::(nat\<times>nat) set. \<not>((\<forall>s. \<exists>s'. (s, s')\<in>C) \<longleftrightarrow> C\<subseteq>P) \<and> C\<noteq>{}"
proof -
  obtain f :: "nat \<Rightarrow> nat" where f1: "\<forall>n. n < f n" by auto
  then have "\<forall>p. p \<in> Field pair_leq"
    by (metis (no_types) Field_iff pair_leqI1 surj_pair)
  then show ?thesis
    using f1 by (metis Id_on_iff equals0D lessThan_iff less_not_refl3 subrelI subset_antisym)
qed
lemma
  assumes "\<forall>C. (\<forall>s. \<exists>s'\<noteq>Div. (s, s')\<in>C) \<longrightarrow> C\<subseteq>P"
  shows "{(s, s')| s s'. s'\<noteq>Div}\<subseteq>P"
  using assms by blast

lemma 
  assumes "\<forall>C. (\<forall>s. \<exists>s'\<noteq>Div. (s, s')\<in>C) \<longrightarrow> C\<subseteq>P"
      and "Div\<noteq>nDiv"
  shows "P=UNIV" using assms by fast 

lemma 
  assumes "P=UNIV"
      and "Div\<noteq>nDiv"
    shows "\<exists>C\<subseteq>P. (\<forall>s. \<exists>s'. (s, s')\<in>C) \<and> (\<exists>s. \<forall>s'\<noteq>Div. (s, s')\<notin>C)" 
proof(-)
  let ?C = "{(s, s')| (s::'a) (s'::'b). s'=Div}"
  have "\<forall>s. \<exists>s'. (s, s')\<in>?C" using assms (2) by force
  moreover have "\<exists>s. \<forall>s'\<noteq>Div. (s, s')\<notin>?C" by blast
  ultimately show ?thesis using assms(1) by blast
qed*)

(*There is no difference for SSIL between \<inter> and \<otimes>, but there is a difference for OSL between them*)
definition join (infix "\<otimes>" 52) where
 "P \<otimes> Q = {S\<^sub>P\<union>S\<^sub>Q| S\<^sub>P S\<^sub>Q. S\<^sub>P\<in>P \<and> S\<^sub>Q\<in>Q}"
lemma join_SL_encs:
  "{S. \<forall>s\<in>S. s x=1}\<otimes>{S. \<forall>s\<in>S. s x=2}={S. \<forall>s\<in>S. s x=1 \<or> s x=2}"
proof
  show "{S. \<forall>s\<in>S. s x=1}\<otimes>{S. \<forall>s\<in>S. s x=2}\<subseteq>{S. \<forall>s\<in>S. s x=1 \<or> s x=2}" unfolding join_def by fast
  show "{S. \<forall>s\<in>S. s x=1 \<or> s x=2}\<subseteq>{S. \<forall>s\<in>S. s x=1}\<otimes>{S. \<forall>s\<in>S. s x=2}"
  proof
    fix S :: "('a \<Rightarrow> 'c) set"
    assume asm: "S\<in>{S. \<forall>s\<in>S. s x=1 \<or> s x=2}"
    then obtain S\<^sub>p S\<^sub>q where "S\<^sub>p=S\<inter>{s. s x=1}" "S\<^sub>q=S\<inter>{s. s x=2}" by presburger
    then have "S\<^sub>p\<in>{S. \<forall>s\<in>S. s x=1}" "S\<^sub>q\<in>{S. \<forall>s\<in>S. s x=2}" "S=S\<^sub>p\<union>S\<^sub>q" using asm by auto
    then show "S\<in>{S. \<forall>s\<in>S. s x=1}\<otimes>{S. \<forall>s\<in>S. s x=2}" unfolding join_def by blast
  qed
qed
lemma "\<lbrakk>ExS s (Asrt s p)\<rbrakk>\<^sub>\<Sigma> = {S. S\<inter>\<lbrakk>p\<rbrakk>\<^sub>a\<noteq>{}}" by auto
lemma "\<lbrakk>And (ExS s (Asrt s p)) (ExS s (Asrt s q))\<rbrakk>\<^sub>\<Sigma> = {S. S\<inter>\<lbrakk>p\<rbrakk>\<^sub>a\<noteq>{} \<and> S\<inter>\<lbrakk>q\<rbrakk>\<^sub>a\<noteq>{}}" by auto
lemma "{S. S\<inter>p\<noteq>{}}\<otimes>{S. S\<inter>q\<noteq>{}}={S. S\<inter>p\<noteq>{} \<and> S\<inter>q\<noteq>{}}" unfolding join_def by auto

lemma "\<lbrakk>And (FaS s (Asrt s p)) (ExS s (Asrt s p))\<rbrakk>\<^sub>\<Sigma> = {S. S\<subseteq>\<lbrakk>p\<rbrakk>\<^sub>a \<and> S\<inter>\<lbrakk>p\<rbrakk>\<^sub>a\<noteq>{}}" by auto
lemma "\<lbrakk>And (And (FaS s (Asrt s p)) (ExS s (Asrt s p))) (And (FaS s (Asrt s q)) (ExS s (Asrt s q)))\<rbrakk>\<^sub>\<Sigma> = {S. S\<subseteq>\<lbrakk>p\<rbrakk>\<^sub>a\<inter>\<lbrakk>q\<rbrakk>\<^sub>a \<and> S\<inter>\<lbrakk>p\<rbrakk>\<^sub>a\<noteq>{} \<and> S\<inter>\<lbrakk>q\<rbrakk>\<^sub>a\<noteq>{}}" by auto
lemma join_OSL_encs: "{S. S\<subseteq>p \<and> S\<inter>p\<noteq>{}}\<otimes>{S. S\<subseteq>q \<and> S\<inter>q\<noteq>{}}={S. S\<subseteq>p\<union>q \<and> S\<inter>p\<noteq>{} \<and> S\<inter>q\<noteq>{}}"
proof
  show "{S. S\<subseteq>p \<and> S\<inter>p\<noteq>{}}\<otimes>{S. S\<subseteq>q \<and> S\<inter>q\<noteq>{}}\<subseteq>{S. S\<subseteq>p\<union>q \<and> S\<inter>p\<noteq>{} \<and> S\<inter>q\<noteq>{}}" unfolding join_def by auto
  show "{S. S\<subseteq>p\<union>q \<and> S\<inter>p\<noteq>{} \<and> S\<inter>q\<noteq>{}}\<subseteq>{S. S\<subseteq>p \<and> S\<inter>p\<noteq>{}}\<otimes>{S. S\<subseteq>q \<and> S\<inter>q\<noteq>{}}"
  proof
    fix S
    assume asm: "S\<in>{S. S\<subseteq>p\<union>q \<and> S\<inter>p\<noteq>{} \<and> S\<inter>q\<noteq>{}}"
    then obtain S\<^sub>p S\<^sub>q where "S\<^sub>p=S\<inter>p" "S\<^sub>q=S\<inter>q" by presburger
    then have "S\<^sub>p\<in>{S. S\<subseteq>p \<and> S\<inter>p\<noteq>{}}" "S\<^sub>q\<in>{S. S\<subseteq>q \<and> S\<inter>q\<noteq>{}}" "S=S\<^sub>p\<union>S\<^sub>q" using asm by auto
    then show "S\<in>{S. S\<subseteq>p \<and> S\<inter>p\<noteq>{}}\<otimes>{S. S\<subseteq>q \<and> S\<inter>q\<noteq>{}}" unfolding join_def by blast
  qed
qed

end