theory HyperFrameRule
  imports HyperStar
begin

(*Note that the condition fv(f)\<inter>rd(C)={} is not added*)
definition HSL_validity :: "hassertion \<Rightarrow> var cmd => hassertion \<Rightarrow> bool" ("\<Turnstile>HSL {_} _ {_}" [0,0,0] 81) where
  "\<Turnstile>HSL {P} C {Q} \<longleftrightarrow> (\<forall>f. \<forall>S\<in>P\<star>{S. S\<subseteq>f}. \<lbrakk>C\<rbrakk>``S\<in>Q\<star>{S. S\<subseteq>f})"

lemma HSL_validityI:
  assumes "\<And>f S. S\<in>P\<star>{S. S\<subseteq>f} \<Longrightarrow> \<lbrakk>C\<rbrakk>``S\<in>Q\<star>{S. S\<subseteq>f}"
    shows "\<Turnstile>HSL {P} C {Q}"
  using HSL_validity_def assms by blast

lemma HSL_validityE:
  assumes "\<Turnstile>HSL {P} C {Q}"
      and "S\<in>P\<star>{S. S\<subseteq>f}"
    shows "\<lbrakk>C\<rbrakk>``S\<in>Q\<star>{S. S\<subseteq>f}"
  using HSL_validity_def assms by blast

lemma HSL_validity_useful:
  assumes "\<Turnstile>HSL {P} C {Q}"
      and "S\<in>P"
    shows "\<lbrakk>C\<rbrakk>``S\<in>Q"
proof(-)
  have "S\<in>P\<star>{S. S\<subseteq>{(s, fmap_empty)| s. True}}" using assms(2) hstar_unit by simp
  then have "\<lbrakk>C\<rbrakk>``S\<in>Q\<star>{S. S\<subseteq>{(s, fmap_empty)| s. True}}" using assms(1) HSL_validityE by force
  then show ?thesis using hstar_unit by auto
qed

lemma Empty_axiom_sound:
  "\<Turnstile>HSL {{{}}} C {{{}}}"
proof(rule HSL_validityI)
  fix S f
  assume "S\<in>{{}}\<star>{S. S\<subseteq>f}"
  then have "S={}" unfolding hstar_def left_lives_def right_lives_equiv star_def hAdd_def by fast
  then have "\<lbrakk>C\<rbrakk>``S={}" by fast
  then show "\<lbrakk>C\<rbrakk>``S\<in>{{}}\<star>{S. S\<subseteq>f}" unfolding hstar_def left_lives_def right_lives_equiv star_def hAdd_def by blast
qed

lemma Cons_rule_sound:
  assumes"\<Turnstile>HSL {P} C {Q}"
      and "P'\<subseteq>P"
      and "Q\<subseteq>Q'"
    shows "\<Turnstile>HSL {P'} C {Q'}"
proof(rule HSL_validityI)
  fix S f
  assume "S\<in>P'\<star>{S. S\<subseteq>f}"
  then have "S\<in>P\<star>{S. S\<subseteq>f}" by (metis (mono_tags, lifting) Collect_mono assms(2) in_mono hstar_mono)
  then have "\<lbrakk>C\<rbrakk>``S\<in>Q\<star>{S. S\<subseteq>f}" using assms(1) HSL_validityE by fast
  then show "\<lbrakk>C\<rbrakk>``S\<in>Q'\<star>{S. S\<subseteq>f}" by (metis (mono_tags, lifting) Collect_mono assms(3) in_mono hstar_mono)
qed

lemma Union_rule_sound:
  assumes "\<Turnstile>HSL {P} C {Q}"
      and "\<Turnstile>HSL {P'} C {Q'}"
    shows "\<Turnstile>HSL {P\<union>P'} C {Q\<union>Q'}"
proof(rule HSL_validityI)
  fix S f
  assume "S\<in>(P\<union>P')\<star>{S. S\<subseteq>f}"
  then have 0: "S\<in>(P\<star>{S. S\<subseteq>f})\<union>(P'\<star>{S. S\<subseteq>f})" by (simp add: hstar_distr_over_union)
  then show "\<lbrakk>C\<rbrakk>``S\<in>(Q\<union>Q')\<star>{S. S\<subseteq>f}"
  proof(cases "S\<in>P\<star>{S. S\<subseteq>f}")
    case True
    then have "\<lbrakk>C\<rbrakk>``S\<in>Q\<star>{S. S\<subseteq>f}" using assms(1) HSL_validityE by blast
    then show ?thesis using hstar_mono by (meson subset_iff sup.cobounded1)
  next
    case False
    then have "S\<in>P'\<star>{S. S\<subseteq>f}" using "0" by fast
    then have "\<lbrakk>C\<rbrakk>``S\<in>Q'\<star>{S. S\<subseteq>f}" using assms(2) HSL_validityE by blast
    then show ?thesis using hstar_mono by (meson subset_iff sup.cobounded2)
  qed
qed

lemma Indexed_union_rule_sound:
  assumes "\<And>i. i\<in>I \<Longrightarrow> \<Turnstile>HSL {Ps i} C {Qs i}"
  shows "\<Turnstile>HSL {\<Union>i\<in>I. Ps i} C {\<Union>i\<in>I. Qs i}"
proof(rule HSL_validityI)
  fix S f
  assume "S\<in>(\<Union>i\<in>I. Ps i)\<star>{S. S\<subseteq>f}"
  then obtain i where i_def: "i\<in>I" "S\<in>(Ps i)\<star>{S. S\<subseteq>f}" using hstar_def by auto
  then obtain P Q where PQ_def: "P=Ps i" "Q=Qs i" by blast
  then have "S\<in>(Ps i)\<star>{S. S\<subseteq>f}" by (simp add: i_def(2)) 
  then have "\<lbrakk>C\<rbrakk>``S\<in>Q\<star>{S. S\<subseteq>f}" using HSL_validity_def PQ_def(2) assms i_def(1) by blast
  then show "\<lbrakk>C\<rbrakk>``S\<in>(\<Union>i\<in>I. Qs i)\<star>{S. S\<subseteq>f}" using hstar_distr_over_indexed_union PQ_def(2) i_def(1) by auto
qed


lemma hstar_distr_over_indexed_union_of_powersets: "P\<star>(\<Union>q\<in>U. {S. S\<subseteq>q})=(\<Union>q\<in>U. P\<star>{S. S\<subseteq>q})" using hstar_def hstar_distr_over_indexed_union by auto

(*Note that the condition fv(f)\<inter>rd(C)={} is not added*)
lemma Frame_rule_sound:
  assumes "\<Turnstile>HSL {P} C {Q}"
      and "\<And>S S'. S\<in>F \<Longrightarrow> S'\<subseteq>S \<Longrightarrow> S'\<in>F"
    shows "\<Turnstile>HSL {P\<star>F} C {Q\<star>F}"
proof(-)
  have "F=(\<Union>f\<in>F. {S. S\<subseteq>f})" using assms(2) by blast
  have "\<forall>f\<in>F. \<Turnstile>HSL {P\<star>{S. S\<subseteq>f}} C {Q\<star>{S. S\<subseteq>f}}" using hstar_assoc hstar_expresses_SL_star assms(1) unfolding HSL_validity_def by simp
  moreover obtain Ps Qs where PsQs_def: "Ps=(\<lambda>f. P\<star>{S. S\<subseteq>f})" "Qs=(\<lambda>f. Q\<star>{S. S\<subseteq>f})" by simp
  ultimately have "\<And>f. f\<in>F \<Longrightarrow> \<Turnstile>HSL {Ps f} C {Qs f}" by fast
  then have "\<Turnstile>HSL {\<Union>f\<in>F. Ps f} C {\<Union>f\<in>F. Qs f}" using Indexed_union_rule_sound by blast
  moreover have "(\<Union>f\<in>F. Ps f)=(\<Union>f\<in>F. P\<star>{S. S\<subseteq>f})" using PsQs_def by blast
  then have "(\<Union>f\<in>F. Ps f)=P\<star>(\<Union>f\<in>F. {S. S\<subseteq>f})" using hstar_distr_over_indexed_union_of_powersets by presburger
  then have "(\<Union>f\<in>F. Ps f)=P\<star>F" using \<open>F=(\<Union>f\<in>F. {S. S\<subseteq>f})\<close> by auto
  moreover have "(\<Union>f\<in>F. Qs f)=(\<Union>f\<in>F. Q\<star>{S. S\<subseteq>f})" using PsQs_def by blast
  then have "(\<Union>f\<in>F. Qs f)=Q\<star>(\<Union>f\<in>F. {S. S\<subseteq>f})" using hstar_distr_over_indexed_union_of_powersets by presburger
  then have "(\<Union>f\<in>F. Qs f)=Q\<star>F" using \<open>F=(\<Union>f\<in>F. {S. S\<subseteq>f})\<close> by auto
  ultimately show ?thesis by argo
qed

lemma Join_rule_sound:
  assumes "\<Turnstile>HSL {P} C {Q}"
      and "\<Turnstile>HSL {P'} C {Q'}"
    shows "\<Turnstile>HSL {P\<otimes>P'} C {Q\<otimes>Q'}"
proof(rule HSL_validityI)
  fix S f
  assume asm: "S\<in>(P\<otimes>P')\<star>{S. S\<subseteq>f}"
  then have "S\<in>(P\<star>{S. S\<subseteq>f})\<otimes>(P'\<star>{S. S\<subseteq>f})" using hstar_distr_over_join_powerset by blast
  then obtain S\<^sub>P S\<^sub>P' where "S=S\<^sub>P\<union>S\<^sub>P'" "S\<^sub>P\<in>P\<star>{S. S\<subseteq>f}" "S\<^sub>P'\<in>P'\<star>{S. S\<subseteq>f}" unfolding join_def by auto
  then have "\<lbrakk>C\<rbrakk>``S\<^sub>P\<in>Q\<star>{S. S\<subseteq>f}" "\<lbrakk>C\<rbrakk>``S\<^sub>P'\<in>Q'\<star>{S. S\<subseteq>f}" using HSL_validityE assms(1) assms(2) by auto
  moreover have "\<lbrakk>C\<rbrakk>``S = \<lbrakk>C\<rbrakk>``S\<^sub>P \<union> \<lbrakk>C\<rbrakk>``S\<^sub>P'" by (simp add: Image_Un \<open>S = S\<^sub>P \<union> S\<^sub>P'\<close>)
  ultimately show "\<lbrakk>C\<rbrakk>``S \<in> (Q \<otimes> Q') \<star> {S. S \<subseteq> f}" using hstar_distr_over_join_powerset[of Q Q' f] unfolding join_def by blast
qed

lemma Indexed_join_rule_sound:
  assumes "\<And>i. i\<in>I \<Longrightarrow> \<Turnstile>HSL {Ps i} C {Qs i}"
    shows "\<Turnstile>HSL {\<Otimes>\<^sub>I Ps} C {\<Otimes>\<^sub>I Qs}"
proof(rule HSL_validityI)
  fix S f
  assume "S\<in>(\<Otimes>\<^sub>I Ps)\<star>{S. S\<subseteq>f}"
  then have "S\<in>\<Otimes>\<^sub>I (\<lambda> i. Ps i\<star>{S. S\<subseteq>f})" using hstar_distr_over_indexed_join_powerset by auto
  then obtain g where g_def: "S=(\<Union>i\<in>I. g i)" "\<forall>i\<in>I. g i\<in>Ps i\<star>{S. S\<subseteq>f}" unfolding join_indexed_def by blast
  then have "\<forall>i\<in>I. \<lbrakk>C\<rbrakk>``(g i)\<in>Qs i\<star>{S. S\<subseteq>f}" using assms HSL_validityE by fast
  moreover have "\<lbrakk>C\<rbrakk>``S = (\<Union>i\<in>I. \<lbrakk>C\<rbrakk>``(g i))" using g_def(1) by blast
  ultimately have "\<lbrakk>C\<rbrakk>``S\<in>\<Otimes>\<^sub>I (\<lambda> i. Qs i\<star>{S. S\<subseteq>f})" unfolding join_indexed_def by blast
  then show "\<lbrakk>C\<rbrakk>``S\<in>(\<Otimes>\<^sub>I Qs)\<star>{S. S\<subseteq>f}" using hstar_distr_over_indexed_join_powerset by auto
qed

lemma Joinion_precondition_rule_1D_sound: \<comment> \<open>From all 1x1 subrelations of p, shows for all (rectangular) subrelations of p.. in 1D all subsets are rectangular\<close>
  assumes "\<Turnstile>HSL {{{\<sigma>}| \<sigma>. \<sigma>\<in>p}} C {{S. S\<subseteq>q}}"
    shows "\<Turnstile>HSL {{S. S\<subseteq>p}} C {{S. S\<subseteq>q}}"
proof(-)
  have "\<Turnstile>HSL {{{\<sigma>}| \<sigma>. \<sigma>\<in>p}\<union>{{}}} C {{S. S\<subseteq>q}\<union>{{}}}" using assms Empty_axiom_sound Union_rule_sound by blast 
  moreover have "{S. S\<subseteq>q}\<union>{{}}={S. S\<subseteq>q}" by fastforce
  ultimately have "\<Turnstile>HSL {{{\<sigma>}| \<sigma>. \<sigma>\<in>p}\<union>{{}}} C {{S. S\<subseteq>q}}" by fastforce
  moreover obtain Ps Qs where PsQs_def:"Ps=(\<lambda>\<sigma>::nstate. {{}, {\<sigma>}})" "Qs=(\<lambda>\<sigma>::nstate. {S. S\<subseteq>q})" by simp
  moreover have "\<And>\<sigma>. \<sigma>\<in>p \<Longrightarrow> Ps \<sigma>\<subseteq>{{\<sigma>}| \<sigma>. \<sigma>\<in>p}\<union>{{}}" unfolding PsQs_def(1) by blast
  ultimately have "\<And>\<sigma>. \<sigma>\<in>p \<Longrightarrow> \<Turnstile>HSL {Ps \<sigma>} C {Qs \<sigma>}" using Cons_rule_sound by auto
  then have "\<Turnstile>HSL {\<Otimes>\<^sub>p Ps} C {\<Otimes>\<^sub>p Qs}" using Indexed_join_rule_sound by blast
  moreover have "{S. S\<subseteq>p}\<subseteq>\<Otimes>\<^sub>p Ps" (*they're even equal*)
  proof
    fix S
    assume "S\<in>{S. S\<subseteq>p}"
    then have "S\<subseteq>p" by fast
    obtain f where f_def: "f=(\<lambda>\<sigma>::nstate. if \<sigma>\<in>S then {\<sigma>} else {})" by force
    then have "S=(\<Union>\<sigma>\<in>p. f \<sigma>)" "\<forall>\<sigma>\<in>p. f \<sigma>\<in>Ps \<sigma>" using \<open>S\<subseteq>p\<close> unfolding PsQs_def(1) f_def by auto
    then show "S\<in>\<Otimes>\<^sub>p Ps" unfolding join_indexed_def by blast
  qed
  moreover have "\<Otimes>\<^sub>p Qs\<subseteq>{S. S\<subseteq>q}" unfolding PsQs_def(2) join_indexed_def by fast (*it could be {}\<subset>{S. S\<subseteq>q} if p={}*)
  ultimately show ?thesis using Cons_rule_sound by simp
qed

lemma Joinion_precondition_rule_1D_sound': \<comment> \<open>From all 1x1 subrelations of p, shows for all (rectangular) subrelations of p.. in 1D all subsets are rectangular\<close>
  assumes "\<Turnstile>HSL {{{\<sigma>}| \<sigma>. \<sigma>\<in>p}} C {{S. S\<subseteq>q}}"
    shows "\<Turnstile>HSL {{S. S\<subseteq>p}} C {{S. S\<subseteq>q}}"
proof(-)
  have "\<And>p\<^sub>0. p\<^sub>0\<subseteq>p \<Longrightarrow> \<Turnstile>HSL {{p\<^sub>0}} C {{S. S\<subseteq>q}}"
  proof(-)
    fix p\<^sub>0
    assume asm: "p\<^sub>0\<subseteq>p"
    moreover obtain Ps Qs where PsQs_def:"Ps=(\<lambda>\<sigma>::nstate. {{\<sigma>}})" "Qs=(\<lambda>\<sigma>::nstate. {S. S\<subseteq>q})" by simp
    ultimately have "\<And>\<sigma>. \<sigma>\<in>p\<^sub>0 \<Longrightarrow> \<Turnstile>HSL {Ps \<sigma>} C {Qs \<sigma>}" using assms Cons_rule_sound by auto 
    then have "\<Turnstile>HSL {\<Otimes>\<^sub>p\<^sub>0 Ps} C {\<Otimes>\<^sub>p\<^sub>0 Qs}" using Indexed_join_rule_sound by blast
    moreover have "\<Otimes>\<^sub>p\<^sub>0 Ps={p\<^sub>0}" unfolding join_indexed_def PsQs_def(1) by auto
    moreover have "\<Otimes>\<^sub>p\<^sub>0 Qs\<subseteq>{S. S\<subseteq>q}" unfolding join_indexed_def PsQs_def(2) by fast
    ultimately show "\<Turnstile>HSL {{p\<^sub>0}} C {{S. S\<subseteq>q}}" using Cons_rule_sound by force
  qed
  then have "\<And>p\<^sub>0. p\<^sub>0\<in>{S. S\<subseteq>p} \<Longrightarrow> \<Turnstile>HSL {{p\<^sub>0}} C {{S. S\<subseteq>q}}" by force
  moreover obtain Ps Qs where PsQs_def: "Ps=(\<lambda>p::assertion. {p})" "Qs=(\<lambda>p::assertion. {S. S\<subseteq>q})" by force
  ultimately have "\<And>p\<^sub>0. p\<^sub>0\<in>{S. S\<subseteq>p} \<Longrightarrow> \<Turnstile>HSL {Ps p\<^sub>0} C {Qs p\<^sub>0}" by blast
  then have "\<Turnstile>HSL {\<Union>p\<^sub>0\<in>{S. S\<subseteq>p}. Ps p\<^sub>0} C {\<Union>p\<^sub>0\<in>{S. S\<subseteq>p}. Qs p\<^sub>0}" using Indexed_union_rule_sound[of "{S. S\<subseteq>p}" Ps C Qs] by fast
  moreover have "(\<Union>p\<^sub>0\<in>{S. S\<subseteq>p}. Ps p\<^sub>0)={S. S\<subseteq>p}" unfolding PsQs_def(1) by fastforce
  moreover have "(\<Union>p\<^sub>0\<in>{S. S\<subseteq>p}. Qs p\<^sub>0)\<subseteq>{S. S\<subseteq>q}" unfolding PsQs_def(2) by blast
  ultimately show ?thesis using Cons_rule_sound[of "\<Union>p\<^sub>0\<in>{S. S\<subseteq>p}. Ps p\<^sub>0" C "\<Union>p\<^sub>0\<in>{S. S\<subseteq>p}. Qs p\<^sub>0" "{S. S\<subseteq>p}" "{S. S\<subseteq>q}"] by blast
qed

lemma "(x1 \<longrightarrow> em) \<and> (x2 \<longrightarrow> mem2) \<equiv> (x1 \<or> x2) \<longrightarrow> (mem2 \<longrightarrow> x1 \<and> \<not>em) \<longrightarrow> (em \<and> \<not>x2)" by linarith

lemma Joinion_precondition_rule_2D_sound:
  assumes "\<Turnstile>HSL {{{\<sigma>\<^sub>1, \<sigma>\<^sub>2}| \<sigma>\<^sub>1 \<sigma>\<^sub>2. (\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p}} C {{S. {\<sigma>\<in>S. fst \<sigma> t=1}\<times>{\<sigma>\<in>S. fst \<sigma> t=2}\<subseteq>q}}"
      and "\<And>\<sigma>\<^sub>1 \<sigma>\<^sub>2. (\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p \<Longrightarrow> fst \<sigma>\<^sub>1 t=1 \<and> fst \<sigma>\<^sub>2 t=2"
      and "t\<notin>md C"
    shows "\<Turnstile>HSL {{S. {\<sigma>\<in>S. fst \<sigma> t=1}\<times>{\<sigma>\<in>S. fst \<sigma> t=2}\<subseteq>p}} C {{S. {\<sigma>\<in>S. fst \<sigma> t=1}\<times>{\<sigma>\<in>S. fst \<sigma> t=2}\<subseteq>q}}"
proof(cases "p={}")
  have "\<Turnstile>HSL {{{\<sigma>\<^sub>1, \<sigma>\<^sub>2}| \<sigma>\<^sub>1 \<sigma>\<^sub>2. (\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p}\<union>{{}}} C {{S. {\<sigma>\<in>S. fst \<sigma> t=1}\<times>{\<sigma>\<in>S. fst \<sigma> t=2}\<subseteq>q}\<union>{{}}}" using assms Empty_axiom_sound Union_rule_sound by blast
  moreover have "{S. {\<sigma>\<in>S. fst \<sigma> t=1}\<times>{\<sigma>\<in>S. fst \<sigma> t=2}\<subseteq>q}\<union>{{}}={S. {\<sigma>\<in>S. fst \<sigma> t=1}\<times>{\<sigma>\<in>S. fst \<sigma> t=2}\<subseteq>q}" by fast
  ultimately have "\<Turnstile>HSL {{{\<sigma>\<^sub>1, \<sigma>\<^sub>2}| \<sigma>\<^sub>1 \<sigma>\<^sub>2. (\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p}\<union>{{}}} C {{S. {\<sigma>\<in>S. fst \<sigma> t=1}\<times>{\<sigma>\<in>S. fst \<sigma> t=2}\<subseteq>q}}" by argo
  moreover have "\<And>\<sigma>\<^sub>1 \<sigma>\<^sub>2. (\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p \<Longrightarrow> {{}, {\<sigma>\<^sub>1, \<sigma>\<^sub>2}}\<subseteq>{{\<sigma>\<^sub>1, \<sigma>\<^sub>2}| \<sigma>\<^sub>1 \<sigma>\<^sub>2. (\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p}\<union>{{}}" by blast
  ultimately have 0: "\<And>\<sigma>\<^sub>1 \<sigma>\<^sub>2. (\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p \<Longrightarrow> \<Turnstile>HSL {{{}, {\<sigma>\<^sub>1, \<sigma>\<^sub>2}}} C {{S. {\<sigma>\<in>S. fst \<sigma> t=1}\<times>{\<sigma>\<in>S. fst \<sigma> t=2}\<subseteq>q}}" using Cons_rule_sound by auto

  have "\<And>\<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>Domain p \<Longrightarrow> True"
  proof(-)
    fix \<sigma>\<^sub>1
    assume asm: "\<sigma>\<^sub>1\<in>Domain p"
    then have "fst \<sigma>\<^sub>1 t=1" using assms(2) by fast
    obtain S\<^sub>1 where S\<^sub>1_def: "S\<^sub>1=\<lbrakk>C\<rbrakk>``{\<sigma>\<^sub>1}" by blast
    have "\<And>\<sigma>\<^sub>2. (\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p \<Longrightarrow> \<Turnstile>HSL {{{}, {\<sigma>\<^sub>1, \<sigma>\<^sub>2}}} C {{S. {\<sigma>\<in>S. fst \<sigma> t=1}\<times>{\<sigma>\<in>S. fst \<sigma> t=2}\<subseteq>q \<and> {\<sigma>\<in>S. fst \<sigma> t=1}\<subseteq>S\<^sub>1}}"
    proof(rule HSL_validityI)
      fix \<sigma>\<^sub>2 S f
      assume asm': "(\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p" "S\<in>{{}, {\<sigma>\<^sub>1, \<sigma>\<^sub>2}}\<star>{S. S\<subseteq>f}"
      then obtain S\<^sub>p where S\<^sub>p_def: "S\<subseteq>S\<^sub>p\<^emph>f" "S\<^sub>p\<in>{{}, {\<sigma>\<^sub>1, \<sigma>\<^sub>2}}" "left_lives S S\<^sub>p f" using hstar_char_powerset by meson
      then show "\<lbrakk>C\<rbrakk>``S\<in>{S. {\<sigma>\<in>S. fst \<sigma> t=1}\<times>{\<sigma>\<in>S. fst \<sigma> t=2}\<subseteq>q \<and> {\<sigma>\<in>S. fst \<sigma> t=1}\<subseteq>S\<^sub>1}\<star>{S. S\<subseteq>f}"
      proof(cases "S\<^sub>p={}")
        case True
        then have "S={}" using S\<^sub>p_def(1) unfolding star_def by auto
        then have "\<lbrakk>C\<rbrakk>``S={}" unfolding dsem_def by fastforce
        then show ?thesis unfolding hstar_def star_def left_lives_def right_lives_equiv hAdd_def by blast
      next
        case False
        then have "S\<^sub>p={\<sigma>\<^sub>1, \<sigma>\<^sub>2}" using S\<^sub>p_def(2) by fast
        have "fst \<sigma>\<^sub>2 t=2" using asm'(1) assms(2) by blast
        then have "\<forall>\<sigma>\<in>{\<sigma>\<^sub>2}\<^emph>f. fst \<sigma> t=2" unfolding star_def by force
        moreover have "S\<subseteq>({\<sigma>\<^sub>1}\<^emph>f)\<union>({\<sigma>\<^sub>2}\<^emph>f)" using S\<^sub>p_def(1) \<open>S\<^sub>p = {\<sigma>\<^sub>1, \<sigma>\<^sub>2}\<close> star_distr_over_union by force
        moreover obtain S\<^sub>1 S\<^sub>2 where S\<^sub>1S\<^sub>2_def: "S\<^sub>1={\<sigma>\<in>S. fst \<sigma> t=1}" "S\<^sub>2={\<sigma>\<in>S. fst \<sigma> t=2}" by force
        ultimately have "S\<^sub>1\<subseteq>{\<sigma>\<^sub>1}\<^emph>f" by force
        moreover have "left_lives S\<^sub>1 {\<sigma>\<^sub>1} f" using S\<^sub>p_def(3) unfolding left_lives_def hAdd_def sledgehammer
        then show ?thesis sorry
      qed
      
      
      
      then have "\<lbrakk>C\<rbrakk>``S\<in>{S. {\<sigma>\<in>S. fst \<sigma> t=1}\<times>{\<sigma>\<in>S. fst \<sigma> t=2}\<subseteq>q}\<star>{S. S\<subseteq>f}" using "0" HSL_validityE by blast
      then obtain S\<^sub>q where S\<^sub>q_def: "\<lbrakk>C\<rbrakk>``S\<subseteq>S\<^sub>q\<^emph>f" "S\<^sub>q\<in>{S. {\<sigma>\<in>S. fst \<sigma> t=1}\<times>{\<sigma>\<in>S. fst \<sigma> t=2}\<subseteq>q}" "left_lives (\<lbrakk>C\<rbrakk>``S) S\<^sub>q f" using hstar_char_powerset by auto

















lemma Union_precondition_rule_1D_sound: \<comment> \<open>From all 1x1 subrelations of p, shows for all (rectangular) subrelations of p.. in 1D all subsets are rectangular\<close>
  assumes "\<Turnstile>HHL {{{\<sigma>}| \<sigma>. \<sigma>\<in>p}} C {{S. S\<subseteq>q}}"
    shows "\<Turnstile>HHL {{S. S\<subseteq>p}} C {{S. S\<subseteq>q}}"
proof(rule HHL_validityI)
  fix S
  assume asm: "S\<in>{S. S \<subseteq> p}"
  then have "S\<subseteq>p" by fast
  then have "\<forall>x\<in>{{\<sigma>}| \<sigma>. \<sigma>\<in>S}. (sem C x)\<in>{S. S\<subseteq>q}" using HHL_validityE assms by blast
  then have "\<forall>x\<in>{{\<sigma>}| \<sigma>. \<sigma>\<in>S}. (sem C x)\<subseteq>q" by blast
  then have "(\<Union>x\<in>{{\<sigma>}| \<sigma>. \<sigma>\<in>S}. (sem C x))\<subseteq>q" by fast
  then have "(sem C (\<Union>x\<in>{{\<sigma>}| \<sigma>. \<sigma>\<in>S}. x))\<subseteq>q" using sem_split by metis
  moreover have "S = (\<Union>x\<in>{{\<sigma>}| \<sigma>. \<sigma>\<in>S}. x)" by blast
  ultimately show "sem C S \<in> {S. S \<subseteq> q}" by auto
qed

lemma non_1_2_remove:
  assumes "\<And>\<sigma>\<^sub>1 \<sigma>\<^sub>2. \<sigma>\<^sub>1\<in>S \<Longrightarrow> \<sigma>\<^sub>2\<in>S \<Longrightarrow> fst \<sigma>\<^sub>1 t = 1 \<Longrightarrow> fst \<sigma>\<^sub>2 t = 2 \<Longrightarrow> P \<sigma>\<^sub>1 \<sigma>\<^sub>2"
    shows "\<And>\<sigma>\<^sub>1 \<sigma>\<^sub>2. \<sigma>\<^sub>1\<in>{\<sigma>. \<sigma>\<in>S \<and> (fst \<sigma> t = 1 \<or> fst \<sigma> t = 2)} \<Longrightarrow> \<sigma>\<^sub>2\<in>{\<sigma>. \<sigma>\<in>S \<and> (fst \<sigma> t = 1 \<or> fst \<sigma> t = 2)} \<Longrightarrow> fst \<sigma>\<^sub>1 t = 1 \<Longrightarrow> fst \<sigma>\<^sub>2 t = 2 \<Longrightarrow> P \<sigma>\<^sub>1 \<sigma>\<^sub>2"
  using assms by auto

lemma non_1_2_add:
  assumes "\<And>\<sigma>\<^sub>1 \<sigma>\<^sub>2. \<sigma>\<^sub>1\<in>S \<Longrightarrow> \<sigma>\<^sub>2\<in>S \<Longrightarrow> fst \<sigma>\<^sub>1 t = 1 \<Longrightarrow> fst \<sigma>\<^sub>2 t = 2 \<Longrightarrow> P \<sigma>\<^sub>1 \<sigma>\<^sub>2"
      and "\<And>\<sigma>. \<sigma>\<in>S' \<Longrightarrow> fst \<sigma> t \<noteq> 1 \<and> fst \<sigma> t \<noteq> 2"
    shows "\<And>\<sigma>\<^sub>1 \<sigma>\<^sub>2. \<sigma>\<^sub>1\<in>S\<union>S' \<Longrightarrow> \<sigma>\<^sub>2\<in>S\<union>S' \<Longrightarrow> fst \<sigma>\<^sub>1 t = 1 \<Longrightarrow> fst \<sigma>\<^sub>2 t = 2 \<Longrightarrow> P \<sigma>\<^sub>1 \<sigma>\<^sub>2"
  using assms by blast

lemma Union_precondition_rule_2D_sound_helper:
  assumes "\<Turnstile>HHL {{{\<sigma>\<^sub>1, \<sigma>\<^sub>2}|\<sigma>\<^sub>1 \<sigma>\<^sub>2. (\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p}} C {{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>q}}"
      and "t\<notin>md C"
  shows "\<Turnstile>HHL {{S. (\<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>p) \<and> (\<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>1'\<in>S. fst \<sigma>\<^sub>1 t = 1 \<longrightarrow> fst \<sigma>\<^sub>1' t = 1 \<longrightarrow> \<sigma>\<^sub>1=\<sigma>\<^sub>1')}} C {{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>q}}"
proof(rule HHL_validityI)
  fix S
  assume asm: "S\<in>{S. (\<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1, \<sigma>\<^sub>2) \<in> p) \<and> (\<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>1'\<in>S. fst \<sigma>\<^sub>1 t = 1 \<longrightarrow> fst \<sigma>\<^sub>1' t = 1 \<longrightarrow> \<sigma>\<^sub>1 = \<sigma>\<^sub>1')}"
  then have 0: "\<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1, \<sigma>\<^sub>2) \<in> p"
               "\<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>1'\<in>S. fst \<sigma>\<^sub>1 t = 1 \<longrightarrow> fst \<sigma>\<^sub>1' t = 1 \<longrightarrow> \<sigma>\<^sub>1 = \<sigma>\<^sub>1'" by auto
  then obtain \<sigma>\<^sub>1 where \<sigma>\<^sub>1_def: "{\<sigma>\<^sub>1}={\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 1} \<or> {}={\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 1}" by blast
  then show "sem C S \<in> {S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1, \<sigma>\<^sub>2) \<in> q}"
  proof(cases "{}={\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 1}")
    case True
    then have "{}={\<sigma>. \<sigma>\<in>sem C S \<and> fst \<sigma> t = 1}" by (smt (verit, best) assms(2) empty_Collect_eq sem_perserves_non_md_element)
      (*by (smt (verit, del_insts) all_not_in_conv assms(2) diff_less_Suc emptyE empty_Collect_eq le_refl less_eq_Suc_le mem_Collect_eq not_less_eq_eq sem_perserves_non_md) - when \<Turnstile> typeless*)
    then show ?thesis by blast
  next
    case False
    then have "\<forall>\<sigma>\<^sub>2\<in>{\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}. (\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p" using "0"(1) \<sigma>\<^sub>1_def by blast
    then have "\<forall>\<sigma>\<^sub>2\<in>{\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}. {\<sigma>\<^sub>1, \<sigma>\<^sub>2}\<in>{{\<sigma>\<^sub>1, \<sigma>\<^sub>2}|\<sigma>\<^sub>1 \<sigma>\<^sub>2. (\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p}" by blast
    then have "\<forall>\<sigma>\<^sub>2\<in>{\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}. sem C {\<sigma>\<^sub>1, \<sigma>\<^sub>2} \<in> {S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>q}" using HHL_validityE assms(1) by blast
    then have "\<forall>\<sigma>\<^sub>2\<in>{\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}. \<forall>\<sigma>\<^sub>1'\<in>sem C {\<sigma>\<^sub>1, \<sigma>\<^sub>2}. \<forall>\<sigma>\<^sub>2'\<in>sem C {\<sigma>\<^sub>1, \<sigma>\<^sub>2}. fst \<sigma>\<^sub>1' t = 1 \<and> fst \<sigma>\<^sub>2' t = 2 \<longrightarrow> (\<sigma>\<^sub>1',\<sigma>\<^sub>2')\<in>q" by blast
    moreover have "fst \<sigma>\<^sub>1 t = 1" using False \<sigma>\<^sub>1_def by blast
    then have "\<forall>\<sigma>\<^sub>1'\<in>sem C {\<sigma>\<^sub>1}. fst \<sigma>\<^sub>1' t = 1" using assms(2) sem_perserves_non_md_singleton by fast
    moreover have "\<forall>\<sigma>\<^sub>2\<in>{\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}. fst \<sigma>\<^sub>2 t = 2" by blast
    then have "\<forall>\<sigma>\<^sub>2\<in>{\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}. \<forall>\<sigma>\<^sub>2'\<in>sem C {\<sigma>\<^sub>2}. fst \<sigma>\<^sub>2' t = 2" using assms(2) sem_perserves_non_md_singleton by fast
    ultimately have "\<forall>\<sigma>\<^sub>2\<in>{\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}. \<forall>\<sigma>\<^sub>1'\<in>sem C {\<sigma>\<^sub>1}. \<forall>\<sigma>\<^sub>2'\<in>sem C {\<sigma>\<^sub>2}. fst \<sigma>\<^sub>1' t = 1 \<and> fst \<sigma>\<^sub>2' t = 2" by blast
    have "\<forall>\<sigma>\<^sub>2\<in>{\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}. \<forall>\<sigma>\<^sub>1'\<in>sem C {\<sigma>\<^sub>1}. \<forall>\<sigma>\<^sub>2'\<in>sem C {\<sigma>\<^sub>2}. (\<sigma>\<^sub>1',\<sigma>\<^sub>2')\<in>q"
    proof
      fix \<sigma>\<^sub>2
      assume "\<sigma>\<^sub>2\<in>{\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}"
      show "\<forall>\<sigma>\<^sub>1'\<in>sem C {\<sigma>\<^sub>1}. \<forall>\<sigma>\<^sub>2'\<in>sem C {\<sigma>\<^sub>2}. (\<sigma>\<^sub>1',\<sigma>\<^sub>2')\<in>q"
      proof
        fix \<sigma>\<^sub>1'
        assume "\<sigma>\<^sub>1'\<in>sem C {\<sigma>\<^sub>1}"
        then have "\<sigma>\<^sub>1'\<in>sem C {\<sigma>\<^sub>1, \<sigma>\<^sub>2}" using sem_union by (metis UnCI insert_is_Un)
        show "\<forall>\<sigma>\<^sub>2'\<in>sem C {\<sigma>\<^sub>2}. (\<sigma>\<^sub>1',\<sigma>\<^sub>2')\<in>q"
        proof
          fix \<sigma>\<^sub>2'
          assume "\<sigma>\<^sub>2'\<in>sem C {\<sigma>\<^sub>2}"
          then have "\<sigma>\<^sub>2'\<in>sem C {\<sigma>\<^sub>1, \<sigma>\<^sub>2}" using sem_union by (metis UnCI insert_is_Un)
          moreover note \<open>\<sigma>\<^sub>1'\<in>sem C {\<sigma>\<^sub>1, \<sigma>\<^sub>2}\<close>
          moreover have "fst \<sigma>\<^sub>1' t = 1" using \<open>\<forall>\<sigma>\<^sub>1'\<in>sem C {\<sigma>\<^sub>1}. fst \<sigma>\<^sub>1' t = 1\<close> \<open>\<sigma>\<^sub>1' \<in> sem C {\<sigma>\<^sub>1}\<close> by simp
          moreover have "fst \<sigma>\<^sub>2' t = 2" using \<open>\<forall>\<sigma>\<^sub>2\<in>{\<sigma> \<in> S. fst \<sigma> t = 2}. \<forall>\<sigma>\<^sub>2'\<in>sem C {\<sigma>\<^sub>2}. fst \<sigma>\<^sub>2' t = 2\<close> \<open>\<sigma>\<^sub>2 \<in> {\<sigma> \<in> S. fst \<sigma> t = 2}\<close> \<open>\<sigma>\<^sub>2' \<in> sem C {\<sigma>\<^sub>2}\<close> by blast
          ultimately show "(\<sigma>\<^sub>1',\<sigma>\<^sub>2')\<in>q"
            using \<open>\<forall>\<sigma>\<^sub>2\<in>{\<sigma> \<in> S. fst \<sigma> t = 2}. \<forall>\<sigma>\<^sub>1' \<in>sem C {\<sigma>\<^sub>1, \<sigma>\<^sub>2}. \<forall>\<sigma>\<^sub>2' \<in>sem C {\<sigma>\<^sub>1, \<sigma>\<^sub>2}. fst \<sigma>\<^sub>1' t = 1 \<and> fst \<sigma>\<^sub>2' t = 2 \<longrightarrow> (\<sigma>\<^sub>1', \<sigma>\<^sub>2') \<in> q\<close> \<open>\<sigma>\<^sub>2 \<in> {\<sigma> \<in> S. fst \<sigma> t = 2}\<close> by blast
        qed
      qed
    qed
    then have "\<forall>\<sigma>\<^sub>2\<in>{\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}. (sem C {\<sigma>\<^sub>1})\<times>(sem C {\<sigma>\<^sub>2})\<subseteq>q" by blast
    then have "\<forall>x\<in>{{\<sigma>}| \<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}. (sem C {\<sigma>\<^sub>1})\<times>(sem C x)\<subseteq>q" by blast
    then have "(sem C {\<sigma>\<^sub>1})\<times>(\<Union>x\<in>{{\<sigma>}| \<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}. sem C x)\<subseteq>q" by blast
    then have "(sem C {\<sigma>\<^sub>1})\<times>(sem C (\<Union>x\<in>{{\<sigma>}| \<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}. x))\<subseteq>q" using sem_split by metis
    moreover have "(\<Union>x\<in>{{\<sigma>}| \<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}. x) = {\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}" by blast
    ultimately have "(sem C {\<sigma>\<^sub>1})\<times>(sem C {\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2})\<subseteq>q" by simp
    have "\<forall>\<sigma>\<^sub>1'\<in>sem C S. \<forall>\<sigma>\<^sub>2'\<in>sem C S. fst \<sigma>\<^sub>1' t = 1 \<and> fst \<sigma>\<^sub>2' t = 2 \<longrightarrow> (\<sigma>\<^sub>1', \<sigma>\<^sub>2') \<in> q"
    proof
      fix \<sigma>\<^sub>1'
      assume "\<sigma>\<^sub>1'\<in>sem C S"
      show "\<forall>\<sigma>\<^sub>2'\<in>sem C S. fst \<sigma>\<^sub>1' t = 1 \<and> fst \<sigma>\<^sub>2' t = 2 \<longrightarrow> (\<sigma>\<^sub>1', \<sigma>\<^sub>2') \<in> q"
      proof
        fix \<sigma>\<^sub>2'
        assume "\<sigma>\<^sub>2'\<in>sem C S"
        show "fst \<sigma>\<^sub>1' t = 1 \<and> fst \<sigma>\<^sub>2' t = 2 \<longrightarrow> (\<sigma>\<^sub>1', \<sigma>\<^sub>2') \<in> q"
        proof
          assume "fst \<sigma>\<^sub>1' t = 1 \<and> fst \<sigma>\<^sub>2' t = 2"
          moreover have "\<And>\<sigma>. \<sigma>\<in>S \<Longrightarrow> fst \<sigma> t = 1 \<Longrightarrow> \<sigma> = \<sigma>\<^sub>1" using \<sigma>\<^sub>1_def by blast
          moreover have "\<sigma>\<^sub>1\<in>S" using False \<sigma>\<^sub>1_def by blast
          then have "sem C S = sem C {\<sigma>\<^sub>1} \<union> sem C (S-{\<sigma>\<^sub>1})" using sem_union by (metis insert_Diff insert_is_Un)
          moreover have "\<forall>\<sigma>\<in>S-{\<sigma>\<^sub>1}. fst \<sigma> t \<noteq> 1" using calculation(2) by blast
          then have "\<forall>\<sigma>'\<in>sem C (S-{\<sigma>\<^sub>1}). fst \<sigma>' t \<noteq> 1" using assms(2) sem_perserves_non_md_element by fast
          ultimately have "\<sigma>\<^sub>1'\<in>sem C {\<sigma>\<^sub>1}" using \<open>\<sigma>\<^sub>1' \<in> sem C S\<close> by blast
          note \<open>fst \<sigma>\<^sub>1' t = 1 \<and> fst \<sigma>\<^sub>2' t = 2\<close>
          moreover have "S = {\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2} \<union> (S-{\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2})" by blast
          then have "sem C S = sem C {\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2} \<union> sem C (S-{\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2})" using sem_union by metis
          moreover have "\<forall>\<sigma>\<in>S-{\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}. fst \<sigma> t \<noteq> 2" by blast
          then have "\<forall>\<sigma>'\<in>sem C (S-{\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}). fst \<sigma>' t \<noteq> 2" using assms(2) sem_perserves_non_md_element by fast
          ultimately have "\<sigma>\<^sub>2'\<in>sem C {\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}" using \<open>\<sigma>\<^sub>2' \<in> sem C S\<close> by blast
          then show "(\<sigma>\<^sub>1', \<sigma>\<^sub>2') \<in> q" using \<open>\<sigma>\<^sub>1'\<in>sem C {\<sigma>\<^sub>1}\<close> \<open>sem C {\<sigma>\<^sub>1} \<times> sem C {\<sigma> \<in> S. fst \<sigma> t = 2} \<subseteq> q\<close> by auto
        qed
      qed
    qed
    then show ?thesis by blast
  qed
qed

lemma Union_precondition_rule_2D_sound: \<comment> \<open>From all 1x1 subrelations of p, shows for all rectangular subrelations of p\<close>
  assumes "\<Turnstile>HHL {{{\<sigma>\<^sub>1, \<sigma>\<^sub>2}|\<sigma>\<^sub>1 \<sigma>\<^sub>2. (\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p}} C {{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>q}}"
      and "t\<notin>md C"
  shows "\<Turnstile>HHL {{S. (\<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>p)}} C {{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>q}}"
proof(rule HHL_validityI)
  fix S
  assume "S\<in>{S. (\<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>p)}"
  then have S_def: "\<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>p" by blast
  obtain S\<^sub>2 where S\<^sub>2_def: "S\<^sub>2={\<sigma>. \<sigma>\<in>S \<and> fst \<sigma> t = 2}" by blast
  then have "\<And>\<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>S \<Longrightarrow> S\<^sub>2\<union>{\<sigma>\<^sub>1}\<in>{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>p}" using S_def by blast
  moreover have "\<And>\<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>S \<Longrightarrow> S\<^sub>2\<union>{\<sigma>\<^sub>1}\<in>{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>1'\<in>S. fst \<sigma>\<^sub>1 t = 1 \<longrightarrow> fst \<sigma>\<^sub>1' t = 1 \<longrightarrow> \<sigma>\<^sub>1=\<sigma>\<^sub>1'}" using S\<^sub>2_def by simp
  ultimately have 0: "\<And>\<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>S \<Longrightarrow> S\<^sub>2\<union>{\<sigma>\<^sub>1}\<in>{S. (\<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>p) \<and> (\<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>1'\<in>S. fst \<sigma>\<^sub>1 t = 1 \<longrightarrow> fst \<sigma>\<^sub>1' t = 1 \<longrightarrow> \<sigma>\<^sub>1=\<sigma>\<^sub>1')}" by blast
  then have "\<And>\<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>S \<Longrightarrow> sem C (S\<^sub>2\<union>{\<sigma>\<^sub>1})\<in>{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>q}"
    using Union_precondition_rule_2D_sound_helper[of p C t q] by (metis (mono_tags, lifting) HHL_validityE assms)
  then have "\<And>\<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>S \<Longrightarrow> \<forall>\<sigma>\<^sub>1'\<in>sem C (S\<^sub>2\<union>{\<sigma>\<^sub>1}). \<forall>\<sigma>\<^sub>2'\<in>sem C (S\<^sub>2\<union>{\<sigma>\<^sub>1}). fst \<sigma>\<^sub>1' t = 1 \<and> fst \<sigma>\<^sub>2' t = 2 \<longrightarrow> (\<sigma>\<^sub>1',\<sigma>\<^sub>2')\<in>q" by blast
  then have "\<And>\<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>S-S\<^sub>2 \<Longrightarrow> {\<sigma>\<^sub>1'. \<sigma>\<^sub>1'\<in>sem C (S\<^sub>2\<union>{\<sigma>\<^sub>1}) \<and> fst \<sigma>\<^sub>1' t = 1}\<times>{\<sigma>\<^sub>2'. \<sigma>\<^sub>2'\<in>sem C (S\<^sub>2\<union>{\<sigma>\<^sub>1}) \<and> fst \<sigma>\<^sub>2' t = 2}\<subseteq>q" by blast
  moreover have "\<And>\<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>S-S\<^sub>2 \<Longrightarrow> fst \<sigma>\<^sub>1 t \<noteq> 2" using S\<^sub>2_def by blast
  then have "\<And>\<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>S-S\<^sub>2 \<Longrightarrow> \<forall>\<sigma>\<^sub>1'\<in>sem C {\<sigma>\<^sub>1}. fst \<sigma>\<^sub>1' t \<noteq> 2" using assms(2) sem_perserves_non_md_singleton by metis
  ultimately have "\<And>\<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>S-S\<^sub>2 \<Longrightarrow> {\<sigma>\<^sub>1'. \<sigma>\<^sub>1'\<in>sem C (S\<^sub>2\<union>{\<sigma>\<^sub>1}) \<and> fst \<sigma>\<^sub>1' t = 1}\<times>{\<sigma>\<^sub>2'. \<sigma>\<^sub>2'\<in>sem C S\<^sub>2 \<and> fst \<sigma>\<^sub>2' t = 2}\<subseteq>q" using sem_union[of C S\<^sub>2] by fast
  moreover have "\<forall>\<sigma>\<^sub>2'\<in>sem C S\<^sub>2. fst \<sigma>\<^sub>2' t = 2" by (metis (mono_tags, lifting) S\<^sub>2_def assms(2) mem_Collect_eq sem_perserves_non_md_element)
  ultimately have "\<And>\<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>S-S\<^sub>2 \<Longrightarrow> {\<sigma>\<^sub>1'. \<sigma>\<^sub>1'\<in>sem C (S\<^sub>2\<union>{\<sigma>\<^sub>1}) \<and> fst \<sigma>\<^sub>1' t = 1}\<times>(sem C S\<^sub>2)\<subseteq>q" using sem_perserves_non_md_singleton by blast
  moreover note \<open>\<forall>\<sigma>\<^sub>2'\<in>sem C S\<^sub>2. fst \<sigma>\<^sub>2' t = 2\<close>
  then have "\<And>\<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>S-S\<^sub>2 \<Longrightarrow> \<forall>\<sigma>\<^sub>1'\<in>sem C (S\<^sub>2\<union>{\<sigma>\<^sub>1}). fst \<sigma>\<^sub>1' t = 1 \<longrightarrow> \<sigma>\<^sub>1'\<in>sem C {\<sigma>\<^sub>1}" using sem_union by fastforce
  moreover have "\<And>\<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>{\<sigma>\<^sub>1\<in>S. fst \<sigma>\<^sub>1 t = 1} \<Longrightarrow> \<sigma>\<^sub>1\<in>S-S\<^sub>2" using S\<^sub>2_def by simp
  moreover have "\<And>\<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>{\<sigma>\<^sub>1\<in>S. fst \<sigma>\<^sub>1 t = 1} \<Longrightarrow> {\<sigma>\<^sub>1'. \<sigma>\<^sub>1'\<in>sem C (S\<^sub>2\<union>{\<sigma>\<^sub>1}) \<and> fst \<sigma>\<^sub>1' t = 1}={\<sigma>\<^sub>1'. \<sigma>\<^sub>1'\<in>sem C {\<sigma>\<^sub>1}}" by (metis (mono_tags, lifting) Un_iff assms(2) mem_Collect_eq sem_perserves_non_md_singleton sem_union calculation(2) calculation(3))
  ultimately have "\<And>\<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>{\<sigma>\<^sub>1\<in>S. fst \<sigma>\<^sub>1 t = 1} \<Longrightarrow> (sem C {\<sigma>\<^sub>1})\<times>(sem C S\<^sub>2)\<subseteq>q" by blast
  then have "\<forall>x\<in>{{\<sigma>\<^sub>1}| \<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>{\<sigma>\<^sub>1\<in>S. fst \<sigma>\<^sub>1 t = 1}}. (sem C x)\<times>(sem C S\<^sub>2)\<subseteq>q" by blast
  then have "(\<Union>x\<in>{{\<sigma>\<^sub>1}| \<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>{\<sigma>\<^sub>1\<in>S. fst \<sigma>\<^sub>1 t = 1}}. sem C x)\<times>(sem C S\<^sub>2)\<subseteq>q" by blast
  then have "(sem C (\<Union>x\<in>{{\<sigma>\<^sub>1}| \<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>{\<sigma>\<^sub>1\<in>S. fst \<sigma>\<^sub>1 t = 1}}. x))\<times>(sem C S\<^sub>2)\<subseteq>q" using sem_split by metis
  moreover have "(\<Union>x\<in>{{\<sigma>\<^sub>1}| \<sigma>\<^sub>1. \<sigma>\<^sub>1\<in>{\<sigma>\<^sub>1\<in>S. fst \<sigma>\<^sub>1 t = 1}}. x) = {\<sigma>\<^sub>1\<in>S. fst \<sigma>\<^sub>1 t = 1}" by blast
  ultimately have "(sem C {\<sigma>\<^sub>1\<in>S. fst \<sigma>\<^sub>1 t = 1})\<times>(sem C S\<^sub>2)\<subseteq>q" by argo
  then have "(sem C {\<sigma>\<^sub>1\<in>S. fst \<sigma>\<^sub>1 t = 1})\<times>(sem C {\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>2 t = 2})\<subseteq>q" using S\<^sub>2_def by fast
  have "\<forall>\<sigma>\<^sub>1'\<in>sem C S. \<forall>\<sigma>\<^sub>2'\<in>sem C S. fst \<sigma>\<^sub>1' t = 1 \<and> fst \<sigma>\<^sub>2' t = 2 \<longrightarrow> (\<sigma>\<^sub>1', \<sigma>\<^sub>2') \<in> q" 
  proof
    fix \<sigma>\<^sub>1'
    assume "\<sigma>\<^sub>1'\<in>sem C S"
    show "\<forall>\<sigma>\<^sub>2'\<in>sem C S. fst \<sigma>\<^sub>1' t = 1 \<and> fst \<sigma>\<^sub>2' t = 2 \<longrightarrow> (\<sigma>\<^sub>1', \<sigma>\<^sub>2') \<in> q"
    proof
      fix \<sigma>\<^sub>2'
      assume "\<sigma>\<^sub>2'\<in>sem C S"
      show "fst \<sigma>\<^sub>1' t = 1 \<and> fst \<sigma>\<^sub>2' t = 2 \<longrightarrow> (\<sigma>\<^sub>1', \<sigma>\<^sub>2') \<in> q"
      proof
        assume asm: "fst \<sigma>\<^sub>1' t = 1 \<and> fst \<sigma>\<^sub>2' t = 2"
        then have "\<sigma>\<^sub>1'\<in>sem C {\<sigma>\<^sub>1\<in>S. fst \<sigma>\<^sub>1 t = 1}" using \<open>\<sigma>\<^sub>1' \<in> sem C S\<close> assms(2) sem_perserves_non_md by fast
        moreover from asm have "\<sigma>\<^sub>2'\<in>sem C {\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>2 t = 2}" using \<open>\<sigma>\<^sub>2' \<in> sem C S\<close> assms(2) sem_perserves_non_md by fast
        ultimately show "(\<sigma>\<^sub>1', \<sigma>\<^sub>2') \<in> q" using \<open>sem C {\<sigma>\<^sub>1 \<in> S. fst \<sigma>\<^sub>1 t = 1} \<times> sem C {\<sigma>\<^sub>2 \<in> S. fst \<sigma>\<^sub>2 t = 2} \<subseteq> q\<close> by auto
      qed
    qed
  qed
  then show "sem C S \<in> {S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1, \<sigma>\<^sub>2) \<in> q}" by blast
qed

theorem
  assumes "\<Turnstile>HHL {{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>p}\<star>{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>r}} C {{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>q}\<star>{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>r}}"
      and "\<And>\<sigma>\<^sub>1 \<sigma>\<^sub>2. (\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p \<Longrightarrow> fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2" (*in fact, we have all 3 relations in this form*)
      and "t\<notin>md C"
shows "\<Turnstile>HHL {{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>p\<^emph>\<^sub>rr}} C {{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>q\<^emph>\<^sub>rr}}"
proof(-)
  have "\<Turnstile>HHL {{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>p}\<star>{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>r}} C {{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>q\<^emph>\<^sub>rr}}"
    using  assms Cons_rule_sound subsumes_RSL_first_direction entails_def entails_refl by meson
  moreover have "\<And>\<sigma>\<^sub>1 \<sigma>\<^sub>2. (\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p\<^emph>\<^sub>rr \<Longrightarrow> fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2"
  proof(rule ccontr)
    fix \<sigma>\<^sub>1 \<sigma>\<^sub>2
    assume "(\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p\<^emph>\<^sub>rr"
    assume "\<not>(fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2)"
    then have "fst \<sigma>\<^sub>1 t \<noteq> 1 \<or> fst \<sigma>\<^sub>2 t \<noteq> 2" by blast
    then show "False"
    proof(cases "fst \<sigma>\<^sub>1 t \<noteq> 1")
      case True
      then show ?thesis using \<open>(\<sigma>\<^sub>1, \<sigma>\<^sub>2) \<in> p \<^emph>\<^sub>r r\<close> assms(2) fst_conv rtimes_char by auto
    next
      case False
      then have "fst \<sigma>\<^sub>2 t \<noteq> 2" using \<open>fst \<sigma>\<^sub>1 t \<noteq> 1 \<or> fst \<sigma>\<^sub>2 t \<noteq> 2\<close> by auto
      then show ?thesis using \<open>(\<sigma>\<^sub>1, \<sigma>\<^sub>2) \<in> p \<^emph>\<^sub>r r\<close> assms(2) fst_conv rtimes_char by auto
    qed
  qed
  then have "{{\<sigma>\<^sub>1, \<sigma>\<^sub>2}|\<sigma>\<^sub>1 \<sigma>\<^sub>2. (\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p\<^emph>\<^sub>rr}\<subseteq>{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>p}\<star>{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>r}"
    using subsumes_RSL_second_direction by blast
  ultimately have "\<Turnstile>HHL {{{\<sigma>\<^sub>1, \<sigma>\<^sub>2}|\<sigma>\<^sub>1 \<sigma>\<^sub>2. (\<sigma>\<^sub>1, \<sigma>\<^sub>2)\<in>p\<^emph>\<^sub>rr}} C {{S. \<forall>\<sigma>\<^sub>1\<in>S. \<forall>\<sigma>\<^sub>2\<in>S. fst \<sigma>\<^sub>1 t = 1 \<and> fst \<sigma>\<^sub>2 t = 2 \<longrightarrow> (\<sigma>\<^sub>1,\<sigma>\<^sub>2)\<in>q\<^emph>\<^sub>rr}}"
    using Cons_rule_sound entails_def entails_refl by meson
  then show ?thesis using Union_precondition_rule_2D_sound[of "p\<^emph>\<^sub>rr" C t "q\<^emph>\<^sub>rr"] assms(3) by blast
qed

end