theory ProgramCommands
  imports FiniteMap
  abbrevs "(=" = "\<subseteq>\<^sub>f"
begin

type_synonym val = nat
type_synonym loc = nat
type_synonym hval = "val + unit"
abbreviation hOcc :: "nat \<Rightarrow> hval option" where "hOcc n \<equiv> Some (Inl n)"
abbreviation hFrd :: "hval option" where "hFrd \<equiv> Some (Inr ())"
abbreviation hUkw :: "hval option" where "hUkw \<equiv> None"
type_synonym 'var sstate = "'var \<Rightarrow> val"
type_synonym hstate = "loc \<rightharpoonup>\<^sub>f hval"
type_synonym 'var state = "'var sstate \<times> hstate"

type_synonym 'var aexp = "'var sstate \<Rightarrow> val"
type_synonym 'var bexp = "'var sstate set"
type_synonym 'var asrt = "'var bexp"

datatype 'var cmd = 
 Assign 'var "'var aexp"
|Seq "'var cmd" "'var cmd"
|If "'var cmd" "'var cmd"
|Skip
|Havoc 'var
|Assume "'var bexp"
|While "'var cmd"
|Alloc 'var
|Write 'var "'var aexp"
|Read 'var 'var
|Free 'var

type_synonym 'var conf = "'var cmd \<times> 'var state"

inductive red :: "'var conf \<Rightarrow> 'var conf \<Rightarrow> bool" ("_ \<rightarrow> _" [0,0] 81) where
 RedAssign: "(Assign x e, (s, h)) \<rightarrow> (Skip, (s(x:=(e s)), h))"
|RedSeq1: "(Seq Skip C\<^sub>2, \<sigma>) \<rightarrow> (C\<^sub>2, \<sigma>)"
|RedSeq2: "(C\<^sub>1, \<sigma>) \<rightarrow> (C\<^sub>1', \<sigma>') \<Longrightarrow> (Seq C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow> (Seq C\<^sub>1' C\<^sub>2, \<sigma>')"
|RedIf1: "(If C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow> (C\<^sub>1, \<sigma>)"
|RedIf2: "(If C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow> (C\<^sub>2, \<sigma>)"
|RedHavoc: "(Havoc x, (s, h)) \<rightarrow> (Skip, (s(x:=v), h))"
|RedAssume: "(fst \<sigma>)\<in>b \<Longrightarrow> (Assume b, \<sigma>) \<rightarrow> (Skip, \<sigma>)"
|RedWhile: "(While C, \<sigma>) \<rightarrow> (If (Seq C (While C)) Skip, \<sigma>)"
|RedAlloc: "fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd \<Longrightarrow> l\<noteq>0 \<Longrightarrow> (Alloc x, (s, h)) \<rightarrow> (Skip, (s(x:=l),fmap_update h l (hOcc n)))" (*{y\<mapsto>\<bottom>} x:=alloc(5) {y\<mapsto>5 \<or> y\<mapsto>\<bottom>}*)
|RedWrite: "fmap_lookup h (s x)=hOcc n \<Longrightarrow> s x\<noteq>0 \<Longrightarrow> (Write x e, (s, h)) \<rightarrow> (Skip, (s, fmap_update h (s x) (hOcc (e s))))" (*h(s(x))=hOcc v \<Longrightarrow> s(x)\<in>dom h.. the opposite is not true, since it could be h(s(x))=hUkw*)
|RedRead: "fmap_lookup h (s x)=hOcc n \<Longrightarrow> s x\<noteq>0 \<Longrightarrow> (Read y x, (s, h)) \<rightarrow> (Skip, (s(y:=n), h))"         \<comment>\<open>y:=[x]\<close>
|RedFree: "fmap_lookup h (s x)=hOcc n \<Longrightarrow> s x\<noteq>0 \<Longrightarrow> (Free x, (s, h)) \<rightarrow> (Skip, (s, fmap_update h (s x) hFrd))"

inductive_cases red_Seq_elim[elim!]: "(Seq C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow> (C', \<sigma>')"
inductive_cases red_Seq_elim2[elim!]: "(Seq Skip C\<^sub>2, \<sigma>) \<rightarrow> (C\<^sub>2, \<sigma>')"
inductive_cases red_While_elim: "(While C, \<sigma>) \<rightarrow> (C', \<sigma>')"
inductive_cases red_If_elim[elim!]: "(If C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow> (C', \<sigma>')"
inductive_cases red_Assume_elim[elim!]: "(Assume b, \<sigma>) \<rightarrow> (Skip, \<sigma>')"
inductive_cases red_Assign_elim[elim!]: "(Assign x e, \<sigma>) \<rightarrow> (Skip, \<sigma>')"
inductive_cases red_Havoc_elim[elim!]: "(Havoc x, \<sigma>) \<rightarrow> (Skip, \<sigma>')"
inductive_cases red_Alloc_elim[elim!]: "(Alloc x, \<sigma>) \<rightarrow> (Skip, \<sigma>')"
inductive_cases red_Write_elim[elim!]: "(Write x e, \<sigma>) \<rightarrow> (Skip, \<sigma>')"
inductive_cases red_Read_elim[elim!]: "(Read y x, \<sigma>) \<rightarrow> (Skip, \<sigma>')"
inductive_cases red_Free_elim[elim!]: "(Free x, \<sigma>) \<rightarrow> (Skip, \<sigma>')"

lemma skip_deadend: 
  "c \<rightarrow> c' \<Longrightarrow> fst c \<noteq> Skip"
  apply(induction rule: red.induct) by auto

lemma red_rhs:
  "(Assign x e, (s, h)) \<rightarrow> c' \<Longrightarrow> c'=(Skip, (s(x:=(e s)), h))"
  "(Seq Skip C\<^sub>2, \<sigma>) \<rightarrow> c' \<Longrightarrow> c'=(C\<^sub>2, \<sigma>)"
  "(Seq C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow> c' \<Longrightarrow> C\<^sub>1\<noteq>Skip \<Longrightarrow> \<exists>C\<^sub>1' \<sigma>'. c'=(Seq C\<^sub>1' C\<^sub>2, \<sigma>') \<and> ((C\<^sub>1, \<sigma>) \<rightarrow> (C\<^sub>1', \<sigma>'))"
  "(If C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow> c' \<Longrightarrow> c'=(C\<^sub>1, \<sigma>) \<or> c'=(C\<^sub>2, \<sigma>)"
  "(Havoc x, (s, h)) \<rightarrow> c' \<Longrightarrow> \<exists>n. c'=(Skip, (s(x:=n), h))"
  "(Assume b, \<sigma>) \<rightarrow> c' \<Longrightarrow> c'=(Skip, \<sigma>) \<and> (fst \<sigma>)\<in>b"
  "(While C, \<sigma>) \<rightarrow> c' \<Longrightarrow> c'=(If (Seq C (While C)) Skip, \<sigma>)"
  "(Alloc x, (s, h)) \<rightarrow> c' \<Longrightarrow> \<exists>l n. c'=(Skip, (s(x:=l),fmap_update h l (hOcc n))) \<and> (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0"
  "(Write x e, (s, h)) \<rightarrow> c' \<Longrightarrow> \<exists>n. c'=(Skip, (s, fmap_update h (s x) (hOcc (e s)))) \<and> fmap_lookup h (s x)=hOcc n \<and> s(x)\<noteq>0"
  "(Read y x, (s, h)) \<rightarrow> c' \<Longrightarrow> \<exists>n. c'=(Skip, (s(y:=n), h)) \<and> fmap_lookup h (s x)=hOcc n \<and> s(x)\<noteq>0"
  "(Free x, (s, h)) \<rightarrow> c' \<Longrightarrow> \<exists>n. c'=(Skip, (s, fmap_update h (s x) hFrd)) \<and> fmap_lookup h (s x)=hOcc n \<and> s(x)\<noteq>0"
proof(-)
  show "(Assign x e, (s, h)) \<rightarrow> c' \<Longrightarrow> c'=(Skip, (s(x:=(e s)), h))"
    apply(induction "(Assign x e, (s, h))" c' rule: red.induct) by blast
  show "(Havoc x, (s, h)) \<rightarrow> c' \<Longrightarrow> \<exists>n. c'=(Skip, (s(x:=n), h))"
    apply(induction "(Havoc x, (s, h))" c' rule: red.induct) by blast
  show "(Assume b, \<sigma>) \<rightarrow> c' \<Longrightarrow> c'=(Skip, \<sigma>) \<and> (fst \<sigma>)\<in>b"
    apply(induction "(Assume b, \<sigma>)" c' rule: red.induct) by blast
  show "(If C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow> c' \<Longrightarrow> c'=(C\<^sub>1, \<sigma>) \<or> c'=(C\<^sub>2, \<sigma>)"
    apply(induction "(If C\<^sub>1 C\<^sub>2, \<sigma>)" c' rule: red.induct) by auto
  show "(While C, \<sigma>) \<rightarrow> c' \<Longrightarrow> c'=(If (Seq C (While C)) Skip, \<sigma>)"
    apply(induction "(While C, \<sigma>)" c' rule: red.induct) by blast
  show "(Seq Skip C\<^sub>2, \<sigma>) \<rightarrow> c' \<Longrightarrow> c'=(C\<^sub>2, \<sigma>)"
    apply(induction "(Seq Skip C\<^sub>2, \<sigma>)" c' rule: red.induct) apply simp by (meson fst_conv skip_deadend)
  show "(Seq C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow> c' \<Longrightarrow> C\<^sub>1\<noteq>Skip \<Longrightarrow> \<exists>C\<^sub>1' \<sigma>'. c'=(Seq C\<^sub>1' C\<^sub>2, \<sigma>') \<and> ((C\<^sub>1, \<sigma>) \<rightarrow> (C\<^sub>1', \<sigma>'))"
    apply(induction "(Seq C\<^sub>1 C\<^sub>2, \<sigma>)" c' rule: red.induct) by auto
  show "(Alloc x, (s, h)) \<rightarrow> c' \<Longrightarrow> \<exists>l n. c'=(Skip, (s(x:=l),fmap_update h l (hOcc n))) \<and> (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0"
    apply(induction "(Alloc x, (s, h))" c' rule: red.induct) by blast
  show "(Write x e, (s, h)) \<rightarrow> c' \<Longrightarrow> \<exists>n. c'=(Skip, (s, fmap_update h (s x) (hOcc (e s)))) \<and> fmap_lookup h (s x)=hOcc n \<and> s(x)\<noteq>0"
    apply(induction "(Write x e, (s, h))" c' rule: red.induct) by blast
  show "(Read y x, (s, h)) \<rightarrow> c' \<Longrightarrow> \<exists>n. c'=(Skip, (s(y:=n), h)) \<and> fmap_lookup h (s x)=hOcc n \<and> s(x)\<noteq>0"
    apply(induction "(Read y x, (s, h))" c' rule: red.induct) by blast
  show "(Free x, (s, h)) \<rightarrow> c' \<Longrightarrow> \<exists>n. c'=(Skip, (s, fmap_update h (s x) hFrd)) \<and> fmap_lookup h (s x)=hOcc n \<and> s(x)\<noteq>0"
    apply(induction "(Free x, (s, h))" c' rule: red.induct) by blast
qed

inductive rtred :: "'var conf \<Rightarrow> 'var conf \<Rightarrow> bool" ("_ \<rightarrow>* _" [0,0] 81) where
 RedRefl: "c \<rightarrow>* c"
|RedTran: "c \<rightarrow> c' \<Longrightarrow> c' \<rightarrow>* c'' \<Longrightarrow> c \<rightarrow>* c''"

lemma rtred_trans: "c \<rightarrow>* c' \<Longrightarrow> c' \<rightarrow>* c'' \<Longrightarrow> c \<rightarrow>* c''"
  apply(induction rule: rtred.induct)
   apply (simp add: RedRefl RedTran)
  by (simp add: RedTran)

lemma rtred_backwards:
  assumes "c \<rightarrow>* c''"
    shows "c=c'' \<or> (\<exists>c'. (c\<rightarrow>*c') \<and> (c'\<rightarrow>c''))"
  using assms
proof(induction rule: rtred.induct)
  case (RedRefl c)
  then show ?case by simp
next
  case (RedTran c c' c'')
  then show ?case using RedRefl rtred.RedTran by blast
qed

fun rSeq :: "'var cmd \<Rightarrow> nat" where
 "rSeq (Seq C\<^sub>1 C\<^sub>2) = Suc (rSeq C\<^sub>2)"
|"rSeq (If C\<^sub>1 C\<^sub>2) = max (rSeq C\<^sub>1) (rSeq C\<^sub>2)"
|"rSeq _ = 0"

fun right_most :: "'var cmd \<Rightarrow> 'var cmd set" where
 "right_most (Seq C\<^sub>1 C\<^sub>2) = right_most C\<^sub>2"
|"right_most (If C\<^sub>1 C\<^sub>2) = (right_most C\<^sub>1) \<union> (right_most C\<^sub>2)"
|"right_most C = {C}"

lemma right_most_is_not_seq: "\<forall>C\<^sub>1 C\<^sub>2. Seq C\<^sub>1 C\<^sub>2 \<notin> right_most C"
  apply(induction C)
  by auto

lemma rSeq_monotonic_decreasing:
  assumes "(C, \<sigma>) \<rightarrow> (C', \<sigma>')"
      and "\<nexists>C\<^sub>0. C=While C\<^sub>0"
    shows "rSeq C' \<le> rSeq C"
  using assms
  apply(induction "(C, \<sigma>)" "(C', \<sigma>')" rule: red.induct)
  by auto

lemma while_predecessor:
  assumes "(C, \<sigma>) \<rightarrow> (While C\<^sub>0, \<sigma>')"
    shows "(C = Seq Skip (While C\<^sub>0)) \<or> (\<exists>C'. C = If C' (While C\<^sub>0)) \<or> (\<exists>C'. C= If (While C\<^sub>0) C')"
  using assms 
  apply(induction "(C, \<sigma>)" "(While C\<^sub>0, \<sigma>')" rule: red.induct)
  by auto

lemma if_predecessor:
  assumes "(C, \<sigma>) \<rightarrow> (If C\<^sub>1 C\<^sub>2, \<sigma>')"
  shows "(C = Seq Skip (If C\<^sub>1 C\<^sub>2)) \<or> (\<exists>C'. C = If C' (If C\<^sub>1 C\<^sub>2)) \<or> (\<exists>C'. C = If (If C\<^sub>1 C\<^sub>2) C') \<or> (\<exists>C\<^sub>0. C = While C\<^sub>0 \<and>  C\<^sub>1 = Seq C\<^sub>0 C \<and> C\<^sub>2 = Skip)"
  using assms
  apply(induction "(C, \<sigma>)" "(If C\<^sub>1 C\<^sub>2, \<sigma>')" rule: red.induct)
  by auto

lemma right_most_contains_while_inv:
  assumes "(C, \<sigma>) \<rightarrow> (C', \<sigma>')"
      and "\<forall>C\<^sub>0. While C\<^sub>0 \<notin> right_most C"
    shows "\<forall>C\<^sub>0. While C\<^sub>0 \<notin> right_most C'"
  using assms
  apply(induction "(C, \<sigma>)" "(C', \<sigma>')")
  by auto

lemma right_most_contains_while_inv_rt:
  assumes "(C, \<sigma>) \<rightarrow>* (C', \<sigma>')"
      and "\<forall>C\<^sub>0. While C\<^sub>0 \<notin> right_most C"
    shows "\<forall>C\<^sub>0. While C\<^sub>0 \<notin> right_most C'"
  using assms
proof(induction "(C, \<sigma>)" "(C', \<sigma>')" arbitrary: C \<sigma>)
  case (RedRefl)
  then show ?case by blast
next
  case (RedTran c')
  then show ?case by (metis prod_cases3 right_most_contains_while_inv)
qed

lemma rSeq_monotonic_decreasing_rt:
  assumes "(C, \<sigma>) \<rightarrow>* (C'', \<sigma>'')"
      and "\<forall>C\<^sub>0. While C\<^sub>0 \<notin> (right_most C)"
    shows "rSeq C'' \<le> rSeq C"
  using assms
proof(induction "(C, \<sigma>)" "(C'', \<sigma>'')" arbitrary: C \<sigma> rule: rtred.induct)
  case (RedRefl)
  then show ?case by auto
next
  case (RedTran c')
  moreover obtain C' \<sigma>' where "c'=(C', \<sigma>')" using prod.exhaust_sel by blast
  ultimately have "(C, \<sigma>) \<rightarrow> (C', \<sigma>')" "(C', \<sigma>') \<rightarrow>* (C'', \<sigma>'')" by auto
  then have "rSeq C'' \<le> rSeq C'" by (simp add: RedTran.hyps(3) RedTran.prems \<open>c'=(C', \<sigma>')\<close> right_most_contains_while_inv)
  moreover have "\<nexists>C\<^sub>0. C = While C\<^sub>0" using RedTran.prems by auto
  then have "rSeq C' \<le> rSeq C" using \<open>(C, \<sigma>) \<rightarrow> (C', \<sigma>')\<close> rSeq_monotonic_decreasing[of C \<sigma> C' \<sigma>'] by blast
  ultimately show ?case by simp
qed

lemma seq_rev:
  assumes "(Seq C\<^sub>1 C\<^sub>2, s) \<rightarrow> (Seq C\<^sub>1' C\<^sub>2, s')"
    shows "(C\<^sub>1, s) \<rightarrow> (C\<^sub>1', s')"
  using assms
  apply(induction "(Seq C\<^sub>1 C\<^sub>2, s)" "(Seq C\<^sub>1' C\<^sub>2, s')" rule: red.induct)
  by auto

lemma seq_rev_rt:
  assumes "(Seq C\<^sub>1 C\<^sub>2, s) \<rightarrow>* (Seq C\<^sub>1'' C\<^sub>2, s'')"
      and "\<forall>C\<^sub>0. While C\<^sub>0 \<notin> right_most C\<^sub>2"
    shows " (C\<^sub>1, s) \<rightarrow>* (C\<^sub>1'', s'')"
  using assms
proof(induction "(Seq C\<^sub>1 C\<^sub>2, s)" "(Seq C\<^sub>1'' C\<^sub>2, s'')" arbitrary: C\<^sub>1 s)
  case (RedRefl)
  then show ?case by (simp add: rtred.RedRefl)
next
  case (RedTran c')
  moreover obtain C' \<sigma>' where "c'=(C', \<sigma>')" using prod.exhaust_sel by blast
  ultimately have C'\<sigma>'_def: "(Seq C\<^sub>1 C\<^sub>2, s) \<rightarrow> (C', \<sigma>')" "(C', \<sigma>') \<rightarrow>* (Seq C\<^sub>1'' C\<^sub>2, s'')" by auto
  moreover have "\<forall>C\<^sub>0. While C\<^sub>0 \<notin> right_most (Seq C\<^sub>1 C\<^sub>2)" using RedTran.prems by simp
  ultimately have "\<forall>C\<^sub>0. While C\<^sub>0 \<notin> right_most C'" using right_most_contains_while_inv by auto
  then have "\<exists>C\<^sub>1'. C' = Seq C\<^sub>1' C\<^sub>2" using C'\<sigma>'_def assms(1) rSeq_monotonic_decreasing_rt by (metis not_less_eq_eq red_Seq_elim rSeq.simps(1) right_most.simps(1))
  then show ?case by (metis (no_types, lifting) RedTran.hyps(1) RedTran.hyps(3) RedTran.prems \<open>c'=(C', \<sigma>')\<close> rtred.RedTran seq_rev)
qed

fun red_iter :: "'var conf \<Rightarrow> nat \<Rightarrow> 'var conf \<Rightarrow> bool" where
 "red_iter c 0 c' \<longleftrightarrow> c=c'"
|"red_iter c (Suc n) c'' \<longleftrightarrow> (\<exists>c'. (c \<rightarrow> c') \<and> red_iter c' n c'')"

lemma rtred_eq_red_iter_bkw: "red_iter c n c' \<Longrightarrow> c \<rightarrow>* c'"
  apply(induction n arbitrary: c c')
   apply(simp add: RedRefl)
  by (meson red_iter.simps(2) rtred.simps)

definition red_iter_closed :: "'var conf \<Rightarrow> 'var conf \<Rightarrow> bool" where
 "red_iter_closed c c' \<longleftrightarrow> (\<exists>n. red_iter c n c')"

lemma rtred_eq_red_iter_closed_fwd: "c \<rightarrow>* c' \<Longrightarrow> red_iter_closed c c'"
  apply(induction rule: rtred.induct)
  using red_iter.simps(1) red_iter_closed_def apply blast
  by (meson red_iter.simps(2) red_iter_closed_def)

lemma rtred_eq_red_iter_closed_bkw: "red_iter_closed c c' \<Longrightarrow> c \<rightarrow>* c'"
  using red_iter_closed_def rtred_eq_red_iter_bkw by blast

lemma rtred_eq_red_iter_closed: "(c \<rightarrow>* c') \<longleftrightarrow> red_iter_closed c c'"
  using rtred_eq_red_iter_closed_bkw rtred_eq_red_iter_closed_fwd by blast

definition dsem :: "'var cmd \<Rightarrow> ('var state \<times> 'var state) set" ("\<lbrakk>_\<rbrakk>" [0] 91) where
 "\<lbrakk>C\<rbrakk> = {(\<sigma>, \<sigma>')| \<sigma> \<sigma>'. ((C, \<sigma>) \<rightarrow>* (Skip, \<sigma>'))}"

lemma dsem_char:
  "(\<sigma>, \<sigma>')\<in>\<lbrakk>C\<rbrakk> \<longleftrightarrow> ((C, \<sigma>) \<rightarrow>* (Skip, \<sigma>'))"
  unfolding dsem_def by blast

(*tacitly: a valid path with length n, note that there are paths with length 0*)
definition a_path :: "'var conf list \<Rightarrow> nat \<Rightarrow> bool" where
 "a_path l n \<longleftrightarrow> length l = n+1 \<and> (\<forall>k<n. (l!k \<rightarrow> l!(k+1)))"

lemma a_pathI:
  assumes "length l = n+1"
      and "\<And>k. k<n \<Longrightarrow> l!k \<rightarrow> l!(k+1)"
    shows "a_path l n"
  using assms a_path_def by blast

(*tacitly: a successful path with length n, note that there are paths with length 0*)
definition is_path :: "'var conf list \<Rightarrow> 'var conf \<Rightarrow> nat \<Rightarrow> 'var conf \<Rightarrow> bool" where
 "is_path l c n c' \<longleftrightarrow> a_path l n \<and> l!0 = c \<and> l!n = c'"

lemma red_iter's_path:
  assumes "red_iter (C, \<sigma>) n (C'', \<sigma>'')"
  shows "\<exists>l. is_path l (C, \<sigma>) n (C'', \<sigma>'')" 
  using assms
proof(induction "(C, \<sigma>)" n "(C'', \<sigma>'')" arbitrary: C \<sigma> \<sigma>'' rule: red_iter.induct)
  case 1
  then have "(C, \<sigma>) = (C'', \<sigma>'')" by simp
  moreover obtain l where "l=[(C, \<sigma>)]" by auto
  ultimately show ?case using a_path_def is_path_def by (metis Suc_eq_plus1 length_Cons list.size(3) not_less_zero nth_Cons_0)
next
  case (2 n)
  then obtain C' \<sigma>' where 0: "(C, \<sigma>)\<rightarrow>(C', \<sigma>')" "red_iter (C', \<sigma>') n (C'', \<sigma>'')" by auto
  then obtain l where l_def: "is_path l (C', \<sigma>') n (C'', \<sigma>'')" using "2.hyps" by blast
  then obtain l' where l'_def: "l'=(C, \<sigma>)#l" by auto
  have "a_path l' (Suc n)"
  proof(rule a_pathI)
    have "length l = n+1" using l_def a_path_def is_path_def by blast
    then show "length l' = Suc n + 1" using l'_def by simp
  next
    fix k
    assume "k<Suc n"
    show "l' ! k \<rightarrow> l' ! (k + 1)"
    proof(cases "k=0")
      case True
      then show ?thesis using "0"(1) by (metis Suc_eq_plus1 is_path_def l'_def l_def nth_Cons_0 nth_Cons_Suc)
    next
      case False
      then show ?thesis using a_path_def is_path_def l_def by (smt (verit, ccfv_SIG) \<open>k < Suc n\<close> add_Suc_shift add_cancel_right_right add_diff_cancel_left' l'_def less_Suc_eq_0_disj nth_Cons_pos plus_1_eq_Suc)
    qed
  qed
  then show ?case using l_def l'_def is_path_def by (metis nth_Cons_0 nth_Cons_Suc)
qed

lemma dsem_path_left_fixed_long:
  assumes "length l = n+1"
      and "\<forall>k'<n. (l!k' \<rightarrow> l!(k'+1))"
      and "k\<le>n"
    shows "l!0 \<rightarrow>* l!k"
  using assms
proof(induction n arbitrary: l k)
  case 0
  then show ?case
    by (simp add: RedRefl)
next
  case (Suc n)
  then obtain c l' where l'_def: "l = c # l'"
    by (metis Suc_eq_plus1 length_Suc_conv)
  then have same: "\<forall>k\<le>n. l'!k = l!(k+1)"
    by auto
  then have "length l' = n + 1" "\<forall>k'<n. (l'!k' \<rightarrow> l'!(k'+1))" "n \<le> Suc n" 
      apply (metis Suc.prems(1) Suc_eq_plus1 diff_Suc_1 l'_def length_Cons)
     using Suc.prems(2) Suc_eq_plus1 Suc_leI Suc_less_eq less_or_eq_imp_le same apply presburger
     by simp
   then have "l'!0 \<rightarrow>* l'!(k-1)"
     using Suc.IH Suc.prems(3) by force
  then show ?case
    by (metis (no_types) Suc.prems(2) l'_def le_add2 le_add_same_cancel2 nth_Cons' rtred.simps same zero_less_Suc)
qed

lemma dsem_path_long:
  assumes "length l = n+1"
      and "\<forall>k'<n. (l!k' \<rightarrow> l!(k'+1))"
      and "k\<^sub>l\<le>k\<^sub>r"
      and "k\<^sub>r\<le>n"
    shows "l!k\<^sub>l \<rightarrow>* l!k\<^sub>r"
  using assms
proof(-)
  obtain l' where l'_def: "l' = drop k\<^sub>l l"
    by simp
  then have same: "\<forall>k\<le>n-k\<^sub>l. l'!k = l!(k+k\<^sub>l)"
    by (metis (no_types, lifting) add.commute assms(1) assms(3) assms(4) le_add1 le_trans nth_drop)
  have props: "length l' = n-k\<^sub>l+1" "\<forall>k'<n-k\<^sub>l. (l'!k' \<rightarrow> l'!(k'+1))" "k\<^sub>r-k\<^sub>l\<le>n-k\<^sub>l" 
      apply (metis Nat.add_diff_assoc2 assms(1) assms(3) assms(4) l'_def le_trans length_drop)
    using same assms(1) assms(2) assms(4) diff_le_mono by auto
  then have "l'!0 \<rightarrow>* l'!(k\<^sub>r-k\<^sub>l)"
    by (meson eq_imp_le dsem_path_left_fixed_long)
  thus "l!k\<^sub>l \<rightarrow>* l!k\<^sub>r" using same
    by (simp add: assms(3) props(3))
qed

lemma dsem_path:
  assumes "a_path l n"
      and "k\<^sub>l\<le>k\<^sub>r"
      and "k\<^sub>r\<le>n"
    shows "l!k\<^sub>l \<rightarrow>* l!k\<^sub>r"
  using assms by (simp add: a_path_def dsem_path_long)

lemma dsem_path_left_fixed_same_len_long:
  assumes "length l = n+1"
      and "\<forall>k'<n. (l!k' \<rightarrow> l!(k'+1))"
      and "k\<le>n"
    shows "red_iter (l!0) k (l!k)"
  using assms
proof(induction n arbitrary: l k)
  case 0
  then show ?case
    by (simp add: RedRefl)
next
  case (Suc n)
  then obtain c l' where l'_def: "l = c # l'"
    by (metis Suc_eq_plus1 length_Suc_conv)
  then have same: "\<forall>k\<le>n. l'!k = l!(k+1)"
    by auto
  then have "length l' = n + 1" "\<forall>k'<n. (l'!k' \<rightarrow> l'!(k'+1))" "n \<le> Suc n" 
      apply (metis Suc.prems(1) Suc_eq_plus1 diff_Suc_1 l'_def length_Cons)
     using Suc.prems(2) Suc_eq_plus1 Suc_leI Suc_less_eq less_or_eq_imp_le same apply presburger
     by simp
   then have "red_iter (l'!0) (k-1) (l'!(k-1))"
     using Suc.IH Suc.prems(3) by simp
  then show ?case by (metis (no_types, lifting) Suc.prems(2) diff_Suc_1 l'_def le_add2 le_add_same_cancel2 nth_Cons' red_iter.elims(3) red_iter.simps(1) same zero_less_Suc)
qed

lemma dsem_path_same_len_long:
  assumes "length l = n+1"
      and "\<forall>k'<n. (l!k' \<rightarrow> l!(k'+1))"
      and "k\<^sub>l\<le>k\<^sub>r"
      and "k\<^sub>r\<le>n"
    shows "red_iter (l!k\<^sub>l) (k\<^sub>r-k\<^sub>l) (l!k\<^sub>r)"
  using assms
proof(-)
  obtain l' where l'_def: "l' = drop k\<^sub>l l"
    by simp
  then have same: "\<forall>k\<le>n-k\<^sub>l. l'!k = l!(k+k\<^sub>l)"
    by (metis (no_types, lifting) add.commute assms(1) assms(3) assms(4) le_add1 le_trans nth_drop)
  have props: "length l' = n-k\<^sub>l+1" "\<forall>k'<n-k\<^sub>l. (l'!k' \<rightarrow> l'!(k'+1))" "k\<^sub>r-k\<^sub>l\<le>n-k\<^sub>l" 
      apply (metis Nat.add_diff_assoc2 assms(1) assms(3) assms(4) l'_def le_trans length_drop)
    using same assms(1) assms(2) assms(4) diff_le_mono by auto
  then have "red_iter (l'!0) (k\<^sub>r-k\<^sub>l) (l'!(k\<^sub>r-k\<^sub>l))" using dsem_path_left_fixed_same_len_long by blast
  thus "red_iter (l!k\<^sub>l) (k\<^sub>r-k\<^sub>l) (l!k\<^sub>r)" using same
    by (simp add: assms(3) props(3))
qed

lemma dsem_path_same_len:
  assumes "a_path l n"
      and "k\<^sub>l\<le>k\<^sub>r"
      and "k\<^sub>r\<le>n"
    shows "red_iter (l!k\<^sub>l) (k\<^sub>r-k\<^sub>l) (l!k\<^sub>r)"
  using assms by (simp add: a_path_def dsem_path_same_len_long)

lemma skip_deadend_rt:
  assumes "(Skip, \<sigma>) \<rightarrow>* c'"
  shows "c'=(Skip, \<sigma>)"
  using assms skip_deadend by (metis fst_conv rtred.cases)

lemma dsem_skip: "\<lbrakk>Skip\<rbrakk> = {(\<sigma>, \<sigma>)| \<sigma>. True}"
proof
  show "\<lbrakk>Skip\<rbrakk> \<supseteq> {(\<sigma>, \<sigma>)| \<sigma>. True}" using RedRefl dsem_def by fast
  show "\<lbrakk>Skip\<rbrakk> \<subseteq> {(\<sigma>, \<sigma>)| \<sigma>. True}" unfolding dsem_def using skip_deadend_rt by fastforce
qed

lemma seq1_unique: (*in the new iteration used commonly with fst_conv*)
  assumes "(Seq Skip C\<^sub>2, \<sigma>) \<rightarrow> c'"
    shows "c' = (C\<^sub>2, \<sigma>)"
  using assms
  apply(induction "(Seq Skip C\<^sub>2, \<sigma>)" c' rule: red.induct)
             apply(auto)
  by (meson fst_conv skip_deadend)

lemma seq1_preserves_dsem:
  "((Seq Skip C\<^sub>2, \<sigma>) \<rightarrow>* (Skip, \<sigma>')) \<longleftrightarrow> ((C\<^sub>2, \<sigma>) \<rightarrow>* (Skip, \<sigma>'))" 
  by (metis (no_types, lifting) Pair_inject RedSeq1 RedTran cmd.distinct(23) rtred.cases seq1_unique)

lemma dsem_seq1: "\<lbrakk>Seq Skip C\<^sub>2\<rbrakk> = \<lbrakk>C\<^sub>2\<rbrakk>" using dsem_def seq1_preserves_dsem by blast

lemma seq_trcl:
  assumes "(C\<^sub>1, \<sigma>) \<rightarrow>* (C\<^sub>1', \<sigma>')"
    shows "(Seq C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow>* (Seq C\<^sub>1' C\<^sub>2, \<sigma>')"
  using assms
  apply(induction "(C\<^sub>1, \<sigma>)" "(C\<^sub>1', \<sigma>')" arbitrary: C\<^sub>1 \<sigma> \<sigma>' rule: rtred.induct)
   apply (simp add: RedRefl)
  using RedSeq2 rtred.simps by fastforce

(* Opposite direction not true: (x++;(x++)*, (x=0)) \<rightarrow>\<^sup>5 (skip;(x++), (x=2)) *)
(* (x++;(x++)*, (x=0)) \<rightarrow> (skip;(x++)*, (x=1)) \<rightarrow> ((x++)*, (x=1)) \<rightarrow> (x++;(x++)* + skip, (x=1)) \<rightarrow> (x++;(x++)*, (x=1)) \<rightarrow> (skip;(x++)*, (x=2)) *)
(* However, check seq_rev_path for particular case, where the opposite direction holds *)
lemma seq_trcl_same_len:
  "red_iter (C\<^sub>1, \<sigma>) n (C\<^sub>1'', s'') \<Longrightarrow> red_iter (Seq C\<^sub>1 C\<^sub>2, \<sigma>) n (Seq C\<^sub>1'' C\<^sub>2, s'')" 
proof(induction "(C\<^sub>1, \<sigma>)" n "(C\<^sub>1'', s'')" arbitrary: C\<^sub>1 \<sigma> s'' rule: red_iter.induct)
  case 1
  then show ?case by fastforce
next
  case (2 n)
  then obtain C\<^sub>1' \<sigma>' where "(C\<^sub>1, \<sigma>) \<rightarrow> (C\<^sub>1', \<sigma>')" "red_iter (C\<^sub>1', \<sigma>') n (C\<^sub>1'', s'')" using red_iter.simps(2) by auto
  then have "red_iter (Seq C\<^sub>1' C\<^sub>2, \<sigma>') n (Seq C\<^sub>1'' C\<^sub>2, s'')" by (simp add: "2.hyps")
  moreover have "(Seq C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow> (Seq C\<^sub>1' C\<^sub>2, \<sigma>')" using \<open>(C\<^sub>1, \<sigma>) \<rightarrow> (C\<^sub>1', \<sigma>')\<close> by (simp add: RedSeq2)
  ultimately show ?case by (meson red_iter.simps(2))
qed

lemma seq2_non_skip:
  assumes "(Seq C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow> c'"
      and "C\<^sub>1 \<noteq> Skip"
    shows "\<exists>C\<^sub>1'. fst c' = Seq C\<^sub>1' C\<^sub>2"
  using assms
  apply(induction "(Seq C\<^sub>1 C\<^sub>2, \<sigma>)" c' rule: red.induct)
  by auto

lemma seq_path:
  assumes "length l = n+1"
      and "l!0 = (Seq C\<^sub>1 C\<^sub>2, \<sigma>)"
      and "\<forall>k<n. (l!k \<rightarrow> l!(k+1))"
    shows "(\<exists>k\<le>n. fst (l!k) = Seq Skip C\<^sub>2 \<and> (\<forall>k'<k. \<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip)) \<or> (\<forall>k\<le>n. \<exists>C\<^sub>1'. fst (l!k) = Seq C\<^sub>1' C\<^sub>2)"
  using assms
proof(induction n arbitrary: l C\<^sub>1 \<sigma>)
  case 0
  then show ?case by simp
next
  case (Suc n)
  then have "length l>1" by linarith 
  then show ?case
  proof(cases "C\<^sub>1=Skip")
    case True
    then show ?thesis using Suc.prems(2) by auto
  next
    case False
    then have "C\<^sub>1\<noteq>Skip" by blast
    then obtain C\<^sub>1' \<sigma>' where C1'_def: "fst (l!1) = Seq C\<^sub>1' C\<^sub>2" "snd (l!1) = \<sigma>'" using Suc.prems(2) Suc.prems(3) by (metis add_0 seq2_non_skip zero_less_Suc)
    obtain l' where l'_def: "l = (Seq C\<^sub>1 C\<^sub>2, \<sigma>) # l'"
      by (metis Cons_nth_drop_Suc Suc.prems(1) Suc.prems(2) add_gr_0 drop0 gr0_conv_Suc)
    then have same: "\<forall>k\<le>n. l'!k = l!(k+1)"
      by auto
    then have l'_prop: "length l' = n+1" "l'!0 = (Seq C\<^sub>1' C\<^sub>2, \<sigma>')" "\<forall>k<n. (l'!k \<rightarrow> l'!(k+1))" 
        apply (metis l'_def Suc.prems(1) Suc_eq_plus1 diff_Suc_1 length_Cons)
       apply (metis C1'_def(1) C1'_def(2) One_nat_def \<open>l = (Seq C\<^sub>1 C\<^sub>2, \<sigma>) # l'\<close> nth_Cons_Suc prod.collapse)
      using Suc.prems(3) l'_def by force
    then have cs: "(\<exists>k\<le>n. fst (l'!k) = Seq Skip C\<^sub>2 \<and> (\<forall>k'<k. \<exists>C\<^sub>1'. fst (l'!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip)) \<or> (\<forall>k\<le>n. \<exists>C\<^sub>1'. fst (l'!k) = Seq C\<^sub>1' C\<^sub>2)" using Suc.IH by simp
    then show ?thesis 
    proof(cases "\<forall>k\<le>n. \<exists>C\<^sub>1'. fst (l'!k) = Seq C\<^sub>1' C\<^sub>2")
      case True
      then have "\<forall>k\<le>Suc n. \<exists>C\<^sub>1'. fst (l!k) = Seq C\<^sub>1' C\<^sub>2" using l'_def same by (metis (no_types, opaque_lifting) Suc.prems(2) Suc_eq_plus1 add_diff_cancel_right fst_conv gr0_conv_Suc gr_implies_not0 linorder_less_linear nless_le nth_Cons_Suc order_refl zero_less_diff)
      then show ?thesis by fast
    next
      case False
      then obtain k where "k\<le>n" "fst (l'!k) = Seq Skip C\<^sub>2" "\<forall>k'<k. \<exists>C\<^sub>1'. fst (l'!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip" using cs by blast
      moreover have "\<And>k'. k'<Suc k \<Longrightarrow> \<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip"
      proof(-)
        fix k'
        assume "k'<Suc k"
        then show "\<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip"
        proof(cases "k'=0")
          case True
          then show ?thesis using Suc.prems(2) \<open>C\<^sub>1 \<noteq> Skip\<close> by auto
        next
          case False
          then show ?thesis using \<open>\<forall>k'<k. \<exists>C\<^sub>1'. fst (l' ! k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip\<close> \<open>k' < Suc k\<close> l'_def by force
        qed
      qed
      then have "\<forall>k'<Suc k. \<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip" by fastforce
      ultimately show ?thesis using l'_def by (metis Suc_le_mono nth_Cons_Suc)
    qed
  qed
qed

lemma seq_rhs_rt:
  assumes "(Seq C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow>* (C'', \<sigma>'')"
    shows "(\<exists>C\<^sub>1'. C''=Seq C\<^sub>1' C\<^sub>2) \<or> (\<exists>\<sigma>'. ((Seq C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow>* (Seq Skip C\<^sub>2, \<sigma>')) \<and> ((C\<^sub>2, \<sigma>') \<rightarrow>* (C'', \<sigma>'')))"
proof(-)
  obtain n where n_def: "red_iter (Seq C\<^sub>1 C\<^sub>2, \<sigma>) n (C'', \<sigma>'')" using red_iter_closed_def rtred_eq_red_iter_closed_fwd assms(1) by blast
  then obtain l where l_def: "is_path l (Seq C\<^sub>1 C\<^sub>2, \<sigma>) n (C'', \<sigma>'')" using red_iter's_path by blast
  then have "length l = n+1" "l!0 = (Seq C\<^sub>1 C\<^sub>2, \<sigma>)" "\<forall>k<n. (l!k \<rightarrow> l!(k+1))" unfolding is_path_def a_path_def by auto
  then have cs: "(\<exists>k\<le>n. fst (l!k) = Seq Skip C\<^sub>2 \<and> (\<forall>k'<k. \<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip)) \<or> (\<forall>k\<le>n. \<exists>C\<^sub>1'. fst (l!k) = Seq C\<^sub>1' C\<^sub>2)" using seq_path[of l n C\<^sub>1 C\<^sub>2 \<sigma>] by argo
  then show ?thesis
  proof(cases "\<exists>k\<le>n. fst (l!k) = Seq Skip C\<^sub>2 \<and> (\<forall>k'<k. \<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip)")
    case True
    then obtain k where k_def: "k\<le>n" "fst (l!k) = Seq Skip C\<^sub>2" "\<forall>k'<k. \<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip" by blast
    then show ?thesis
    proof(cases "k=n")
      case True
      then show ?thesis by (metis fst_conv is_path_def k_def(2) l_def)
    next
      case False
      moreover obtain \<sigma>' where \<sigma>'_def: "\<sigma>' = snd (l!k)" by simp
      ultimately have 0: "(Seq C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow>* (Seq Skip C\<^sub>2, \<sigma>')" by (metis is_path_def k_def(1) k_def(2) l_def prod.collapse dsem_path zero_le)
      have "l!k \<rightarrow> l!(k+1)" using k_def(1) False l_def unfolding is_path_def a_path_def by force
      then have "l!(k+1) = (C\<^sub>2, \<sigma>')" using k_def(2) \<sigma>'_def by (metis prod.collapse seq1_unique)
      moreover have "k+1\<le>n" using k_def(1) False by fastforce
      then have "l!(k+1) \<rightarrow>* l!n" by (meson \<open>\<forall>k<n. (l!k \<rightarrow> l!(k + 1))\<close> \<open>length l = n+1\<close> order_refl dsem_path_long)
      ultimately have "(C\<^sub>2, \<sigma>') \<rightarrow>* (C'', \<sigma>'')" using l_def unfolding is_path_def by simp
      then show ?thesis using "0" by blast
    qed
  next
    case False
    then show ?thesis by (metis (mono_tags, lifting) cs fst_conv is_path_def l_def order_le_less)
  qed
qed

lemma seq_has_skip_in_path_long:
  assumes "length l = n+1"
      and "l!0 = (Seq C\<^sub>1 C\<^sub>2, \<sigma>)"
      and "l!n = c''" 
      and "\<forall>k<n. (l!k \<rightarrow> l!(k+1))"
      and "\<forall>C\<^sub>1''. fst c'' \<noteq> Seq C\<^sub>1'' C\<^sub>2"
    shows "\<exists>k<n. fst (l!k) = Seq Skip C\<^sub>2 \<and> (\<forall>k'<k. \<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip)"
proof(cases "C\<^sub>1=Skip")
  case True
  then show ?thesis using assms(2) by (metis assms(3) assms(5) fst_conv less_nat_zero_code neq0_conv)
next
  case False
  then have "\<not>(\<forall>k\<le>n. \<exists>C\<^sub>1'. fst (l!k) = Seq C\<^sub>1' C\<^sub>2)" using assms(3) assms(5) by auto
  moreover have "(\<exists>k\<le>n. fst (l!k) = Seq Skip C\<^sub>2 \<and> (\<forall>k'<k. \<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip)) \<or> (\<forall>k\<le>n. \<exists>C\<^sub>1'. fst (l!k) = Seq C\<^sub>1' C\<^sub>2)" using assms seq_path by blast
  ultimately show ?thesis using le_eq_less_or_eq by auto
qed

lemma seq_has_skip_in_path:
  assumes "is_path l (Seq C\<^sub>1 C\<^sub>2, \<sigma>) n c''" 
      and "\<forall>C\<^sub>1''. fst c'' \<noteq> Seq C\<^sub>1'' C\<^sub>2"
    shows "\<exists>k<n. fst (l!k) = Seq Skip C\<^sub>2 \<and> (\<forall>k'<k. \<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip)"
proof(cases "C\<^sub>1=Skip")
  case True
  then show ?thesis by (metis assms(1) assms(2) fst_conv is_path_def less_nat_zero_code neq0_conv)
next
  case False
  then show ?thesis using assms seq_has_skip_in_path_long a_path_def is_path_def by blast
qed

lemma seq_rev_path_long:
  assumes "length l = n+1"
      and "l!0 = (Seq C\<^sub>1 C\<^sub>2, \<sigma>)"
      and "l!n = (Seq C\<^sub>1'' C\<^sub>2, \<sigma>'')" 
      and "\<forall>k<n. (l!k \<rightarrow> l!(k+1))"
      and "\<forall>k<n. \<exists>C\<^sub>1'. fst (l!k) = Seq C\<^sub>1' C\<^sub>2"
    shows "red_iter (C\<^sub>1, \<sigma>) n (C\<^sub>1'', \<sigma>'')"
  using assms
proof(induction n arbitrary: l C\<^sub>1 \<sigma> \<sigma>'')
  case 0
  then show ?case by simp
next
  case (Suc n)
  then obtain C\<^sub>1' \<sigma>' where l1: "l!1 = (Seq C\<^sub>1' C\<^sub>2, \<sigma>')"
    by (metis Suc_eq_plus1 fst_conv linorder_neqE_nat not_add_less2 surj_pair)
  have "length l > 0"
    by (simp add: Suc.prems(1))
  then obtain l' where l'_def: "l = (Seq C\<^sub>1 C\<^sub>2, \<sigma>) # l'"
    by (metis Suc.prems(2) hd_conv_nth length_greater_0_conv list.collapse)
  then have same: "\<forall>k\<le>n. l'!k = l!(k+1)"
    by auto
  then have l'_prop: "\<forall>k<n. (l'!k \<rightarrow> l'!(k+1))" "\<forall>k<n. \<exists>C\<^sub>1'. fst (l'!k) = Seq C\<^sub>1' C\<^sub>2"
     apply (metis Suc.prems(4) Suc_eq_plus1 Suc_less_eq less_eq_Suc_le less_or_eq_imp_le)
    using Suc.prems(5) l'_def by auto
  then have "red_iter (C\<^sub>1', \<sigma>') n (C\<^sub>1'', \<sigma>'')"
    by (metis One_nat_def Suc.IH Suc.prems(1) Suc.prems(3) Suc_eq_plus1 add_left_cancel l'_def l1 length_Cons nth_Cons_Suc plus_1_eq_Suc)
  then show ?case
    by (metis One_nat_def Suc.prems(2) Suc.prems(4) add.commute l1 plus_1_eq_Suc red_iter.simps(2) seq_rev zero_less_Suc)
qed

lemma seq_rev_path:
  assumes "is_path l (Seq C\<^sub>1 C\<^sub>2, \<sigma>) n (Seq C\<^sub>1'' C\<^sub>2, \<sigma>'')"
      and "\<forall>k<n. \<exists>C\<^sub>1'. fst (l!k) = Seq C\<^sub>1' C\<^sub>2"
    shows "red_iter (C\<^sub>1, \<sigma>) n (C\<^sub>1'', \<sigma>'')"
  using assms seq_rev_path_long a_path_def is_path_def by blast

lemma seq_has_skip:
  assumes "(Seq C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow>* (Skip, \<sigma>'')"
    shows "\<exists>\<sigma>'. ((C\<^sub>1, \<sigma>) \<rightarrow>* (Skip, \<sigma>')) \<and> ((Seq Skip C\<^sub>2, \<sigma>') \<rightarrow>* (Skip, \<sigma>''))"
proof(-)
  have "red_iter_closed (Seq C\<^sub>1 C\<^sub>2, \<sigma>) (Skip, \<sigma>'')" using assms by (simp add: rtred_eq_red_iter_closed)
  then obtain n where n_def: "red_iter (Seq C\<^sub>1 C\<^sub>2, \<sigma>) n (Skip, \<sigma>'')" using red_iter_closed_def by blast
  then obtain l where l_def: "is_path l (Seq C\<^sub>1 C\<^sub>2, \<sigma>) n (Skip, \<sigma>'')" using red_iter's_path by blast
  then obtain k where k_def: "k<n" "fst (l!k) = Seq Skip C\<^sub>2" "\<forall>k'<k. \<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' C\<^sub>2 \<and> C\<^sub>1' \<noteq> Skip" using seq_has_skip_in_path by (metis cmd.distinct(23) fst_conv)
  obtain l' where l'_def: "l'=take (k+1) l" by blast
  have "\<forall>k'<k. (l!k' \<rightarrow> l!(k'+1))" using l_def is_path_def a_path_def k_def(1) by (metis add_lessD1 canonically_ordered_monoid_add_class.lessE)
  moreover have "length l = n+1" using l_def is_path_def a_path_def by blast
  then have "length l' = k+1" using k_def(1) l'_def \<open>length l = n + 1\<close> by force
  ultimately have "a_path l' k" using a_path_def[of l' k] by (simp add: l'_def)
  moreover have "l'!0 = (Seq C\<^sub>1 C\<^sub>2, \<sigma>)" using is_path_def l'_def l_def by fastforce
  ultimately have "is_path l' (Seq C\<^sub>1 C\<^sub>2, \<sigma>) k (l!k)" using is_path_def[of l' "(Seq C\<^sub>1 C\<^sub>2, \<sigma>)" k] using l'_def by simp
  moreover have "\<forall>k'<k. \<exists>C\<^sub>1'. fst (l'!k') = Seq C\<^sub>1' C\<^sub>2" using k_def(3) l'_def by auto
  then have "red_iter (C\<^sub>1, \<sigma>) k (Skip, snd (l!k))" using seq_rev_path by (metis calculation k_def(2) prod.collapse)
  then show ?thesis by (metis (no_types, lifting) eq_imp_le is_path_def k_def(1) k_def(2) l_def less_imp_le_nat prod.collapse rtred_eq_red_iter_bkw dsem_path)
qed

lemma seq1'_preserves_dsem:
 "((Seq C Skip, \<sigma>) \<rightarrow>* (Skip, \<sigma>')) \<longleftrightarrow> ((C, \<sigma>) \<rightarrow>* (Skip, \<sigma>'))"
  by (meson RedRefl rtred_trans seq1_preserves_dsem seq_has_skip seq_trcl)

lemma dsem_seq1': "\<lbrakk>Seq C\<^sub>2 Skip\<rbrakk> = \<lbrakk>C\<^sub>2\<rbrakk>" using dsem_def seq1'_preserves_dsem by blast

lemma skip_comm_: "\<lbrakk>Seq C Skip\<rbrakk> = \<lbrakk>Seq Skip C\<rbrakk>" using dsem_seq1 dsem_seq1' by blast

lemma dsem_seq:
  "\<lbrakk>Seq C\<^sub>1 C\<^sub>2\<rbrakk> = {(\<sigma>, \<sigma>'')| \<sigma> \<sigma>''. \<exists>\<sigma>'. (\<sigma>, \<sigma>')\<in>\<lbrakk>C\<^sub>1\<rbrakk> \<and> (\<sigma>', \<sigma>'')\<in>\<lbrakk>C\<^sub>2\<rbrakk>}" 
proof(-)
  have "\<lbrakk>Seq C\<^sub>1 C\<^sub>2\<rbrakk> = {(\<sigma>, \<sigma>'')| \<sigma> \<sigma>''. ((Seq C\<^sub>1 C\<^sub>2), \<sigma>) \<rightarrow>* (Skip, \<sigma>'')}" using dsem_def by blast
  then have "\<lbrakk>Seq C\<^sub>1 C\<^sub>2\<rbrakk> = {(\<sigma>, \<sigma>'')| \<sigma> \<sigma>''. \<exists>\<sigma>'. ((C\<^sub>1, \<sigma>) \<rightarrow>* (Skip, \<sigma>')) \<and> ((Seq Skip C\<^sub>2, \<sigma>') \<rightarrow>* (Skip, \<sigma>''))}" by (smt (verit, ccfv_SIG) Collect_cong case_prodI2 eq_fst_iff fst_eqD rtred_trans seq_has_skip seq_trcl)
  then show ?thesis using dsem_char by (smt (verit, ccfv_SIG) Collect_cong dsem_seq1)
qed

lemma seq_assoc:
  "\<lbrakk>Seq (Seq C\<^sub>1 C\<^sub>2) C\<^sub>3\<rbrakk> = \<lbrakk>Seq C\<^sub>1 (Seq C\<^sub>2 C\<^sub>3)\<rbrakk>" (is "?X=?Y")
proof(-)
  have "?X = {(\<sigma>, \<sigma>''')| \<sigma> \<sigma>'''. \<exists>\<sigma>''. (\<sigma>, \<sigma>'')\<in>\<lbrakk>Seq C\<^sub>1 C\<^sub>2\<rbrakk> \<and> (\<sigma>'', \<sigma>''')\<in>\<lbrakk>C\<^sub>3\<rbrakk>}" using dsem_seq by blast
  then have "?X = {(\<sigma>, \<sigma>''')| \<sigma> \<sigma>'''. \<exists>\<sigma>''. (\<sigma>, \<sigma>'')\<in>{(\<sigma>, \<sigma>'')| \<sigma> \<sigma>''. \<exists>\<sigma>'. (\<sigma>, \<sigma>')\<in>\<lbrakk>C\<^sub>1\<rbrakk> \<and> (\<sigma>', \<sigma>'')\<in>\<lbrakk>C\<^sub>2\<rbrakk>} \<and> (\<sigma>'', \<sigma>''')\<in>\<lbrakk>C\<^sub>3\<rbrakk>}" unfolding dsem_seq by blast
  then have "?X = {(\<sigma>, \<sigma>''')| \<sigma> \<sigma>' \<sigma>'' \<sigma>'''. (\<sigma>, \<sigma>')\<in>\<lbrakk>C\<^sub>1\<rbrakk> \<and> (\<sigma>', \<sigma>'')\<in>\<lbrakk>C\<^sub>2\<rbrakk> \<and> (\<sigma>'', \<sigma>''')\<in>\<lbrakk>C\<^sub>3\<rbrakk>}" by blast
  moreover have "?Y = {(\<sigma>, \<sigma>''')| \<sigma> \<sigma>'''. \<exists>\<sigma>'. (\<sigma>, \<sigma>')\<in>\<lbrakk>C\<^sub>1\<rbrakk> \<and> (\<sigma>', \<sigma>''')\<in>\<lbrakk>Seq C\<^sub>2 C\<^sub>3\<rbrakk>}" using dsem_seq by blast
  then have "?Y = {(\<sigma>, \<sigma>''')| \<sigma> \<sigma>'''. \<exists>\<sigma>'. (\<sigma>, \<sigma>')\<in>\<lbrakk>C\<^sub>1\<rbrakk> \<and> (\<sigma>', \<sigma>''')\<in>{(\<sigma>', \<sigma>''')| \<sigma>' \<sigma>'''. \<exists>\<sigma>''. (\<sigma>', \<sigma>'')\<in>\<lbrakk>C\<^sub>2\<rbrakk> \<and> (\<sigma>'', \<sigma>''')\<in>\<lbrakk>C\<^sub>3\<rbrakk>}}" using dsem_seq by fast
  then have "?Y = {(\<sigma>, \<sigma>''')| \<sigma> \<sigma>' \<sigma>'' \<sigma>'''. (\<sigma>, \<sigma>')\<in>\<lbrakk>C\<^sub>1\<rbrakk> \<and> (\<sigma>', \<sigma>'')\<in>\<lbrakk>C\<^sub>2\<rbrakk> \<and> (\<sigma>'', \<sigma>''')\<in>\<lbrakk>C\<^sub>3\<rbrakk>}" by blast
  ultimately show ?thesis by argo
qed

lemma seq_skip_equiv: "((Seq C Skip, \<sigma>) \<rightarrow>* (Seq C' Skip, \<sigma>')) \<longleftrightarrow> ((C, \<sigma>) \<rightarrow>* (C', \<sigma>'))" (is "?X \<longleftrightarrow> ?Y")
proof
  show "?Y \<Longrightarrow> ?X" using seq_trcl by blast
  have "right_most Skip = {Skip}" by fastforce
  then have "\<forall>C\<^sub>0. While C\<^sub>0 \<notin> right_most Skip" by simp
  then show "?X \<Longrightarrow> ?Y" using seq_rev_rt[of C Skip \<sigma> C' \<sigma>'] by fast
qed

lemma assume_preserves_dsem:
  "((Assume b, \<sigma>)\<rightarrow>*(Skip, \<sigma>')) \<longleftrightarrow> \<sigma>=\<sigma>' \<and> fst \<sigma>\<in>b"
proof
  show "\<sigma>=\<sigma>' \<and> fst \<sigma>\<in>b \<Longrightarrow> ((Assume b, \<sigma>)\<rightarrow>*(Skip, \<sigma>'))" by (meson RedAssume RedRefl RedTran)
  show "((Assume b, \<sigma>)\<rightarrow>*(Skip, \<sigma>')) \<Longrightarrow> \<sigma>=\<sigma>' \<and> fst \<sigma>\<in>b"
  proof(-)
    have "\<And>\<sigma> \<sigma>'. ((Assume b, \<sigma>)\<rightarrow>(Skip, \<sigma>')) \<longleftrightarrow> \<sigma>=\<sigma>' \<and> fst \<sigma>\<in>b" using RedAssume by blast
    assume "(Assume b, \<sigma>)\<rightarrow>*(Skip, \<sigma>')"
    then have "red_iter_closed (Assume b, \<sigma>) (Skip, \<sigma>')" by (simp add: rtred_eq_red_iter_closed)
    then obtain n where n_def: "red_iter (Assume b, \<sigma>) n (Skip, \<sigma>')" using red_iter_closed_def by blast
    then obtain l where l_def: "is_path l (Assume b, \<sigma>) n (Skip, \<sigma>')" using red_iter's_path by blast
    then have 0: "\<forall>k<n. (l!k\<rightarrow>l!(k+1))" using is_path_def a_path_def by blast
    moreover have "l!0 = (Assume b, \<sigma>)" using is_path_def l_def by fast
    ultimately have "fst (l!1) = Skip" by (metis (no_types, opaque_lifting) add_0 cmd.distinct(57) fst_eqD is_path_def l_def not_gr_zero red_rhs(6))
    then have "n=1" by (metis "0" cmd.distinct(57) is_path_def l_def less_one nat_neq_iff prod.inject skip_deadend)
    then show "\<sigma>=\<sigma>' \<and> fst \<sigma>\<in>b" by (metis "0" add_0 fact_0 fact_gt_zero is_path_def l_def red_Assume_elim)
  qed
qed

lemma dsem_assume: "\<lbrakk>Assume b\<rbrakk> = {(\<sigma>, \<sigma>)| \<sigma>. fst \<sigma>\<in>b}"
  unfolding dsem_def by (smt (verit, ccfv_SIG) Collect_cong assume_preserves_dsem)

lemma if_preserves_dsem:
  "((If C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow>* (Skip, \<sigma>')) \<longleftrightarrow> ((C\<^sub>1, \<sigma>) \<rightarrow>* (Skip, \<sigma>')) \<or> ((C\<^sub>2, \<sigma>) \<rightarrow>* (Skip, \<sigma>'))" (is "?X \<longleftrightarrow> ?Y")
proof
  show "?X \<Longrightarrow> ?Y" using rtred.cases by fastforce
  show "?Y \<Longrightarrow> ?X" by (meson RedIf1 RedIf2 RedTran)
qed

lemma dsem_if:
  "\<lbrakk>If C\<^sub>1 C\<^sub>2\<rbrakk> = \<lbrakk>C\<^sub>1\<rbrakk> \<union> \<lbrakk>C\<^sub>2\<rbrakk>" 
proof(-)
  have "\<lbrakk>If C\<^sub>1 C\<^sub>2\<rbrakk> = {(\<sigma>, \<sigma>')| \<sigma> \<sigma>'. ((If C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow>* (Skip, \<sigma>'))}" using dsem_def by blast
  then have "\<lbrakk>If C\<^sub>1 C\<^sub>2\<rbrakk> = {(\<sigma>, \<sigma>')| \<sigma> \<sigma>'. ((C\<^sub>1, \<sigma>) \<rightarrow>* (Skip, \<sigma>')) \<or> ((C\<^sub>2, \<sigma>) \<rightarrow>* (Skip, \<sigma>'))}" using if_preserves_dsem by fastforce
  then have "\<lbrakk>If C\<^sub>1 C\<^sub>2\<rbrakk> = {(\<sigma>, \<sigma>')| \<sigma> \<sigma>'. ((C\<^sub>1, \<sigma>) \<rightarrow>* (Skip, \<sigma>'))} \<union> {(\<sigma>, \<sigma>')| \<sigma> \<sigma>'. ((C\<^sub>2, \<sigma>) \<rightarrow>* (Skip, \<sigma>'))}" by auto
  then show ?thesis unfolding dsem_def by fast
qed

lemma Iter_0_step: "red_iter (While C, \<sigma>) 0 c' \<Longrightarrow> c'=(While C, \<sigma>)" by simp

lemma Iter_1_step: "red_iter (While C, \<sigma>) 1 c' \<Longrightarrow> c'=(If (Seq C (While C)) Skip, \<sigma>)"
  by (metis (no_types, opaque_lifting) One_nat_def red_While_elim red_iter.simps(1) red_iter.simps(2) surj_pair)

lemma Iter_2_step: "red_iter (While C, \<sigma>) 2 c' \<Longrightarrow> c'=(Skip, \<sigma>) \<or> c'=(Seq C (While C), \<sigma>)"
  using Iter_1_step by (smt (verit) One_nat_def Suc_1 old.prod.exhaust red_If_elim red_iter.simps(1) red_iter.simps(2))

lemma Successfull_iter_is_ge2: "red_iter (While C, \<sigma>) n (Skip, \<sigma>') \<Longrightarrow> n>1"
  using Iter_0_step Iter_1_step using nat_neq_iff by blast

lemma Iter_ge2_steps: "is_path l (While C, \<sigma>) n c'' \<Longrightarrow> n>2 \<Longrightarrow> l!2=(Seq C (While C), \<sigma>)"
proof(-)
  assume asm: "is_path l (While C, \<sigma>) n c''" "n>2"
  then have "red_iter (While C, \<sigma>) 2 (l!2)" using is_path_def dsem_path_same_len by (metis (no_types, lifting) diff_zero less_or_eq_imp_le zero_le_numeral)
  moreover have "is_path l (While C, \<sigma>) n c'' \<Longrightarrow> n>2 \<Longrightarrow> l!2\<noteq>(Skip, \<sigma>)"
  proof(-)
    assume asm: "is_path l (While C, \<sigma>) n c''" "n>2"
    then have "length l > 3" using is_path_def a_path_def by (metis add_less_cancel_right nat_1_add_1 numeral_Bit1 numerals(1))
    then have "l!2 \<rightarrow> l!3" using asm is_path_def a_path_def by (metis (no_types, lifting) inc.simps(2) numeral_inc)
    then show ?thesis by (metis fst_conv skip_deadend)
  qed
  ultimately show ?thesis using Iter_2_step asm by blast
qed

(* Seems useless for now *)
lemma iter_has_next:
  assumes "is_path l (While C, \<sigma>) n (Skip, \<sigma>'')"
      and "n>2"
    shows "\<exists>k<n. k>2 \<and> fst (l!k) = (While C)" (*k<n.. even k<n-1 or n-2 or n-3.. w/e*)
proof(-)
  obtain l' where l'_def: "l' = drop 2 l" by fastforce
  moreover have "length l > 3" using a_path_def is_path_def assms by (metis add.commute nat_add_left_cancel_less numeral_Bit0 numeral_Bit1)
  ultimately have shifted_by_2: "\<forall>k. l'!k = l!(k+2)" by simp
  moreover have l_path: "\<forall>k<n. (l!k \<rightarrow> l!(k+1))" using is_path_def a_path_def assms(1) by blast
  ultimately have l'_path: "\<forall>k<n-2. (l'!k \<rightarrow> l'!(k+1))" by simp
  moreover have l_second: "l!2=(Seq C (While C), \<sigma>)" using Iter_ge2_steps assms by blast
  then have "l'!0=(Seq C (While C), \<sigma>)" using shifted_by_2 by (metis add_0)
  moreover have "l!n=(Skip, \<sigma>'')" using assms(1) is_path_def by blast
  then have l'_last: "l'!(n-2)=(Skip, \<sigma>'')" using shifted_by_2 by (metis assms(2) le_add_diff_inverse2 less_imp_le_nat)
  ultimately have "is_path l' (Seq C (While C), \<sigma>) (n-2) (Skip, \<sigma>'')"
    using is_path_def a_path_def by (smt (verit, del_insts) Nat.diff_diff_right \<open>l' = drop 2 l\<close> assms diff_diff_left le_add_diff_inverse2 length_drop less_diff_conv less_or_eq_imp_le nat_1_add_1)
  moreover have "\<forall>C\<^sub>1''. fst (Skip, \<sigma>'') \<noteq> Seq C\<^sub>1'' (While C)" by auto
  show ?thesis
  proof(cases "C=Skip")
    case True
    then show ?thesis by (metis One_nat_def Suc_1 Suc_lessI \<open>l ! n = (Skip, \<sigma>'')\<close> assms(2) cmd.distinct(59) fst_eqD l_path l_second less_add_same_cancel1 less_numeral_extra(1) numeral_3_eq_3 numeral_Bit1 numerals(1) plus_1_eq_Suc seq1_unique)
  next
    case False
    obtain k where k_def: "fst (l'!k) = Seq Skip (While C)" "\<forall>k'<k. \<exists>C\<^sub>1'. fst (l'!k') = Seq C\<^sub>1' (While C) \<and> C\<^sub>1' \<noteq> Skip" "k<n-2"
      using seq_has_skip_in_path[of l' C "While C" \<sigma> "n-2" "(Skip, \<sigma>'')"] using False calculation by auto
    then have "fst (l'!(k+1)) = (While C)" by (metis eq_fst_iff l'_path seq1_unique)
    moreover have "k+3\<le>n" using k_def(3) by linarith
    ultimately have "\<exists>k\<le>n. k>2 \<and> fst (l!k) = (While C)" using \<open>3 < length l\<close> l'_def verit_comp_simplify1(3) k_def(1) by fastforce
    then show ?thesis using \<open>l ! n = (Skip, \<sigma>'')\<close> fst_conv by fastforce
  qed
qed

lemma iter_one:
  assumes "is_path l (While C, \<sigma>) n (Skip, \<sigma>'')"
      and "n>2"
    shows "\<exists>k<n. k>2 \<and> fst (l!(k-1)) = Seq Skip (While C) \<and> fst (l!k) = While C \<and> (\<forall>k'<k. k'=0 \<or> fst (l!k')\<noteq>While C) \<and> (\<forall>k'<k. k'\<ge>2 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' (While C)))"
proof(-)
  obtain l' where l'_def: "l' = drop 2 l" by fastforce
  moreover have "length l > 3" using a_path_def is_path_def assms by (metis add.commute nat_add_left_cancel_less numeral_Bit0 numeral_Bit1)
  ultimately have shifted_by_2: "\<forall>k. l'!k = l!(k+2)" by simp
  moreover have l_path: "\<forall>k<n. (l!k \<rightarrow> l!(k+1))" using is_path_def a_path_def assms(1) by blast
  ultimately have l'_path: "\<forall>k<n-2. (l'!k \<rightarrow> l'!(k+1))" by simp
  moreover have l_second: "l!2=(Seq C (While C), \<sigma>)" using Iter_ge2_steps assms by blast
  then have "l'!0=(Seq C (While C), \<sigma>)" using shifted_by_2 by (metis add_0)
  moreover have "l!n=(Skip, \<sigma>'')" using assms(1) is_path_def by blast
  then have l'_last: "l'!(n-2)=(Skip, \<sigma>'')" using shifted_by_2 by (metis assms(2) le_add_diff_inverse2 less_imp_le_nat)
  ultimately have "is_path l' (Seq C (While C), \<sigma>) (n-2) (Skip, \<sigma>'')"
    using is_path_def a_path_def by (smt (verit, del_insts) Nat.diff_diff_right \<open>l' = drop 2 l\<close> assms diff_diff_left le_add_diff_inverse2 length_drop less_diff_conv less_or_eq_imp_le nat_1_add_1)
  moreover have "\<forall>C\<^sub>1''. fst (Skip, \<sigma>'') \<noteq> Seq C\<^sub>1'' (While C)" by auto
  show ?thesis
  proof(cases "C=Skip")
    case True
    then have "fst (l ! 3) = While C" using Suc_eq_plus1 Suc_lessI \<open>l ! n = (Skip, \<sigma>'')\<close> assms(2) l_path l_second lessI seq1_unique by (metis One_nat_def Suc_1 fst_conv numeral_3_eq_3)
    moreover have "fst (l!1) \<noteq> While C" by (metis add_lessD1 assms(2) cmd.distinct(21) eq_fst_iff l_path l_second one_add_one red_While_elim)
    moreover have "fst (l!2) \<noteq> While C" by (simp add: l_second)
    moreover have "3<n" by (metis One_nat_def Suc_1 Suc_lessI assms(1) assms(2) iter_has_next not_less_eq numeral_3_eq_3)
    moreover have "\<forall>k'<(3::nat). k'=0 \<or> k'=1 \<or> k'=2" by auto
    ultimately have "3<n \<and> 3>(2::nat) \<and> fst (l!3) = While C \<and> (\<forall>k'<3. k'=0 \<or> fst (l!k')\<noteq>While C) \<and> (\<forall>k'<3. k'\<ge>2 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' (While C)))" using l_second numeral_Bit1 by auto
    then show ?thesis using True l_second by auto
  next
    case False
    obtain k where k_def: "fst (l'!k) = Seq Skip (While C)" "\<forall>k'<k. \<exists>C\<^sub>1'. fst (l'!k') = Seq C\<^sub>1' (While C) \<and> C\<^sub>1' \<noteq> Skip" "k<n-2"
      using seq_has_skip_in_path[of l' C "While C" \<sigma> "n-2" "(Skip, \<sigma>'')"] using False calculation by auto
    then have "fst (l'!(k+1)) = (While C)" by (metis eq_fst_iff l'_path seq1_unique)
    moreover have "k+3\<le>n" using k_def(3) by linarith
    moreover have "\<forall>k'\<le>k+2. k'=0 \<or> fst (l!k')\<noteq>While C"
    proof(-)
      have "\<forall>k'\<le>k. fst (l'!k')\<noteq>While C" using k_def(2) k_def(1) nat_less_le by fastforce
      then have "\<forall>k'\<le>k+2. k'<2 \<or> fst (l!k')\<noteq>While C" by (metis add.commute le_diff_conv linordered_semidom_class.add_diff_inverse shifted_by_2)
      moreover have "fst (l!1) \<noteq> While C" by (metis add_lessD1 assms(2) cmd.distinct(21) eq_fst_iff l_path l_second one_add_one red_While_elim)
      moreover have "fst (l!2) \<noteq> While C" by (simp add: l_second)
      ultimately show ?thesis by (metis diff_is_0_eq le_antisym less_diff_conv2 less_one one_add_one verit_comp_simplify1(3))
    qed
    moreover have "fst (l!(k+3)) = (While C)" by (metis add.assoc calculation(1) nat_1_add_1 numeral_Bit1 numerals(1) shifted_by_2)
    moreover have "fst (l!(k+2)) = Seq Skip (While C)" using k_def(1) shifted_by_2 by auto
    moreover have "\<forall>k'<k+2. k'\<ge>2 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' (While C))" using k_def(2) by (metis le_add_diff_inverse2 less_diff_conv2 shifted_by_2)
    then have "\<forall>k'<k+3. k'\<ge>2 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' (While C))" using \<open>fst (l!(k+2)) = Seq Skip (While C)\<close> by (smt (verit, ccfv_threshold) add.commute add.left_commute add_diff_inverse_nat less_diff_conv2 less_one numeral_Bit0 numeral_Bit1 plus_nat.add_0 verit_comp_simplify1(3))
    ultimately have "k+3\<le>n \<and> k+3>0 \<and> fst (l!(k+3)) = (While C) \<and> (\<forall>k'<k+3. k'=0 \<or> fst (l!k')\<noteq>While C) \<and> fst (l!(k+2)) = Seq Skip (While C) \<and> (\<forall>k'<k+3. k'\<ge>2 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' (While C)))" by force
    then have "\<exists>k\<le>n. k>2 \<and> fst (l!k) = (While C) \<and> (\<forall>k'<k. k'=0 \<or> fst (l!k')\<noteq>While C) \<and> fst (l!(k-1)) = Seq Skip (While C) \<and> (\<forall>k'<k. k'\<ge>2 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' (While C)))" by auto
    then show ?thesis using \<open>l ! n = (Skip, \<sigma>'')\<close> fst_conv by (metis cmd.distinct(59) le_neq_implies_less)
  qed
qed

(* the world's ugliest proof *)
lemma iter_preserves_prop_intermediate:
  assumes "is_path l (While C, \<sigma>) n (While C, \<sigma>'')"
      and "P \<sigma>"
      and "\<And>\<sigma> \<sigma>'. P \<sigma> \<Longrightarrow> (C, \<sigma>) \<rightarrow>* (Skip, \<sigma>') \<Longrightarrow> P \<sigma>'"
    shows "P \<sigma>''"
  using assms
proof(induction n arbitrary: l \<sigma> rule: nat_less_induct)
  case (1 n)
  then show ?case
  proof(cases "n=0")
    case True
    then show ?thesis by (metis "1.prems"(1) "1.prems"(2) Pair_inject is_path_def)
  next
    case False
    then have "red_iter (l!0) 1 (l!1)" using dsem_path_same_len[of l n 0 1] "1.prems"(1) is_path_def by (metis bot_nat_0.extremum diff_zero leI less_one)
    then have "l!1 = (If (Seq C (While C)) Skip, \<sigma>)" by (metis "1.prems"(1) Iter_1_step is_path_def)
    then have "n>1" by (metis "1.prems"(1) False cmd.distinct(45) is_path_def leI less_one nat_less_le prod.inject)
    then have "red_iter (l!0) 2 (l!2)" using dsem_path_same_len[of l n 0 2] "1.prems"(1) is_path_def by fastforce
    then have l_seq_beg: "l!2 = (Seq C (While C), \<sigma>)" by (metis (mono_tags, lifting) "1.prems"(1) Iter_2_step Iter_ge2_steps RedIf2 Suc_1 Suc_lessI \<open>1 < n\<close> cmd.distinct(59) is_path_def red_If_elim)
    then have "n>2" by (metis "1.prems"(1) \<open>1 < n\<close> add_less_imp_less_left canonically_ordered_monoid_add_class.lessE cmd.distinct(30) is_path_def less_one linorder_neqE_nat nat_1_add_1 prod.inject)
    
    obtain l' where l'_def: "l' = drop 2 l" by fast
    moreover have "length l = n+1" using is_path_def a_path_def "1.prems"(1) by blast
    then have "length l > 3" using is_path_def a_path_def "1.prems"(1) \<open>n>2\<close> by linarith
    ultimately have shifted_by_2: "\<forall>k. l'!k = l!(k+2)" by simp
    then have l_path: "\<forall>k<n. (l!k \<rightarrow> l!(k+1))" using is_path_def a_path_def "1.prems"(1) by blast

    have l'_beg: "l'!0 = (Seq C (While C), \<sigma>)" using l_seq_beg shifted_by_2 \<open>n>1\<close> by (metis add_0)
    moreover have "l'!(n-2) = (While C, \<sigma>'')" using l_seq_beg shifted_by_2 by (metis "1.prems"(1) \<open>2 < n\<close> is_path_def le_add_diff_inverse2 less_imp_le_nat)
    moreover have l'_path: "\<forall>k<n-2. (l'!k \<rightarrow> l'!(k+1))" using l_path shifted_by_2 by fastforce
    ultimately have l'_is_path: "is_path l' (Seq C (While C), \<sigma>) (n-2) (While C, \<sigma>'')"
      using a_path_def is_path_def by (smt (verit) \<open>2 < n\<close> \<open>length l = n + 1\<close> add.commute add.left_commute add_diff_cancel_left' canonically_ordered_monoid_add_class.lessE l'_def length_drop)
    moreover have "\<forall>C\<^sub>1''. fst (While C, \<sigma>'') \<noteq> Seq C\<^sub>1'' (While C)" by fastforce
    ultimately obtain k where k_def: "fst (l'!k) = Seq Skip (While C)" "\<forall>k'<k. \<exists>C\<^sub>1'. fst (l'!k') = Seq C\<^sub>1' (While C) \<and> C\<^sub>1' \<noteq> Skip" "k<n-2"
      using seq_has_skip_in_path[of l' C "(While C)" \<sigma> "n-2" "(While C, \<sigma>'')"] by blast
    then obtain l'' where l''_def: "l'' = take (k+1) l'" by fast
    then have l''_prefix: "\<forall>n<k+1. l'!n=l''!n" by simp

    have "l''!0 = (Seq C (While C), \<sigma>)" using l''_prefix l'_beg by auto
    moreover have "l''!k = l'!k" using l''_prefix by fastforce
    moreover have "\<forall>k'<k. (l''!k' \<rightarrow> l''!(k'+1))"
      using l'_path l''_prefix \<open>k<n-2\<close> by (smt (verit, ccfv_SIG) add.commute le_add_diff_inverse nat_add_left_cancel_less nat_less_le one_add_one trans_less_add2)
    moreover have "length l'' = k+1" using \<open>k < n - 2\<close> \<open>length l = n + 1\<close> l'_def l''_def by force
    ultimately have l''_path: "is_path l'' (Seq C (While C), \<sigma>) k (l''!k)" using is_path_def a_path_def by blast
    then have "red_iter (C, \<sigma>) k (Skip, snd (l'!k))"
      using seq_rev_path[of l'' C "While C" \<sigma> k] by (metis \<open>l'' ! k = l' ! k\<close> add.commute k_def(1) k_def(2) l''_prefix prod.collapse trans_less_add2)
    then have "P (snd (l'!k))" using "1.prems"(2) "1.prems"(3) rtred_eq_red_iter_bkw by blast
    moreover have "fst (l'!(k+1)) = While C" by (metis k_def(1) k_def(3) l'_path prod.collapse seq1_unique fst_conv)
    moreover have "snd (l'!k) = snd (l'!(k+1))" using k_def(1) k_def(3) l'_path by (metis calculation(2) prod.collapse red_Seq_elim2)
    then have "P (snd (l'!(k+1)))" using \<open>P (snd (l'!k))\<close> by simp
    ultimately have 0: "fst (l'!(k+1)) = While C \<and> P (snd (l'!(k+1)))" by fast

    obtain l\<^sub>0 where l\<^sub>0_def: "l\<^sub>0 = drop (k+1) l'" by blast
    moreover have "length l' > k+1" using is_path_def a_path_def \<open>length l = n + 1\<close> l'_def using k_def(3) by auto
    ultimately have shifted: "\<forall>k'. l\<^sub>0!k' = l'!(k'+k+1)" by (simp add: add.commute)
    then have "\<forall>k'<n-2-(k+1). (l\<^sub>0!k' \<rightarrow> l\<^sub>0!(k'+1))" using l'_path add.commute by auto
    moreover have "length l\<^sub>0 = n-1 - (k+1)" using is_path_def a_path_def \<open>length l = n + 1\<close> l'_def l\<^sub>0_def by simp
    ultimately have "a_path l\<^sub>0 (n-2 - (k+1))" using a_path_def[of l\<^sub>0 "n-2 - (k+1)"] k_def(3) by auto
    moreover have "n-2\<ge>k+1 \<and> (\<forall>k'\<ge>k+1. l\<^sub>0!(k'-(k+1)) = l'!(k'))" using \<open>k<n-2\<close> \<open>\<forall>k'. l\<^sub>0!k' = l'!(k'+k+1)\<close> by force
    then have "l\<^sub>0!(n-2-(k+1)) = l'!(n-2)" by blast
    moreover have "l\<^sub>0!0 = l'!(k+1)" using \<open>\<forall>k'. l\<^sub>0!k' = l'!(k'+k+1)\<close> by fastforce
    ultimately have "is_path l\<^sub>0 (l'!(k+1)) (n-2-(k+1)) (l'!(n-2))" using is_path_def[of l\<^sub>0 "l'!(k+1)" "n-2-(k+1)" "l'!(n-2)"] by fast
    moreover have \<open>\<forall>m<n. \<forall>\<sigma>'. is_path l\<^sub>0 (While C, \<sigma>') m (While C, \<sigma>'') \<longrightarrow> P \<sigma>' \<longrightarrow> (\<forall>x. P x \<longrightarrow> (\<forall>xb. ((C, x) \<rightarrow>* (Skip, xb)) \<longrightarrow> P xb)) \<longrightarrow> P \<sigma>''\<close> using "1.IH" by blast
    
    ultimately show ?thesis 
      by (metis (no_types, lifting) "1.prems"(3) False "0" \<open>l' ! (n - 2) = (While C, \<sigma>'')\<close> add_gr_0 bot_nat_0.not_eq_extremum diff_diff_left diff_less prod.collapse zero_less_one)
  qed
qed

lemma iter_preserves_prop:
  assumes "is_path l (While C, \<sigma>) n (Skip, \<sigma>'')"
      and "P \<sigma>"
      and "\<And>\<sigma> \<sigma>'. P \<sigma> \<Longrightarrow> (C, \<sigma>) \<rightarrow>* (Skip, \<sigma>') \<Longrightarrow> P \<sigma>'"
    shows "P \<sigma>''"
  using assms
proof(induction n arbitrary: l \<sigma> rule: nat_less_induct)
  case (1 n)
  have "red_iter (l!0) 1 (l!1)" using dsem_path_same_len[of l n 0 1] "1.prems"(1) is_path_def by (metis bot_nat_0.extremum cmd.distinct(59) diff_zero leI less_one prod.inject)
  then have "l!1 = (If (Seq C (While C)) Skip, \<sigma>)" by (metis "1.prems"(1) Iter_1_step is_path_def)
  then have "n>1" by (metis "1.prems"(1) cmd.distinct(39) cmd.distinct(59) is_path_def less_one linorder_neqE_nat prod.inject)
  then have "red_iter (l!0) 2 (l!2)" using dsem_path_same_len[of l n 0 2] "1.prems"(1) is_path_def by fastforce
  then have l_two: "l!2 = (Seq C (While C), \<sigma>) \<or> l!2 = (Skip, \<sigma>)" by (metis (no_types, lifting) "1.prems"(1) Iter_2_step is_path_def)
  then show ?case 
  proof(cases "l!2 = (Skip, \<sigma>)")
    case True
    then show ?thesis
      by (metis (no_types, lifting) "1.prems"(1) "1.prems"(2) Iter_ge2_steps Pair_inject Suc_1 Suc_lessI \<open>1 < n\<close> cmd.distinct(23) is_path_def)
  next
    case False
    then have "l!2 = (Seq C (While C), \<sigma>)" using l_two by presburger
    then have "n>2" by (metis "1.prems"(1) Suc_1 Suc_lessI \<open>1 < n\<close> cmd.distinct(23) fst_conv is_path_def)
    then obtain k where k_def:  "k<n" "k>2" "fst (l!k) = (While C)" using iter_has_next using "1.prems"(1) by blast
    then obtain l' where l'_def: "l' = drop k l" by blast
    moreover have "length l = n+1" using is_path_def a_path_def "1.prems"(1) by blast
    ultimately have l'_shifted: "\<forall>k'. l'!k' = l!(k'+k)" by (metis add.commute k_def(1) less_imp_le_nat nth_drop trans_le_add2)
    
    then have "\<forall>k'<n-k. (l'!k' \<rightarrow> l'!(k'+1))" using is_path_def a_path_def "1.prems"(1) by (smt (verit) add.assoc add.commute less_diff_conv)
    moreover have "l'!(n-k) = (Skip, \<sigma>'')" using l'_shifted "1.prems"(1) by (simp add: is_path_def k_def(1) less_imp_le_nat)
    moreover have "length l' = n+1-k" by (simp add: \<open>length l = n + 1\<close> l'_def)
    ultimately have "is_path l' (l'!0) (n-k) (Skip, \<sigma>'')" using a_path_def is_path_def by (metis Nat.add_diff_assoc2 k_def(1) less_imp_le_nat)

    moreover have "fst (l'!0) = While C" using l'_shifted k_def(3) by simp
    moreover have "P (snd (l!k))"
    proof(-)
      obtain l\<^sub>0 where l\<^sub>0_def: "l\<^sub>0 = take (k+1) l" by blast
      then have l\<^sub>0_shifted: "\<forall>k'\<le>k. l\<^sub>0!k'=l!k'" by auto
      then have "\<forall>k'<k. (l\<^sub>0!k' \<rightarrow> l\<^sub>0!(k'+1))" using is_path_def a_path_def "1.prems"(1) by (smt (verit) add.commute k_def(1) l\<^sub>0_def le_antisym le_trans nat_add_left_cancel_less nat_less_le nth_take)
      moreover have "k\<noteq>n" using \<open>fst (l' ! 0) = While C\<close> \<open>l' ! (n - k) = (Skip, \<sigma>'')\<close> by auto
      then have "length l \<ge> k+1" using \<open>k<n\<close> \<open>length l = n + 1\<close> by simp
      then have "length l\<^sub>0 = k+1" using is_path_def a_path_def "1.prems"(1) l\<^sub>0_def by simp
      ultimately have "a_path l\<^sub>0 k" using a_path_def by blast
      then have "is_path l\<^sub>0 (l\<^sub>0!0) k (l\<^sub>0!k)" using is_path_def[of l\<^sub>0 "l\<^sub>0!0" k "l\<^sub>0!k"] by blast
      moreover have "snd (l\<^sub>0!0) = \<sigma>" by (metis "1.prems"(1) is_path_def l\<^sub>0_shifted less_eq_nat.simps(1) snd_conv)
      then have "P (snd (l\<^sub>0!0))" using "1.prems"(2) by blast
      moreover have "fst (l\<^sub>0!0) = While C" by (metis "1.prems"(1) bot_nat_0.extremum fst_eqD is_path_def l\<^sub>0_shifted)
      moreover have "fst (l\<^sub>0!k) = While C" using k_def(3) l\<^sub>0_shifted by fastforce
      ultimately show ?thesis by (smt (verit) "1.prems"(3) iter_preserves_prop_intermediate l\<^sub>0_shifted order_refl prod.collapse)
    qed
    then have "P (snd (l'!0))" using l'_shifted by simp
    ultimately show ?thesis
      by (metis (no_types, lifting) "1.IH" "1.prems"(3) \<open>1 < n\<close> diff_less k_def(2) nat_1_add_1 order_less_trans prod.collapse zero_less_one zero_less_two)
  qed
qed

lemma dsem_add_seqL:
  assumes "\<lbrakk>C'\<rbrakk> = \<lbrakk>C''\<rbrakk>"
    shows "\<lbrakk>Seq C C'\<rbrakk> = \<lbrakk>Seq C C''\<rbrakk>"
  using assms dsem_def by (simp add: dsem_seq)

lemma dsem_add_seqR:
  assumes "\<lbrakk>C'\<rbrakk> = \<lbrakk>C''\<rbrakk>"
    shows "\<lbrakk>Seq C' C\<rbrakk> = \<lbrakk>Seq C'' C\<rbrakk>"
  using assms dsem_def by (simp add: dsem_seq)

fun iter :: "'var cmd \<Rightarrow> nat \<Rightarrow> 'var cmd" where
 "iter C 0 = Skip"
|"iter C (Suc n) = (Seq C (iter C n))"

fun iter' :: "'var cmd \<Rightarrow> nat \<Rightarrow> 'var cmd" where
 "iter' C 0 = Skip"
|"iter' C (Suc n) = (Seq (iter' C n) C)"

lemma iter_inject: "(iter C m) = (iter C m') \<Longrightarrow> m=m'"
proof(induction m arbitrary: m')
  case 0
  then show ?case using cmd.distinct(24) by (metis iter.elims iter.simps(1))
next
  case (Suc m)
  then show ?case by (smt (verit) iter.elims iter.simps(2) cmd.distinct(23) cmd.inject(2))
qed

lemma add_Seq_to_iterL:
  "\<lbrakk>iter C (Suc n)\<rbrakk> = \<lbrakk>Seq C (iter C n)\<rbrakk>" by simp

lemma add_Seq_to_iterR:
  "\<lbrakk>iter C (Suc n)\<rbrakk> = \<lbrakk>Seq (iter C n) C\<rbrakk>"
proof(induction n)
  case 0
  then show ?case using dsem_seq1 dsem_seq1' iter.simps by metis
next
  case (Suc n)
  then show ?case by (smt (z3) iter.simps(2) dsem_add_seqL seq_assoc)
qed

lemma iter_iter':
  "\<lbrakk>iter C n\<rbrakk> = \<lbrakk>iter' C n\<rbrakk>"
proof(induction n)
  case 0
  then show ?case by simp
next
  case (Suc n)
  then show ?case by (metis add_Seq_to_iterR iter'.simps(2) dsem_add_seqR)
qed

lemma "\<lbrakk>iter C 0\<rbrakk> = \<lbrakk>Skip\<rbrakk>" by simp
lemma "\<lbrakk>iter C 1\<rbrakk> = \<lbrakk>C\<rbrakk>" using dsem_seq1' by auto
lemma "\<lbrakk>iter C 2\<rbrakk> = \<lbrakk>Seq C C\<rbrakk>" using iter.simps dsem_seq1' seq_assoc by (metis numeral_2_eq_2)

lemma iter_breakdown:
  assumes "(iter C (Suc n), \<sigma>) \<rightarrow>* (Skip, \<sigma>'')"
    shows "\<exists>\<sigma>'. ((iter C n, \<sigma>) \<rightarrow>* (Skip, \<sigma>')) \<and> ((C, \<sigma>') \<rightarrow>* (Skip, \<sigma>''))"
  using assms by (metis add_Seq_to_iterR dsem_char seq1_preserves_dsem seq_has_skip)

lemma iter_breakdown2:
  assumes "(iter C (Suc n), \<sigma>) \<rightarrow>* (Skip, \<sigma>'')"
    shows "\<exists>\<sigma>'. ((C, \<sigma>) \<rightarrow>* (Skip, \<sigma>')) \<and> ((iter C n, \<sigma>') \<rightarrow>* (Skip, \<sigma>''))"
  by (metis add_Seq_to_iterL assms dsem_char seq1_preserves_dsem seq_has_skip)


lemma every_iter_is_while_skip:
  assumes "(iter C n, \<sigma>) \<rightarrow>* (Skip, \<sigma>'')"
    shows "(While C, \<sigma>) \<rightarrow>* (Skip, \<sigma>'')"
  using assms
proof(induction n arbitrary: \<sigma>)
  case 0
  then show ?case using RedTran RedWhile if_preserves_dsem by (metis (mono_tags, opaque_lifting) iter.simps(1))
next
  case (Suc n)
  then obtain \<sigma>' where s'_def: "(C, \<sigma>) \<rightarrow>* (Skip, \<sigma>')" "(iter C n, \<sigma>') \<rightarrow>* (Skip, \<sigma>'')" using iter_breakdown2 by blast
  then have "(While C, \<sigma>') \<rightarrow>* (Skip, \<sigma>'')" using Suc.IH by fast
  moreover have "\<And>\<sigma> \<sigma>' \<sigma>'' C. (C, \<sigma>) \<rightarrow>* (Skip, \<sigma>') \<Longrightarrow> (While C, \<sigma>') \<rightarrow>* (Skip, \<sigma>'') \<Longrightarrow> (While C, \<sigma>) \<rightarrow>* (Skip, \<sigma>'')" by (meson RedIf1 RedSeq1 RedTran RedWhile rtred_trans seq_trcl)
  ultimately show ?case using s'_def(1) by meson
qed

lemma "\<lbrakk>iter C 0\<rbrakk> = \<lbrakk>Skip\<rbrakk>" by simp
lemma "\<lbrakk>iter C 1\<rbrakk> = \<lbrakk>C\<rbrakk>" using dsem_seq1' by auto
lemma "\<lbrakk>iter C 2\<rbrakk> = \<lbrakk>Seq C C\<rbrakk>" using iter.simps dsem_seq1' seq_assoc by (metis numeral_2_eq_2)

lemma seq_rev_rt_iter:
  assumes "(Seq C\<^sub>1 (iter C n), s) \<rightarrow>* (Seq C\<^sub>1'' (iter C n), s'')"
  shows " (C\<^sub>1, s) \<rightarrow>* (C\<^sub>1'', s'')"
proof(-)
  have "right_most (iter C n) = {Skip}" apply(induction n) by auto
  then show ?thesis using assms seq_rev_rt by fast
qed

lemma iter_split:
  assumes "(iter C (Suc n), \<sigma>) \<rightarrow>* (C'', \<sigma>'')"
      and "\<nexists>C\<^sub>1'. C'' = Seq C\<^sub>1' (iter C n)"
    shows "\<exists>\<sigma>'. ((C, \<sigma>) \<rightarrow>* (Skip, \<sigma>')) \<and> ((iter C n, \<sigma>') \<rightarrow>* (C'', \<sigma>''))"
  using assms seq_rev_rt_iter seq_rhs_rt by (metis iter.simps(2))

lemma iter_split_useless:
  assumes "(iter C (Suc n), \<sigma>) \<rightarrow>* (C'', \<sigma>'')"
      and "\<nexists>C\<^sub>1'. C'' = Seq C\<^sub>1' (iter C n)"
      and "n>0"
    shows "\<exists>\<sigma>'. ((iter C (Suc n), \<sigma>) \<rightarrow>* (iter C n, \<sigma>')) \<and> ((iter C n, \<sigma>') \<rightarrow>* (C'', \<sigma>''))"
proof(-)
  have "\<exists>\<sigma>'. ((iter C (Suc n), \<sigma>) \<rightarrow>* (Seq Skip (iter C n), \<sigma>')) \<and> ((iter C n, \<sigma>') \<rightarrow>* (C'', \<sigma>''))" using assms seq_rhs_rt[of C "(iter C n)" \<sigma> C'' \<sigma>''] gr0_conv_Suc by auto
  then obtain \<sigma>' where \<sigma>'_def: "(iter C (Suc n), \<sigma>) \<rightarrow>* (Seq Skip (iter C n), \<sigma>')" "(iter C n, \<sigma>') \<rightarrow>* (C'', \<sigma>'')" by blast
  moreover have "(Seq Skip (iter C n), \<sigma>') \<rightarrow>* (iter C n, \<sigma>')" by (meson RedSeq1 rtred.simps)
  ultimately show ?thesis using rtred_trans by blast
qed

lemma seq_ending_in_skip: "(Seq C Skip, \<sigma>) \<rightarrow>* c' \<Longrightarrow> fst c'\<noteq>Skip \<Longrightarrow> \<exists>C'. fst c'=Seq C' Skip"
proof(induction "(Seq C Skip, \<sigma>)" c' arbitrary: C \<sigma> rule: rtred.induct)
  case (RedRefl)
  then show ?case by simp
next
  case (RedTran c' c'')
  then show ?case by (metis fst_conv red_rhs(3) skip_deadend_rt seq1_unique)
qed

lemma iter_rhs_rt_aux:
  assumes "(Seq C\<^sub>0 (iter C n), \<sigma>) \<rightarrow>* (C', \<sigma>'')"
    shows "\<exists>m\<le>n. C' = iter C m \<or> (\<exists>C\<^sub>0'. C' = Seq C\<^sub>0' (iter C m))"
  using assms
proof(induction "(Seq C\<^sub>0 (iter C n), \<sigma>)" "(C', \<sigma>'')" arbitrary: C\<^sub>0 n \<sigma>)
  case (RedRefl)
  then show ?case by fastforce
next
  case (RedTran c')
  then have cs: "(\<exists>C\<^sub>0' \<sigma>'. c' = (Seq C\<^sub>0' (iter C n), \<sigma>')) \<or> (C\<^sub>0=Skip \<and> c'=(iter C n, \<sigma>))" using red_rhs(3) seq1_unique by metis
  then show ?case
  proof(cases "\<exists>C\<^sub>0' \<sigma>'. c' = (Seq C\<^sub>0' (iter C n), \<sigma>')")
    case True
    then show ?thesis using RedTran.hyps(3) by blast
  next
    case False
    then have "C\<^sub>0=Skip \<and> c'=(iter C n, \<sigma>)" using cs by blast
    then show ?thesis
    proof(cases "n=0")
      case True
      then show ?thesis by (metis False iter.simps(1) RedTran.hyps(2) cs fst_conv le0 skip_deadend_rt)
    next
      case False
      then have "c' = (Seq C (iter C (n-1)), \<sigma>)" by (metis iter.simps(2) \<open>C\<^sub>0=Skip \<and> c'=(iter C n, \<sigma>)\<close> diff_Suc_1 old.nat.exhaust)
      then show ?thesis by (meson RedTran.hyps(3) diff_le_self order_trans)
    qed
  qed
qed

lemma iter_rhs_rt:
  assumes "(iter C n, \<sigma>) \<rightarrow>* (C', \<sigma>')"
  shows "\<exists>m\<le>n. C' = iter C m \<or> (\<exists>C\<^sub>0'. C' = Seq C\<^sub>0' (iter C m))"
  using assms iter_rhs_rt_aux  RedSeq1 RedTran by meson

lemma iter_rhs_rt':
  assumes "(iter C n, \<sigma>) \<rightarrow>* (C', \<sigma>')"
      and "C' \<noteq> Skip"
    shows "\<exists>m\<le>n. \<exists>C\<^sub>0'. C' = Seq C\<^sub>0' (iter C m)"
  using assms iter_rhs_rt by (metis iter.elims Suc_leD)

lemma iter_rhs_rt_decreasing:
  assumes "(Seq C\<^sub>0 (iter C n), \<sigma>) \<rightarrow>* (Seq C\<^sub>0'' (iter C m), \<sigma>'')"
  shows "n\<ge>m"
  using assms iter_rhs_rt' by (metis Orderings.order_eq_iff cmd.distinct(23) cmd.inject(2) iter_inject seq_rhs_rt)

lemma iter_rhs_unique:
  assumes "m\<le>n"
      and "C' = Seq C\<^sub>0' (iter C m)"
      and "m'\<le>n"
      and "C' = Seq C\<^sub>0'' (iter C m')"
    shows "C\<^sub>0'=C\<^sub>0'' \<and> m=m'"
  using assms iter_inject by auto

lemma every_iter_is_while_seq_aux:
  assumes "(Seq C\<^sub>0 (iter C n), \<sigma>) \<rightarrow>* (Seq C\<^sub>0'' (iter C m), \<sigma>'')"
    shows "(Seq C\<^sub>0 (While C), \<sigma>) \<rightarrow>* (Seq C\<^sub>0'' (While C), \<sigma>'')"
  using assms
proof(induction "(Seq C\<^sub>0 (iter C n), \<sigma>)" "(Seq C\<^sub>0'' (iter C m), \<sigma>'')" arbitrary: C\<^sub>0 n \<sigma>)
  case (RedRefl)
  then show ?case by (simp add: rtred.RedRefl)
next
  case (RedTran c')
  moreover obtain C' \<sigma>' where "c'=(C', \<sigma>')" using old.prod.exhaust by blast
  ultimately have C'\<sigma>'_def: "(Seq C\<^sub>0 (iter C n), \<sigma>) \<rightarrow> (C', \<sigma>')" "(C', \<sigma>') \<rightarrow>* (Seq C\<^sub>0'' (iter C m), \<sigma>'')" by auto
  then have cs: "C'=iter C n \<or> (\<exists>C\<^sub>0'. C'=Seq C\<^sub>0' (iter C n))" by blast
  then show ?case
  proof(cases "n=0")
    case True
    then show ?thesis by (metis (no_types, opaque_lifting) iter.simps(1) RedTran.hyps(1) RedTran.hyps(2) iter_rhs_rt_decreasing le0 nle_le rtred.RedTran seq_skip_equiv seq_trcl)
  next
    case False
    then have "n>0" by fast
    then show ?thesis
    proof(cases "C'=iter C n")
      case True
      then have " C'=Seq C (iter C (n-1))" using Suc_pred' \<open>0 < n\<close> iter.simps(2) by metis
      then have "(Seq C (iter C (n-1)), \<sigma>') \<rightarrow>* (Seq C\<^sub>0'' (iter C m), \<sigma>'')" using C'\<sigma>'_def(2) by auto
      then have "(Seq C (While C), \<sigma>') \<rightarrow>* (Seq C\<^sub>0'' (While C), \<sigma>'')" by (simp add: RedTran.hyps(3) \<open>C' = Seq C (iter C (n-1))\<close> \<open>c' = (C', \<sigma>')\<close>)
      moreover have "(While C, \<sigma>') \<rightarrow>* (Seq C (While C), \<sigma>')" using RedWhile by (metis (no_types, lifting) RedIf1 rtred.simps)
      ultimately have "(While C, \<sigma>') \<rightarrow>* (Seq C\<^sub>0'' (While C), \<sigma>'')" using rtred_trans by fast
      moreover have "(Seq C\<^sub>0 (iter C n), \<sigma>) \<rightarrow> (Seq C (iter C (n-1)), \<sigma>')" using C'\<sigma>'_def(1) by (simp add: \<open>C' = Seq C (iter C (n - 1))\<close>)
      then have "C\<^sub>0 = Skip \<and> \<sigma>=\<sigma>'" using \<open>0 < n\<close> iter_inject by fastforce
      ultimately show ?thesis by (meson RedSeq1 rtred.simps)
    next
      case False
      then obtain C\<^sub>0' where "C'=Seq C\<^sub>0' (iter C n)" using cs by blast
      then have "(Seq C\<^sub>0' (iter C n), \<sigma>') \<rightarrow>* (Seq C\<^sub>0'' (iter C m), \<sigma>'')" using C'\<sigma>'_def(2) by auto
      then have "(Seq C\<^sub>0' (While C), \<sigma>') \<rightarrow>* (Seq C\<^sub>0'' (While C), \<sigma>'')" by (simp add: RedTran.hyps(3) \<open>C' = Seq C\<^sub>0' (iter C n)\<close> \<open>c' = (C', \<sigma>')\<close>)
      moreover have "(Seq C\<^sub>0 (iter C n), \<sigma>) \<rightarrow> (Seq C\<^sub>0' (iter C n), \<sigma>')" using C'\<sigma>'_def(1) \<open>C' = Seq C\<^sub>0' (iter C n)\<close> by auto
      ultimately show ?thesis by (meson RedSeq2 rtred.simps seq_rev)
    qed
  qed
qed

lemma every_iter_is_while_seq:
  assumes "(iter C n, \<sigma>) \<rightarrow>* (Seq C\<^sub>0'' (iter C m), \<sigma>'')"
    shows "(While C, \<sigma>) \<rightarrow>* (Seq C\<^sub>0'' (While C), \<sigma>'')"
proof(-)
  have "n>0" using assms by (metis iter.simps(1) cmd.distinct(23) gr0I prod.inject skip_deadend_rt)
  then have "(Seq C (iter C (n-1)), \<sigma>) \<rightarrow>* (Seq C\<^sub>0'' (iter C m), \<sigma>'')" by (metis iter.simps(2) Suc_pred' assms)
  then have "(Seq C (While C), \<sigma>) \<rightarrow>* (Seq C\<^sub>0'' (While C), \<sigma>'')" using every_iter_is_while_seq_aux by meson
  moreover have "(While C, \<sigma>) \<rightarrow>* (Seq C (While C), \<sigma>)" using RedWhile by (metis (no_types, lifting) RedIf1 rtred.simps)
  ultimately show ?thesis using rtred_trans by blast
qed

(* (If (Seq C (While C)) Skip, \<sigma>)*)

(*lemma
  assumes "(While C, \<sigma>) \<rightarrow> (C', \<sigma>')"
  shows "C'=If (Seq C (While C)) Skip"
*)

lemma while_rhs:
  assumes "(\<C>, \<sigma>) \<rightarrow> (\<C>', \<sigma>')"
    shows "(\<forall>C. \<C>=While C \<longrightarrow> \<C>'=(If (Seq C (While C)) Skip))
         \<and> (\<forall>C C\<^sub>0. \<C>=Seq C\<^sub>0 (While C) \<longrightarrow> \<C>'=While C \<or> (\<exists>C\<^sub>0'. \<C>'=Seq C\<^sub>0' (While C)))
         \<and> (\<forall>C. \<C>=(If (Seq C (While C)) Skip) \<longrightarrow> \<C>'=Skip \<or> (\<exists>C\<^sub>0'. \<C>'=Seq C\<^sub>0' (While C)))"
  using assms red_While_elim by blast

lemma while_rhs_rt_aux:
  assumes "(\<C>, \<sigma>) \<rightarrow>* (\<C>', \<sigma>')"
    shows "(\<forall>C. \<C>=While C \<longrightarrow> \<C>'=Skip \<or> \<C>'=While C \<or> (\<exists>C\<^sub>0'. \<C>'=Seq C\<^sub>0' (While C)) \<or> \<C>'=(If (Seq C (While C)) Skip))
         \<and> (\<forall>C C\<^sub>0. \<C>=Seq C\<^sub>0 (While C) \<longrightarrow> \<C>'=Skip \<or> \<C>'=While C \<or> (\<exists>C\<^sub>0'. \<C>'=Seq C\<^sub>0' (While C)) \<or> \<C>'=(If (Seq C (While C)) Skip))
         \<and> (\<forall>C. \<C>=(If (Seq C (While C)) Skip) \<longrightarrow> \<C>'=Skip \<or> \<C>'=While C \<or> (\<exists>C\<^sub>0'. \<C>'=Seq C\<^sub>0' (While C)) \<or> \<C>'=(If (Seq C (While C)) Skip))"
  using assms
proof(induction "(\<C>, \<sigma>)" "(\<C>', \<sigma>')" arbitrary: \<C> \<sigma>)
  case (RedRefl)
  then show ?case by blast
next
  case (RedTran c')
  then show ?case using while_rhs by (smt (verit, del_insts) red_rhs(3) red_rhs(4) red_rhs(7) seq1_unique skip_deadend_rt)
qed

lemma while_rhs_rt:
  assumes "(While C, \<sigma>) \<rightarrow>* (C', \<sigma>')"
    shows "C'=Skip \<or> C'=While C \<or> (\<exists>C\<^sub>0'. C'=Seq C\<^sub>0' (While C)) \<or> C'=If (Seq C (While C)) Skip"
  using assms while_rhs_rt_aux by blast

lemma while_self_loop:
  assumes "(While C, \<sigma>) \<rightarrow>* (Skip, \<sigma>''')"
  shows "(While C, \<sigma>) \<rightarrow>* (While C, \<sigma>''')"
proof(-)
  obtain C'' \<sigma>'' where "(While C, \<sigma>)\<rightarrow>*(C'', \<sigma>'')" "(C'', \<sigma>'')\<rightarrow>(Skip, \<sigma>''')" using rtred_backwards assms old.prod.exhaust by (metis cmd.distinct(59) prod.inject)
  then have "C''=Skip \<or> C''=While C \<or> (\<exists>C\<^sub>0''. C''=Seq C\<^sub>0'' (While C)) \<or> C''=If (Seq C (While C)) Skip" using while_rhs_rt by blast
  then have "C''=If (Seq C (While C)) Skip" using \<open>(C'', \<sigma>'')\<rightarrow>(Skip, \<sigma>''')\<close> by (metis cmd.distinct(23) cmd.distinct(39) cmd.distinct(59) fst_eqD skip_deadend while_rhs)
  then obtain C' \<sigma>' where "(While C, \<sigma>)\<rightarrow>*(C', \<sigma>')" "(C', \<sigma>')\<rightarrow>(C'', \<sigma>'')" using \<open>(While C, \<sigma>)\<rightarrow>*(C'', \<sigma>'')\<close> rtred_backwards assms old.prod.exhaust by (metis cmd.distinct(45) prod.inject)
  then have "C'=Skip \<or> C'=While C \<or> (\<exists>C\<^sub>0'. C'=Seq C\<^sub>0' (While C)) \<or> C'=If (Seq C (While C)) Skip" using while_rhs_rt by blast
  then have "C'=While C" using \<open>(C', \<sigma>')\<rightarrow>(C'', \<sigma>'')\<close> \<open>C''=If (Seq C (While C)) Skip\<close> by (metis cmd.distinct(21) cmd.distinct(45) fst_eqD skip_deadend while_rhs)
  moreover have "\<sigma>'''=\<sigma>'" using \<open>(C'', \<sigma>'')\<rightarrow>(Skip, \<sigma>''')\<close> \<open>(C', \<sigma>')\<rightarrow>(C'', \<sigma>'')\<close> calculation red_While_elim by blast
  ultimately show ?thesis using \<open>(While C, \<sigma>)\<rightarrow>*(C', \<sigma>')\<close> by auto
qed

fun iter_count :: "'var cmd \<Rightarrow> 'var conf list \<Rightarrow> nat" where
 "iter_count C [] = 0"
|"iter_count C (c#l) = (if fst c=While C then Suc (iter_count C l) else (iter_count C l))"

lemma iter_count_mono_aux:
  "iter_count C l \<le> iter_count C (c#l)" 
  using iter_count.cases by simp

lemma iter_count_mono:
  "iter_count C (drop n l) \<le> iter_count C l"
proof(induction n)
  case 0
  then show ?case by simp
next
  case (Suc n)
  then show ?case using iter_count_mono_aux by (metis (no_types, lifting) Cons_nth_drop_Suc drop_eq_Nil2 dual_order.trans le_SucI linorder_le_less_linear)
qed

lemma iter_count_ge2:
  assumes "is_path l (While C, \<sigma>) len (Skip, \<sigma>')"
    shows "len>2 \<longleftrightarrow> iter_count C l \<ge> 2"
proof
  show "len>2 \<Longrightarrow> iter_count C l \<ge> 2"
  proof(-)
    assume "len>2"
    then obtain k where k_def: "k<len" "k>2" "fst (l!k) = (While C)" using assms iter_has_next by meson
    moreover obtain l' where l'_def: "l' = drop k l" by blast
    then have "fst (l'!0) = While C" using k_def(3) assms(1) is_path_def a_path_def by (metis (mono_tags, lifting) add.right_neutral add_mono_thms_linordered_semiring(1) k_def(1) le0 nth_drop less_imp_le_nat) 
    moreover have "length l' > 0" using l'_def k_def(1) is_path_def a_path_def by (metis add_leE assms(1) length_drop verit_comp_simplify1(3) zero_less_diff)
    moreover obtain l'' where "l'' = drop 1 l'" by force
    then have "l'=(l'!0)#l''" using Cons_nth_drop_Suc calculation(5) by force
    ultimately have "iter_count C l' > 0" by (metis iter_count.simps(2) less_Suc_eq_0_disj)
    moreover have "l' = drop (k-1) (drop 1 l)" using k_def(2) l'_def by force
    then have "iter_count C (drop 1 l) \<ge> iter_count C l'" using iter_count_mono by fast
    ultimately have "iter_count C (drop 1 l) \<ge> 1" by linarith
    moreover have "l=(l!0)#(drop 1 l)" by (metis Cons_nth_drop_Suc One_nat_def \<open>0 < length l'\<close> diff_le_self drop0 l'_def length_drop less_eq_Suc_le not_gr_zero not_less_eq_eq)
    moreover have "fst (l!0) = While C" using assms(1) is_path_def by (metis fst_conv)
    ultimately show ?thesis using iter_count.simps(2) by (metis Suc_1 Suc_leI le_imp_less_Suc)
  qed
  show "iter_count C l \<ge> 2 \<Longrightarrow> len>2"
  proof(rule ccontr)
    assume cntr: "iter_count C l \<ge> 2" "\<not> len>2"
    then show "False" 
    proof(cases "len<2")
      case True
      moreover have "red_iter (While C, \<sigma>) len (Skip, \<sigma>')" using assms dsem_path_same_len unfolding is_path_def by fastforce
      ultimately show ?thesis using Successfull_iter_is_ge2 by fastforce
    next
      case False
      then have "len=2" using cntr(2) by auto
      then have "red_iter (While C, \<sigma>) 2 (Skip, \<sigma>')" using assms dsem_path_same_len unfolding is_path_def by fastforce
      then have 2: "fst (l!2) = Skip" by (metis \<open>len = 2\<close> assms fst_conv is_path_def)
      have "fst (l!0) = While C" using assms unfolding is_path_def by force
      moreover have "l!0 \<rightarrow> l!1" using assms unfolding is_path_def a_path_def by (metis (no_types, lifting) \<open>len = 2\<close> add_0 nat_1_add_1 zero_less_two)
      ultimately have 1: "fst (l!1) = (If (Seq C (While C)) Skip)" by (metis eq_fst_iff red_While_elim)
      have 0: "fst (l!0) = While C" using assms unfolding is_path_def by force
      have "length (drop 2 l) = 1" using \<open>len=2\<close> assms is_path_def a_path_def by (metis add.commute add_diff_cancel_right' length_drop)
      moreover have "l!2 = (drop 2 l)!0" using calculation by auto
      ultimately have "drop 2 l = [l!2]" by (metis (no_types, lifting) Cons_nth_drop_Suc One_nat_def diff_Suc_1 drop0 length_0_conv length_drop less_numeral_extra(1))
      moreover have "drop 1 l = (l!1)#(drop 2 l)" by (metis Cons_nth_drop_Suc Suc_1 \<open>length (drop 2 l) = 1\<close> diff_Suc_1 length_drop less_imp_diff_less less_numeral_extra(1) zero_less_diff)
      ultimately have "drop 1 l = [l!1, l!2]" by presburger
      moreover have "l=(l!0)#(drop 1 l)" by (metis Cons_nth_drop_Suc \<open>length (drop 2 l) = 1\<close> drop_0 length_drop less_nat_zero_code not_gr_zero numeral_1_eq_Suc_0 numeral_One zero_less_diff zero_less_one)
      ultimately have "l=[l!0, l!1, l!2]" by presburger
      then have "iter_count C l = 1" using "0" "1" "2" by (metis One_nat_def cmd.distinct(45) cmd.distinct(59) iter_count.simps(1) iter_count.simps(2))
      then show ?thesis using cntr(1) by auto
    qed
  qed
qed

lemma iter_count_1:
  assumes "is_path l (While C, \<sigma>) len (Skip, \<sigma>')"
      and "iter_count C l = 1"
    shows "len=2 \<and> \<sigma>=\<sigma>'"
proof(-)
  have "len\<le>2" using assms iter_count_ge2 by fastforce
  moreover have "len\<noteq>0" using assms by (metis cmd.distinct(59) fst_eqD is_path_def)
  moreover have "red_iter (While C, \<sigma>) len (Skip, \<sigma>')" using assms(1) dsem_path_same_len unfolding is_path_def by fastforce
  moreover have "len\<noteq>1" using calculation(3) Iter_1_step by fast
  ultimately show ?thesis using Iter_2_step[of C \<sigma> "(Skip, \<sigma>')"] assms(1) antisym_conv2 by fastforce
qed

lemma concat_successfull_paths:
  assumes "(C, \<sigma>) \<rightarrow>* (Skip, \<sigma>')"
      and "(C', \<sigma>') \<rightarrow>* (Skip, \<sigma>'')"
    shows "(Seq C C', \<sigma>) \<rightarrow>* (Skip, \<sigma>'')"
  using assms by (meson RedSeq1 RedTran rtred_trans seq_trcl)

lemma iter_combine:
  assumes "(iter C n, \<sigma>) \<rightarrow>* (Skip, \<sigma>')"
      and "(C, \<sigma>') \<rightarrow>* (Skip, \<sigma>'')"
    shows "(iter C (Suc n), \<sigma>) \<rightarrow>* (Skip, \<sigma>'')"
  using add_Seq_to_iterR assms(1) assms(2) concat_successfull_paths dsem_char by blast

lemma iter_combine2:
  assumes "(C, \<sigma>) \<rightarrow>* (Skip, \<sigma>')"
      and "(iter C n, \<sigma>') \<rightarrow>* (Skip, \<sigma>'')"
    shows "(iter C (Suc n), \<sigma>) \<rightarrow>* (Skip, \<sigma>'')"
  using assms concat_successfull_paths add_Seq_to_iterL dsem_char by blast

(*non-precise*)
lemma while_to_seq_and_while:
  assumes "is_path l (While C, \<sigma>) len (Skip, \<sigma>'')"
      and "iter_count C l = n"
      and "n\<ge>2"
    shows "\<exists>k<len. ((C, \<sigma>) \<rightarrow>* (Skip, snd (l!k))) \<and> (is_path (drop k l) (While C, snd (l!k)) (len-k) (Skip, \<sigma>'')) \<and> iter_count C (drop k l) \<le> n-1"
proof(-)
  have "len>2" using assms iter_count_ge2 by fast
  then obtain k where k_def: "k<len" "k>2" "fst (l!k) = While C"
                             "\<forall>k'<k. k'=0 \<or> fst (l!k')\<noteq>While C"
                             "fst (l!(k-1)) = Seq Skip (While C)"
                             "\<forall>k'<k. k'\<ge>2 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' (While C))" using assms(1) iter_one by blast
  obtain l' where l'_def: "l' = take (Suc k) l" by fast
  moreover have l_len: "length l = len+1" using assms(1) is_path_def a_path_def by blast
  ultimately have l'_len: "length l' = k+1" using k_def(1) by fastforce
  have ll'_same: "\<forall>k'\<le>k. l!k'=l'!k'" using l'_def by auto
  moreover have "\<forall>k'<len. (l!k' \<rightarrow> l!(k'+1))" using assms(1) is_path_def a_path_def by blast
  ultimately have "\<forall>k'<k. (l'!k' \<rightarrow> l'!(k'+1))" using k_def(1) l'_def by auto
  then have "a_path l' k" using l'_len a_path_def by blast
  then have "is_path l' (While C, \<sigma>) k (l!k)" using ll'_same by (metis assms(1) bot_nat_0.extremum eq_imp_le is_path_def)

  then have "fst (l'!2) = (Seq C (While C))" using Iter_ge2_steps by (metis fst_conv k_def(2))
  obtain l'\<^sub>2 where l'\<^sub>2_def: "l'\<^sub>2 = drop 2 l'" by blast
  then have l'\<^sub>2_len: "length l'\<^sub>2 = k-1" using l'_len by fastforce
  moreover have l'l'\<^sub>2_same: "\<forall>k'\<le>k-2. l'\<^sub>2!k'=l'!(2+k')" by (metis add_increasing2 k_def(2) l'\<^sub>2_def l'_len less_SucI less_Suc_eq_le linordered_nonzero_semiring_class.zero_le_one nth_drop)
  then have "\<forall>k'<k-2. (l'\<^sub>2!k' \<rightarrow> l'\<^sub>2!(k'+1))" using \<open>\<forall>k'<k. (l'!k' \<rightarrow> l'!(k'+1))\<close> by fastforce
  ultimately have "a_path l'\<^sub>2 (k-2)" using a_path_def by (metis Nat.add_diff_assoc2 k_def(2) l'\<^sub>2_def l'_len length_drop less_imp_le_nat)
  then have "is_path l'\<^sub>2 (Seq C (While C), snd (l!2)) (k-2) (l!k)" by (metis (no_types, opaque_lifting) Nat.add_0_right \<open>fst (l' ! 2) = Seq C (While C)\<close> add_diff_inverse_nat is_path_def k_def(2) l'l'\<^sub>2_same le0 less_or_eq_imp_le ll'_same prod.collapse verit_comp_simplify1(3))

  obtain l'\<^sub>3 where l'\<^sub>3_def: "l'\<^sub>3 = butlast l'\<^sub>2" by simp
  then have l'\<^sub>3_len: "length l'\<^sub>3 = k-2" by (simp add: l'\<^sub>2_len)
  moreover have l'\<^sub>2l'\<^sub>3_same: "\<forall>k'\<le>k-3. l'\<^sub>2!k'=l'\<^sub>3!k'" by (smt (verit) add_diff_cancel_right cancel_comm_monoid_add_class.diff_cancel diff_add diff_is_0_eq' k_def(2) l'\<^sub>3_def l'\<^sub>3_len less_add_one less_imp_diff_less nat_1_add_1 not_less_iff_gr_or_eq nth_butlast numeral_Bit1 numerals(1) order.strict_implies_order zero_less_diff)
  then have "\<forall>k'<k-3. (l'\<^sub>3!k' \<rightarrow> l'\<^sub>3!(k'+1))" using \<open>\<forall>k'<k-2. (l'\<^sub>2!k' \<rightarrow> l'\<^sub>2!(k'+1))\<close> by (smt (verit) ab_semigroup_add_class.add_ac(1) add.commute add_lessD1 l'\<^sub>3_def l'\<^sub>3_len less_diff_conv nth_butlast numeral_Bit0 numeral_Bit1)
  ultimately have "a_path l'\<^sub>3 (k-3)" using a_path_def by (smt (verit, best) diff_add diff_diff_left diff_is_0_eq k_def(2) less_one nat_1_add_1 numeral_Bit1 numerals(1) verit_comp_simplify1(3))
  moreover have "l'\<^sub>3!0 = (Seq C (While C), snd (l!2))" by (metis \<open>is_path l'\<^sub>2 (Seq C (While C), snd (l ! 2)) (k - 2) (l ! k)\<close> is_path_def l'\<^sub>2l'\<^sub>3_same zero_le)
  moreover have "l'\<^sub>3!(k-3) = l!(k-1)" by (smt (verit) \<open>a_path l'\<^sub>2 (k - 2)\<close> a_path_def add.commute add_numeral_left calculation(1) diff_le_self l'\<^sub>2_len l'\<^sub>3_def l'\<^sub>3_len l'l'\<^sub>2_same less_add_one less_imp_le_nat ll'_same nth_butlast numeral_One semiring_norm(2))
  ultimately have l'\<^sub>3_path: "is_path l'\<^sub>3 (Seq C (While C), snd (l!2)) (k-3) (Seq Skip (While C), snd (l!(k-1)))" using k_def(5) by (metis is_path_def prod.collapse)
  have "\<forall>k'\<le>k-3. l'\<^sub>3!k'=l'!(2+k')" by (metis add.commute add_lessD1 k_def(2) l'\<^sub>2_def l'\<^sub>2l'\<^sub>3_same l'_len less_or_eq_imp_le nat_1_add_1 nat_add_left_cancel_le nth_drop)
  have "\<forall>k'<k. k'\<ge>2 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l!k') = Seq C\<^sub>1' (While C))" using k_def(6) by blast
  have "\<forall>k'<k-3. \<exists>C\<^sub>1'. fst (l'\<^sub>3!k') = Seq C\<^sub>1' (While C)" 
  proof
    fix k'
    show "k'<k-3 \<longrightarrow> (\<exists>C\<^sub>1'. fst (l'\<^sub>3!k') = Seq C\<^sub>1' (While C))"
    proof
      assume "k'<k-3"
      have "k'+2\<ge>2" by simp
      moreover have "\<exists>C\<^sub>1'. fst (l!(k'+2)) = Seq C\<^sub>1' (While C)" using \<open>k' < k - 3\<close> k_def(6) by force
      then show "\<exists>C\<^sub>1'. fst (l'\<^sub>3!k') = Seq C\<^sub>1' (While C)" using \<open>\<forall>k'\<le>k - 3. l'\<^sub>3 ! k' = l' ! (2 + k')\<close> \<open>k' < k - 3\<close> ll'_same by force
    qed
  qed
  then have "red_iter (C, snd (l!2)) (k-3) (Skip, snd (l!(k-1)))" using seq_rev_path l'\<^sub>3_path by blast
  note \<open>fst (l!(k-1)) = Seq Skip (While C)\<close>
  moreover note \<open>fst (l!k) = While C\<close>
  moreover have "l!(k-1) \<rightarrow> l!k" by (metis (no_types, opaque_lifting) \<open>\<forall>k'<k. (l'!k' \<rightarrow> l'!(k'+1))\<close> \<open>a_path l'\<^sub>2 (k-2)\<close> a_path_def add.commute add_diff_cancel_right' diff_diff_left diff_le_self l'\<^sub>2_len less_numeral_extra(1) ll'_same ordered_cancel_comm_monoid_diff_class.add_diff_inverse zero_less_diff)
  ultimately have "snd (l!(k-1)) = snd (l!k)" by (metis prod.collapse red_Seq_elim2)
  have "l!0 = (While C, \<sigma>)" using assms(1) is_path_def by blast
  moreover have "l!2=(Seq C (While C), \<sigma>)" using Iter_ge2_steps \<open>2 < len\<close> assms(1) by blast
  ultimately have "snd (l!0) = snd (l!2)" by auto
  then have "red_iter (C, snd (l!0)) (k-3) (Skip, snd (l!k))" using \<open>red_iter (C, snd (l!2)) (k-3) (Skip, snd (l!(k-1)))\<close> \<open>snd (l ! (k - 1)) = snd (l ! k)\<close> by auto
  then have 0: "(C, \<sigma>) \<rightarrow>* (Skip, snd (l!k))" by (simp add: \<open>l ! 0 = (While C, \<sigma>)\<close> rtred_eq_red_iter_bkw)

  obtain l'' where l''_def: "l'' = drop k l" by fast
  have l''_len: "length l'' = len-k+1" by (metis Nat.add_diff_assoc2 k_def(1) l''_def l_len length_drop less_or_eq_imp_le)
  have ll''_same: "\<forall>k'\<le>len-k. l!(k'+k)=l''!k'" by (metis add.commute k_def(1) l''_def l_len less_or_eq_imp_le nth_drop trans_le_add2)
  moreover note \<open>\<forall>k'<len. (l!k' \<rightarrow> l!(k'+1))\<close>
  ultimately have "\<forall>k'<len-k. (l''!k' \<rightarrow> l''!(k'+1))" using k_def(1) by (simp add: l''_def l_len)
  then have "a_path l'' (len-k)" using a_path_def l''_len by auto
  then have 1: "is_path l'' (l!k) (len-k) (Skip, \<sigma>'')" by (metis (no_types, lifting) assms(1) is_path_def k_def(1) le_add_diff_inverse2 le_add_same_cancel2 le_numeral_extra(3) less_imp_le_nat ll''_same plus_nat.add_0)

  have "fst (l!0) = While C" using assms(1) unfolding is_path_def by force
  then have "iter_count C (drop 1 l) = n-1" using assms(2) by (metis Suc_diff_1 diff_0_eq_0 diff_Suc_1 diff_self_eq_0 drop0 drop_Nil drop_Suc_Cons iter_count.elims less_one nth_Cons_0)
  moreover have "drop k l = drop (k-1) (drop 1 l)" using k_def(2) by force
  ultimately have "iter_count C (drop k l) \<le> n-1" using iter_count_mono by metis

  then have "((C, \<sigma>) \<rightarrow>* (Skip, snd (l!k))) \<and> (is_path (drop k l) (While C, snd (l!k)) (len-k) (Skip, \<sigma>'')) \<and> iter_count C (drop k l) \<le> n-1" by (metis "0" "1" k_def(3) l''_def prod.collapse)
  then show ?thesis using k_def(1) by blast
qed

lemma every_while_is_some_iter_path:
  assumes "is_path l (While C, \<sigma>) len (Skip, \<sigma>'')"
      and "iter_count C l = Suc n"
    shows "\<exists>N. ((iter C N, \<sigma>) \<rightarrow>* (Skip, \<sigma>''))"
  using assms
proof(induction n arbitrary: l \<sigma> len rule: less_induct)
  case (less n)
  then show ?case
  proof(cases "n=0")
    case True
    then show ?thesis using assms iter_count_1 RedRefl less.prems by (metis iter.simps(1) One_nat_def)
  next
    case False
    
    then obtain k where k_def: "k<len" "(C, \<sigma>) \<rightarrow>* (Skip, snd (l!k))"
                               "is_path (drop k l) (While C, snd (l!k)) (len-k) (Skip, \<sigma>'')"
                               "iter_count C (drop k l) \<le> n"
      using while_to_seq_and_while by (metis Suc_1 diff_Suc_1 diff_is_0_eq' less.prems(1) less.prems(2) not_less_eq_eq)
    moreover have "iter_count C (drop k l) \<le> n \<longleftrightarrow> iter_count C (drop k l) - 1 < n" using False by linarith
    have 0: "length l>0 \<Longrightarrow> fst (l!0) = While C \<Longrightarrow> iter_count C l > 0" using iter_count.elims by force
    have 1: "length (drop k l) > 0" using k_def(3) is_path_def a_path_def by (metis Suc_eq_plus1 Zero_not_Suc bot_nat_0.not_eq_extremum)
    have 2: "fst ((drop k l)!0) = While C" using k_def(3) unfolding is_path_def by force
    moreover have "iter_count C (drop k l)>0" using "0" "1" "2" by (metis (no_types, lifting) iter_count.elims length_greater_0_conv less_Suc_eq_0_disj nth_Cons_0)
    then have "iter_count C (drop k l) = Suc (iter_count C (drop k l) - 1)" by force
    ultimately obtain N where "(iter C N, snd (l ! k)) \<rightarrow>* (Skip, \<sigma>'')" using less.IH[of "iter_count C (drop k l) - 1" "drop k l" "snd (l!k)" "len-k"] by fastforce
    then show ?thesis using k_def(2) iter_combine2 by blast
  qed
qed

theorem while_iff_union_iter:
  "((While C, \<sigma>) \<rightarrow>* (Skip, \<sigma>')) \<longleftrightarrow> (\<exists>n. ((iter C n, \<sigma>) \<rightarrow>* (Skip, \<sigma>')))"
proof
  show "(While C, \<sigma>) \<rightarrow>* (Skip, \<sigma>') \<Longrightarrow> \<exists>n. ((iter C n, \<sigma>) \<rightarrow>* (Skip, \<sigma>'))"
  proof(-)
    assume "(While C, \<sigma>) \<rightarrow>* (Skip, \<sigma>')"
    then obtain len where "red_iter (While C, \<sigma>) len (Skip, \<sigma>')" using red_iter_closed_def rtred_eq_red_iter_closed_fwd by blast
    then obtain l where "is_path l (While C, \<sigma>) len (Skip, \<sigma>')" using red_iter's_path by blast
    then have "fst (l!0) = While C" by (simp add: is_path_def)
    moreover have "length l>0"
    proof(rule ccontr)
      assume "\<not> (length l>0)"
      then have "length l=0" by fast
      then have "(While C, \<sigma>) = (Skip, \<sigma>')" by (metis Suc_eq_plus1 Zero_not_Suc \<open>is_path l (While C, \<sigma>) len (Skip, \<sigma>')\<close> a_path_def is_path_def)
      then show "False" by simp
    qed
    ultimately have "iter_count C l \<ge> 1" by (metis One_nat_def Suc_leI iter_count.elims length_greater_0_conv less_Suc_eq_0_disj nth_Cons_0)
    then show "\<exists>n. ((iter C n, \<sigma>) \<rightarrow>* (Skip, \<sigma>'))" by (metis One_nat_def \<open>is_path l (While C, \<sigma>) len (Skip, \<sigma>')\<close> every_while_is_some_iter_path gr0_implies_Suc linorder_not_less not_less_eq_eq)
  qed
  show "\<exists>n. ((iter C n, \<sigma>) \<rightarrow>* (Skip, \<sigma>')) \<Longrightarrow> (While C, \<sigma>) \<rightarrow>* (Skip, \<sigma>')" using every_iter_is_while_skip by blast
qed

corollary dsem_while:
  "(\<Union>n. \<lbrakk>iter C n\<rbrakk>) = \<lbrakk>While C\<rbrakk>" 
proof(-)
  have "(\<Union>n. \<lbrakk>iter C n\<rbrakk>) = (\<Union>n. {(\<sigma>, \<sigma>')| \<sigma> \<sigma>'. (iter C n, \<sigma>) \<rightarrow>* (Skip, \<sigma>')})" using dsem_def by fast
  then have "(\<Union>n. \<lbrakk>iter C n\<rbrakk>) = {(\<sigma>, \<sigma>')| \<sigma> \<sigma>'. \<exists>n. ((iter C n, \<sigma>) \<rightarrow>* (Skip, \<sigma>'))}" by blast
  then have "(\<Union>n. \<lbrakk>iter C n\<rbrakk>) = {(\<sigma>, \<sigma>')| \<sigma> \<sigma>'. (While C, \<sigma>) \<rightarrow>* (Skip, \<sigma>')}" unfolding while_iff_union_iter by blast
  then show ?thesis unfolding dsem_def by auto
qed

lemma dsem_iter:
  "\<lbrakk>iter C (Suc n)\<rbrakk> = {(\<sigma>, \<sigma>'')| \<sigma> \<sigma>' \<sigma>''. (\<sigma>, \<sigma>')\<in>\<lbrakk>iter C n\<rbrakk> \<and> (\<sigma>', \<sigma>'')\<in>\<lbrakk>C\<rbrakk>}"
  by (smt (verit, best) Collect_cong add_Seq_to_iterR dsem_seq)

lemma dsem_iter':
  "\<lbrakk>iter C (Suc n)\<rbrakk> = {(\<sigma>, \<sigma>'')| \<sigma> \<sigma>' \<sigma>''. (\<sigma>, \<sigma>')\<in>\<lbrakk>C\<rbrakk> \<and> (\<sigma>', \<sigma>'')\<in>\<lbrakk>iter C n\<rbrakk>}"
  by (smt (verit, best) Collect_cong add_Seq_to_iterL dsem_seq)

lemma "{\<sigma>'. \<exists>\<sigma>\<in>p. (\<sigma>, \<sigma>')\<in>\<lbrakk>Skip\<rbrakk>} = p" unfolding dsem_skip by auto

(*As a future idea of how to more clearly prove those:
lemma rtred_rhs:
  "(Assign x e, (s, h)) \<rightarrow>* (skip, \<sigma>'') \<Longrightarrow> skip=Skip \<Longrightarrow> \<sigma>''=(s(x:=(e s)), h)"
  "(Seq Skip C\<^sub>2, \<sigma>) \<rightarrow> c' \<Longrightarrow> c'=(C\<^sub>2, \<sigma>)"
  "(Seq C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow> c' \<Longrightarrow> C\<^sub>1\<noteq>Skip \<Longrightarrow> \<exists>C\<^sub>1' \<sigma>'. c'=(Seq C\<^sub>1' C\<^sub>2, \<sigma>') \<and> ((C\<^sub>1, \<sigma>) \<rightarrow> (C\<^sub>1', \<sigma>'))"
  "(If C\<^sub>1 C\<^sub>2, \<sigma>) \<rightarrow> c' \<Longrightarrow> c'=(C\<^sub>1, \<sigma>) \<or> c'=(C\<^sub>2, \<sigma>)"
  "(Havoc x, (s, h)) \<rightarrow> c' \<Longrightarrow> \<exists>n. c'=(Skip, (s(x:=n), h))"
  "(Assume b, \<sigma>) \<rightarrow> c' \<Longrightarrow> c'=(Skip, \<sigma>) \<and> (fst \<sigma>)\<in>b"
  "(While C, \<sigma>) \<rightarrow> c' \<Longrightarrow> c'=(If (Seq C (While C)) Skip, \<sigma>)"
  "(Alloc x, (s, h)) \<rightarrow> c' \<Longrightarrow> \<exists>l n. c'=(Skip, (s(x:=l),fmap_update h l (hOcc n))) \<and> (h(l)=hUkw \<or> h(l)=hFrd) \<and> l\<noteq>0 \<and> l\<notin>(s, h)"
  "(Write x e, (s, h)) \<rightarrow> c' \<Longrightarrow> \<exists>n. c'=(Skip, (s, h(s(x):=hOcc (e s)))) \<and> h(s(x))=hOcc n \<and> s(x)\<noteq>0"
  "(Read y x, (s, h)) \<rightarrow> c' \<Longrightarrow> \<exists>n. c'=(Skip, (s(y:=n), h)) \<and> h(s(x))=hOcc n \<and> s(x)\<noteq>0"
  "(Free x, (s, h)) \<rightarrow> c' \<Longrightarrow> \<exists>n. c'=(Skip, (s, h(s(x):=hUkw))) \<and> h(s(x))=hOcc n \<and> s(x)\<noteq>0"
proof(-)
  show "(Assign x e, (s, h)) \<rightarrow>* (skip, \<sigma>'') \<Longrightarrow> skip=Skip \<Longrightarrow> \<sigma>''=(s(x:=(e s)), h)"
    apply(induction "(Assign x e, (s, h))" "(skip, \<sigma>'')" rule: rtred.induct) apply blast sledgehammer*)

lemma assign_rt_:
  assumes "(Assign x e, \<sigma>) \<rightarrow>* (Skip, \<sigma>'')"
  shows "fst \<sigma>''=(fst \<sigma>)(x:=e (fst \<sigma>)) \<and> snd \<sigma>''=snd \<sigma>"
  using assms by (smt (verit, ccfv_SIG) cmd.distinct(5) fst_conv prod.swap_def red_rhs(1) rtred.simps skip_deadend_rt snd_conv swap_swap)

lemma dsem_assign:
  "\<lbrakk>Assign x e\<rbrakk> = {((s, h), (s(x:=e s), h))| s h. True}" 
proof(-)
  have "\<lbrakk>Assign x e\<rbrakk> = {((s, h), (s', h'))| s h s' h'. (Assign x e, (s, h)) \<rightarrow>* (Skip, (s', h'))}" unfolding dsem_def by simp
  moreover have "\<And>s h s' h'. ((Assign x e, (s, h)) \<rightarrow>* (Skip, (s', h'))) \<longleftrightarrow> s'=s(x:=e s) \<and> h=h'" by (metis (no_types, opaque_lifting) RedAssign RedRefl RedTran assign_rt_ fst_conv snd_conv)
  ultimately show ?thesis by presburger
qed

lemma havoc_rt_:
  assumes "(Havoc x, \<sigma>) \<rightarrow>* (Skip, \<sigma>'')"
  shows "\<exists>n. fst \<sigma>''=(fst \<sigma>)(x:=n) \<and> snd \<sigma>''=snd \<sigma>"
  using assms by (smt (verit, del_insts) cmd.distinct(55) prod.collapse prod.inject red_rhs(5) rtred.cases skip_deadend_rt)

lemma dsem_havoc:
  "\<lbrakk>Havoc x\<rbrakk> = {((s, h), (s(x:=n), h))| s h n. True}"
proof(-)
  have "\<lbrakk>Havoc x\<rbrakk> = {((s, h), (s', h'))| s h s' h'. (Havoc x, (s, h)) \<rightarrow>* (Skip, (s', h'))}" unfolding dsem_def by force
  moreover have "\<And>s h s' h'. ((Havoc x, (s, h)) \<rightarrow>* (Skip, (s', h'))) \<longleftrightarrow> (\<exists>n. s'=s(x:=n)) \<and> h=h'" by (metis (no_types, opaque_lifting) RedHavoc fst_conv havoc_rt_ rtred.simps snd_conv)
  ultimately show ?thesis by auto
qed

lemma alloc_rt:
  assumes "(Alloc x, \<sigma>) \<rightarrow>* (Skip, \<sigma>'')"
  shows "\<exists>l n. (fmap_lookup (snd \<sigma>) l=hUkw \<or> fmap_lookup (snd \<sigma>) l=hFrd) \<and> l\<noteq>0 \<and> fst \<sigma>''=(fst \<sigma>)(x:=l) \<and> snd \<sigma>''=fmap_update (snd \<sigma>) l (hOcc n)"
proof(-)
  obtain c' where c'_def: "(Alloc x, \<sigma>) \<rightarrow> c'" "c' \<rightarrow>* (Skip, \<sigma>'')" by (metis assms cmd.distinct(61) prod.inject rtred.cases)
  then show ?thesis by (smt (z3) old.prod.inject prod.collapse red_rhs(8) skip_deadend_rt)
qed

lemma dsem_alloc:
  "\<lbrakk>Alloc x\<rbrakk> = {((s, h), (s(x:=l), fmap_update h l (hOcc n)))| s h l n. (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0}"
proof(-)
  have "\<lbrakk>Alloc x\<rbrakk> = {((s, h), (s', h'))| s h s' h'. (Alloc x, (s, h)) \<rightarrow>* (Skip, (s', h'))}" unfolding dsem_def by auto
  moreover have "\<And>s h s' h'. ((Alloc x, (s, h)) \<rightarrow>* (Skip, (s', h'))) \<longleftrightarrow> (\<exists>l n. (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0 \<and> s'=s(x:=l) \<and> h'=fmap_update h l (hOcc n))"
  proof
    show "\<And>s h s' h'. (Alloc x, (s, h)) \<rightarrow>* (Skip, (s', h')) \<Longrightarrow> \<exists>l n. (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0 \<and> s'=s(x:=l) \<and> h'=fmap_update h l (hOcc n)" using alloc_rt by fastforce
    show "\<And>s h s' h'. \<exists>l n. (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0 \<and> s'=s(x:=l) \<and> h'=fmap_update h l (hOcc n) \<Longrightarrow> (Alloc x, (s, h)) \<rightarrow>* (Skip, (s', h'))" by (meson RedAlloc RedRefl RedTran)
  qed
  ultimately have "\<lbrakk>Alloc x\<rbrakk> = {((s, h), (s', h'))| s h s' h'. \<exists>l n. (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0 \<and> s'=s(x:=l) \<and> h'=fmap_update h l (hOcc n)}" by presburger
  then show ?thesis by fastforce
qed

lemma write_rt:
  assumes "(Write x e, \<sigma>) \<rightarrow>* (Skip, \<sigma>'')"
  shows "fst \<sigma>''=fst \<sigma> \<and> snd \<sigma>''=fmap_update (snd \<sigma>) ((fst \<sigma>) x) (hOcc (e (fst \<sigma>))) \<and> (fst \<sigma>) x\<noteq>0 \<and> (\<exists>n. fmap_lookup (snd \<sigma>) ((fst \<sigma>) x)=hOcc n)"
  using assms by (smt (verit) Pair_inject cmd.distinct(63) prod.collapse red_rhs(9) rtred.cases skip_deadend_rt)

lemma dsem_write:
  "\<lbrakk>Write x e\<rbrakk> = {((s, h), (s, fmap_update h (s x) (hOcc (e s))))| s h n. fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}" 
proof(-)
  have "\<lbrakk>Write x e\<rbrakk> = {((s, h), (s', h'))| s h s' h'. (Write x e, (s, h)) \<rightarrow>* (Skip, (s', h'))}" unfolding dsem_def by simp
  moreover have "\<And>s h s' h'. ((Write x e, (s, h)) \<rightarrow>* (Skip, (s', h'))) \<longleftrightarrow> s'=s \<and> h'=fmap_update h (s x) (hOcc (e s)) \<and> s x\<noteq>0 \<and> (\<exists>n. fmap_lookup h (s x)=hOcc n)"
  proof
    show "\<And>s h s' h'. (Write x e, (s, h)) \<rightarrow>* (Skip, (s', h')) \<Longrightarrow> s'=s \<and> h'=fmap_update h (s x) (hOcc (e s)) \<and> s x\<noteq>0 \<and> (\<exists>n. fmap_lookup h (s x)=hOcc n)" using write_rt by fastforce
    show "\<And>s h s' h'. s'=s \<and> h'=fmap_update h (s x) (hOcc (e s)) \<and> s x\<noteq>0 \<and> (\<exists>n. fmap_lookup h (s x)=hOcc n) \<Longrightarrow> (Write x e, (s, h)) \<rightarrow>* (Skip, (s', h'))" by (meson RedRefl RedTran RedWrite)
  qed
  ultimately show ?thesis by auto
qed

lemma read_rt:
  assumes "(Read y x, \<sigma>) \<rightarrow>* (Skip, \<sigma>'')"                \<comment>\<open>y:=[x]\<close>
  shows "\<exists>n. fmap_lookup (snd \<sigma>) (fst \<sigma> x)=hOcc n \<and> fst \<sigma>''=(fst \<sigma>)(y:=n) \<and> snd \<sigma>''=snd \<sigma> \<and> fst \<sigma> x\<noteq>0"
  using assms by (smt (verit, del_insts) Pair_inject cmd.distinct(65) prod.collapse red_rhs(10) rtred.cases skip_deadend_rt)

lemma dsem_read:
  "\<lbrakk>Read y x\<rbrakk> = {((s, h), (s(y:=n), h))| s h n. fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}" 
proof(-)
  have "\<lbrakk>Read y x\<rbrakk> = {((s, h), (s', h'))| s h s' h'. (Read y x, (s, h)) \<rightarrow>* (Skip, (s', h'))}" unfolding dsem_def by simp
  moreover have "\<And>s h s' h'. ((Read y x, (s, h)) \<rightarrow>* (Skip, (s', h'))) \<longleftrightarrow> (\<exists>n. s'=s(y:=n) \<and> h'=h \<and> s x\<noteq>0 \<and> fmap_lookup h (s x)=hOcc n)"
  proof
    show "\<And>s h s' h'. (Read y x, (s, h)) \<rightarrow>* (Skip, (s', h')) \<Longrightarrow> \<exists>n. s'=s(y:=n) \<and> h'=h \<and> s x\<noteq>0 \<and> fmap_lookup h (s x)=hOcc n" using read_rt by fastforce
    show "\<And>s h s' h'. \<exists>n. s'=s(y:=n) \<and> h'=h \<and> s x\<noteq>0 \<and>  fmap_lookup h (s x)=hOcc n \<Longrightarrow> (Read y x, (s, h)) \<rightarrow>* (Skip, (s', h'))" by (meson RedRefl RedTran RedRead)
  qed
  ultimately show ?thesis by auto
qed

lemma free_rt:
  assumes "(Free x, \<sigma>) \<rightarrow>* (Skip, \<sigma>'')"
  shows "fst \<sigma>''=fst \<sigma> \<and> snd \<sigma>''=fmap_update (snd \<sigma>) ((fst \<sigma>) x) hFrd \<and> (fst \<sigma>) x\<noteq>0 \<and> (\<exists>n. fmap_lookup (snd \<sigma>) ((fst \<sigma>) x)=hOcc n)"
  using assms by (smt (verit) cmd.distinct(67) fst_conv prod.swap_def red_rhs(11) rtred.cases skip_deadend_rt snd_conv swap_swap)

lemma dsem_free:
  "\<lbrakk>Free x\<rbrakk> = {((s, h), (s, fmap_update h (s x) hFrd))| s h n. fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}" 
proof(-)
  have "\<lbrakk>Free x\<rbrakk> = {((s, h), (s', h'))| s h s' h'. (Free x, (s, h)) \<rightarrow>* (Skip, (s', h'))}" unfolding dsem_def by simp
  moreover have "\<And>s h s' h'. ((Free x, (s, h)) \<rightarrow>* (Skip, (s', h'))) \<longleftrightarrow> s'=s \<and> h'=fmap_update h (s x) hFrd \<and> s x\<noteq>0 \<and> (\<exists>n. fmap_lookup h (s x)=hOcc n)"
  proof
    show "\<And>s h s' h'. (Free x, (s, h)) \<rightarrow>* (Skip, (s', h')) \<Longrightarrow> s'=s \<and> h'=fmap_update h (s x) hFrd \<and> s x\<noteq>0 \<and> (\<exists>n. fmap_lookup h (s x)=hOcc n)" using free_rt by fastforce
    show "\<And>s h s' h'. s'=s \<and> h'=fmap_update h (s x) hFrd \<and> s x\<noteq>0 \<and> (\<exists>n. fmap_lookup h (s x)=hOcc n) \<Longrightarrow> (Free x, (s, h)) \<rightarrow>* (Skip, (s', h'))" by (meson RedRefl RedTran RedFree)
  qed
  ultimately show ?thesis by auto
qed

theorem dsem_properties:
  "\<lbrakk>Skip\<rbrakk> = {(\<sigma>, \<sigma>)| \<sigma>. True}"
  "\<lbrakk>Assign x e\<rbrakk> = {((s, h), (s(x:=e s), h))| s h. True}"
  "\<lbrakk>Havoc x\<rbrakk> = {((s, h), (s(x:=n), h))| s h n. True}"
  "\<lbrakk>Assume b\<rbrakk> = {(\<sigma>, \<sigma>)| \<sigma>. fst \<sigma>\<in>b}"
  "\<lbrakk>Seq C\<^sub>1 C\<^sub>2\<rbrakk> = \<lbrakk>C\<^sub>1\<rbrakk> O \<lbrakk>C\<^sub>2\<rbrakk>"
  "\<lbrakk>If C\<^sub>1 C\<^sub>2\<rbrakk> = \<lbrakk>C\<^sub>1\<rbrakk> \<union> \<lbrakk>C\<^sub>2\<rbrakk>"
  "\<lbrakk>While C\<rbrakk> = (\<Union>n. \<lbrakk>iter C n\<rbrakk>)"
  "\<lbrakk>Alloc x\<rbrakk> = {((s, h), (s(x:=l), fmap_update h l (hOcc n)))| s h l n. (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0}"
  "\<lbrakk>Write x e\<rbrakk> = {((s, h), (s, fmap_update h (s x) (hOcc (e s))))| s h n. fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}"
  "\<lbrakk>Read y x\<rbrakk> = {((s, h), (s(y:=n), h))| s h n. fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}"
  "\<lbrakk>Free x\<rbrakk> = {((s, h), (s, fmap_update h (s x) hFrd))| s h n. fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}"
            apply (auto simp add: dsem_skip dsem_assign dsem_havoc dsem_assume dsem_seq dsem_if dsem_while dsem_alloc dsem_write dsem_read dsem_free)
  by fast

fun md :: "'var cmd \<Rightarrow> 'var set" where
 "md (Assign x e) = {x}"
|"md (Seq C\<^sub>1 C\<^sub>2) = md C\<^sub>1 \<union> md C\<^sub>2"
|"md (If C\<^sub>1 C\<^sub>2) = md C\<^sub>1 \<union> md C\<^sub>2"
|"md Skip = {}"
|"md (Havoc x) = {x}"
|"md (Assume _) = {}"
|"md (While C) = md C"
|"md (Alloc x) = {x}"
|"md (Write _ _) = {}"
|"md (Read x _) = {x}"
|"md (Free _) = {}"

lemma red_perserves_non_md:
  "(C, \<sigma>) \<rightarrow> (C', \<sigma>') \<Longrightarrow> t\<notin>md C \<Longrightarrow> (fst \<sigma>) t = n \<Longrightarrow> (fst \<sigma>') t = n"
  apply(induction "(C, \<sigma>)" "(C', \<sigma>')" arbitrary: C \<sigma> C' \<sigma>' rule: red.induct)
  by auto

lemma red_step_doesnt_introduce_new_md: 
  "(C, \<sigma>) \<rightarrow> (C', \<sigma>') \<Longrightarrow> t\<notin>md C \<Longrightarrow> t\<notin>md C'"
  apply(induction "(C, \<sigma>)" "(C', \<sigma>')" arbitrary: C \<sigma> C' \<sigma>' rule: red.induct)
  by auto
  

lemma rtred_perserves_non_md:
  "(C, \<sigma>) \<rightarrow>* (C'', \<sigma>'') \<Longrightarrow> t\<notin>md C \<Longrightarrow> (fst \<sigma>) t = n \<Longrightarrow> (fst \<sigma>'') t = n"
proof(induction "(C, \<sigma>)" "(C'', \<sigma>'')" arbitrary: C \<sigma> C'' \<sigma>'' rule: rtred.induct)
  case (RedRefl)
  then show ?case by fastforce
next
  case (RedTran c')
  moreover obtain C' \<sigma>' where "c'=(C', \<sigma>')" by (meson prod.exhaust_sel)
  moreover have "(fst \<sigma>') t = n" using red_perserves_non_md by (metis calculation(1) calculation(4) calculation(5) calculation(6))
  moreover have "t\<notin>md C'" using red_step_doesnt_introduce_new_md by (metis calculation(1) calculation(4) calculation(6))
  ultimately show ?case by blast
qed

lemma sem_perserves_non_md:
  "t\<notin>md C \<Longrightarrow> \<lbrakk>C\<rbrakk>``{\<sigma>\<in>X. (fst \<sigma>) t = n} = {\<sigma>\<in>(\<lbrakk>C\<rbrakk>``X). (fst \<sigma>) t = n}"
  using rtred_perserves_non_md dsem_char by fastforce 

lemma sem_perserves_non_md_singleton:
  "t\<notin>md C \<Longrightarrow> fst \<sigma> t = n \<Longrightarrow> \<forall>\<sigma>'\<in>\<lbrakk>C\<rbrakk>``{\<sigma>}. fst \<sigma>' t = n"
  using sem_perserves_non_md[of t C "{\<sigma>}" n] by blast

lemma sem_perserves_non_md_element:
  assumes "t\<notin>md C"
      and "\<And>\<sigma>. \<sigma>\<in>S \<Longrightarrow> fst \<sigma> t \<noteq> n"
    shows "\<And>\<sigma>'. \<sigma>'\<in>\<lbrakk>C\<rbrakk>``S \<Longrightarrow> fst \<sigma>' t \<noteq> n"
proof(rule ccontr)
  fix \<sigma>'
  show "\<sigma>'\<in>\<lbrakk>C\<rbrakk>``S \<Longrightarrow> \<not>fst \<sigma>' t \<noteq> n \<Longrightarrow> False"
  proof(-)
    assume asm: "\<sigma>'\<in>\<lbrakk>C\<rbrakk>``S" "\<not>fst \<sigma>' t \<noteq> n"
    then have "fst \<sigma>' t = n" by auto
    then obtain \<sigma> where "\<sigma>\<in>S" "(C, \<sigma>) \<rightarrow>* (Skip, \<sigma>')" using asm(1) unfolding dsem_def by blast
    then have "\<sigma>'\<in>\<lbrakk>C\<rbrakk>``{\<sigma>}" unfolding dsem_def by blast
    then show "False" by (metis \<open>\<sigma> \<in> S\<close> \<open>fst \<sigma>' t = n\<close> assms(1) assms(2) sem_perserves_non_md_singleton)
  qed
qed







inductive aborts where
 "s x=0 \<or> fmap_lookup h (s x)=hUkw \<or> fmap_lookup h (s x)=hFrd \<Longrightarrow> aborts (Write x e, (s, h))"
|"s x=0 \<or> fmap_lookup h (s x)=hUkw \<or> fmap_lookup h (s x)=hFrd \<Longrightarrow> aborts (Read y x, (s, h))"
|"s x=0 \<or> fmap_lookup h (s x)=hUkw \<or> fmap_lookup h (s x)=hFrd \<Longrightarrow> aborts (Free x, (s, h))"
|"aborts (C\<^sub>1, \<sigma>) \<Longrightarrow> aborts (Seq C\<^sub>1 C\<^sub>2, \<sigma>)"

lemma aborts_in_head:
  assumes "aborts (C, \<sigma>)"
      and "C = Seq C\<^sub>1 C\<^sub>2"
  shows "aborts (C\<^sub>1, \<sigma>)"
  using assms
proof(induction "(C, \<sigma>)" rule: aborts.induct)
  case (1 s x h e)
  then show ?case by fast
next
  case (2 s x h y)
  then show ?case by force
next
  case (3 s x h)
  then show ?case by simp
next
  case (4 C\<^sub>1 C\<^sub>2)
  then show ?case by fast
qed

lemma aborts_mono:
  assumes "aborts (C, (s, h++\<^sub>fh'))"
      and "h\<bottom>h'" (*otherwise (x\<mapsto>\<bottom>)++(x\<mapsto>1) idk what happens*)
  shows "aborts (C, (s, h))"
  using assms
proof(induction "(C, (s, h++\<^sub>fh'))" arbitrary: C rule: aborts.induct)
  case (1 x e)
  then show ?case by (metis aborts.intros(1) compat_heap_union_comm domIff map_add_dom_app_simps(1) fmap_add_simp fmap_lookup_simp)
next
  case (2 x y)
  then show ?case by (metis aborts.intros(2) compat_heap_union_comm domIff map_add_dom_app_simps(1) fmap_add_simp fmap_lookup_simp)
next
  case (3 x)
  then show ?case by (metis aborts.intros(3) compat_heap_union_comm domIff map_add_dom_app_simps(1) fmap_add_simp fmap_lookup_simp)
next
  case (4 C\<^sub>1 C\<^sub>2)
  then show ?case by (simp add: aborts.intros(4))
qed

(*We can strenghten it to say that h\<^sub>f' is h\<^sub>f with potentially one \<bottom> replaced with \<emptyset>*)
lemma heap_extension_independence:
  assumes "(C, (s, h\<^sub>p++\<^sub>fh\<^sub>f)) \<rightarrow> (C', (s', h'))"
      and "\<And>c'. (C, (s, h\<^sub>p)) \<rightarrow>* c' \<Longrightarrow> \<not> aborts c'"
      and "h\<^sub>p\<bottom>h\<^sub>f"
    shows "\<exists>h\<^sub>p' h\<^sub>f'. h'=h\<^sub>p'++\<^sub>fh\<^sub>f' \<and> h\<^sub>p'\<bottom>h\<^sub>f' \<and> ((C, (s, h\<^sub>p)) \<rightarrow> (C', (s', h\<^sub>p')))"
  using assms
proof(induction "(C, (s, h\<^sub>p++\<^sub>fh\<^sub>f))" "(C', (s', h'))" arbitrary: C C' rule: red.induct)
  case (RedAssign x e)
  then show ?case using red.RedAssign by metis
next
  case (RedSeq1)
  then show ?case using red.RedSeq1 by blast
next
  case (RedSeq2 C\<^sub>1 C\<^sub>1' C\<^sub>2)
  then show ?case using aborts.intros(4) seq_trcl red.RedSeq2 by (metis old.prod.exhaust)
next
  case (RedIf1 C\<^sub>1 C\<^sub>2)
  then show ?case using red.RedIf1 by blast
next
  case (RedIf2 C\<^sub>1 C\<^sub>2)
  then show ?case using red.RedIf2 by blast
next
  case (RedHavoc x v)
  then show ?case using red.RedHavoc by metis
next
  case (RedAssume b)
  then show ?case using red.RedAssume by (metis fst_eqD)
next
  case (RedWhile C)
  then show ?case using red.RedWhile by blast
next
  case (RedAlloc l x n)
  moreover obtain h\<^sub>p' where "h\<^sub>p'=fmap_update h\<^sub>p l (hOcc n)" by force
  ultimately show ?case
  proof(cases "fmap_lookup (h\<^sub>p++\<^sub>fh\<^sub>f) l=hUkw")
    case True
    then have "fmap_lookup h\<^sub>p l=hUkw" "fmap_lookup h\<^sub>f l=hUkw" by auto
    then have "h\<^sub>p'\<bottom>h\<^sub>f" using \<open>h\<^sub>p'=fmap_update h\<^sub>p l (hOcc n)\<close> assms(3) unfolding compat_def by auto
    moreover have "h'=h\<^sub>p'++\<^sub>fh\<^sub>f"
    proof(-)
      have "h'\<subseteq>\<^sub>fh\<^sub>p'++\<^sub>fh\<^sub>f" using RedAlloc.hyps(4) \<open>h\<^sub>p'=fmap_update h\<^sub>p l (hOcc n)\<close> assms(3) \<open>h\<^sub>p'\<bottom>h\<^sub>f\<close> compat_heap_union_comm by (metis fmap_add_upd fmap_le_refl)
      moreover have "h\<^sub>p'++\<^sub>fh\<^sub>f\<subseteq>\<^sub>fh'" using RedAlloc.hyps(4) \<open>h\<^sub>p'=fmap_update h\<^sub>p l (hOcc n)\<close> assms(3) \<open>h\<^sub>p'\<bottom>h\<^sub>f\<close> compat_heap_union_comm by (metis fmap_add_upd fmap_le_refl)
      ultimately show ?thesis by (meson Rep_fmap_inject fmap_le_simp map_le_antisym)
    qed
    ultimately show ?thesis using red.RedAlloc by (metis RedAlloc.hyps(2) RedAlloc.hyps(3) \<open>fmap_lookup h\<^sub>p l=hUkw\<close> \<open>h\<^sub>p'=fmap_update h\<^sub>p l (hOcc n)\<close>)
  next
    case False
    then have "fmap_lookup (h\<^sub>p++\<^sub>fh\<^sub>f) l=hFrd" using RedAlloc.hyps(1) by argo
    then have cs: "(fmap_lookup h\<^sub>p l=hFrd \<and> fmap_lookup h\<^sub>f l=hUkw) \<or> (fmap_lookup h\<^sub>p l=hUkw \<and> fmap_lookup h\<^sub>f l=hFrd)" using RedAlloc.prems(2) unfolding compat_def by auto
    then show ?thesis
    proof(cases "fmap_lookup h\<^sub>p l=hFrd \<and> fmap_lookup h\<^sub>f l=hUkw")
      case True
      then have "h\<^sub>p'\<bottom>h\<^sub>f" using \<open>h\<^sub>p'=fmap_update h\<^sub>p l (hOcc n)\<close> assms(3) unfolding compat_def by auto
      moreover have "h'=h\<^sub>p'++\<^sub>fh\<^sub>f" 
      proof(-)
        have "h'\<subseteq>\<^sub>fh\<^sub>p'++\<^sub>fh\<^sub>f" using RedAlloc.hyps(4) \<open>h\<^sub>p'=fmap_update h\<^sub>p l (hOcc n)\<close> assms(3) \<open>h\<^sub>p'\<bottom>h\<^sub>f\<close> compat_heap_union_comm by (metis fmap_add_upd fmap_le_refl)
        moreover have "h\<^sub>p'++\<^sub>fh\<^sub>f\<subseteq>\<^sub>fh'" using RedAlloc.hyps(4) \<open>h\<^sub>p'=fmap_update h\<^sub>p l (hOcc n)\<close> assms(3) \<open>h\<^sub>p'\<bottom>h\<^sub>f\<close> compat_heap_union_comm by (metis fmap_add_upd fmap_le_refl)
        ultimately show ?thesis by (meson Rep_fmap_inject fmap_le_simp map_le_antisym)
      qed
      ultimately show ?thesis using red.RedAlloc by (metis RedAlloc.hyps(2) RedAlloc.hyps(3) \<open>fmap_lookup h\<^sub>p l=hFrd \<and> fmap_lookup h\<^sub>f l=hUkw\<close> \<open>h\<^sub>p'=fmap_update h\<^sub>p l (hOcc n)\<close>)
    next
      case False
      then have "fmap_lookup h\<^sub>p l=hUkw \<and> fmap_lookup h\<^sub>f l=hFrd" using cs by blast
      moreover obtain h\<^sub>f' where "h\<^sub>f'=fmap_update h\<^sub>f l (hUkw)" by simp
      then have "h\<^sub>p'\<bottom>h\<^sub>f'" using \<open>h\<^sub>p'=fmap_update h\<^sub>p l (hOcc n)\<close> assms(3) unfolding compat_def by auto
      moreover have "h'=h\<^sub>p'++\<^sub>fh\<^sub>f'" using RedAlloc.hyps(4) \<open>h\<^sub>f'=fmap_update h\<^sub>f l (hUkw)\<close> \<open>h\<^sub>p'=fmap_update h\<^sub>p l (hOcc n)\<close> domIff fun_upd_same fun_upd_upd map_add_upd map_add_upd_left by (smt (verit) Rep_fmap_inject fmap_add_simp fmap_update_simp)
      ultimately show ?thesis by (metis RedAlloc.hyps(2,3) \<open>h\<^sub>p'=fmap_update h\<^sub>p l (hOcc n)\<close> red.RedAlloc)
    qed
  qed
next
  case (RedWrite x n e)
  then have "s x\<noteq>0 \<and> fmap_lookup h\<^sub>p (s x)\<noteq>hUkw \<and> fmap_lookup h\<^sub>p (s x)\<noteq>hFrd" by (meson RedRefl aborts.intros(1))
  moreover obtain h\<^sub>p' where "h\<^sub>p'=fmap_update h\<^sub>p (s x) (hOcc (e s))" by fast
  ultimately have "h\<^sub>p'\<bottom>h\<^sub>f" using assms(3) unfolding compat_def by auto
  have "h'=h\<^sub>p'++\<^sub>fh\<^sub>f"
  proof(-)
    have "h'\<subseteq>\<^sub>fh\<^sub>p'++\<^sub>fh\<^sub>f" using RedWrite.hyps(4) \<open>h\<^sub>p'=fmap_update h\<^sub>p (s x) (hOcc (e s))\<close> \<open>h\<^sub>p'\<bottom>h\<^sub>f\<close> assms(3) compat_heap_union_comm by (metis fmap_add_upd fmap_le_refl)
    moreover have "h\<^sub>p'++\<^sub>fh\<^sub>f\<subseteq>\<^sub>fh'" using RedWrite.hyps(4) \<open>h\<^sub>p'=fmap_update h\<^sub>p (s x) (hOcc (e s))\<close> \<open>h\<^sub>p'\<bottom>h\<^sub>f\<close> assms(3) compat_heap_union_comm by (metis fmap_add_upd fmap_le_refl)
    ultimately show ?thesis by (meson Rep_fmap_inject fmap_le_simp map_le_antisym)
  qed
  then show ?case using \<open>h\<^sub>p'\<bottom>h\<^sub>f\<close> using red.RedWrite RedWrite.hyps(1-4) \<open>h\<^sub>p'=fmap_update h\<^sub>p(s x) (hOcc (e s))\<close> assms(3) map_add_SomeD by (smt (z3) \<open>s x \<noteq> 0 \<and> fmap_lookup h\<^sub>p (s x) \<noteq> hUkw \<and> fmap_lookup h\<^sub>p (s x) \<noteq> hFrd\<close> compat_heap_union_comm fmap_add_simp fmap_lookup_simp)
next
  case (RedRead x n y)
  then show ?case using red.RedRead RedRefl compat_heap_union_comm by (smt (z3) aborts.intros(2) fmap_add_simp fmap_lookup_simp map_add_Some_iff red.RedRead rtred.simps)
next
  case (RedFree x n)
  then have "s x\<noteq>0 \<and> fmap_lookup h\<^sub>p (s x)\<noteq>hUkw \<and> fmap_lookup h\<^sub>p (s x)\<noteq>hFrd" by (meson RedRefl aborts.intros(3))
  moreover obtain h\<^sub>p' where "h\<^sub>p'=fmap_update h\<^sub>p (s x) hFrd" by fast
  ultimately have "h\<^sub>p'\<bottom>h\<^sub>f" using assms(3) unfolding compat_def by auto
  have "h'=h\<^sub>p'++\<^sub>fh\<^sub>f" 
  proof(-)
    have "h'\<subseteq>\<^sub>fh\<^sub>p'++\<^sub>fh\<^sub>f" using RedFree.hyps(1) RedFree.hyps(4) \<open>h\<^sub>p'\<bottom>h\<^sub>f\<close> \<open>h\<^sub>p'=fmap_update h\<^sub>p (s x) hFrd\<close> assms(3) compat_heap_union_comm by (metis fmap_add_upd fmap_le_refl)
    moreover have "h\<^sub>p'++\<^sub>fh\<^sub>f\<subseteq>\<^sub>fh'" using RedFree.hyps(1) RedFree.hyps(4) \<open>h\<^sub>p'\<bottom>h\<^sub>f\<close> \<open>h\<^sub>p'=fmap_update h\<^sub>p (s x) hFrd\<close> assms(3) compat_heap_union_comm by (metis fmap_add_upd fmap_le_refl)
    ultimately show ?thesis by (meson Rep_fmap_inject fmap_le_simp map_le_antisym)
  qed
  then show ?case using \<open>h\<^sub>p'\<bottom>h\<^sub>f\<close> using red.RedFree RedFree.hyps(1-4) \<open>h\<^sub>p'=fmap_update h\<^sub>p(s x) hFrd\<close> assms(3) map_add_SomeD by (smt (z3) \<open>s x \<noteq> 0 \<and> fmap_lookup h\<^sub>p (s x) \<noteq> hUkw \<and> fmap_lookup h\<^sub>p (s x) \<noteq> hFrd\<close> compat_heap_union_comm fmap_add_simp fmap_lookup_simp)
qed

(*We can strenghten it to say that h\<^sub>f' is h\<^sub>f with some \<bottom> replaced with \<emptyset>*)
lemma heap_extension_independence_rt:
  assumes "(C, (s, h\<^sub>p++\<^sub>fh\<^sub>f)) \<rightarrow>* (C'', (s'', h''))"
      and "\<And>c'. (C, (s, h\<^sub>p)) \<rightarrow>* c' \<Longrightarrow> \<not> aborts c'"
      and "h\<^sub>p\<bottom>h\<^sub>f"
    shows "\<exists>h\<^sub>p'' h\<^sub>f''. h''=h\<^sub>p''++\<^sub>fh\<^sub>f'' \<and> h\<^sub>p''\<bottom>h\<^sub>f'' \<and> ((C, (s, h\<^sub>p)) \<rightarrow>* (C'', (s'', h\<^sub>p'')))"
  using assms
proof(induction "(C, (s, h\<^sub>p++\<^sub>fh\<^sub>f))" "(C'', (s'', h''))" arbitrary: C s h\<^sub>p h\<^sub>f rule: rtred.induct)
  case RedRefl
  then show ?case using rtred.RedRefl by blast
next
  case (RedTran c')
  then obtain C' s' h\<^sub>p' h\<^sub>f' where "c' = (C', (s', h\<^sub>p'++\<^sub>fh\<^sub>f'))" "h\<^sub>p'\<bottom>h\<^sub>f'" "(C, (s, h\<^sub>p)) \<rightarrow> (C', (s', h\<^sub>p'))" using heap_extension_independence[of C s h\<^sub>p h\<^sub>f] by (metis prod_cases3)
  then show ?case by (meson RedTran.hyps(3) RedTran.prems(1) rtred.RedTran) 
qed

definition provides where
 "provides p C \<equiv> \<forall>\<sigma>\<in>p. \<forall>c'. ((C, \<sigma>)\<rightarrow>*c') \<longrightarrow> \<not>(aborts c')"

lemma provides_seq_left:
  assumes "provides p (Seq C\<^sub>1 C\<^sub>2)"
  shows "provides p C\<^sub>1"
  using assms unfolding provides_def by (metis aborts.intros(4) seq_trcl surj_pair)

lemma provides_if_left:
  assumes "provides p (If C\<^sub>1 C\<^sub>2)"
  shows "provides p C\<^sub>1"
  using assms unfolding provides_def by (meson RedIf1 RedTran)

lemma provides_if_right:
  assumes "provides p (If C\<^sub>1 C\<^sub>2)"
  shows "provides p C\<^sub>2"
  using assms unfolding provides_def by (meson RedIf2 RedTran)

lemma provides_while:
  assumes "provides p (While C)"
  shows "provides p C"
proof(rule ccontr)
  assume "\<not> provides p C"
  then obtain \<sigma> C' \<sigma>' where "\<sigma>\<in>p" "(C, \<sigma>)\<rightarrow>*(C', \<sigma>')" "aborts (C', \<sigma>')" unfolding provides_def by fast
  then have "aborts (Seq C' (While C), \<sigma>')" by (simp add: aborts.intros(4))
  moreover have "(While C, \<sigma>)\<rightarrow>*(Seq C (While C), \<sigma>)" by (metis RedRefl iter.simps(2) every_iter_is_while_seq)
  then have "(While C, \<sigma>)\<rightarrow>*(Seq C' (While C), \<sigma>')" using \<open>(C, \<sigma>)\<rightarrow>*(C', \<sigma>')\<close> rtred_trans seq_trcl by blast
  ultimately show "False" using \<open>\<sigma>\<in>p\<close> assms unfolding provides_def by blast
qed
  
lemma provides_while_general:
  assumes "provides p (While C)"
  shows "provides p (iter C n)"
proof(rule ccontr)
  assume "\<not> provides p (iter C n)"
  then obtain \<sigma> C' \<sigma>' where "\<sigma>\<in>p" "(iter C n, \<sigma>)\<rightarrow>*(C', \<sigma>')" "aborts (C', \<sigma>')" unfolding provides_def by fast
  then obtain m where m_def: "m\<le>n" "C' = iter C m \<or> (\<exists>C\<^sub>0'. C' = Seq C\<^sub>0' (iter C m))" using iter_rhs_rt by blast
  show "False"
  proof(cases "C' = iter C m")
    case True
    then show ?thesis using aborts.intros(4) aborts_in_head assms every_iter_is_while_seq every_iter_is_while_skip iter_rhs_rt' provides_def by (metis \<open>\<And>thesis. (\<And>\<sigma> C' \<sigma>'. \<lbrakk>\<sigma> \<in> p; (iter C n, \<sigma>) \<rightarrow>* (C', \<sigma>'); aborts (C', \<sigma>')\<rbrakk> \<Longrightarrow> thesis) \<Longrightarrow> thesis\<close>)
  next
    case False
    then obtain C\<^sub>0' where "C' = Seq C\<^sub>0' (iter C m)" using m_def(2) by fast
    then have "aborts (C\<^sub>0', \<sigma>')" using \<open>aborts (C', \<sigma>')\<close> aborts_in_head by blast
    then have "aborts (Seq C\<^sub>0' (While C), \<sigma>')" by (simp add: aborts.intros(4))
    moreover have "(While C, \<sigma>) \<rightarrow>* (Seq C\<^sub>0' (While C), \<sigma>')" using \<open>(iter C n, \<sigma>)\<rightarrow>*(C', \<sigma>')\<close> \<open>C'=Seq C\<^sub>0' (iter C m)\<close> by (simp add: every_iter_is_while_seq)
    ultimately show "False" using \<open>\<sigma>\<in>p\<close> assms unfolding provides_def by blast
  qed
qed

end