theory Star
  imports StrongestPostcondition
begin

type_synonym var = nat
type_synonym nstate = "var state"
type_synonym assertion = "nstate set"
type_synonym rassertion = "(nstate\<times>nstate) set"
type_synonym hassertion = "assertion set"

definition star :: "assertion \<Rightarrow> assertion \<Rightarrow> assertion" (infix "\<^emph>" 51) where
 "p\<^emph>q \<equiv> {(s, h\<^sub>p++\<^sub>fh\<^sub>q)| s h\<^sub>p h\<^sub>q. (s, h\<^sub>p)\<in>p \<and> (s, h\<^sub>q)\<in>q \<and> h\<^sub>p\<bottom>h\<^sub>q}"

definition hAdd :: "nstate \<Rightarrow> nstate \<Rightarrow> nstate \<Rightarrow> bool" where
 "hAdd \<sigma> \<sigma>\<^sub>1 \<sigma>\<^sub>2 \<longleftrightarrow> fst \<sigma>=fst \<sigma>\<^sub>1 \<and> fst \<sigma>=fst \<sigma>\<^sub>2 \<and> snd \<sigma>=snd \<sigma>\<^sub>1++\<^sub>fsnd \<sigma>\<^sub>2 \<and> snd \<sigma>\<^sub>1\<bottom>snd \<sigma>\<^sub>2"

lemma hAdd_prop:
  "hAdd (s, h) (s, h\<^sub>1) (s, h\<^sub>2) \<longleftrightarrow> h=h\<^sub>1++\<^sub>fh\<^sub>2 \<and> h\<^sub>1\<bottom>h\<^sub>2"
  unfolding hAdd_def by auto

lemma hAdd_is_function:
  "hAdd \<sigma> \<sigma>\<^sub>1 \<sigma>\<^sub>2 \<Longrightarrow> hAdd \<sigma>' \<sigma>\<^sub>1 \<sigma>\<^sub>2 \<Longrightarrow> \<sigma>=\<sigma>'"
  unfolding hAdd_def by (simp add: prod.expand)

lemma hAdd_comm:
  "hAdd \<sigma> \<sigma>\<^sub>1 \<sigma>\<^sub>2 \<longleftrightarrow> hAdd \<sigma> \<sigma>\<^sub>2 \<sigma>\<^sub>1"
  unfolding hAdd_def using compat_comm compat_heap_union_comm by metis

lemma hAdd_assoc:
  "hAdd \<sigma>\<^sub>1\<^sub>2\<^sub>3 \<sigma>\<^sub>1\<^sub>2 \<sigma>\<^sub>3 \<Longrightarrow> hAdd \<sigma>\<^sub>1\<^sub>2 \<sigma>\<^sub>1 \<sigma>\<^sub>2 \<Longrightarrow> \<exists>\<sigma>\<^sub>2\<^sub>3. hAdd \<sigma>\<^sub>1\<^sub>2\<^sub>3 \<sigma>\<^sub>1 \<sigma>\<^sub>2\<^sub>3 \<and> hAdd \<sigma>\<^sub>2\<^sub>3 \<sigma>\<^sub>2 \<sigma>\<^sub>3"
proof(-)
  assume asm: "hAdd \<sigma>\<^sub>1\<^sub>2 \<sigma>\<^sub>1 \<sigma>\<^sub>2" "hAdd \<sigma>\<^sub>1\<^sub>2\<^sub>3 \<sigma>\<^sub>1\<^sub>2 \<sigma>\<^sub>3"
  have "snd \<sigma>\<^sub>1 ++\<^sub>f snd \<sigma>\<^sub>2 = snd \<sigma>\<^sub>1\<^sub>2" using asm(1) hAdd_def by auto
  then have "snd \<sigma>\<^sub>1 \<bottom> snd \<sigma>\<^sub>3" using asm(2) by (metis compat_distr hAdd_def)
  moreover have "fst \<sigma>\<^sub>1\<^sub>2\<^sub>3 = fst \<sigma>\<^sub>1\<^sub>2" using asm(2) hAdd_def by blast 
  moreover have "hAdd (fst \<sigma>\<^sub>1\<^sub>2, snd \<sigma>\<^sub>2 ++\<^sub>f snd \<sigma>\<^sub>3) \<sigma>\<^sub>2 \<sigma>\<^sub>3" using asm unfolding hAdd_def by (metis compat_distr fst_conv sndI)
  ultimately show ?thesis using asm by (simp add: compat_union hAdd_def)
qed

lemma hAdd_assoc':
  "hAdd \<sigma>\<^sub>1\<^sub>2\<^sub>3 \<sigma>\<^sub>1 \<sigma>\<^sub>2\<^sub>3 \<and> hAdd \<sigma>\<^sub>2\<^sub>3 \<sigma>\<^sub>2 \<sigma>\<^sub>3 \<Longrightarrow> \<exists>\<sigma>\<^sub>1\<^sub>2. hAdd \<sigma>\<^sub>1\<^sub>2\<^sub>3 \<sigma>\<^sub>1\<^sub>2 \<sigma>\<^sub>3 \<and> hAdd \<sigma>\<^sub>1\<^sub>2 \<sigma>\<^sub>1 \<sigma>\<^sub>2"
  by (meson hAdd_assoc hAdd_comm)

lemma hAdd_lookup_disjoint:
  assumes "hAdd (s, h) (s\<^sub>1, h\<^sub>1) (s\<^sub>2, h\<^sub>2)"
    shows "(fmap_lookup h l=fmap_lookup h\<^sub>1 l \<and> fmap_lookup h\<^sub>2 l=None) \<or> (fmap_lookup h l=fmap_lookup h\<^sub>2 l \<and> fmap_lookup h\<^sub>1 l=None)"
proof(-)
  have "h=h\<^sub>1++\<^sub>fh\<^sub>2" "h\<^sub>1\<bottom>h\<^sub>2" using assms unfolding hAdd_def by auto
  then show ?thesis unfolding compat_def by (metis \<open>h\<^sub>1 \<bottom> h\<^sub>2\<close> compat_heap_union_comm disjnt_def disjnt_iff empty_fmap_add fmap_add_dom_app_simps(3) fmap_empty_simp fmap_lookup_simp)
qed

lemma star_via_hAdd:
  "p\<^emph>q = {\<sigma>| \<sigma> \<sigma>\<^sub>p \<sigma>\<^sub>q. \<sigma>\<^sub>p\<in>p \<and> \<sigma>\<^sub>q\<in>q \<and> hAdd \<sigma> \<sigma>\<^sub>p \<sigma>\<^sub>q}" unfolding star_def hAdd_def by fastforce

lemma star_mono:
  assumes "p\<subseteq>p'"
      and "q\<subseteq>q'"
  shows "p\<^emph>q\<subseteq>p'\<^emph>q'"
  using assms unfolding star_def by blast

lemma provides_mono:
  assumes "provides p C"
    shows "provides (p\<^emph>f) C"
proof(rule ccontr)
  assume "\<not> provides (p\<^emph>f) C"
  then obtain s h\<^sub>p h\<^sub>f c' where "(s, h\<^sub>p)\<in>p" "(s, h\<^sub>f)\<in>f" "h\<^sub>p\<bottom>h\<^sub>f" "(C, (s, h\<^sub>p++\<^sub>fh\<^sub>f))\<rightarrow>*c'" "aborts c'" unfolding provides_def star_def by blast
  moreover obtain C' s' h' where "c'=(C', (s', h'))" using prod_cases3 by blast
  ultimately have "(C, (s, h\<^sub>p++\<^sub>fh\<^sub>f))\<rightarrow>*(C', (s', h'))" by blast
  moreover have "\<And>c'. ((C, (s, h\<^sub>p))\<rightarrow>*c') \<Longrightarrow> \<not> aborts c'" using assms(1)\<open>(s, h\<^sub>p)\<in>p\<close> unfolding provides_def by blast
  moreover note \<open>h\<^sub>p\<bottom>h\<^sub>f\<close>
  ultimately obtain h\<^sub>p' h\<^sub>f' where "h'=h\<^sub>p'++\<^sub>fh\<^sub>f'" "h\<^sub>p'\<bottom>h\<^sub>f'" "(C, (s, h\<^sub>p)) \<rightarrow>* (C', (s', h\<^sub>p'))" using heap_extension_independence_rt by blast
  then have "aborts (C', (s', h\<^sub>p'))" using \<open>aborts c'\<close> \<open>c'=(C', (s', h'))\<close> aborts_mono by auto
  then show "False" by (simp add: \<open>(C, (s, h\<^sub>p)) \<rightarrow>* (C', (s', h\<^sub>p'))\<close> \<open>\<And>c'. (C, (s, h\<^sub>p)) \<rightarrow>* c' \<Longrightarrow> \<not> aborts c'\<close>)
qed

lemma sp_frame:
  assumes "\<And>s h x n. (s, h)\<in>f \<Longrightarrow> x\<in>md C \<Longrightarrow> (s(x:=n), h)\<in>f"
      and "provides p C"
      and "\<And>s h l. (s, h)\<in>f \<Longrightarrow> fmap_lookup h l=hFrd \<Longrightarrow> (s, fmap_update h l hUkw)\<in>f" (*no syntactic \<bottom> in f*)
    shows "sp (p\<^emph>f) C = (sp p C)\<^emph>f"
  using assms
proof(induction C arbitrary: p)
  case (Assign x e)
  have "{(s(x:=e s), h)| s h. (s, h)\<in>p}\<^emph>f \<subseteq> {(s(x:=e s), h)| s h. (s, h)\<in>p\<^emph>f}"
  proof
    fix \<sigma>'
    assume asm: "\<sigma>'\<in>{(s(x:=e s), h)| s h. (s, h)\<in>p}\<^emph>f"
    then obtain s h\<^sub>p h\<^sub>f where sh\<^sub>ph\<^sub>f_def: "(s, h\<^sub>p)\<in>p" "(s(x:=e s), h\<^sub>f)\<in>f" "h\<^sub>p\<bottom>h\<^sub>f" "\<sigma>'=(s(x:=e s), h\<^sub>p++\<^sub>fh\<^sub>f)" unfolding star_def by blast
    then have "(s, h\<^sub>f)\<in>f" using Assign.prems(1) by (metis fun_upd_triv fun_upd_upd insertI1 md.simps(1))
    then have "(s, h\<^sub>p++\<^sub>fh\<^sub>f)\<in>p\<^emph>f" using sh\<^sub>ph\<^sub>f_def(1,3) unfolding star_def by blast
    then show "\<sigma>'\<in>{(s(x:=e s), h)| s h. (s, h)\<in>p\<^emph>f}" using sh\<^sub>ph\<^sub>f_def(4) by fast
  qed
  moreover have "{(s(x:=e s), h)| s h. (s, h)\<in>p\<^emph>f} \<subseteq> {(s(x:=e s), h)| s h. (s, h)\<in>p}\<^emph>f"
  proof
    fix \<sigma>'
    assume asm: "\<sigma>'\<in>{(s(x:=e s), h)| s h. (s, h)\<in>p\<^emph>f}"
    then obtain s h\<^sub>p h\<^sub>f where sh\<^sub>ph\<^sub>f_def: "(s, h\<^sub>p)\<in>p" "(s, h\<^sub>f)\<in>f" "h\<^sub>p\<bottom>h\<^sub>f" "\<sigma>'=(s(x:=e s), h\<^sub>p++\<^sub>fh\<^sub>f)" unfolding star_def by blast
    then have "(s(x:=e s), h\<^sub>f)\<in>f" using Assign.prems(1) by fastforce
    moreover have "(s(x:=e s), h\<^sub>p)\<in>{(s(x:=e s), h)| s h. (s, h)\<in>p}" using sh\<^sub>ph\<^sub>f_def(1) by fast
    ultimately show "\<sigma>'\<in>{(s(x:=e s), h)| s h. (s, h)\<in>p}\<^emph>f" using sh\<^sub>ph\<^sub>f_def(3,4) unfolding star_def by blast
  qed
  ultimately show ?case unfolding subset_antisym sp_assign by fastforce
next
  case (Seq C\<^sub>1 C\<^sub>2)
  moreover have "provides p C\<^sub>1" "provides (sp p C\<^sub>1) C\<^sub>2" using calculation(4) provides_seq_left provides_seq_right by auto
  ultimately show ?case by (simp add: sp_seq)
next
  case (If C\<^sub>1 C\<^sub>2)
  moreover have "provides p C\<^sub>1" "provides p C\<^sub>2" using calculation(4) provides_if_left provides_if_right by auto
  ultimately show ?case unfolding sp_if star_def by auto
next
  case Skip
  then show ?case unfolding sp_skip by meson
next
  case (Havoc x)
  have "{(s(x:=n), h)| s h n. (s, h)\<in>p}\<^emph>f \<subseteq> {(s(x:=n), h)| s h n. (s, h)\<in>p\<^emph>f}"
  proof
    fix \<sigma>'
    assume asm: "\<sigma>'\<in>{(s(x:=n), h)| s h n. (s, h)\<in>p}\<^emph>f"
    then obtain s h\<^sub>p h\<^sub>f n where sh\<^sub>ph\<^sub>f_def: "(s, h\<^sub>p)\<in>p" "(s(x:=n), h\<^sub>f)\<in>f" "h\<^sub>p\<bottom>h\<^sub>f" "\<sigma>'=(s(x:=n), h\<^sub>p++\<^sub>fh\<^sub>f)" unfolding star_def by blast
    moreover obtain m where m_def: "s(x:=n, x:=m)=s" by force
    ultimately have "(s(x:=n, x:=m), h\<^sub>f)\<in>f" using Havoc.prems(1)[of "s(x:=n)" h\<^sub>f x] by fastforce
    then have "(s, h\<^sub>f)\<in>f" using m_def by fastforce
    then have "(s, h\<^sub>p++\<^sub>fh\<^sub>f)\<in>p\<^emph>f" using sh\<^sub>ph\<^sub>f_def(1,3) unfolding star_def by blast
    then show "\<sigma>'\<in>{(s(x:=n), h)| s h n. (s, h)\<in>p\<^emph>f}" using sh\<^sub>ph\<^sub>f_def(4) by fast
  qed
  moreover have "{(s(x:=n), h)| s h n. (s, h)\<in>p\<^emph>f} \<subseteq> {(s(x:=n), h)| s h n. (s, h)\<in>p}\<^emph>f"
  proof
    fix \<sigma>'
    assume asm: "\<sigma>'\<in>{(s(x:=n), h)| s h n. (s, h)\<in>p\<^emph>f}"
    then obtain s h\<^sub>p h\<^sub>f n where sh\<^sub>ph\<^sub>f_def: "(s, h\<^sub>p)\<in>p" "(s, h\<^sub>f)\<in>f" "h\<^sub>p\<bottom>h\<^sub>f" "\<sigma>'=(s(x:=n), h\<^sub>p++\<^sub>fh\<^sub>f)" unfolding star_def by blast
    then have "(s(x:=n), h\<^sub>f)\<in>f" using Havoc.prems(1) by fastforce
    moreover have "(s(x:=n), h\<^sub>p)\<in>{(s(x:=n), h)| s h n. (s, h)\<in>p}" using sh\<^sub>ph\<^sub>f_def(1) by fast
    ultimately show "\<sigma>'\<in>{(s(x:=n), h)| s h n. (s, h)\<in>p}\<^emph>f" using sh\<^sub>ph\<^sub>f_def(3,4) unfolding star_def by blast
  qed
  ultimately show ?case unfolding subset_antisym sp_havoc by blast
next
  case (Assume b)
  then show ?case unfolding sp_assume star_def by blast
next
  case (While C)
  have "\<And>n. sp p (iter C n) = sp p (iter' C n)" unfolding sp_def iter_iter' by blast
  have "\<And>n. sp (p \<^emph> f) (iter' C n) = sp p (iter' C n) \<^emph> f"
  proof(-)
    fix n
    show "sp (p \<^emph> f) (iter' C n) = sp p (iter' C n) \<^emph> f"
    proof(induction n)
      case 0
      then show ?case by (simp add: sp_skip)
    next
      case (Suc n)
      moreover have "\<And>p. provides p C \<Longrightarrow> sp (p \<^emph> f) C = sp p C \<^emph> f" using While.IH While.prems(1,3) by auto 
      moreover have "\<And>n. provides (sp p (iter C n)) C" by (simp add: While.prems(2) provides_while_specific)
      ultimately have "\<And>n. sp ((sp p (iter C n)) \<^emph> f) C = sp (sp p (iter C n)) C \<^emph> f" by simp
      then show ?case unfolding iter'.simps(2) sp_seq by (simp add: Suc \<open>\<And>n. sp p (iter C n) = sp p (iter' C n)\<close>)
    qed
  qed
  then have "\<And>n. {s'. \<exists>s\<in>p \<^emph> f. (s, s') \<in> \<lbrakk>iter C n\<rbrakk>} = {s'. \<exists>s\<in>p. (s, s') \<in> \<lbrakk>iter C n\<rbrakk>} \<^emph> f" unfolding sp_def by (simp add: iter_iter')
  then have "{s'. \<exists>s\<in>p \<^emph> f. (s, s') \<in> (\<Union>n. \<lbrakk>iter C n\<rbrakk>)} = {s'. \<exists>s\<in>p. (s, s') \<in> (\<Union>n. \<lbrakk>iter C n\<rbrakk>)} \<^emph> f" unfolding star_def by fast
  then show ?case by (simp add: dsem_properties(7) sp_def)
next
  case (Alloc x)
  have "{(s(x:=l), fmap_update h l (hOcc n))| s h l n. (s, h)\<in>p \<and> (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0}\<^emph>f \<subseteq> {(s(x:=l), fmap_update h l (hOcc n))| s h l n. (s, h)\<in>p\<^emph>f \<and> (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0}"
  proof
    fix \<sigma>'
    assume asm: "\<sigma>'\<in>{(s(x:=l), fmap_update h l (hOcc n))| s h l n. (s, h)\<in>p \<and> (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0}\<^emph>f"
    then obtain s' h\<^sub>p' h\<^sub>f where s'h\<^sub>p'h\<^sub>f_def: "(s', h\<^sub>p')\<in>{(s(x:=l), fmap_update h l (hOcc n))| s h l n. (s, h)\<in>p \<and> (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0}" "(s', h\<^sub>f)\<in>f" "h\<^sub>p'\<bottom>h\<^sub>f" "\<sigma>'=(s', h\<^sub>p'++\<^sub>fh\<^sub>f)" unfolding star_def by blast
    then obtain s h\<^sub>p l n where sh\<^sub>ph\<^sub>f_def: "(s, h\<^sub>p)\<in>p" "fmap_lookup h\<^sub>p l=hUkw \<or> fmap_lookup h\<^sub>p l=hFrd" "l\<noteq>0" "s'=s(x:=l)" "h\<^sub>p'=fmap_update h\<^sub>p l (hOcc n)" by blast
    moreover have "(s, h\<^sub>f)\<in>f" using \<open>(s', h\<^sub>f)\<in>f\<close> \<open>s'=s(x:=l)\<close> Alloc.prems(1) by (metis fun_upd_triv fun_upd_upd insertI1 md.simps(8))
    ultimately have "(s, h\<^sub>p++\<^sub>fh\<^sub>f)\<in>p\<^emph>f" unfolding star_def using \<open>h\<^sub>p'\<bottom>h\<^sub>f\<close> compat_distr map_add_upd map_le_refl mem_Collect_eq by (smt (verit) Rep_fmap_inject fmap_add_simp fmap_update_simp fun_upd_triv fun_upd_upd map_add_assoc map_add_empty map_add_subsumed1 map_le_empty map_le_upd)
    moreover have "fmap_lookup (h\<^sub>p++\<^sub>fh\<^sub>f) l=hUkw \<or> fmap_lookup (h\<^sub>p++\<^sub>fh\<^sub>f) l=hFrd" by (smt (z3) compat_comm compat_distr compat_heap_added_equal compat_heap_union_comm fmap_upd_triv fmap_add_None fmap_add_SomeD fmap_add_subsumed1 fmap_add_upd fmap_le_refl fmap_upd_Some_unfold s'h\<^sub>p'h\<^sub>f_def(3) sh\<^sub>ph\<^sub>f_def(2,5))
    moreover have "(fmap_update h\<^sub>p l (hOcc n))++\<^sub>fh\<^sub>f = fmap_update (h\<^sub>p++\<^sub>fh\<^sub>f) l (hOcc n)" by (metis compat_distr compat_heap_union_comm fmap_add_subsumed1 fmap_add_upd fmap_le_refl s'h\<^sub>p'h\<^sub>f_def(3) sh\<^sub>ph\<^sub>f_def(5))
    ultimately show "\<sigma>'\<in>{(s(x:=l), fmap_update h l (hOcc n))| s h l n. (s, h)\<in>p\<^emph>f \<and> (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0}" by (smt (verit, best) mem_Collect_eq s'h\<^sub>p'h\<^sub>f_def(4) sh\<^sub>ph\<^sub>f_def(3-5))
  qed
  moreover have "{(s(x:=l), fmap_update h l (hOcc n))| s h l n. (s, h)\<in>p\<^emph>f \<and> (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0} \<subseteq> {(s(x:=l), fmap_update h l (hOcc n))| s h l n. (s, h)\<in>p \<and> (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0}\<^emph>f"
  proof
    fix \<sigma>'
    assume asm: "\<sigma>'\<in>{(s(x:=l), fmap_update h l (hOcc n))| s h l n. (s, h)\<in>p\<^emph>f \<and> (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0}"
    then obtain s h\<^sub>p h\<^sub>f l n where sh\<^sub>ph\<^sub>fln_def: "(s, h\<^sub>p)\<in>p" "(s, h\<^sub>f)\<in>f" "h\<^sub>p\<bottom>h\<^sub>f" "fmap_lookup (h\<^sub>p++\<^sub>fh\<^sub>f) l=hUkw \<or> fmap_lookup (h\<^sub>p++\<^sub>fh\<^sub>f) l=hFrd" "l\<noteq>0" "\<sigma>'=(s(x:=l), fmap_update (h\<^sub>p++\<^sub>fh\<^sub>f) l (hOcc n))" unfolding star_def by force
    then have "(s(x:=l), h\<^sub>f)\<in>f" using Alloc.prems(1)[of s h\<^sub>f x] by fastforce
    moreover have 0: "(s(x:=l), fmap_update h\<^sub>p l (hOcc n))\<in>{(s(x:=l), fmap_update h l (hOcc n))| s h l n. (s, h)\<in>p \<and> (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0}" using compat_heap_union_comm sh\<^sub>ph\<^sub>fln_def(1) sh\<^sub>ph\<^sub>fln_def(3-5) by (smt (z3) fmap_add_None fmap_add_Some_iff mem_Collect_eq)
    show "\<sigma>'\<in>{(s(x:=l), fmap_update h l (hOcc n))| s h l n. (s, h)\<in>p \<and> (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0}\<^emph>f"
    proof(cases "fmap_lookup h\<^sub>f l=hUkw")
      case True
      then have "fmap_update (h\<^sub>p++\<^sub>fh\<^sub>f) l (hOcc n) = (fmap_update h\<^sub>p l (hOcc n))++\<^sub>fh\<^sub>f" by (metis compat_heap_union_comm fmap_upd_triv fmap_add_upd fmap_le_iff_fmap_add_commute fmap_le_imp_upd_le sh\<^sub>ph\<^sub>fln_def(3))
      moreover have "(fmap_update h\<^sub>p l (hOcc n))\<bottom>h\<^sub>f" using sh\<^sub>ph\<^sub>fln_def(3) True unfolding compat_def by auto
      ultimately show "\<sigma>'\<in>{(s(x:=l), fmap_update h l (hOcc n))| s h l n. (s, h)\<in>p \<and> (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0}\<^emph>f" using "0" \<open>(s(x:=l), h\<^sub>f)\<in>f\<close> unfolding star_def sh\<^sub>ph\<^sub>fln_def(6) by blast
    next
      case False
      then have "fmap_lookup h\<^sub>f l=hFrd" using sh\<^sub>ph\<^sub>fln_def(4) by force
      moreover obtain h\<^sub>f' where "h\<^sub>f'=fmap_update h\<^sub>f l hUkw" by simp
      ultimately have "(s(x:=l), h\<^sub>f')\<in>f" using \<open>(s(x:=l), h\<^sub>f)\<in>f\<close> using Alloc.prems(3) by blast
      moreover have "fmap_lookup (h\<^sub>p++\<^sub>fh\<^sub>f') l=hUkw \<or> fmap_lookup (h\<^sub>p++\<^sub>fh\<^sub>f') l=hFrd" by (metis \<open>h\<^sub>f' = fmap_update h\<^sub>f l hUkw\<close> compat_heap_union_comm fmap_add_None fmap_add_dom_app_simps(1,3) fmap_lookup_simp fmap_update_simp fun_upd_same sh\<^sub>ph\<^sub>fln_def(3,4))
      then have "fmap_update (h\<^sub>p++\<^sub>fh\<^sub>f') l (hOcc n) = (fmap_update h\<^sub>p l (hOcc n))++\<^sub>fh\<^sub>f'" by (simp add: \<open>h\<^sub>f'=fmap_update h\<^sub>f l hUkw\<close> fmap_add_upd_left)
      moreover have "(fmap_update h\<^sub>p l (hOcc n))\<bottom>h\<^sub>f'" using sh\<^sub>ph\<^sub>fln_def(3) \<open>h\<^sub>f'=fmap_update h\<^sub>f l hUkw\<close> unfolding compat_def by auto
      moreover have "fmap_update (h\<^sub>p++\<^sub>fh\<^sub>f) l (hOcc n) = fmap_update (h\<^sub>p++\<^sub>fh\<^sub>f') l (hOcc n)" using \<open>h\<^sub>f'=fmap_update h\<^sub>f l hUkw\<close> fun_upd_upd fmap_add_upd by (metis Rep_fmap_inject fmap_update_simp)
      ultimately show "\<sigma>'\<in>{(s(x:=l), fmap_update h l (hOcc n))| s h l n. (s, h)\<in>p \<and> (fmap_lookup h l=hUkw \<or> fmap_lookup h l=hFrd) \<and> l\<noteq>0}\<^emph>f" using "0" unfolding star_def sh\<^sub>ph\<^sub>fln_def(6) by fastforce
    qed
  qed
  ultimately show ?case unfolding sp_alloc by blast
next
  case (Write x e)
  have "{(s, fmap_update h (s x) (hOcc (e s)))| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}\<^emph>f \<subseteq> {(s, fmap_update h (s x) (hOcc (e s)))| s h n. (s, h)\<in>p\<^emph>f \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}"
  proof
    fix \<sigma>'
    assume asm: "\<sigma>'\<in>{(s, fmap_update h (s x) (hOcc (e s)))| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}\<^emph>f"
    then obtain s h\<^sub>p' h\<^sub>f where sh\<^sub>p'h\<^sub>f_def: "(s, h\<^sub>p')\<in>{(s, fmap_update h (s x) (hOcc (e s)))| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}" "(s, h\<^sub>f)\<in>f" "h\<^sub>p'\<bottom>h\<^sub>f" "\<sigma>'=(s, h\<^sub>p'++\<^sub>fh\<^sub>f)" unfolding star_def by blast
    then obtain h\<^sub>p n where sh\<^sub>pn_def: "(s, h\<^sub>p)\<in>p" "fmap_lookup h\<^sub>p (s x)=hOcc n" "s x\<noteq>0" "h\<^sub>p'=fmap_update h\<^sub>p (s x) (hOcc (e s))" by blast
    then have "(s, h\<^sub>p++\<^sub>fh\<^sub>f)\<in>p\<^emph>f" unfolding star_def using sh\<^sub>p'h\<^sub>f_def(2,3) unfolding compat_def by fastforce
    moreover have "fmap_lookup (h\<^sub>p++\<^sub>fh\<^sub>f) (s x)=hOcc n" using compat_heap_union_comm fmap_add_dom_app_simps(3) sh\<^sub>p'h\<^sub>f_def(3) sh\<^sub>pn_def(2,4) unfolding compat_def by fastforce
    ultimately show "\<sigma>'\<in>{(s, fmap_update h (s x) (hOcc (e s)))| s h n. (s, h)\<in>p\<^emph>f \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}" by (smt (verit, del_insts) compat_distr compat_heap_union_comm fmap_add_empty fmap_add_upd mem_Collect_eq sh\<^sub>p'h\<^sub>f_def(3,4) sh\<^sub>pn_def(3,4))
  qed
  moreover have "{(s, fmap_update h (s x) (hOcc (e s)))| s h n. (s, h)\<in>p\<^emph>f \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0} \<subseteq> {(s, fmap_update h (s x) (hOcc (e s)))| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}\<^emph>f"
  proof
    fix \<sigma>'
    assume asm: "\<sigma>'\<in>{(s, fmap_update h (s x) (hOcc (e s)))| s h n. (s, h)\<in>p\<^emph>f \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}"
    then obtain s h\<^sub>p h\<^sub>f n where "(s, h\<^sub>p)\<in>p" "(s, h\<^sub>f)\<in>f" "h\<^sub>p\<bottom>h\<^sub>f" "fmap_lookup (h\<^sub>p++\<^sub>fh\<^sub>f) (s x)=hOcc n" "s x\<noteq>0" "\<sigma>'=(s, fmap_update (h\<^sub>p++\<^sub>fh\<^sub>f) (s x) (hOcc (e s)))" unfolding star_def by auto
    then have "fmap_lookup h\<^sub>p (s x)=hOcc n" using Write.prems(2) unfolding provides_def by (metis RedRefl aborts.intros(1) compat_heap_union_comm fmap_add_SomeD)
    then have "(s, fmap_update h\<^sub>p (s x) (hOcc (e s)))\<in>{(s, fmap_update h (s x) (hOcc (e s)))| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}" "fmap_lookup h\<^sub>f (s x)=hUkw" using \<open>(s, h\<^sub>p)\<in>p\<close> \<open>s x\<noteq>0\<close> \<open>h\<^sub>p\<bottom>h\<^sub>f\<close> unfolding compat_def by auto
    moreover have "\<sigma>'=(s, (fmap_update h\<^sub>p (s x) (hOcc (e s)))++\<^sub>fh\<^sub>f)" by (metis \<open>\<sigma>'=(s, fmap_update (h\<^sub>p++\<^sub>fh\<^sub>f) (s x) (hOcc (e s)))\<close> \<open>h\<^sub>p\<bottom>h\<^sub>f\<close> calculation(2) compat_heap_union_comm fmap_upd_triv fmap_add_upd fmap_le_iff_fmap_add_commute fmap_le_imp_upd_le)
    moreover have "fmap_update h\<^sub>p (s x) (hOcc (e s))\<bottom>h\<^sub>f" using \<open>h\<^sub>p\<bottom>h\<^sub>f\<close> \<open>fmap_lookup h\<^sub>p (s x)=hOcc n\<close> unfolding compat_def by auto
    ultimately show "\<sigma>'\<in>{(s, fmap_update h (s x) (hOcc (e s)))| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}\<^emph>f" using \<open>(s, h\<^sub>f)\<in>f\<close> unfolding star_def by blast
  qed
  ultimately show ?case unfolding sp_write by blast
next
  case (Read y x)
  have "{(s(y:=n), h)| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}\<^emph>f \<subseteq> {(s(y:=n), h)| s h n. (s, h)\<in>p\<^emph>f \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}"
  proof
    fix \<sigma>'
    assume asm: "\<sigma>'\<in>{(s(y:=n), h)| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}\<^emph>f"
    then obtain s' h\<^sub>p h\<^sub>f where s'h\<^sub>ph\<^sub>f_def: "(s', h\<^sub>p)\<in>{(s(y:=n), h)| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}" "(s', h\<^sub>f)\<in>f" "h\<^sub>p\<bottom>h\<^sub>f" "\<sigma>'=(s', h\<^sub>p++\<^sub>fh\<^sub>f)" unfolding star_def by blast
    then obtain s n where "(s, h\<^sub>p)\<in>p" "fmap_lookup h\<^sub>p (s x)=hOcc n" "s x\<noteq>0" "s'=s(y:=n)" by blast
    moreover have "(s, h\<^sub>f)\<in>f" using Read.prems(1)[of s' h\<^sub>f y] \<open>(s', h\<^sub>f)\<in>f\<close> \<open>s'=s(y:=n)\<close> by (metis fun_upd_triv fun_upd_upd insertI1 md.simps(10))
    ultimately have "(s, h\<^sub>p++\<^sub>fh\<^sub>f)\<in>p\<^emph>f" unfolding star_def using compat_def s'h\<^sub>ph\<^sub>f_def(3) by blast
    moreover have "fmap_lookup (h\<^sub>p++\<^sub>fh\<^sub>f) (s x)=hOcc n" using compat_heap_union_comm \<open>fmap_lookup h\<^sub>p (s x)=hOcc n\<close> s'h\<^sub>ph\<^sub>f_def(3) by fastforce
    ultimately show "\<sigma>'\<in>{(s(y:=n), h)| s h n. (s, h)\<in>p\<^emph>f \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}" unfolding star_def by (smt (z3) \<open>s x\<noteq>0\<close> \<open>s'=s(y:=n)\<close> mem_Collect_eq s'h\<^sub>ph\<^sub>f_def(4))
  qed
  moreover have "{(s(y:=n), h)| s h n. (s, h)\<in>p\<^emph>f \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0} \<subseteq> {(s(y:=n), h)| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}\<^emph>f"
  proof
    fix \<sigma>'
    assume asm: "\<sigma>'\<in>{(s(y:=n), h)| s h n. (s, h)\<in>p\<^emph>f \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}"
    then obtain s h\<^sub>p h\<^sub>f n where "(s, h\<^sub>p)\<in>p" "(s, h\<^sub>f)\<in>f" "h\<^sub>p\<bottom>h\<^sub>f" "fmap_lookup (h\<^sub>p++\<^sub>fh\<^sub>f)(s x)=hOcc n" "s x\<noteq>0" "\<sigma>'=(s(y:=n), h\<^sub>p++\<^sub>fh\<^sub>f)" unfolding star_def by auto
    then have "fmap_lookup h\<^sub>p (s x)=hOcc n" using Read.prems(2) unfolding provides_def by (metis RedRefl aborts.intros(2) compat_heap_union_comm fmap_add_SomeD)
    then have "(s(y:=n), h\<^sub>p)\<in>{(s(y:=n), h)| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}" using \<open>(s, h\<^sub>p)\<in>p\<close> \<open>s x\<noteq>0\<close> by auto
    moreover have "(s(y:=n), h\<^sub>f)\<in>f" using Read.prems(1)[of s h\<^sub>f y] \<open>(s, h\<^sub>f)\<in>f\<close> by fastforce
    ultimately show "\<sigma>'\<in>{(s(y:=n), h)| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}\<^emph>f" using \<open>\<sigma>'=(s(y:=n), h\<^sub>p++\<^sub>fh\<^sub>f)\<close> \<open>h\<^sub>p\<bottom>h\<^sub>f\<close> unfolding star_def by auto
  qed
  ultimately show ?case unfolding sp_read by blast
next
  case (Free x)
  have "{(s, fmap_update h (s x) hFrd)| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}\<^emph>f \<subseteq> {(s, fmap_update h (s x) hFrd)| s h n. (s, h)\<in>p\<^emph>f \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}"
  proof
    fix \<sigma>'
    assume asm: "\<sigma>'\<in>{(s, fmap_update h (s x) hFrd)| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}\<^emph>f"
    then obtain s h\<^sub>p' h\<^sub>f where sh\<^sub>p'h\<^sub>f_def: "(s, h\<^sub>p')\<in>{(s, fmap_update h (s x) hFrd)| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}" "(s, h\<^sub>f)\<in>f" "h\<^sub>p'\<bottom>h\<^sub>f" "\<sigma>'=(s, h\<^sub>p'++\<^sub>fh\<^sub>f)" unfolding star_def by blast
    then obtain h\<^sub>p n where "(s, h\<^sub>p)\<in>p" "fmap_lookup h\<^sub>p (s x)=hOcc n" "s x\<noteq>0" "h\<^sub>p'=fmap_update h\<^sub>p (s x) hFrd" by blast
    moreover have "h\<^sub>p\<bottom>h\<^sub>f" using \<open>h\<^sub>p'\<bottom>h\<^sub>f\<close> \<open>h\<^sub>p'=fmap_update h\<^sub>p (s x) hFrd\<close> unfolding compat_def by auto
    ultimately have "(s, h\<^sub>p++\<^sub>fh\<^sub>f)\<in>p\<^emph>f" using sh\<^sub>p'h\<^sub>f_def(2) unfolding star_def by blast
    moreover have "fmap_lookup (h\<^sub>p++\<^sub>fh\<^sub>f) (s x)=hOcc n" using compat_heap_union_comm \<open>fmap_lookup h\<^sub>p (s x)=hOcc n\<close> \<open>h\<^sub>p\<bottom>h\<^sub>f\<close> by fastforce
    ultimately show "\<sigma>'\<in>{(s, fmap_update h (s x) hFrd)| s h n. (s, h)\<in>p\<^emph>f \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}" by (smt (verit, ccfv_SIG) \<open>h\<^sub>p \<bottom> h\<^sub>f\<close> \<open>h\<^sub>p'=fmap_update h\<^sub>p (s x) hFrd\<close> \<open>s x \<noteq> 0\<close> compat_heap_union_comm fmap_add_upd mem_Collect_eq sh\<^sub>p'h\<^sub>f_def(3,4))
  qed
  moreover have "{(s, fmap_update h (s x) hFrd)| s h n. (s, h)\<in>p\<^emph>f \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0} \<subseteq> {(s, fmap_update h (s x) hFrd)| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}\<^emph>f"
  proof
    fix \<sigma>'
    assume asm: "\<sigma>'\<in>{(s, fmap_update h (s x) hFrd)| s h n. (s, h)\<in>p\<^emph>f \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}"
    then obtain s h\<^sub>p h\<^sub>f n where "(s, h\<^sub>p)\<in>p" "(s, h\<^sub>f)\<in>f" "h\<^sub>p\<bottom>h\<^sub>f" "fmap_lookup (h\<^sub>p++\<^sub>fh\<^sub>f)(s x)=hOcc n" "s x\<noteq>0" "\<sigma>'=(s, fmap_update (h\<^sub>p++\<^sub>fh\<^sub>f) (s x) hFrd)" unfolding star_def by auto
    then have "fmap_lookup h\<^sub>p (s x)=hOcc n" using Free.prems(2) unfolding provides_def by (metis RedRefl aborts.intros(3) compat_heap_union_comm fmap_add_SomeD)
    then have "(s, fmap_update h\<^sub>p (s x) hFrd)\<in>{(s, fmap_update h (s x) hFrd)| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}" using \<open>(s, h\<^sub>p)\<in>p\<close> \<open>s x\<noteq>0\<close> \<open>h\<^sub>p\<bottom>h\<^sub>f\<close> unfolding compat_def by auto
    moreover have "fmap_lookup h\<^sub>f (s x)=hUkw" using \<open>h\<^sub>p\<bottom>h\<^sub>f\<close> \<open>fmap_lookup h\<^sub>p (s x)=hOcc n\<close> unfolding compat_def by auto
    then have "\<And>l. fmap_lookup ((fmap_update h\<^sub>p (s x) hFrd)++\<^sub>fh\<^sub>f) l = fmap_lookup (fmap_update (h\<^sub>p++\<^sub>fh\<^sub>f) (s x) hFrd) l"
    proof(-)
      fix l
      show "fmap_lookup ((fmap_update h\<^sub>p (s x) hFrd)++\<^sub>fh\<^sub>f) l = fmap_lookup (fmap_update (h\<^sub>p++\<^sub>fh\<^sub>f) (s x) hFrd) l"
      proof(cases "l=s x")
        case True
        then show ?thesis by (metis \<open>fmap_lookup (h\<^sub>p++\<^sub>fh\<^sub>f) (s x)=hOcc n\<close> \<open>fmap_lookup h\<^sub>f (s x)=hUkw\<close> fmap_add_dom_app_simps(1) fmap_add_upd_left option.discI)
      next
        case False
        then show ?thesis by (metis \<open>fmap_lookup (h\<^sub>p++\<^sub>fh\<^sub>f) (s x)=hOcc n\<close> \<open>fmap_lookup h\<^sub>f (s x)=hUkw\<close> fmap_add_dom_app_simps(1) fmap_add_upd_left option.discI)
      qed
    qed
    then have "\<sigma>'=(s, (fmap_update h\<^sub>p (s x) hFrd)++\<^sub>fh\<^sub>f)" using \<open>\<sigma>'=(s, fmap_update (h\<^sub>p++\<^sub>fh\<^sub>f) (s x) hFrd)\<close> using Rep_fmap_inject by fastforce
    moreover have "fmap_update h\<^sub>p (s x) hFrd\<bottom>h\<^sub>f" using \<open>h\<^sub>p\<bottom>h\<^sub>f\<close> \<open>fmap_lookup h\<^sub>p (s x)=hOcc n\<close> unfolding compat_def by auto
    ultimately show "\<sigma>'\<in>{(s, fmap_update h (s x) hFrd)| s h n. (s, h)\<in>p \<and> fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}\<^emph>f" using \<open>(s, h\<^sub>f)\<in>f\<close> unfolding star_def by blast
  qed
  ultimately show ?case unfolding sp_free by blast
qed

lemma
  assumes "provides p (Write x e)"
    shows "\<exists>p'. p\<subseteq>{(s,h)| s h n. fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}\<^emph>p'"
proof(-)
  have "\<forall>(s, h)\<in>p. s x\<noteq>0 \<and> fmap_lookup h(s x)\<noteq>hUkw \<and> fmap_lookup h (s x)\<noteq>hFrd" using assms unfolding provides_def by (smt (verit, del_insts) RedRefl aborts.intros(1) case_prodI2)
  moreover have "\<And>h l. h l\<noteq>hUkw \<Longrightarrow> h l\<noteq>hFrd \<Longrightarrow> \<exists>n. h l=hOcc n" by (metis (mono_tags, opaque_lifting) not_None_eq old.unit.exhaust sum.exhaust_sel)
  ultimately have "\<forall>(s, h)\<in>p. s x\<noteq>0 \<and> (\<exists>n. fmap_lookup h (s x)=hOcc n)" by fast
  obtain p' where p'_def: "p'={(s, fmap_update h (s x) hUkw)| s h. (s, h)\<in>p}" by simp
  have "p\<subseteq>{(s,h)| s h n. fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}\<^emph>p'" 
  proof
    fix \<sigma>
    assume "\<sigma>\<in>p"
    then obtain s h n where shn_def: "\<sigma>=(s, h)" "s x\<noteq>0" "fmap_lookup h (s x)=hOcc n" using \<open>\<forall>(s, h)\<in>p. s x\<noteq>0 \<and> (\<exists>n. fmap_lookup h (s x)=hOcc n)\<close> by blast
    moreover obtain h\<^sub>0 h' where h\<^sub>0h'_def: "h\<^sub>0=fmap_update fmap_empty (s x) (hOcc n)" "h'=fmap_update h (s x) hUkw" by meson
    then have "h\<^sub>0\<bottom>h'" unfolding compat_def by simp
    moreover  have "h=h\<^sub>0++\<^sub>fh'"
    proof(-)
      have "h\<subseteq>\<^sub>fh\<^sub>0++\<^sub>fh'" using map_add_upd_left map_upd_triv calculation(4) compat_heap_union_comm h\<^sub>0h'_def(1,2) shn_def(3) by (simp add: map_add_upd_left map_le_def)
      moreover have "h\<^sub>0++\<^sub>fh'\<subseteq>\<^sub>fh"
      proof -
        have "h = h' ++\<^sub>f h"
          by (metis fmap_add_subsumed1 h\<^sub>0h'_def(2) upd_None_fmap_le)
        then show ?thesis
          by (metis (no_types) fmap_add_dom_app_simps(1) fmap_add_empty fmap_add_subsumed1 fmap_add_upd fmap_add_upd_left fmap_le_iff_fmap_add_commute fmap_upd_triv h\<^sub>0h'_def(1,2) shn_def(3) upd_None_fmap_le)
      qed
      ultimately show ?thesis by (metis fmap_add_subsumed1 fmap_le_iff_fmap_add_commute)
    qed
    ultimately show "\<sigma>\<in>{(s,h)| s h n. fmap_lookup h (s x)=hOcc n \<and> s x\<noteq>0}\<^emph>p'" unfolding p'_def star_def using \<open>\<sigma>\<in>p\<close> h\<^sub>0h'_def shn_def(1,2) by fastforce
  qed
  then show ?thesis by fast
qed

lemma star_comm:
  "p\<^emph>q=q\<^emph>p" unfolding star_def using compat_comm compat_heap_union_comm by blast

lemma star_unit:
  "p\<^emph>{(s, fmap_empty)| s. True}=p" (is "p\<^emph>?u=p")
proof
  show "p\<^emph>?u\<subseteq>p" unfolding star_def by (smt (verit) Pair_inject compat_heap_union_comm empty_fmap_add mem_Collect_eq subsetI)
  show "p\<subseteq>p\<^emph>?u"
  proof(-)
    have "\<And>s h. (s, h)\<in>p \<Longrightarrow> (s, h)\<in>p\<^emph>?u"
    proof(-)
      fix s h
      assume "(s, h)\<in>p"
      moreover have "(s, fmap_empty)\<in>?u" by blast
      moreover have "h\<bottom>fmap_empty" unfolding compat_def by auto
      moreover have "h=h++\<^sub>ffmap_empty" by simp
      ultimately show "(s, h)\<in>p\<^emph>?u" unfolding star_def by blast
    qed
    then show ?thesis by fast
  qed
qed

lemma star_unit':
  "{(s, fmap_empty)| s. True}\<^emph>q=q" using star_comm star_unit by blast

lemma star_distr_over_union: 
  "(p\<union>q)\<^emph>f = (p\<^emph>f)\<union>(q\<^emph>f)"
  unfolding star_def by blast

lemma star_distr_over_union_general:
  "(\<Union>p\<in>P. p)\<^emph>f = (\<Union>p\<in>P. p\<^emph>f)"
  unfolding star_def by fast

lemma star_assoc_aux:
  "abc=ab\<^emph>c \<Longrightarrow> ab=a\<^emph>b \<Longrightarrow> bc=b\<^emph>c \<Longrightarrow> abc=a\<^emph>bc"
proof(-)
  assume asm: "abc=ab\<^emph>c" "ab=a\<^emph>b" "bc=b\<^emph>c"
  have "a\<^emph>b = {\<sigma>\<^sub>a\<^sub>b| \<sigma>\<^sub>a\<^sub>b \<sigma>\<^sub>a \<sigma>\<^sub>b. \<sigma>\<^sub>a\<in>a \<and> \<sigma>\<^sub>b\<in>b \<and> hAdd \<sigma>\<^sub>a\<^sub>b \<sigma>\<^sub>a \<sigma>\<^sub>b}" unfolding star_via_hAdd by fast
  moreover have "ab\<^emph>c = {\<sigma>\<^sub>a\<^sub>b\<^sub>c| \<sigma>\<^sub>a\<^sub>b\<^sub>c \<sigma>\<^sub>a\<^sub>b \<sigma>\<^sub>c. \<sigma>\<^sub>a\<^sub>b\<in>ab \<and> \<sigma>\<^sub>c\<in>c \<and> hAdd \<sigma>\<^sub>a\<^sub>b\<^sub>c \<sigma>\<^sub>a\<^sub>b \<sigma>\<^sub>c}" unfolding star_via_hAdd by blast
  ultimately have "abc = {\<sigma>\<^sub>a\<^sub>b\<^sub>c| \<sigma>\<^sub>a\<^sub>b\<^sub>c \<sigma>\<^sub>a\<^sub>b \<sigma>\<^sub>c. \<sigma>\<^sub>a\<^sub>b\<in>{\<sigma>\<^sub>a\<^sub>b| \<sigma>\<^sub>a\<^sub>b \<sigma>\<^sub>a \<sigma>\<^sub>b. \<sigma>\<^sub>a\<in>a \<and> \<sigma>\<^sub>b\<in>b \<and> hAdd \<sigma>\<^sub>a\<^sub>b \<sigma>\<^sub>a \<sigma>\<^sub>b} \<and> \<sigma>\<^sub>c\<in>c \<and> hAdd \<sigma>\<^sub>a\<^sub>b\<^sub>c \<sigma>\<^sub>a\<^sub>b \<sigma>\<^sub>c}" using asm(1,2) by presburger
  then have "abc = {\<sigma>\<^sub>a\<^sub>b\<^sub>c| \<sigma>\<^sub>a\<^sub>b\<^sub>c \<sigma>\<^sub>a\<^sub>b \<sigma>\<^sub>c \<sigma>\<^sub>a \<sigma>\<^sub>b. \<sigma>\<^sub>a\<in>a \<and> \<sigma>\<^sub>b\<in>b \<and> hAdd \<sigma>\<^sub>a\<^sub>b \<sigma>\<^sub>a \<sigma>\<^sub>b \<and> \<sigma>\<^sub>c\<in>c \<and> hAdd \<sigma>\<^sub>a\<^sub>b\<^sub>c \<sigma>\<^sub>a\<^sub>b \<sigma>\<^sub>c}" by blast
  then have "abc = {\<sigma>\<^sub>a\<^sub>b\<^sub>c| \<sigma>\<^sub>a\<^sub>b\<^sub>c \<sigma>\<^sub>b\<^sub>c \<sigma>\<^sub>c \<sigma>\<^sub>a \<sigma>\<^sub>b. \<sigma>\<^sub>a\<in>a \<and> \<sigma>\<^sub>b\<in>b \<and> hAdd \<sigma>\<^sub>b\<^sub>c \<sigma>\<^sub>b \<sigma>\<^sub>c \<and> \<sigma>\<^sub>c\<in>c \<and> hAdd \<sigma>\<^sub>a\<^sub>b\<^sub>c \<sigma>\<^sub>a \<sigma>\<^sub>b\<^sub>c}" using hAdd_assoc hAdd_assoc' by blast
  then have "abc = {\<sigma>\<^sub>a\<^sub>b\<^sub>c| \<sigma>\<^sub>a\<^sub>b\<^sub>c \<sigma>\<^sub>b\<^sub>c \<sigma>\<^sub>a. \<sigma>\<^sub>a\<in>a \<and> \<sigma>\<^sub>b\<^sub>c\<in>{\<sigma>\<^sub>b\<^sub>c| \<sigma>\<^sub>b\<^sub>c \<sigma>\<^sub>b \<sigma>\<^sub>c. \<sigma>\<^sub>b\<in>b \<and> hAdd \<sigma>\<^sub>b\<^sub>c \<sigma>\<^sub>b \<sigma>\<^sub>c \<and> \<sigma>\<^sub>c\<in>c} \<and> hAdd \<sigma>\<^sub>a\<^sub>b\<^sub>c \<sigma>\<^sub>a \<sigma>\<^sub>b\<^sub>c}" by blast
  then have "abc = {\<sigma>\<^sub>a\<^sub>b\<^sub>c| \<sigma>\<^sub>a\<^sub>b\<^sub>c \<sigma>\<^sub>b\<^sub>c \<sigma>\<^sub>a. \<sigma>\<^sub>a\<in>a \<and> \<sigma>\<^sub>b\<^sub>c\<in>bc \<and> hAdd \<sigma>\<^sub>a\<^sub>b\<^sub>c \<sigma>\<^sub>a \<sigma>\<^sub>b\<^sub>c}" using asm(3) unfolding star_via_hAdd by blast
  then show ?thesis unfolding star_via_hAdd by blast
qed

lemma star_assoc:
  "(a\<^emph>b)\<^emph>c = a\<^emph>(b\<^emph>c)"
  using star_assoc_aux by auto





definition rstar :: "rassertion \<Rightarrow> rassertion \<Rightarrow> rassertion" (infix "\<^emph>\<^sub>r" 51) where
 "p \<^emph>\<^sub>r q = {(\<sigma>, \<sigma>')| \<sigma> \<sigma>\<^sub>p \<sigma>\<^sub>q \<sigma>' \<sigma>\<^sub>p' \<sigma>\<^sub>q'. (\<sigma>\<^sub>p, \<sigma>\<^sub>p')\<in>p \<and> (\<sigma>\<^sub>q, \<sigma>\<^sub>q')\<in>q \<and> hAdd \<sigma> \<sigma>\<^sub>p \<sigma>\<^sub>q \<and> hAdd \<sigma>' \<sigma>\<^sub>p' \<sigma>\<^sub>q'}"





definition intu :: "assertion \<Rightarrow> bool" where
  "intu p \<equiv> \<forall>(s,h\<^sub>0)\<in>p. \<forall>h. h\<supseteq>\<^sub>fh\<^sub>0 \<longrightarrow> (s, h)\<in>p"

definition pure :: "assertion \<Rightarrow> bool" where
  "pure p \<equiv> \<forall>(s, h)\<in>p. \<forall>h'. (s, h')\<in>p"

lemma intuI:
  assumes "\<And>s h\<^sub>0 h. (s, h\<^sub>0)\<in>p \<Longrightarrow> h\<^sub>0\<subseteq>\<^sub>fh \<Longrightarrow> (s, h)\<in>p"
    shows "intu p"
  using assms unfolding intu_def by fast

lemma pureI:
  assumes "\<And>s h h'. (s, h)\<in>p \<Longrightarrow> (s, h')\<in>p"
    shows "pure p"
  using assms unfolding pure_def by fast

lemma pure_implies_intu: "pure p \<Longrightarrow> intu p"
  unfolding pure_def intu_def by blast

lemma star_coincides_inter':
  assumes "intu p"
      and "pure q"
    shows "p\<^emph>q=p\<inter>q"
proof
  show "p\<^emph>q\<subseteq>p\<inter>q"
  proof
    fix \<sigma>
    assume "\<sigma>\<in>p\<^emph>q"
    then obtain s h where "\<sigma>=(s, h)" "(s, h)\<in>p\<^emph>q" by (metis prod.collapse)
    then obtain h\<^sub>p h\<^sub>q where "(s, h\<^sub>p)\<in>p" "(s, h\<^sub>q)\<in>q" "h=h\<^sub>p++\<^sub>fh\<^sub>q" "h\<^sub>p\<bottom>h\<^sub>q" unfolding star_def by blast
    then have "(s, h)\<in>p" using assms(1) unfolding intu_def using compat_heap_union_comm by fastforce
    moreover have "(s, h)\<in>q" using \<open>(s, h\<^sub>q)\<in>q\<close> assms(2) unfolding pure_def by fast
    ultimately show "\<sigma>\<in>p\<inter>q" using \<open>\<sigma>=(s, h)\<close> by fast
  qed
  show "p\<inter>q\<subseteq>p\<^emph>q"
  proof
    fix \<sigma>
    assume "\<sigma>\<in>p\<inter>q"
    then obtain s h where "\<sigma>=(s, h)" "(s, h)\<in>p" "(s, h)\<in>q" by (metis Int_iff surj_pair)
    then have "(s, fmap_empty)\<in>q" using assms(2) unfolding pure_def by fast
    then show "\<sigma>\<in>p\<^emph>q" using \<open>(s, h)\<in>p\<close> unfolding \<open>\<sigma>=(s, h)\<close> star_def compat_def by force
  qed
qed

lemma star_preserves_intu:
  assumes "intu p"
      and "intu q"
    shows "intu (p\<^emph>q)"
proof(rule intuI)
  fix s h H
  assume asm: "(s, h)\<in>p\<^emph>q" "h\<subseteq>\<^sub>fH"
  then obtain h\<^sub>p h\<^sub>q where h\<^sub>ph\<^sub>q_def: "(s, h\<^sub>p)\<in>p" "(s, h\<^sub>q)\<in>q" "h=h\<^sub>p++\<^sub>fh\<^sub>q" "h\<^sub>p\<bottom>h\<^sub>q" unfolding star_def by blast
  obtain H\<^sub>p H\<^sub>q where H\<^sub>pH\<^sub>q_def: "H\<^sub>p=H|`\<^sub>f(fmap_dom h\<^sub>p)" "H\<^sub>q=H|`\<^sub>f(-fmap_dom h\<^sub>p)" by fastforce
  have "\<forall>l\<in>fmap_dom h\<^sub>p. fmap_lookup h\<^sub>p l = fmap_lookup H\<^sub>p l" by (metis (no_types, opaque_lifting) H\<^sub>pH\<^sub>q_def(1) IntI asm(2) compat_heap_union_comm fmap_add_dom_app_simps(1) fmap_add_subsumed1 fmap_dom_restrict fmap_le_iff_fmap_add_commute fmap_le_implies_dom_le h\<^sub>ph\<^sub>q_def(3) h\<^sub>ph\<^sub>q_def(4) in_mono split_fmap)
  then have "h\<^sub>p\<subseteq>\<^sub>fH\<^sub>p" apply simp by (simp add: map_le_def)
  then have "(s, H\<^sub>p)\<in>p" using \<open>(s, h\<^sub>p)\<in>p\<close> assms(1) unfolding intu_def by fast
  moreover have "\<forall>l\<in>fmap_dom h\<^sub>q. fmap_lookup h\<^sub>q l = fmap_lookup H\<^sub>q l" by (metis Diff_Compl Diff_iff FiniteMap.compat_def H\<^sub>pH\<^sub>q_def(2) asm(2) compat_comm emptyE fmap_add_dom_app_simps(1) fmap_add_le_fmapE fmap_add_subsumed1 fmap_le_iff_fmap_add_commute fmap_restrict_in h\<^sub>ph\<^sub>q_def(3) h\<^sub>ph\<^sub>q_def(4))
  then have "h\<^sub>q\<subseteq>\<^sub>fH\<^sub>q" apply simp by (simp add: map_le_def)
  then have "(s, H\<^sub>q)\<in>q" using \<open>(s, h\<^sub>q)\<in>q\<close> assms(2) unfolding intu_def by fast
  moreover from H\<^sub>pH\<^sub>q_def have "H=H\<^sub>p++\<^sub>fH\<^sub>q" by (metis double_complement split_fmap)
  moreover from H\<^sub>pH\<^sub>q_def have "H\<^sub>p\<bottom>H\<^sub>q" unfolding compat_def by simp
  ultimately show "(s, H)\<in>p\<^emph>q" unfolding star_def by blast
qed

lemma star_preserves_pure:
  assumes "pure p"
      and "pure q"
    shows "pure (p\<^emph>q)"
proof(rule pureI)
  fix s h h'
  assume asm: "(s, h)\<in>p\<^emph>q"
  then obtain h\<^sub>p h\<^sub>q where "(s, h\<^sub>p)\<in>p" "(s, h\<^sub>q)\<in>q" unfolding star_def by blast
  then have "(s, h')\<in>p" "(s, fmap_empty)\<in>q" using assms unfolding pure_def by auto
  moreover have "h'=h'++\<^sub>ffmap_empty" by simp
  moreover have "h'\<bottom>fmap_empty" unfolding compat_def by simp
  ultimately show "(s, h')\<in>p\<^emph>q" unfolding star_def by fast
qed









definition larger :: "hstate \<Rightarrow> hstate \<Rightarrow> bool" (infix "\<sqsupseteq>" 52) where
  "h \<sqsupseteq> h\<^sub>0 \<equiv> fmap_dom h\<supseteq>fmap_dom h\<^sub>0 \<and> (\<forall>l\<in>fmap_dom h\<^sub>0. fmap_lookup h\<^sub>0 l=hFrd \<or> fmap_lookup h l=fmap_lookup h\<^sub>0 l)"

abbreviation smaller (infix "\<sqsubseteq>" 52) where
  "h\<^sub>0 \<sqsubseteq> h \<equiv> h \<sqsupseteq> h\<^sub>0"

lemma fmap_le_implies_smaller:
  "h\<subseteq>\<^sub>fh' \<Longrightarrow> h\<sqsubseteq>h'"
  unfolding larger_def by (metis fmap_add_dom_app_simps(1) fmap_add_subsumed1 fmap_le_iff_fmap_add_commute fmap_le_implies_dom_le)

definition intu' :: "assertion \<Rightarrow> bool" where
  "intu' p \<equiv> \<forall>(s,h\<^sub>0)\<in>p. \<forall>h. h\<sqsupseteq>h\<^sub>0 \<longrightarrow> (s, h)\<in>p"

lemma intu'I:
  assumes "\<And>s h\<^sub>0 h. (s, h\<^sub>0)\<in>p \<Longrightarrow> h\<^sub>0\<sqsubseteq>h \<Longrightarrow> (s, h)\<in>p"
    shows "intu' p"
  using assms unfolding intu'_def by fast

lemma pure_implies_intu': "pure p \<Longrightarrow> intu' p"
  unfolding pure_def intu'_def by blast

lemma star_coincides_inter:
  assumes "intu' p"
      and "pure q"
    shows "p\<^emph>q=p\<inter>q"
proof
  show "p\<^emph>q\<subseteq>p\<inter>q"
  proof
    fix \<sigma>
    assume "\<sigma>\<in>p\<^emph>q"
    then obtain s h where "\<sigma>=(s, h)" "(s, h)\<in>p\<^emph>q" by (metis prod.collapse)
    then obtain h\<^sub>p h\<^sub>q where "(s, h\<^sub>p)\<in>p" "(s, h\<^sub>q)\<in>q" "h=h\<^sub>p++\<^sub>fh\<^sub>q" "h\<^sub>p\<bottom>h\<^sub>q" unfolding star_def by blast
    then have "(s, h)\<in>p" using assms(1) unfolding intu'_def using compat_heap_union_comm fmap_le_implies_smaller by fastforce
    moreover have "(s, h)\<in>q" using \<open>(s, h\<^sub>q)\<in>q\<close> assms(2) unfolding pure_def by fast
    ultimately show "\<sigma>\<in>p\<inter>q" using \<open>\<sigma>=(s, h)\<close> by fast
  qed
  show "p\<inter>q\<subseteq>p\<^emph>q"
  proof
    fix \<sigma>
    assume "\<sigma>\<in>p\<inter>q"
    then obtain s h where "\<sigma>=(s, h)" "(s, h)\<in>p" "(s, h)\<in>q" by (metis Int_iff surj_pair)
    then have "(s, fmap_empty)\<in>q" using assms(2) unfolding pure_def by fast
    then show "\<sigma>\<in>p\<^emph>q" using \<open>(s, h)\<in>p\<close> unfolding \<open>\<sigma>=(s, h)\<close> star_def compat_def by force
  qed
qed

lemma star_preserves_intu':
  assumes "intu' p"
      and "intu' q"
    shows "intu' (p\<^emph>q)"
proof(rule intu'I)
  fix s h H
  assume asm: "(s, h)\<in>p\<^emph>q" "h\<sqsubseteq>H"
  then obtain h\<^sub>p h\<^sub>q where h\<^sub>ph\<^sub>q_def: "(s, h\<^sub>p)\<in>p" "(s, h\<^sub>q)\<in>q" "h=h\<^sub>p++\<^sub>fh\<^sub>q" "h\<^sub>p\<bottom>h\<^sub>q" unfolding star_def by blast
  obtain H\<^sub>p H\<^sub>q where H\<^sub>pH\<^sub>q_def: "H\<^sub>p=H|`\<^sub>f(fmap_dom h\<^sub>p)" "H\<^sub>q=H|`\<^sub>f(-fmap_dom h\<^sub>p)" by fastforce
  have "h\<^sub>p\<sqsubseteq>H\<^sub>p" unfolding larger_def
  proof
    show "\<forall>l\<in>fmap_dom h\<^sub>p. fmap_lookup h\<^sub>p l = hFrd \<or> fmap_lookup H\<^sub>p l = fmap_lookup h\<^sub>p l"
    proof
      fix l
      assume asm': "l\<in>fmap_dom h\<^sub>p"
      then obtain v where "fmap_lookup h\<^sub>p l=v" "v\<noteq>hUkw" by (simp add: domIff)
      then have "fmap_lookup h l=v" by (metis asm' compat_heap_union_comm fmap_add_dom_app_simps(1) h\<^sub>ph\<^sub>q_def(3,4))
      then have "v\<noteq>hFrd \<Longrightarrow> fmap_lookup H l=fmap_lookup h l" by (metis asm' asm(2) compat_heap_union_comm fmap_le_iff_fmap_add_commute fmap_le_implies_dom_le h\<^sub>ph\<^sub>q_def(3,4) in_mono larger_def)
      moreover have "v=hFrd \<Longrightarrow> \<exists>v'\<noteq>hUkw. fmap_lookup H l=v'" using \<open>fmap_lookup h l = v\<close> asm(2) larger_def by auto
      ultimately show "fmap_lookup h\<^sub>p l = hFrd \<or> fmap_lookup H\<^sub>p l = fmap_lookup h\<^sub>p l" using H\<^sub>pH\<^sub>q_def(1) \<open>fmap_lookup h l = v\<close> \<open>fmap_lookup h\<^sub>p l = v\<close> asm' by force
    qed
    then show "fmap_dom h\<^sub>p \<subseteq> fmap_dom H\<^sub>p" using H\<^sub>pH\<^sub>q_def(1) asm(2) h\<^sub>ph\<^sub>q_def(3) larger_def by auto
  qed
  then have "(s, H\<^sub>p)\<in>p" using \<open>(s, h\<^sub>p)\<in>p\<close> assms(1) unfolding intu'_def by fast
  moreover have "h\<^sub>q\<sqsubseteq>H\<^sub>q" unfolding larger_def
  proof
    show "\<forall>l\<in>fmap_dom h\<^sub>q. fmap_lookup h\<^sub>q l = hFrd \<or> fmap_lookup H\<^sub>q l = fmap_lookup h\<^sub>q l"
    proof
      fix l
      assume asm': "l\<in>fmap_dom h\<^sub>q"
      then obtain v where "fmap_lookup h\<^sub>q l=v" "v\<noteq>hUkw" by (simp add: domIff)
      then have "fmap_lookup h l=v" by (metis asm' fmap_add_dom_app_simps(1) h\<^sub>ph\<^sub>q_def(3))
      then have "v\<noteq>hFrd \<Longrightarrow> fmap_lookup H l=fmap_lookup h l" by (metis asm' asm(2) compat_heap_union_comm fmap_le_iff_fmap_add_commute fmap_le_implies_dom_le h\<^sub>ph\<^sub>q_def(3,4) in_mono larger_def)
      moreover have "v=hFrd \<Longrightarrow> \<exists>v'\<noteq>hUkw. fmap_lookup H l=v'" using \<open>fmap_lookup h l = v\<close> asm(2) larger_def by auto
      ultimately show "fmap_lookup h\<^sub>q l = hFrd \<or> fmap_lookup H\<^sub>q l = fmap_lookup h\<^sub>q l" using H\<^sub>pH\<^sub>q_def(1) \<open>fmap_lookup h\<^sub>q l = v\<close>  by (metis ComplI H\<^sub>pH\<^sub>q_def(2) \<open>h\<^sub>p \<sqsubseteq> H\<^sub>p\<close> \<open>v \<noteq> hUkw\<close> fmap_restrict_in hAdd_lookup_disjoint hAdd_prop h\<^sub>ph\<^sub>q_def(3) h\<^sub>ph\<^sub>q_def(4) larger_def option.discI)
    qed
    show "fmap_dom h\<^sub>q \<subseteq> fmap_dom H\<^sub>q"
    proof
      fix l
      assume asm': "l\<in>fmap_dom h\<^sub>q"
      then obtain v where v_def: "fmap_lookup h\<^sub>q l=v" "v\<noteq>hUkw" by (simp add: domIff)
      then have "fmap_lookup h l=v" by (metis asm' fmap_add_dom_app_simps(1) h\<^sub>ph\<^sub>q_def(3))
      then have "v\<noteq>hFrd \<Longrightarrow> fmap_lookup H l=fmap_lookup h l" by (metis asm' asm(2) compat_heap_union_comm fmap_le_iff_fmap_add_commute fmap_le_implies_dom_le h\<^sub>ph\<^sub>q_def(3,4) in_mono larger_def)
      moreover have "v=hFrd \<Longrightarrow> \<exists>v'\<noteq>hUkw. fmap_lookup H l=v'" using \<open>fmap_lookup h l = v\<close> asm(2) larger_def by auto
      ultimately show "l\<in>fmap_dom H\<^sub>q" by (metis (no_types, lifting) H\<^sub>pH\<^sub>q_def(2) v_def compat_heap_union_comm fmap_add_dom_app_simps(1) fmap_add_dom_app_simps(2) fmap_restrict_out hAdd_lookup_disjoint hAdd_prop h\<^sub>ph\<^sub>q_def(3) h\<^sub>ph\<^sub>q_def(4) split_fmap)
    qed
  qed
  then have "(s, H\<^sub>q)\<in>q" using \<open>(s, h\<^sub>q)\<in>q\<close> assms(2) unfolding intu'_def by fast
  moreover from H\<^sub>pH\<^sub>q_def have "H=H\<^sub>p++\<^sub>fH\<^sub>q" by (metis double_complement split_fmap)
  moreover from H\<^sub>pH\<^sub>q_def have "H\<^sub>p\<bottom>H\<^sub>q" unfolding compat_def by simp
  ultimately show "(s, H)\<in>p\<^emph>q" unfolding star_def by blast
qed

end